"""
(c) 2023 LLC «DACOM M»
"""

import json
import os
import signal
import re
import shutil
from consul import Consul
from time import sleep
from secrets import token_bytes
from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output, \
    get_result, action_service, is_service_active, is_service_failed, states_result, CONTROLLER_LOCAL_DATA_PATH, \
    WORKER_USER_UID, CONTROLLER_USER
from veilrpc.states.puppet import puppet_modify
from veilrpc.utils.file import get_uid_gid
from veilrpc.utils.mount import active

COROSYNC_PATH = '/etc/corosync/corosync.conf'
COROSYNC_KEY = f'{CONTROLLER_LOCAL_DATA_PATH}/authkey'
DLM_PATH = '/etc/dlm/dlm.conf'
GFS2_LOCKCAPTURE_PATH = '/var/run/gfs2_lockcapture.pid'
SBD_PATH = '/etc/sysconfig/sbd'
CONSUL_PATH = '/etc/consul.d/config.json'


def attach_nodes_to_ocfs2(nodes, **kwargs):
    for node in nodes:
        cmd = f"o2cb_ctl -C -i -n {node['uuid']} -t node -a number={node['number']} " \
              f"-a ip_address={node['address']} -a ip_port=7777 -a cluster=ocfs2"
        cmd_ret = get_output(cmd, **kwargs)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    return f"Successfully attach nodes {', '.join(nodes)} to ocfs2."


@states_response
def configure_cluster_fs(cluster_fs_conf, enabled, **kwargs):
    if enabled:
        with open('/etc/ocfs2/cluster.conf', 'w') as f:
            f.write("\n".join(cluster_fs_conf))
        action, service = 'restart', 'o2cb'
        cmd_ret = action_service(service, action, timeout=300)
        if not cmd_ret['result']:
            raise CommandExecutionError(f"Error in {action} {service}: {cmd_ret['comment']}")
    else:
        action, service = 'stop', 'o2cb'
        cmd_ret = action_service(service, action, timeout=300)
        get_output('rm -f /etc/ocfs2/cluster.conf')
        if not cmd_ret['result']:
            raise CommandExecutionError(f"Error in {action} {service}: {cmd_ret['comment']}")
    from .node import heartbeat_update
    cmd_ret = heartbeat_update(**kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if enabled:
        return 'Successfully enabled ocfs2 config.'
    else:
        return 'Successfully disabled ocfs2 config.'


@states_response
def configure_corosync(corosync_conf, dlm_conf, sbd_conf, enabled, key_path, corosync_key: str,
                       clean_shutdown=False, reload=False, restart=True, **kwargs):
    from .node import heartbeat_update

    def check_svc_action(action, service, exc=True, timeout=300, rty=1, delay=3):
        for loop in range(rty):
            if (_cmd_ret := action_service(service_name=service, action=action, timeout=timeout))['result']:
                return f'{action} {service}: success.\n'
            sleep(delay if loop < rty-1 else 0)
        if exc:
            raise CommandExecutionError(f"Error in {action} {service}: {_cmd_ret['comment']}")
        return f"Error in {action} {service}: {_cmd_ret['comment']}\n"

    def chk_result(_res):
        if not _res['result']:
            raise CommandExecutionError(_res['comment'])
        return f"{_res['comment']}\n"

    # pylint: disable=unsubscriptable-object
    if enabled:
        comment, comm_tail = '', '\nSuccessfully set gfs2 config.'
        from .node import check_watchdog, load_module
        cmd_ret = check_watchdog()
        if not cmd_ret['result']:
            raise CommandExecutionError('Watchdog check error.')
        if not cmd_ret['comment']:
            res = load_module('softdog')
            comment += chk_result(res)
            sleep(1)
            new_cmd_ret = check_watchdog()
            if not new_cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            if not new_cmd_ret['comment']:
                raise CommandExecutionError('Not found watchdog device.')
        with open(COROSYNC_PATH, 'w') as f:
            f.write("\n".join(corosync_conf))
        with open(DLM_PATH, 'w') as f:
            f.write("\n".join(dlm_conf))
        with open(key_path, 'wb') as f:
            f.write(bytes.fromhex(corosync_key))
        with open(SBD_PATH, 'w') as f:  # actually not used for now; default /etc/default/sbd is used
            f.write("\n".join(sbd_conf))
        if not os.path.isfile('/run/sbd.pid'):
            open('/run/sbd.pid', 'w').close()
        puppet_modify(module='space_node', params={'space_node::dlm_ensure': True}, apply=False)
        comment += check_svc_action(action='enable', service='sbd')

        if restart:
            comment += check_svc_action(action='restart', service='corosync.service', rty=2)
        else:
            if not is_service_active('corosync.service'):
                comment += check_svc_action(action='enable', service='corosync.service')
                comment += check_svc_action(action='start', service='corosync.service', exc=False)
        sleep(3)
        comment += check_svc_action(action='enable', service='dlm.service')
        if not is_service_active('dlm'):
            comment += check_svc_action(action='start', service='dlm', exc=False)

        if reload:
            comment += chk_result(reload_corosync_conf()) + chk_result(reload_dlm_conf())

    else:
        comment, comm_tail = '', '\nSuccessfully remove gfs2 config. '
        for sbd in ('sbd.service', 'sbd_remote.service'):
            if is_service_active(sbd):
                sbd_pid = get_output(f'systemctl show --property=MainPID --value {sbd}')
                if sbd_pid['result']:
                    if sbd_pid['comment'] == '0':
                        comment += f'Got PID=0 for {sbd}. Unable to kill SBD.'
                    else:
                        os.kill(int(sbd_pid['comment']), signal.SIGTERM)
                        comment += f'Successfully stopped {sbd}\n'
            get_result(f'systemctl reset-failed {sbd}')
        comment += check_svc_action(action='disable', service='sbd.service')
        comment += check_svc_action(action='disable', service='sbd_remote.service')

        try:
            comment += check_svc_action(action='stop', service='corosync.service')
            comment += check_svc_action(action='disable', service='corosync.service')
            comment += check_svc_action(action='disable', service='dlm.service')

        except CommandExecutionError:
            raise
        finally:
            puppet_modify(module='space_node', params={'space_node::dlm_ensure': False}, apply=False)
            for i in [DLM_PATH, SBD_PATH, key_path, GFS2_LOCKCAPTURE_PATH]:
                get_output(f'rm -f {i}')

    comment += chk_result(heartbeat_update(**kwargs)) + comm_tail
    return comment


@states_response
def write_conf(corosync_conf):
    with open(COROSYNC_PATH, 'w') as f:
        f.write("\n".join(corosync_conf))


@states_response
def write_conf_and_reload(corosync_conf, reload=False):
    """Обновляет файл конфигурации и заставляет перепрочитать его на всех узлах."""
    # pylint: disable=unsubscriptable-object
    cmd_ret = write_conf(corosync_conf)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if not is_service_active('corosync'):
        action, service = 'restart', 'corosync'
        cmd_ret = action_service(service, action, timeout=300)
        if not cmd_ret['result']:
            raise CommandExecutionError(f"Error in {action} {service}: {cmd_ret['comment']}")
    if reload:
        cmd_ret = reload_corosync_conf()
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def reload_corosync_conf(**kwargs):
    """Tell all instances of corosync in this cluster to reload corosync.conf."""
    cmd = 'corosync-cfgtool -R'
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return 'Successfully reload corosync conf.'


@states_response
def shutdown_cleanly_corosync():
    """Shutdown corosync cleanly on this node."""
    if is_service_active('corosync'):
        cmd = 'corosync-cfgtool -H --force'
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return 'Successfully cleanly shutdown corosync.'
    else:
        return 'corosync service is not active.'


@states_response
def kill_corosync_node(node_id):
    """Shutdown corosync cleanly on this node."""
    if is_service_active('corosync'):
        cmd = f"corosync-cfgtool -k {node_id}"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return f'Successfully kill corosync node {node_id}.'
    else:
        return 'corosync service is not active.'


@states_response
def corosync_keygen(filename=None, length=256) -> str:
    """
    Генерация ключа на одном узле для всего кластера
    https://github.com/corosync/corosync/blob/main/tools/corosync-keygen.c о ключе
    """
    if not filename:
        filename = '/etc/corosync/authkey'
    with open(filename, 'wb') as keyfile:
        keyfile.write(token := token_bytes(length))
    _, ctrl_grp = get_uid_gid(CONTROLLER_USER)
    os.chown(filename, WORKER_USER_UID, ctrl_grp)
    os.chmod(filename, 0O640)
    return token.hex()


@states_response
def get_corosync_key(filename) -> str:
    if os.path.exists(filename):
        with open(filename, 'rb') as f:
            key = f.read()
            return key.hex()
    else:
        cmd_ret = corosync_keygen(filename)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return cmd_ret['comment']


@states_response
def corosync_membership(**kwargs):
    """Get corosync membership information."""
    cmd = 'corosync-quorumtool -lp'
    cmd_ret = get_output(cmd, **kwargs)
    if cmd_ret.get('retcode', 1) != 1:
        raise CommandExecutionError(cmd_ret['comment'])
    lines = cmd_ret['comment'].splitlines()[4:]
    res = []
    for line in lines:
        # pylint: disable=unused-variable
        node_id, votes, qdevice, name, *other = line.strip().split()
        res.append({'node_id': node_id,
                    'votes': votes,
                    'qdevice': qdevice,
                    'name': name})
    return res


@states_response
def corosync_quorum(**kwargs):
    """Get corosync quorum information."""
    cmd = 'corosync-quorumtool -sp'
    cmd_ret = get_output(cmd, **kwargs)
    if cmd_ret.get('retcode', 1) != 1:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def quorum_status(**kwargs):
    """Get quorum status."""
    cmd = 'corosync-quorumtool 2>&1 | grep -q -E "^Quorate:.*Yes$"'
    if get_result(cmd) and get_result('corosync-cfgtool -bs | grep -P "^\taddr\t" | grep -q -v "\ 127.0.0.1$"'):
        return True
    raise CommandExecutionError('Not quorate in gfs2 quorum')


@states_response
def dlm_lockspaces(**kwargs):
    """Get active dlm lockspaces."""
    cmd = "dlm_tool ls | grep name | awk '{{print $2}}'"
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment'].strip().splitlines()


@states_response
def dlm_ls(**kwargs):
    """Get dlm status."""
    return states_result(get_output, cmd="dlm_tool ls -n")


@states_response
def gfs2_lockcapture(**kwargs):
    """Get dlm status."""
    return states_result(get_output, cmd="gfs2_lockcapture -i")


@states_response
def leave_lockspace(name, **kwargs):
    """Leave dlm lockspace."""
    cmd = f"dlm_tool leave {name}"
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f'Successfully leave dlm lockspace {name}.'


@states_response
def reload_dlm_conf(**kwargs):
    """Reload dlm_controld config settings from dlm.conf"""
    cmd = 'dlm_tool reload_config'
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return 'Successfully reload dlm conf.'


@states_response
def status_detail(**kwargs):
    result = ''
    if os.path.isdir(sysfs := '/sys/fs/gfs2/'):
        for fs in os.listdir(sysfs):
            with open(os.path.join(sysfs, fs, 'withdraw'), 'r') as wh:
                withdraw = wh.read().strip() == '1'
            with open(os.path.join(sysfs, fs, 'device', 'ro'), 'r') as ro:
                readonly = ro.read().strip() == '1'
            if os.path.isdir(os.path.join(sysfs, fs, 'device', 'dm')):
                with open(os.path.join(sysfs, fs, 'device', 'dm', 'name'), 'r') as d_n:
                    devname = d_n.read().strip()
                with open(os.path.join(sysfs, fs, 'device', 'dm', 'suspended'), 'r') as ssp:
                    suspended = ssp.read().strip() == '1'
            else:
                suspended = False
                with open(os.path.join(sysfs, fs, 'device', 'uevent'), 'r') as uev:
                    devname = [line.split('=')[1] for line in uev.read().splitlines() if line.startswith('DEVNAME=')][0]

            if withdraw or suspended or readonly:
                result += f"*** Warning! FS {fs} on block device {devname} FAILED ***\n"
                result += 'FS WITHDRAWN; FSCK NEEDED\n' if withdraw else ''
                result += 'MPATH DEVICE SUSPENDED; CHECK MULTIPATH STATUS\n' if suspended else ''
                result += 'DEVICE IS READONLY; CHECK DEVICE\n' if readonly else ''
                result += ('-' * 30) + '\n'

    result += 'corosync.service: '
    if is_service_active('corosync.service'):
        result += 'active\n\n'
        corosync_ret0 = corosync_quorum(**kwargs)
        if corosync_ret0['result']:
            result += corosync_ret0['comment'] + '\n'
        else:
            result += f'Error getting corosync quorum result: err="{corosync_ret0["comment"]}"\n\n'
        corosync_ret1 = get_output('corosync-cfgtool -n')
        if corosync_ret1['result']:
            result += corosync_ret1['comment'] + '\n'
        else:
            result += f'Error getting corosync links map: err="{corosync_ret1["comment"]}"\n\n'
        corosync_ret2 = get_output('corosync-cmapctl -m stats | grep -e retries -e timeout -e down -e error -e fail '
                                   '| grep -v =\ 0')
        if corosync_ret2['result']:
            if c_ret2 := corosync_ret2['comment'].strip():
                result += f'KNET errors, timeouts, retries, link down events:\n{c_ret2}\n'
        else:
            result += f'Error getting KNET stats: err="{corosync_ret1["comment"]}"\n'

    else:
        result += 'inactive' + (', failed' if is_service_failed('corosync.service') else '') + '\n'

    dlm_ret = dlm_ls(**kwargs)
    if dlm_ret['result']:
        result += dlm_ret['comment']
    else:
        result += f'Error getting DLM lockspace status: err="{dlm_ret["comment"]}"\ndlm.service: '
        if is_service_active('dlm.service'):
            result += 'active\n'
        else:
            result += 'inactive' + (', failed' if is_service_failed('dlm.service') else '') + '\n'

    for k, v in active().items():
        if "/storages/gfs2/" in k:
            cmd = f"gfs2_edit -p jindex {v['device']} | grep journal"
            gfs2_ret = get_output(cmd, timeout=10)
            if gfs2_ret['result']:
                result += f"\ngfs2 lun {v['device']} journals info:\n"
                result += gfs2_ret['comment']
    return result


@states_response
def write_sbd_conf(sbd_conf):
    with open(SBD_PATH, 'w') as f:
        f.write("\n".join(sbd_conf))


@states_response
def send_sbd_message(devices, node_uuid, message, **kwargs):
    devices = '-d '.join(devices)
    cmd = f"sbd -d {devices} message {node_uuid} {message}"
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f'Successfully send_sbd_message {message} to {node_uuid}.'


@states_response
def create_sbd_device(devices, **kwargs):
    devices = '-d '.join(devices)
    cmd = f"sbd -d {devices} create"
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f'Successfully create_sbd_device {devices}.'


@states_response
def gfs2_info(**kwargs):
    ret = {}
    ret['corosync_quorum'] = corosync_quorum(**kwargs)
    ret['dlm_ls'] = dlm_ls(**kwargs)
    ret['gfs2_lockcapture'] = gfs2_lockcapture(**kwargs)
    return ret


@states_response
def ocfs2_status(**kwargs):
    """Get ocfs2 cluster status."""
    cmd = 'o2cb cluster-status ocfs2'
    if get_result(cmd):
        return True
    raise CommandExecutionError('Not quorate in ocfs2 quorum')


@states_response
def ocfs2_nodes(**kwargs):
    """Get ocfs2 nodes status."""
    return states_result(get_output, cmd='o2cb list-nodes ocfs2')


@states_response
def ocfs2_devices(**kwargs):
    """Get ocfs2 devices."""
    return states_result(get_output, cmd='mounted.ocfs2 -f')


@states_response
def ocfs2_remove_node(node, **kwargs):
    return states_result(get_output, cmd=f"o2cb remove-node ocfs2 {node}")


@states_response
def ocfs2_add_node(node, ip_address=None, port=7777, **kwargs):
    """Get ocfs2 nodes status."""
    if ip_address:
        cmd = f"o2cb add-node ocfs2 {node} --port {port}"
    else:
        cmd = f"o2cb add-node ocfs2 {node} --ip {ip_address} --port {port}"
    return states_result(get_output, cmd=cmd)


def consul_object():
    return Consul()


@states_response
def consul_leader(**kwargs):
    cons = kwargs.get('cons') or Consul()
    status = cons.Status(cons)
    return status.leader()


@states_response
def configure_quorum(consul_conf, quorum, **kwargs):
    for root, dirs, files in os.walk('/var/lib/consul'):
        for f in files:
            os.unlink(os.path.join(root, f))
        for d in dirs:
            shutil.rmtree(os.path.join(root, d))
    if quorum:
        with open(CONSUL_PATH, 'w') as f1, open('/etc/consul.d/consul.hcl', 'w') as f2:
            json.dump(consul_conf, f1)
            json.dump(consul_conf, f2)
        action, service = 'restart', 'consul'
        cmd_ret = action_service(service, action, timeout=300)
        if not cmd_ret['result']:
            raise CommandExecutionError(f"Error in {action} {service}: {cmd_ret['comment']}")
        count = 180
        leader_exists = False
        while count > 0:
            leader_ret = consul_leader(suppress_err=True)
            if leader_ret['result'] and leader_ret['comment']:
                leader_exists = True
                sleep(3)
                break
            else:
                sleep(2)
                count -= 2
        if not leader_exists:
            raise CommandExecutionError('Cluster has no active leader')
    else:
        action, service = 'stop', 'consul'
        cmd_ret = action_service(service, action, timeout=300)
        if not cmd_ret['result']:
            raise CommandExecutionError(f"Error in {action} {service}: {cmd_ret['comment']}")

        cmd = f'rm -f {CONSUL_PATH}'
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    puppet_modify(module='space_node', params={'space_node::consul_ensure': quorum}, apply=False)
    from .node import heartbeat_update
    cmd_ret = heartbeat_update(**kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f"Successfully {'en' if quorum else 'dis'}abled consul config."


@states_response
def consul_status(repeat: int = 2, **kwargs):
    cons = kwargs.get('cons') or Consul()
    status = cons.Status(cons)
    leader, leader_ip, leader_port = None, None, None
    for _ in range(repeat):
        leader = status.leader()
        if leader:
            leader_ip, leader_port = leader.split(':')
            break
    peers = status.peers()
    return {"leader": leader, "leader_ip": leader_ip, "leader_port": leader_port, "peers": peers}


@states_response
def consul_members(**kwargs):
    cons = kwargs.get('cons') or Consul()
    operator = cons.Agent(cons)
    config = operator.members()
    return config


@states_response
def consul_operator(**kwargs):
    cons = kwargs.get('cons') or Consul()
    operator = cons.Operator(cons)
    config = operator.raft_config()
    return config


@states_response
def consul_list_tokens(token=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    acl = cons.ACL(cons)
    return acl.list(token)


@states_response
def consul_info_token(acl_id, token=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    acl = cons.ACL(cons)
    return acl.info(acl_id, token)


# pylint: disable=redefined-builtin)
@states_response
def consul_create_token(name=None, type='management', rules=None, acl_id=None, token=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    acl = cons.ACL(cons)
    return acl.create(name, type, rules, acl_id, token)


# pylint: disable=redefined-builtin)
@states_response
def consul_update_token(acl_id, name=None, type='management', rules=None, token=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    acl = cons.ACL(cons)
    return acl.update(acl_id, name, type, rules, token)


@states_response
def consul_destroy_token(acl_id, token=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    acl = cons.ACL(cons)
    return acl.destroy(acl_id, token)


@states_response
def consul_bootstrap(**kwargs):
    cmd = 'consul acl bootstrap -format=json'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        if 'Permission denied' in cmd_ret['comment'] or 'Invalid bootstrap reset index' in cmd_ret['comment']:
            ints = re.findall(r'\d+', cmd_ret['comment'])
            if ints:
                reset_index = ints[-1]
                cmd_ret = get_output(f"echo {reset_index} > '/var/lib/consul/acl-bootstrap-reset'")
                if not cmd_ret['result']:
                    raise CommandExecutionError(cmd_ret['comment'])
                cmd_ret = get_output(cmd)
                if not cmd_ret['result']:
                    raise CommandExecutionError(cmd_ret['comment'])
                output = cmd_ret['comment']
                return json.loads(output)
        else:
            raise CommandExecutionError(cmd_ret['comment'])
    output = cmd_ret['comment']
    return json.loads(output)


@states_response
def consul_session_create(name=None, node=None,
                          checks=None, lock_delay=15,
                          behavior='release', ttl=None,
                          dc=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    ses = cons.Session(cons)
    session = ses.create(name, node, checks, lock_delay,
                         behavior, ttl, dc)
    return session


@states_response
def consul_session_destroy(session_id, dc=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    ses = cons.Session(cons)
    res = ses.destroy(session_id, dc)
    return res


@states_response
def consul_session_list(index=None, wait=None,
                        consistency=None, dc=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    ses = cons.Session(cons)
    sessions = ses.list(index, wait, consistency, dc)
    return sessions


@states_response
def consul_session_info(session_id, index=None, wait=None,
                        consistency=None, dc=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    ses = cons.Session(cons)
    session_info = ses.info(session_id, index, wait,
                            consistency, dc)
    return session_info


@states_response
def consul_session_renew(session_id, dc=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    ses = cons.Session(cons)
    res = ses.renew(session_id, dc)
    return res


@states_response
def consul_kv_get(key, index=None, recurse=False, wait=None,
                  token=None, consistency=None, keys=False,
                  separator=None, dc=None, to_json=False, **kwargs):
    cons = kwargs.get('cons') or Consul()
    kv = cons.KV(cons)
    res = kv.get(key, index, recurse, wait, token, consistency, keys,
                 separator, dc)
    if to_json:
        return json.loads(res[-1]['Value'].decode()) if res[-1] else {}
    return res


@states_response
def consul_kv_put(key, value, cas=None, flags=None,
                  acquire=None, release=None, token=None, dc=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    kv = cons.KV(cons)
    res = kv.put(key, value, cas, flags, acquire,
                 release, token, dc)
    return res


@states_response
def consul_kv_delete(key, recurse=None, cas=False,
                     token=None, dc=None, **kwargs):
    cons = kwargs.get('cons') or Consul()
    kv = cons.KV(cons)
    res = kv.delete(key, recurse, cas, token, dc)
    return res
