"""
(c) 2023 LLC «DACOM M»
"""
import json
import glob
import errno
import os
import shutil
from uuid import uuid4

from veilrpc.utils.bazalt import states_response, CommandExecutionError, \
    get_output, call_with_args, WORKER_USER, thread_pool, states_result, NFS_PATH, \
    SHARED_STORAGES_PATHS, LOCAL_STORAGES_PATH, CLUSTER_STORAGES_PATHS, \
    ZPOOL_STORAGES_PATH, action_service, \
    PATH_ISO_STORAGE, PATH_FILE_STORAGE, PATH_NODES_HEARTBEAT, PATH_TMP
from veilrpc.states.puppet import puppet_modify
from veilrpc.utils.file import chown_directory

GET_OWNER_CMD = "ls -ld {} | awk 'NR==1 {{print $3}}'"
GET_SIZE = "du -b '{file}' 2>&- | cut -f1"
GET_FOLDER_SIZE = "du -sh -B 1 '{folder}' | cut -f1"
LS_CMD = "ls {storage} 2>&-"
CHECK_DIR_HEALTH_CMD = "ls -d '{path}'"
FILEINFO_CMD = "file -s '{file}' 2>&-"
CHECK_SHARED_STORAGE_MOUNTED = "mount | grep -v autofs | awk '{{print $3}}' | grep {grep_args}"
FIND_WITH_SIZE_CMD = 'find {path} -maxdepth 1 -type f -exec du -b {{}} \\;'
FIND_FILES_CMD = "find {folders} -maxdepth 1 -type f 2>&-"
MINIO_CONFIG_PATH = '/etc/default/minio'


@states_response
def created(path, meta_file, metadata, iso_path, library_path, nodes_heartbeat_path, server_use_no_root_squash=None, **kwargs):
    if 'owner' not in kwargs:
        owner = WORKER_USER
    else:
        owner = kwargs['owner']
    if os.path.exists(path):
        try:
            _clean_meta(path, owner)
        except Exception as e:
            raise CommandExecutionError(f'Can not remove old meta: {str(e)}') from None
        if server_use_no_root_squash:
            try:
                chown_directory(path, owner)
            except Exception as e:
                raise CommandExecutionError(f'An error occurred while changing owner of {path}: {e}') from None
    else:
        try:
            _create_directory(path, server_use_no_root_squash, owner)
        except Exception as e:
            raise CommandExecutionError(str(e)) from None
    try:
        _create_nested_directory(iso_path, owner, server_use_no_root_squash)
        _create_nested_directory(library_path, owner, server_use_no_root_squash)
        _create_nested_directory(nodes_heartbeat_path, owner, server_use_no_root_squash)
        _create_nested_directory(os.path.join(path, PATH_TMP), owner, server_use_no_root_squash)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None

    if os.path.exists(meta_file):
        return f'The datapool with path {path} was successfully created.'
    else:
        try:
            _write_meta(meta_file, metadata, owner)
            return f'The datapool with path {path} was successfully created.'
        except Exception as e:
            raise CommandExecutionError(str(e)) from None


@states_response
def create_default_datapool(**kwargs):
    datapool_id = str(uuid4())
    path = "/storages/local/default"
    data = {'verbose_name': 'Default local datapool ',
            'description': 'Default local datapool ',
            'built_in': True,
            'options': {'path': path},
            'controller_id': None,
            'datapool_id': datapool_id,
            'type': 'local',
            'lun': None,
            'lun_device': None}
    metadata = json.dumps(data)
    fargs = {'path': path,
             'meta_file': os.path.join(path, f'datapool-{datapool_id[:8]}.meta'),
             'metadata': [metadata],
             'iso_path': os.path.join(path, PATH_ISO_STORAGE),
             'library_path': os.path.join(path, PATH_FILE_STORAGE),
             'nodes_heartbeat_path': os.path.join(path, PATH_NODES_HEARTBEAT),
             'owner': WORKER_USER}
    fargs.update(kwargs)
    return created(**fargs)


@states_response
def deleted(path, owner, server_use_no_root_squash=None, **kwargs):
    try:
        _delete_directory(path, owner, server_use_no_root_squash)
        return f'The datapool with path {path} was successfully deleted.'
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while deleting datapool: {e}') from None


@states_response
def write_meta_state(meta_file, metadata, owner, clean=True, prefix="shared_storage"):
    try:
        if clean:
            _clean_meta(os.path.dirname(meta_file), owner, prefix)
        _write_meta(meta_file, metadata, owner)
        return f'The meta_file with path {meta_file} was successfully updated.'
    except Exception as e:
        raise CommandExecutionError(e) from None


def write_meta(meta_file, metadata, owner):
    try:
        _write_meta(meta_file, metadata, owner)
        return f'The meta_file with path {meta_file} was successfully updated.'
    except Exception as e:
        raise CommandExecutionError(e) from None


def _create_directory(path, create_as_root, owner):
    if create_as_root:
        try:
            os.mkdir(path)
            chown_directory(path, owner)
        except OSError as err:
            if err.errno != errno.EEXIST:
                raise
        except Exception as e:
            raise Exception(f'An error occurred while creating directory {path} by root: {e}') from None
    else:
        cmd = "su {} -c 'mkdir -p {}'".format(owner, path)
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise Exception(f"An error occurred while creating directory {path} by {owner}: {cmd_ret['comment']}")


def _write_meta(meta_file, metadata, owner):
    metadata = '\n'.join(metadata)
    if owner == 'root':
        with open(meta_file, 'w') as f:
            f.write(metadata)
    else:
        cmd = ['su', owner, '-c', "echo '{0}' > {1}".format(metadata, meta_file)]
        cmd_ret = call_with_args(cmd)
        if not cmd_ret['result']:
            raise Exception(f"An error occurred while creating meta file {meta_file}: {cmd_ret['comment']}")


def _clean_meta(path, owner, prefix="datapool"):
    meta_files = glob.glob(f'{path}/{prefix}*.meta')
    if meta_files:
        if path.startswith(NFS_PATH):
            cmd = "su {} -c 'rm {}/*.meta'".format(owner, path)
            cmd_ret = get_output(cmd)
            if not cmd_ret['result']:
                raise Exception(cmd_ret['comment'])
        else:
            for meta_file in meta_files:
                os.remove(meta_file)


def _delete_directory(path, owner, delete_as_root):
    if path.startswith(NFS_PATH) and not delete_as_root:
        cmd = "su {} -c 'rm -rf {}'".format(owner, path)
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise Exception(cmd_ret['comment'])
    else:
        try:
            shutil.rmtree(path, ignore_errors=True)
        except Exception as e:
            raise Exception(e) from None


def _create_nested_directory(path, owner, create_as_root):
    if os.path.exists(path):
        if create_as_root:
            try:
                chown_directory(path, owner)
            except Exception as e:
                raise Exception(f'An error occurred while changing owner of {path}: {e}') from None
    else:
        _create_directory(path, create_as_root, owner)


@states_response
def get_files_size(files, **kwargs):
    sizes = []
    for file in files:
        cmd_ret = get_file_size(file, **kwargs)
        if cmd_ret['result']:
            sizes.append(cmd_ret['comment'])
    return sizes


@states_response
def get_file_size(file, **kwargs):
    cmd = GET_SIZE.format(file=file)
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment'].strip()


@states_response
def get_folder_size(folder, **kwargs):
    cmd = GET_FOLDER_SIZE.format(folder=folder)
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment'].strip()


@states_response
def get_free_space(path):
    fsstats = os.statvfs(path)
    blksz = fsstats.f_bsize
    available = fsstats.f_bavail * blksz
    total = fsstats.f_blocks * blksz
    return dict(available=available, total=total)


@states_response
def get_free_space_shell(path, **kwargs):
    """
    Возвращает tuple из доступного и полного места каталога в Mb
    """
    res = states_result(get_output, cmd=f'df -m --output=size,avail {path}', **kwargs)
    try:
        total, available = tuple(res.strip().split()[-2:])
    except Exception as e:
        raise CommandExecutionError(f'Failed to parse output {res}. Error: {e}') from None
    return int(available), int(total)


@states_response
def get_path_usage_percent_shell(path, **kwargs):
    """
    Возвращает int процентов занятого места в каталоге
    """
    res = states_result(get_output, cmd=f'df --output=pcent {path}', **kwargs)
    try:
        pcent = int(res.strip().split()[-1].replace('%', ''))
    except Exception as e:
        raise CommandExecutionError(f'Failed to parse output {res}. Error: {e}') from None
    return pcent


@states_response
def ls_path(storage, **kwargs):
    cmd = LS_CMD.format(storage=storage)
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return [os.path.join(storage, nested.strip()) for nested in cmd_ret['comment'].splitlines() if nested]


@states_response
def check_directory_health(path, **kwargs):
    cmd = CHECK_DIR_HEALTH_CMD.format(path=path)
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment'].strip()


@states_response
def get_file_info(file, **kwargs):
    cmd = FILEINFO_CMD.format(file=file)
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment'].strip()


@states_response
def check_storage_mounted(grep_args, **kwargs):
    cmd = CHECK_SHARED_STORAGE_MOUNTED.format(grep_args=grep_args)
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment'].splitlines()


@states_response
def find_with_size_cmd(path, **kwargs):
    cmd = FIND_WITH_SIZE_CMD.format(path=path)
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment'].splitlines()


@states_response
def find_files(folders: list, **kwargs) -> list:
    files = []
    for folder in folders:
        cmd_ret = get_output(FIND_FILES_CMD.format(folders=folder), **kwargs)
        if cmd_ret['result']:
            files.extend([file for file in cmd_ret['comment'].splitlines() if file])
    return files


@states_response
def ioping(paths: list, cmd: str = None, **kwargs):
    cmd = cmd or "ioping -sync -direct -q -interval 0 -work-time 3 -json"

    def ioping_check(path):
        nonlocal cmd
        io_cmd = f"{cmd} {path}"
        cmd_ret = get_output(io_cmd)
        if cmd_ret['result']:
            comment = json.loads(cmd_ret['comment'])[0]
        else:
            comment = cmd_ret['comment']
        return {'path': path,
                'cmd': io_cmd,
                'result': cmd_ret['result'],
                'comment': comment}
    return thread_pool(ioping_check, paths)


@states_response
def fio(paths=None, remove=True, **kwargs):
    """
    rw can be
    read, write, trim, randread, randwrite, randtrim, rw, readwrite, randrw, trimwrite
    """
    from .vdisk import lvm_create, lvm_full_delete
    if paths is None:
        from .node import discovery_datapool
        storages_dirs = [f"{path}/*" for path in SHARED_STORAGES_PATHS] + [LOCAL_STORAGES_PATH] + \
                        [f"{path}/*" for path in CLUSTER_STORAGES_PATHS] + [f"{ZPOOL_STORAGES_PATH}/*"]
        dps_res = discovery_datapool(storages_dirs=storages_dirs, check_lvm=False)
        if dps_res['result']:
            paths = [dp['path'].replace('meta', 'test') for dp in dps_res['comment']]
        else:
            raise CommandExecutionError('No datapools found.')
    else:
        for path in paths:
            if path.startswith('/dev/') and not path.startswith('/dev/mapper'):
                *_, vgname, vdisk_name = path.split('/')
                lvm_create_res = lvm_create(vdisk_name, vgname, size=kwargs['size'], owner=WORKER_USER)
                if not lvm_create_res['result']:
                    raise CommandExecutionError(lvm_create_res['comment'])
    base_args = {'rw': "randrw",
                 'direct': 1,
                 'buffered': 0,
                 'numjobs': 5,
                 'ioengine': 'libaio',
                 'iodepth': 256,
                 'group_reporting': None,
                 'time_based': None,
                 'name': 'fio_test',
                 "bs": '4K',
                 'runtime': 15,
                 "output-format": 'json'}
    base_args.update(kwargs)
    if kwargs.get('size') is None:
        base_args.update({'size': '10M'})
    base_cmd = 'fio'
    for k, v in base_args.items():
        if v is not None:
            base_cmd = f"{base_cmd} --{k}={v}"
        else:
            base_cmd = f"{base_cmd} --{k}"

    def fio_check(path):
        cmd = f"{base_cmd} --filename={path}"
        cmd_ret = get_output(cmd, timeout=base_args['runtime'] + 15)
        if cmd_ret['result']:
            comment = {}
            res = json.loads(cmd_ret['comment'])['jobs'][0]

            if base_args['rw'] in ['read', 'randread', 'rw', 'readwrite', 'randrw']:
                res_read = res['read']
                comment['iops_read'] = round(res_read['iops'], 1)
                comment['latency_read_ms'] = round(res_read['clat_ns']['mean'] / 1000000, 1)
                comment['speed_read_mib_s'] = round(res_read['bw'] / 1024, 1)
            if base_args['rw'] in ['write', 'randwrite', 'rw', 'readwrite', 'randrw']:
                res_write = res['write']
                comment['iops_write'] = round(res_write['iops'], 1)
                comment['latency_write_ms'] = round(res_write['clat_ns']['mean'] / 1000000, 1)
                comment['speed_write_mib_s'] = round(res_write['bw'] / 1024, 1)
        else:
            comment = cmd_ret['comment']
        if remove:
            if path.startswith('/dev/'):
                *_, vgname, vdisk_name = path.split('/')
                lvm_full_delete_res = lvm_full_delete(False, '', 0, None, WORKER_USER,
                                                      vgname, vdisk_name)
                if not lvm_full_delete_res['result']:
                    raise CommandExecutionError(lvm_create_res['comment'])
            else:
                try:
                    os.remove(path)
                except OSError:
                    pass
        return {'path': path,
                'result': cmd_ret['result'],
                'comment': comment}
    return thread_pool(fio_check, paths)


@states_response
def outside_check(path, **kwargs):
    if not os.path.exists(path):
        raise CommandExecutionError(f'Path: {path} is not exists.')
    elif not os.path.isdir(path):
        raise CommandExecutionError(f'Path: {path} is not directory.')
    else:
        return f'The datapool with path {path} was successfully created.'


@states_response
def get_tree(path, **kwargs):
    from .iso import ISO_TYPES
    from .library import LIBRARY_TYPES, EXTENSION_TYPES, EXTENSION_TEXT_TYPES, additional_flags

    level_deep = kwargs.get('level_deep', 5)
    exts_pattern = "|".join("*" + ext for ext in EXTENSION_TYPES.list())
    cmd = f'tree -x -s -J -f -L {level_deep} --prune --noreport --ignore-case -P "{exts_pattern}" {path}'

    def tree(f, p):
        try:
            if f['type'] == 'file':
                info_ret = get_file_info(f['name'], **kwargs)
                if not info_ret['result'] or \
                        not [_ for _ in LIBRARY_TYPES.list() + ISO_TYPES if _ in info_ret['comment']] or \
                        ((LIBRARY_TYPES.ascii in info_ret['comment'] or
                          LIBRARY_TYPES.unicode in info_ret['comment']) and
                         f'.{f["name"].split(".")[-1].lower()}' not in EXTENSION_TEXT_TYPES.list()):
                    return
                elif f["name"].split(".")[-1].lower() == 'iso' and [_ for _ in ISO_TYPES if _ in info_ret['comment']]:
                    info = {'type': 'iso', 'path': f['name'], 'size': f['size']}
                    p.update(info)
                else:
                    flags = additional_flags(f['name'])
                    info = {'type': f['type'], 'path': f['name'], 'size': f['size'], 'additional_flags': flags}
                    p.update(info)
            elif f['type'] == 'directory':
                info = {'type': f['type'], 'path': f['name'], 'contents': []}
                p.update(info)
                for ff in f['contents']:
                    p['contents'].append({})
                    tree(ff, p['contents'][-1])
                    if not p['contents'][-1]:
                        del p['contents'][-1]
        except Exception:
            return

    ret = get_output(cmd)
    if not ret['result']:
        raise CommandExecutionError(f'An error occurred while get tree filesystem from {path}. Error: {ret["comment"]}')
    try:
        res = json.loads(ret['comment'])
        tree_dict = {}
        tree(res[0], tree_dict)
    except Exception as e:
        raise CommandExecutionError(
            f'An error occurred while get dictionary tree filesystem from json. Error: {e}') from None
    return tree_dict


@states_response
def rename(src, dst):
    if not os.path.exists(src):
        raise CommandExecutionError(f'Path: {src} is not exists.')
    if os.path.exists(dst):
        raise CommandExecutionError(f'Path: {dst} is exists.')
    os.rename(src, dst)
    return f'Successfully renamed {src} to {dst}'


@states_response
def configure_minio(enabled, minio_conf, minio_path, certgen=True, **kwargs):
    if enabled:
        if certgen:
            cmd = "cd /etc/minio; /usr/local/sbin/certgen-linux-amd64 -host $( hostname -I | sed -r \'s/[ ]/,/g\' ) -ca -org-name \"SpaceVM\"; chmod 775 -R /etc/minio/"
            ret = get_output(cmd)
            if not ret['result']:
                raise CommandExecutionError(f'An error occurred while generate certs: {ret["comment"]}')
        with open(MINIO_CONFIG_PATH, 'w') as f:
            f.write("\n".join(minio_conf))
        action, service = 'restart', 'minio'
        cmd_ret = action_service(service, action, timeout=300)
    else:
        action, service = 'stop', 'minio'
        cmd_ret = action_service(service, action, timeout=300)
        get_output(f'rm -f {MINIO_CONFIG_PATH}')
        get_output(f'rm -rf {minio_path}')
    puppet_modify(module='space_node', params={'space_node::minio_ensure': enabled}, apply=False)
    if not cmd_ret['result']:
        raise CommandExecutionError(f"Error in {action} {service}: {cmd_ret['comment']}")
    return f"Successfully {'en' if enabled else 'dis'}abled minio config."
