"""
(c) 2023 LLC «DACOM M»
"""

import json
import logging
import signal
import uuid
from collections import defaultdict
from contextlib import contextmanager
from datetime import datetime as dt
from time import sleep, time
from typing import Any, Union
from xml.etree import ElementTree as ET  # nosec

import os
import re
import shutil
import stat
import tempfile
from base64 import b64decode
from operator import itemgetter

try:
    import ctypes
    from ptrace.debugger import PtraceDebugger
    from ptrace.debugger.process import ptrace_detach

    HAS_PTRACE = True
except (ImportError, NotImplementedError):
    HAS_PTRACE = False

from veilrpc.states.networking import set_vlan_config, set_ovs_mtu
from veilrpc.states.library import LIBRARY_ZIP_TYPES, EXTENSION_ZIP_TYPES
from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output, delete_file, get_output_with_args, \
    DOM_XML_EXT, DOMAINS_FOLDER, ZPOOL_STORAGES_PATH, VDISK_IO_TYPES, clean_guaranteed, PATH_FILE_STORAGE, PATH_TMP
from veilrpc.utils.domain.defaults import DEFAULT_MEMORY_UNIT, USB_CONTROLLER_TYPES, CPU_PERIOD, DEFAULT_MACHINE, \
    DEFAULT_ARCH
from veilrpc.utils.domain.user_exceptions import FileExtentionError, RequiredTagsError
from veilrpc.utils.domain.os_templates import OS_VERSIONS, OS_TYPES
from veilrpc.utils.size import memory_units_convertor
from veilrpc.utils.file import chown_directory
from veilrpc.utils.virt import virt_reboot, virt_reset, virt_resume, virt_suspend, \
    virt_undefine, virt_destroy, virt_is_active, virt_is_paused, virt_is_running, \
    start_from_xml, virt_get_xml, get_domains, screenshot, BASE_QEMU_CHANNEL, \
    virt_get_sched_params, virt_set_sched_params, virt_set_mem_params, virt_memory_stats, virt_memory_params, \
    virt_get_hostname, virt_get_interfaces, STATES, virt_is_pmsuspended, virt_pmwakeup, virt_get_metadata
from veilrpc.utils.domain.constructor import DataOperator, LibvirtDataOperator, XMLDataOperator


LOG_PATH = '/var/log/libvirt/qemu/'
DOMAIN_GA_NODE_PREFIX = 'GA_NODE_DOMAIN:'
DOMAIN_UPTIME_NODE_PREFIX = 'UPTIME_NODE_DOMAIN:'
CONTROLLER_TAG = {'veil': "http://controller/"}
CONTROLLER_PATTERN = 'metadata/veil:instance/veil:{k}'

# Префиксы для детектора зависаний ВМ
GA_FAIL_PREFIX = 'GA_FAIL_START:'
PING_FAIL_PREFIX = 'PING_FAIL_START:'
REBOOT_ATTEMPTS_PREFIX = 'REBOOT_ATTEMPTS:'
HALTPOLL_FAIL_PREFIX = 'HALTPOLL_FAIL_START:'
HALTPOLL_FAIL_VALUE_PREFIX = 'HALTPOLL_FAIL_VALUE:'


def _domain_info(domain_id):
    cmd = f"virsh domstate {domain_id}"
    result = get_output(cmd)
    return result['comment'] if result['result'] else []


def _is_domain_active(domain_id):
    return virt_is_active(domain_id)


def _is_domain_running(domain_id):
    return virt_is_running(domain_id)


def _is_domain_paused(domain_id):
    return virt_is_paused(domain_id)


def _is_domain_pmsuspended(domain_id):
    return virt_is_pmsuspended(domain_id)


@states_response
def check_is_domain_powered(domain_id):
    return virt_is_active(domain_id)


@states_response
def domain_log(domain_id):
    path = f'{LOG_PATH}{domain_id}.log'
    if os.path.exists(path):
        with open(path, 'r') as f:
            return f.read()
    return 'No log found'


def is_domain_xml_exists(domain_id: str) -> bool:
    path_temp = f'{DOMAINS_FOLDER}/{domain_id}'
    if os.path.isfile(f'{path_temp}.xml') or os.path.isfile(f'{path_temp}.{DOM_XML_EXT}'):
        return True
    return False


def xml_path_by_id(domain_id: str) -> str:
    path_temp = f'{DOMAINS_FOLDER}/{domain_id}'
    path = f'{path_temp}.xml'
    if os.path.isfile(path):
        return path
    path = f'{path_temp}.{DOM_XML_EXT}'
    if os.path.isfile(path):
        return path
    raise CommandExecutionError(f'Domains {domain_id} XML file is not found.')


@states_response
# @file_based_lock.lock('/tmp/veil_libivirt_lock')
def start(
        domain_id: str,
        xml: str,
        cpu_priority: int,
        domain_infs_map: list,
        first_boot: bool,
        paused: bool = False,
        **kwargs):
    if _is_domain_active(domain_id):
        if _is_domain_paused(domain_id):
            cmd_ret = resume(domain_id)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            return f"Domain {domain_id} successfully resumed."
        if _is_domain_pmsuspended(domain_id):
            cmd_ret = wakeup(domain_id)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            return f"Domain {domain_id} successfully wakeup."
        if _is_domain_running(domain_id):
            return f"Domain {domain_id} was already running."

    dom = start_from_xml(xml, 1 if paused else 0)
    if not dom:
        comment = "An error occurred while starting domain."
        raise CommandExecutionError(comment)

    for pid in get_pid(domain_id):
        nice_cmd = f"renice -n {cpu_priority} {pid}"
        cmd_nice = get_output(nice_cmd)
        if not cmd_nice['result']:
            comment = f"An error occurred while changing CPU priority for domain: {cmd_nice['comment']}"
            raise CommandExecutionError(comment)

    for entry in domain_infs_map:
        mtu = entry['config'].pop('mtu')
        config = entry['config']
        if config['vlan_mode']:
            set_vlan_config(entry['interface'], **config)
        set_ovs_mtu(entry['interface'], mtu)

    data_operator = CommonDataOperator(domain_id)
    data_operator.create_domain_info(xml)
    data_operator.update_power_state_expected('ON')

    return f"Domain {domain_id} successfully started."


@states_response
def wakeup(domain_name):
    result = virt_pmwakeup(domain_name)
    if not result:
        raise CommandExecutionError(f"Domain {domain_name} was not wake up.")
    return f"Domain {domain_name} successfully wake up."


@states_response
def purge(domain_name):
    cmd = f"virsh purge {domain_name}"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f"Domain {domain_name} successfully purged."


@states_response
def destroy(domain_name):
    result = virt_destroy(domain_name)
    if not result:
        raise CommandExecutionError(f"Domain {domain_name} was not destroyed.")
    return f"Domain {domain_name} successfully destroyed."


@states_response
def undefine(domain_name):
    result = virt_undefine(domain_name)
    if not result:
        raise CommandExecutionError(f"Domain {domain_name} was not undefined.")
    return f"Domain {domain_name} device successfully undefined."


@states_response
def reboot(domain_name):
    result = virt_reboot(domain_name)
    if not result:
        raise CommandExecutionError(f"Domain {domain_name} was not rebooted.")
    return f"Domain {domain_name} successfully rebooted."


@states_response
def reset(domain_name):
    result = virt_reset(domain_name)
    if not result:
        raise CommandExecutionError(f"Domain {domain_name} was not reseted.")
    return f"Domain {domain_name} successfully reseted."


@states_response
def resume(domain_name):
    result = virt_resume(domain_name)
    if not result:
        raise CommandExecutionError(f"Domain {domain_name} was not resumed.")
    return f"Domain {domain_name} successfully resumed."


@states_response
def suspend(domain_name):
    result = virt_suspend(domain_name)
    if not result:
        raise CommandExecutionError(f"Domain {domain_name} was not suspended.")
    return f"Domain {domain_name} successfully suspended."


@states_response
def shutdown(name: str, force: bool):
    if not _is_domain_active(name):
        return "Domain is not active"
    data_operator = CommonDataOperator(name)
    previous_power_state_expected = data_operator.get_field_info(
        data_operator.frequently_used_fields.get("power_state_expected"))
    try:
        data_operator.update_power_state_expected("OFF")
        action = "destroy" if force else "shutdown"
        cmd_ret = get_output(f"virsh {action} {name}", timeout=180)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        if force and _is_domain_active(name):
            raise CommandExecutionError("Domain is still active")
        return f"Domain {name} successfully shutdown."
    except Exception as err:
        if previous_power_state_expected:
            data_operator.update_power_state_expected(previous_power_state_expected)
        raise err


@states_response
def console(domain_name):
    cmd = f"virsh console {domain_name}"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f"Successfully connected to {domain_name}."


def prepare_agent_command(cmd, kwargs=None, indent=None):
    return json.dumps({"execute": f"{cmd}", "arguments": kwargs or {}}, indent=indent)


def parse_agent_response(response):
    lines = response.split("\n")
    for line in lines:
        if not line:
            continue
        try:
            return json.loads(line)["return"]
        except (json.JSONDecodeError, KeyError):
            continue
    raise CommandExecutionError('Unexpected agent response')


def check_agent_response(domain_name, pid, retry=True, retry_timeout=3, retry_attempts=3):
    check_cmd = prepare_agent_command('guest-exec-status', {"pid": pid})
    check_cmd = ["virsh", "qemu-agent-command", domain_name, check_cmd]
    while True:
        retry_attempts -= 1
        check_ret = get_output_with_args(check_cmd)
        if check_ret['result']:
            result = parse_agent_response(check_ret['comment'])
            if result.get('exited'):
                if 'err-data' in result:
                    try:
                        result['err-data'] = b64decode(result['err-data']).decode("utf-8")
                    except Exception:
                        result['err-data'] = b64decode(result['err-data']).decode("cp866")
                if 'out-data' in result:
                    try:
                        result['out-data'] = b64decode(result['out-data']).decode("utf-8")
                    except Exception:
                        result['out-data'] = b64decode(result['out-data']).decode("cp866")
                if result.get('exitcode') != 0:
                    raise CommandExecutionError(f"Error in checking result for pid {pid}:"
                                                f" exitcode {result.get('exitcode')}, "
                                                f"err-data {result.get('err-data')}")
                return result
        else:
            if retry and retry_attempts:
                sleep(retry_timeout)
            else:
                raise CommandExecutionError(
                    f"Error in checking result for pid {pid}: {check_ret['comment']}")


@states_response
def qemu_guest_command(domain_name, command, fargs=None, timeout=5, indent=None, **kwargs):
    if command == 'available_cmds':
        qemu_cmd = 'guest-info'
    else:
        qemu_cmd = command
        if command == 'guest-exec':
            if fargs is None:
                fargs = {"capture-output": True}
            else:
                fargs["capture-output"] = True
    cmd = prepare_agent_command(qemu_cmd, fargs, indent=indent)
    base_cmd = ["virsh", "qemu-agent-command", domain_name, cmd, "--timeout", str(timeout)]
    ret = get_output_with_args(base_cmd, suppress_err=True, timeout=timeout)
    if ret['result']:
        base_result = parse_agent_response(ret['comment'])
        if command == 'available_cmds':
            result = [cmd['name'] for cmd in base_result['supported_commands']]
        elif command == "guest-exec":
            result = check_agent_response(domain_name, base_result['pid'])
        else:
            result = base_result
    else:
        raise CommandExecutionError(f"Error in executing {command}: {ret['comment']}")
    return result


def edit_file(domain_name: str, file_path: str, file_text: str, mode: str = 'w', **kwargs):
    import base64
    fargs_5 = {"path": file_path,
               "mode": mode}
    ret = qemu_guest_command(domain_name, 'guest-file-open', fargs_5, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    handle = ret['comment']

    fargs_6 = {"handle": handle,
               "buf-b64": base64.b64encode(bytes(file_text, 'utf-8')).decode()}
    ret = qemu_guest_command(domain_name, 'guest-file-write', fargs_6, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])

    fargs_7 = {"handle": handle}
    ret = qemu_guest_command(domain_name, 'guest-file-close', fargs_7, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])


@states_response
def linux_to_ad(domain_name, script, packet_manager, **kwargs):
    file_path = '/tmp/to_ad.sh'

    # making file on VM with bash script
    edit_file(domain_name, file_path, script)

    ret = qemu_guest_command(domain_name, 'guest-exec',
                             {"path": packet_manager, "arg": ['-y', 'install', 'dos2unix']},
                             **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])

    ret = qemu_guest_command(domain_name, 'guest-exec',
                             {"path": 'dos2unix', "arg": [file_path]},
                             **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])

    # executing bash script
    ret = qemu_guest_command(domain_name, 'guest-exec',
                             {"path": 'bash', "arg": [file_path]},
                             **kwargs)

    # deleting a bash script, even if it failed
    qemu_guest_command(domain_name, 'guest-exec',
                       {"path": 'rm', "arg": ["-f", file_path]},
                       **kwargs)

    if not ret['result']:
        raise CommandExecutionError(ret['comment'])


@states_response
def is_hostname_same_as_verbose_name(domain_id, domain_name):
    try:
        return domain_name.lower() == virt_get_hostname(domain_id).lower()
    except Exception as e:
        raise CommandExecutionError(f"Error in getting hostname of domain {domain_name} ({domain_id}): {str(e)}")


@states_response
def set_hostname(domain_name, hostname, command, fargs=None, timeout=5, indent=None, **kwargs):
    ret = is_hostname_same_as_verbose_name(domain_id=domain_name, domain_name=hostname)
    if ret['result']:
        if ret['comment']:
            return "New hostname is the same as old. Don't need set new hostname. Skip."
    else:
        raise CommandExecutionError(ret['comment'])
    ret = qemu_guest_command(domain_name, command, fargs=fargs, timeout=timeout, indent=indent, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(f"Error in executing sysprep: {ret['comment']}")
    return ret['comment']


@states_response
def parse_xml(path, powershell):
    with open(path, 'r') as f:
        data = f.read().replace('\n\n', '\n').strip()
        # only for checking valid xml
        ET.fromstring(data)
        if powershell:
            data = data.replace('<', '`<').replace('>', '`>').replace('"', '`"').replace('\n', '`n').replace(' ', '` ')
        else:
            data = data.replace('"', '`"').splitlines()
        if not data:
            raise CommandExecutionError(f"No xml data in path {path}.")
        return data


@states_response
def sysprep(domain_name,
            path="c:\\windows\\system32\\sysprep\\sysprep.exe",
            arg=None,
            timeout=5,
            unattend_src_path=None,
            unattend_dest_path=None,
            powershell=True,
            **kwargs):
    fargs = {"path": path}
    if arg is None:
        arg = ["/generalize", "/quiet", "/quit"]

    if unattend_src_path and unattend_dest_path:
        xml_ret = parse_xml(unattend_src_path, powershell)
        if not xml_ret['result']:
            raise CommandExecutionError(xml_ret['comment'])
        unattend_data = xml_ret['comment']
        unattend_path = f"c:\\{unattend_dest_path}"
        if powershell:
            args = {"path": "powershell.exe",
                    "arg": ["Set-Content",
                            "-Path",
                            unattend_path,
                            "-Value",
                            unattend_data]}
        else:
            # падла cmd.exe за раз пишет тольео одну строку
            args = {"path": "cmd.exe", "arg": ["/c"]}
            for i, dat in enumerate(unattend_data):
                arg = ["echo",
                       dat,
                       ">" if i == 0 else ">>",
                       unattend_path,
                       "|"]
                args['arg'].extend(arg)
            del args['arg'][-1]
        unattend_ret = qemu_guest_command(domain_name=domain_name,
                                          command='guest-exec',
                                          fargs=args,
                                          timeout=timeout,
                                          indent=1)
        if not unattend_ret['result']:
            raise CommandExecutionError(f"Error in creating {unattend_path}: {unattend_ret['comment']}")
        arg.append(f"/unattend:{unattend_path}")
    fargs['arg'] = arg
    ret = qemu_guest_command(domain_name=domain_name, command='guest-exec',
                             fargs=fargs, timeout=timeout)
    if not ret['result']:
        raise CommandExecutionError(f"Error in executing sysprep: {ret['comment']}")
    return ret['comment']


@states_response
def virt_sysprep(vdisk_path, options=None, **kwargs):
    base_cmd = "virt-sysprep"
    cmd = f"{base_cmd} -a {vdisk_path}"
    if options:
        for option in options:
            cmd += f" {option}"
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        comment = f"An error occurred while virt_sysprep domain: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return f"Domain was successfully virt-sysprep: {cmd_ret['comment']}"


@states_response
def virt_customize(vdisk_path, options=None, **kwargs):
    base_cmd = "virt-customize"
    cmd = f"{base_cmd} -a {vdisk_path}"
    if options:
        for option in options:
            cmd += f" {option}"
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        comment = f"An error occurred while virt-customize domain: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return f"Domain was successfully virt-customize: {cmd_ret['comment']}"


@states_response
def ssh_inject(vdisk_path, ssh_user, ssh_key, options: list, **kwargs):
    # keygen_cmd = f'echo "y" | ssh-keygen -b 2048 -t rsa -f {key_path} -q -N ""'
    tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
    with os.fdopen(tmp_fh, 'w') as ssh_key_file:
        ssh_key_file.write(ssh_key)
    options.append(f'--ssh-inject {ssh_user}:file:{tmp_ssh_key_path}')
    cmd_ret = virt_sysprep(vdisk_path=vdisk_path, options=options, **kwargs)
    delete_file(tmp_ssh_key_path)
    if not cmd_ret['result']:
        comment = f"An error occurred while virt_sysprep domain: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return f"Domain was successfully ssh_inject: {cmd_ret['comment']}"


@states_response
def ssh_inject_qemu(domain_name, command, ssh_user, ssh_key, **kwargs):
    # 1 mkdir /home/{ssh_user}/.ssh
    fargs_1 = {"path": "mkdir",
               "arg": ["-p", f"/home/{ssh_user}/.ssh"]}
    ret = qemu_guest_command(domain_name, command, fargs_1, **kwargs)
    if not ret['result']:
        comment = f"An error occurred while mkdir /home/{ssh_user}/.ssh: {ret['comment']}"
        raise CommandExecutionError(comment)
    # 2 chmod 700 /home/{ssh_user}/.ssh
    fargs_2 = {"path": "chmod",
               "arg": ["700", f"/home/{ssh_user}/.ssh"]}
    ret = qemu_guest_command(domain_name, command, fargs_2, **kwargs)
    if not ret['result']:
        comment = f"An error occurred while chmod 700 /home/{ssh_user}/.ssh: {ret['comment']}"
        raise CommandExecutionError(comment)
    # 3 touch /home/{ssh_user}/.ssh/authorized_keys
    fargs_3 = {"path": "touch",
               "arg": [f"/home/{ssh_user}/.ssh/authorized_keys"]}
    ret = qemu_guest_command(domain_name, command, fargs_3, **kwargs)
    if not ret['result']:
        comment = f"An error occurred while touch /home/{ssh_user}/.ssh/authorized_keys: {ret['comment']}"
        raise CommandExecutionError(comment)
    # 4 chmod 600 /home/{ssh_user}/.ssh/authorized_keys
    fargs_4 = {"path": "chmod",
               "arg": ["600", f"/home/{ssh_user}/.ssh/authorized_keys"]}
    ret = qemu_guest_command(domain_name, command, fargs_4, **kwargs)
    if not ret['result']:
        comment = f"An error occurred while chmod 600 /home/{ssh_user}/.ssh/authorized_keys: {ret['comment']}"
        raise CommandExecutionError(comment)
    # 5 guest-file-open /home/{ssh_user}/.ssh/authorized_keys
    fargs_5 = {"path": f"/home/{ssh_user}/.ssh/authorized_keys",
               "mode": "a+"}
    ret = qemu_guest_command(domain_name, 'guest-file-open', fargs_5, **kwargs)
    if not ret['result']:
        comment = f"An error occurred while guest-file-open /home/{ssh_user}/.ssh/authorized_keys: {ret['comment']}"
        raise CommandExecutionError(comment)
    handle = ret['comment']
    ret64 = get_output(f"echo \'{ssh_key}\' | base64 -w 0")
    buf = ret64['comment']
    # 6 guest-file-write /home/{ssh_user}/.ssh/authorized_keys
    fargs_6 = {"handle": handle,
               "buf-b64": buf}
    ret = qemu_guest_command(domain_name, 'guest-file-write', fargs_6, **kwargs)
    if not ret['result']:
        comment = f"An error occurred while guest-file-write /home/{ssh_user}/.ssh/authorized_keys: {ret['comment']}"
        raise CommandExecutionError(comment)
    # 7 guest-file-close /home/{ssh_user}/.ssh/authorized_keys
    fargs_7 = {"handle": handle}
    ret = qemu_guest_command(domain_name, 'guest-file-close', fargs_7, **kwargs)
    if not ret['result']:
        comment = f"An error occurred while guest-file-close /home/{ssh_user}/.ssh/authorized_keys: {ret['comment']}"
        raise CommandExecutionError(comment)
    return f'Successfully inject ssh key to domain {domain_name}'


@states_response
def delete(ids: list):
    """
    Функция безусловной аннигиляции списка ВМ
    """

    if not ids:
        raise CommandExecutionError("Domains list is empty.")
    errors = ''
    for domain_id in ids:
        if _is_domain_active(domain_id):
            cmd_ret = destroy(domain_id)
            if not cmd_ret['result']:
                errors += '\n{}'.format(cmd_ret['comment'])
            if _is_domain_active(domain_id):
                errors += f"\nDomain {domain_id} is still active"
        delete_xml(domain_id)
    if errors:
        raise CommandExecutionError(errors)
    return 'Success in destroying domains.'


@states_response
def update_device(domain_id, xml):
    tmp_file_path = f'/tmp/{uuid.uuid4()}.xml'
    with open(tmp_file_path, 'w') as f:
        f.write(xml)
    virsh_cmd = f"virsh update-device {domain_id} {tmp_file_path}"
    cmd_ret = get_output(virsh_cmd)
    os.remove(tmp_file_path)
    if not cmd_ret['result']:
        comment = f"An error occurred while updating domain device: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return f"Domain {domain_id} device successfully updated."


@states_response
def update_devices(domain_id, xml_list):
    for xml in xml_list:
        res = update_device(domain_id, xml)
        if not res['result']:
            raise CommandExecutionError(res['comment'])
    return f"Domain {domain_id} devices successfully updated."


@states_response
def node_update_graphics_password(domain_id, xml_list, passwd: str):
    update_devices(domain_id, xml_list)
    xml_path = xml_path_by_id(domain_id)
    if os.path.isfile(xml_path):
        ET.register_namespace('veil', "http://controller/")
        ET.register_namespace('qemu', "http://libvirt.org/schemas/domain/qemu/1.0")
        tree = ET.parse(xml_path)
        root = tree.getroot()
        root.find('metadata/veil:instance/veil:graphics_password', {'veil': "http://controller/"}).text = passwd
        for tag in root.iterfind('devices/graphics'):
            tag.attrib['passwd'] = passwd
        ET.indent(root)
        with open(xml_path, 'bw') as f:
            f.write(ET.tostring(root))
        return f"Domain {domain_id} graphics password successfully updated."
    else:
        raise CommandExecutionError(f"Can`t update domains {domain_id} graphics password in XML.")


@states_response
def node_get_graphics_password(domain_id):
    xml_path = xml_path_by_id(domain_id)
    if os.path.isfile(xml_path):
        with open(xml_path, 'r') as f:
            domain_xml = f.read()
    else:
        if _is_domain_active(domain_id):
            res = get_domain_dumpxml(domain_id)
            if res['result']:
                domain_xml = res['comment']
            else:
                raise CommandExecutionError(
                    f"Can`t get domains {domain_id} graphics password. Comment: {res['comment']}")
        else:
            raise CommandExecutionError(
                f"Can`t get domains {domain_id} graphics password.")
    _id, meta_values = get_meta_values(domain_xml, {'graphics_password': '1'})
    return meta_values['graphics_password']


@states_response
def set_permission(path_list, mode):
    for path in path_list:
        os.chmod(path, mode)
    return f"Successfully set mode {mode} to {str(path_list)}."


@states_response
def attach_device(domain_id, xml):
    tmp_file_path = f'/tmp/{uuid.uuid4()}.xml'
    with open(tmp_file_path, 'w+') as f:
        f.write(xml)

    virsh_cmd = f"virsh attach-device --domain {domain_id} --file {tmp_file_path} --live"
    cmd_ret = get_output(virsh_cmd)

    os.remove(tmp_file_path)
    if not cmd_ret['result']:
        comment = f"An error occurred while attaching domain device: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return f"Domain {domain_id} device successfully attached."


@states_response
def attach_devices(domain_id, xml_list):
    result = True
    comment = ''
    for i, xml in enumerate(xml_list):
        cmd_ret = attach_device(domain_id, xml)
        if not cmd_ret['result']:
            comment = f"An error occurred while attaching domain devices: {cmd_ret['comment']}." \
                      f" Checking already attached devices and trying to detach.\n"
            result = False
            attached_devices = xml_list[:i]
            if attached_devices:
                ret_detach = detach_devices(domain_id, attached_devices)
                comment += f"{ret_detach['name']}\n{ret_detach['comment']}"
            else:
                comment += "\nNo attached devices."
            break

    if not result:
        raise CommandExecutionError(comment)
    return f"Domain {domain_id} devices successfully attached."


@states_response
def usb_attach(usb_path, domain_id, xml_list, live=False):
    cmd_ret = set_permission(usb_path, stat.S_IRWXO)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if live:
        cmd_ret = attach_devices(domain_id, xml_list)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def detach_device(domain_id, xml, safe=None):
    disk_source_file = ''
    if safe:
        root = ET.fromstring(xml)
        source = root.find('source')
        if root.get('device') == 'disk' and source is not None:
            disk_source_file = source.get('file')
    tmp_file_path = f'/tmp/{uuid.uuid4()}.xml'
    with open(tmp_file_path, 'w+') as f:
        f.write(xml)
    virsh_cmd = f"virsh detach-device --domain {domain_id} --file {tmp_file_path} --live"
    cmd_ret = get_output(virsh_cmd)
    os.remove(tmp_file_path)
    if not cmd_ret['result']:
        comment = f"An error occurred while detaching domain device: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    if safe and disk_source_file:
        cmd_ret_xml = get_domain_dumpxml(domain_id)
        if not cmd_ret_xml['result']:
            raise CommandExecutionError(cmd_ret_xml['comment'])
        new_root = ET.fromstring(cmd_ret_xml['comment'])
        if new_root.find(f"devices/disk/source/[@file='{disk_source_file}']") is not None:
            raise CommandExecutionError(f"Domain {domain_id} device did not detach.")

    return f"Domain {domain_id} device successfully detached."


@states_response
def detach_devices(domain_id, xml_list):
    for xml in xml_list:
        ret_detach = detach_device(domain_id, xml)
        if not ret_detach['result']:
            comment = f"An error occurred while detaching domain devices: " \
                      f"{ret_detach['comment']}. Power off domain and detach pci devices."
            raise CommandExecutionError(comment)
    return f"Domain {domain_id} devices successfully detached."


@states_response
def detach_spice_channel(domain_id, controller_order):
    cmd_ret = get_domain_dumpxml(domain_id)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    xml = ET.fromstring(cmd_ret['comment'])
    alias = ''
    for redirdev in xml.iter('redirdev'):
        alias = redirdev[0].attrib['name']
        bus = redirdev[1].attrib['bus']
        if int(bus) == controller_order:
            break
    if alias:
        cmd = f"virsh detach-device-alias --domain {domain_id} {alias} --live"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return "Spice usb channel successfully detached."
    else:
        raise CommandExecutionError(f"No spice usb channels on usb controller - {controller_order}")


@states_response
def update_vcpu(domain_id, cpu_count):
    """
    Обновляет кол-во vcpu домена
    """

    virsh_cmd = f"virsh setvcpus {domain_id} {cpu_count} --live"
    cmd_ret = get_output(virsh_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while updating domain vcpu count: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return f"Domain {domain_id} vcpu count successfully updated."


@states_response
def update_vcpu_pinning(domain_id, cpu_map):
    """
    Обновляет привязку vcpu's домена к cpu's ноды.
    """

    # Подготовка. Надо знать актуальную карту привязок, чтобы откатиться, если что.
    actual_topology = {}
    pinned_cpus = []
    vcpupin_cmd = f"virsh vcpupin {domain_id}"
    cmd_ret = get_output(vcpupin_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while getting domain vcpu pinning: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        lines = cmd_ret['comment'].splitlines()[2:-1]
        for line in lines:
            real_vcpu, real_cpuset = line.strip().split()
            actual_cpus = []
            # 0,2-5,3-7
            for cpus in real_cpuset.split(','):
                if '-' not in cpus:
                    actual_cpus.append(cpus)
                else:
                    cpu_start, cpu_end = [int(cpu) for cpu in cpus.split('-')]
                    for i in range(cpu_start, cpu_end + 1):
                        actual_cpus.append(str(i))

            actual_topology[real_vcpu] = ','.join(actual_cpus)

        # Сама привязка domain vcpus -> node cpus
        for vcpu, cpuset in cpu_map.items():
            if cpuset != actual_topology[vcpu]:
                virsh_cmd = "{vcpupin_cmd} --vcpu {vcpu} --cpulist {cpulist} --live".format(vcpupin_cmd=vcpupin_cmd,
                                                                                            vcpu=vcpu,
                                                                                            cpulist=cpuset)
                cmd_pin_ret = get_output(virsh_cmd)
            else:
                continue
            if not cmd_pin_ret['result']:
                # Откат на предыдущую карту в случае ошибки.
                for pin_cpu in pinned_cpus:
                    if pin_cpu in actual_topology:
                        pin_cmd = "{vcpupin_cmd} --vcpu {vcpu} --cpulist {cpulist} --live".\
                            format(vcpupin_cmd=vcpupin_cmd, vcpu=pin_cpu, cpulist=actual_topology[pin_cpu])
                        get_output(pin_cmd)

                comment = "An error occurred while updating domain vcpu ({vcpu}) pinning. " \
                          "for CPU set ({cpuset}). Error: {error}".format(vcpu=vcpu,
                                                                          cpuset=cpuset,
                                                                          error=cmd_ret['comment'])
                raise CommandExecutionError(comment)

            else:
                pinned_cpus.append(vcpu)
    return f"Domain {domain_id} vcpu pinning successfully updated."


@states_response
def update_priority(domain_id, nice):
    """
    Изменяет nice ВМ.
    """

    pids = get_pid(domain_id)
    for pid in pids:
        renice_cmd = f"renice -n {nice} {pid}"
        cmd_ret = get_output(renice_cmd)
        if not cmd_ret['result']:
            comment = f"An error occurred while updating CPU priority for domain: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
    return f"Domain {domain_id} CPU priority was successfully updated."


def get_pid(dom: str) -> list:
    """
    Возвращает pids процесса домена.
    """

    file_name = f"/var/run/libvirt/qemu/{dom}.xml"
    if os.path.exists(file_name):
        cmd = f"grep pid {file_name}"
        try:
            cmd_ret = get_output(cmd)
            if not cmd_ret['result']:
                return []
            output = cmd_ret['comment']
            n = re.compile(r"pid=(\d+)")
            return list(map(int, n.findall(output.replace("'", ''))))
        except Exception:
            return []
    else:
        return []


def get_memory_map(pid, min_mem_block_size) -> tuple:
    """
    Генератор возвращения начального адреса блока памяти и его размера.
    """

    with open(f"/proc/{pid}/maps", "r") as mapping:
        for line in mapping:
            if '/' not in line:
                settings = line.split(' ')
                memory, perms = settings[0], settings[1]
                if perms == 'rw-p':
                    memory_start, memory_end = [int(addr, 16) for addr in memory.split('-')]
                    memory_size = memory_end - memory_start
                    if int(memory_size) >= min_mem_block_size:
                        yield memory_start, memory_size


@states_response
def ptrace(domain_id, clean_type, clean_count, task):
    """
    Очистка оперативной памяти домена.
    """
    if not HAS_PTRACE:
        raise CommandExecutionError('can not import library ptrace')
    data = {'task': task, 'progress': '0', 'parser': 'dd'}

    BLOCKSIZE = 1024 ** 2  # 1Mb Don't change!
    MIN_MEM_BLOCK_SIZE = 128 * (1024 ** 2)  # 128Mb Can up or down
    libwrite_memory = '/usr/lib/libwrite_memory.so'
    os.nice(19)
    percent, mem_map_count = 0, 1
    pids = get_pid(domain_id)
    if not pids:
        return "Not found domain pids."
    else:
        base_pid = pids[0]
    mmap = get_memory_map(base_pid, MIN_MEM_BLOCK_SIZE)
    debugger = PtraceDebugger()
    tracer = debugger.addProcess(pid=base_pid, is_attached=False)
    try:
        # остановка его
        tracer.kill(signal.SIGSTOP)
        # Получение адреса и размера
        for memory_start, memory_size in mmap:
            count = int(memory_size / BLOCKSIZE)
            last_size = memory_size - BLOCKSIZE * count
            # Загрузка библиотеки для записи в память процесса
            lib = ctypes.cdll.LoadLibrary(libwrite_memory)
            with open("/dev/{}".format(clean_type), "rb") as dev:
                line = dev.read(BLOCKSIZE)
            address_txt = ctypes.c_char_p(line)
            pid_t = ctypes.c_int32(base_pid)
            # Цикл по количеству блоков размером BLOCKSIZE
            for _ in range(1, clean_count + 1):
                for i in range(count):
                    if i != count - 1:
                        rw_size = BLOCKSIZE
                    else:
                        if last_size:
                            rw_size = last_size

                    if tracer.readBytes(memory_start + i * BLOCKSIZE, rw_size) != line[:rw_size]:
                        write = True
                    else:
                        write = False

                    if write:
                        mem_st = ctypes.c_ulonglong(memory_start + i * BLOCKSIZE)
                        mem_sz = ctypes.c_ulonglong(rw_size)
                        if lib.clean(address_txt, pid_t, mem_st, mem_sz) > 0:
                            continue
                        else:
                            raise Exception("Can not clean memory map by function process_vm_writev.")
            percent = int(mem_map_count * (100 / clean_count))
            data.update({'progress': str(percent)})
            # TODO __salt__['event.fire_master'](data, tag='salt/progress')

            mem_map_count += 1
    except Exception as e:
        comment = f"An error occurred while tracing domain memory: {e}"
        raise CommandExecutionError(comment) from None
    finally:
        try:
            tracer.kill(signal.SIGCONT)
            # Косяк в библиотеке python_ptrace, через tracer.detach() не отключается от процесса.
            # Поэтому используем ptrace_detach напрямую.
            ptrace_detach(tracer.pid)
            debugger.deleteProcess(process=tracer)
            cmd_ret = destroy(domain_id)
            if not cmd_ret['result']:
                raise Exception()
        except Exception:
            os.kill(base_pid, signal.SIGKILL)
    if percent != 100:
        data.update({'progress': '100'})
        # TODO __salt__['event.fire_master'](data, tag='salt/progress')
    return "Memory was cleaned and domain was destroyed successfully."


@states_response
def get_domain_xml(path):
    """
    Функция делает базовые проверки xml файла, загруженного пользователем,
    и сохраняет xml контент в виде строки в поле ret для последующего анализа и конвертации в структуру,
    на основе которой будет строиться диалог с пользователем по созданию виртуальной машины.
    """

    if not os.path.exists(path):
        comment = f'Source file {path} does not exist.'
        raise CommandExecutionError(comment)
    else:
        try:
            file = path.split('.')
            if len(file) >= 1:
                extension = file[-1].lower()
                if extension != 'xml':
                    raise FileExtentionError('Error file extention.')
            with open(path) as f:
                xml_dump = f.read()
            root = ET.fromstring(xml_dump)
            if root.tag != 'domain':
                raise RequiredTagsError('Tag <domain> is required.')
        except (Exception, ET.ParseError, RequiredTagsError, FileExtentionError) as e:
            raise CommandExecutionError(e) from None
        return xml_dump


@states_response
def get_domain_dumpxml(domain_id, migratable=False):
    if _is_domain_active(domain_id):
        flags = 8 if migratable else 0
        return virt_get_xml(domain_id, flags=flags)
    return _get_local_domain_xml(domain_id)


@states_response
def upload_domain_xml(domain_id):
    cmd_ret = get_domain_dumpxml(domain_id)
    if not cmd_ret['result']:
        comment = f"Can't get domain xml. Error: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    xml_cmd = f'echo \"{cmd_ret["comment"]}\" > /storages/local/default/{PATH_FILE_STORAGE}/{domain_id}.xml'
    cmd_ret = get_output(xml_cmd)
    if not cmd_ret['result']:
        comment = f"Can't upload domain xml. Error: {cmd_ret['comment']}"
    return f"Successfully upload domain {domain_id} xml to /storages/local/default/{PATH_FILE_STORAGE}/{domain_id}.xml"


def _get_local_domain_xml(domain_id):
    with open(xml_path_by_id(domain_id)) as f:
        return f.read()


@states_response
def get_domain_metadata(domain_id):
    return virt_get_metadata(domain_id)


@states_response
def set_xml_power_state_expected(domain_id: str, value: str) -> str:
    """
    Обновление поля power_state_expected ВМ на узле.
    :param domain_id: uuid
    :param value: значение для обновляемого поля
    :return: отчет о удачном завершении
    """
    data_operator = CommonDataOperator(domain_id, libvirt_operator=False)
    data_operator.update_power_state_expected(value)
    return f"Successfully update power_state_expected for domain {domain_id}"


@states_response
def switch_off_ha_enabled(domain_id: str) -> str:
    """
    Обновление поля ha_enabled ВМ на узле.
    :param domain_id: uuid
    :return: отчет о удачном завершении
    """
    data_operator = CommonDataOperator(domain_id, libvirt_operator=False)
    data_operator.update_ha_enabled('False')
    return f"Successfully switch off ha_enabled for domain {domain_id}"


@states_response
def get_xml_power_state_expected(domain_id: str) -> str:
    """
    Получение информации по параметру power_state_expected домена без учета данных в libvirt.
    :param domain_id: id домена.
    :return: состояние поля.
    """
    data_operator = CommonDataOperator(domain_id, libvirt_operator=False)
    return data_operator.get_field_info(data_operator.frequently_used_fields.get("power_state_expected"))


@states_response
def update_metadata(domain_id: str, metadata: str) -> str:
    """
    Обновление группы полей metadata
    :param domain_id: id домена.
    :param metadata: новое значение для все группы.
    :return: отчет о успешном выполнении.
    """
    data_operator = CommonDataOperator(domain_id)
    data_operator.update_metadata(metadata)
    return f"Successfully update metadata for domain {domain_id}"


@states_response
def get_domain_sched_params(domain_id, flags=None):
    return virt_get_sched_params(domain_id, flags)


@states_response
def set_domain_sched_params(domain_id, params, flags=None):
    virt_set_sched_params(domain_id, params, flags)
    return f"Successfully set scheduler params for domain {domain_id}"


@states_response
def set_domain_mem_params(domain_id, params, flags=0):
    virt_set_mem_params(domain_id, params, flags)
    return f"Successfully set memory params for domain {domain_id}"


@states_response
def get_domain_memory_stats(domain_id):
    return virt_memory_stats(domain_id)


@states_response
def get_domain_memory_params(domain_id):
    return virt_memory_params(domain_id)


@states_response
def get_domain_add_info(domain_id, redis):
    result = {}
    if redis.exists(DOMAIN_GA_NODE_PREFIX + domain_id):
        result.update(json.loads(redis.get(DOMAIN_GA_NODE_PREFIX + domain_id)))
    if redis.exists(DOMAIN_UPTIME_NODE_PREFIX + domain_id):
        result.update(json.loads(redis.get(DOMAIN_UPTIME_NODE_PREFIX + domain_id)))
    return result


@states_response
def send_key(domain_id, keycode, codeset="linux", holdtime=None):
    virsh_cmd = f"virsh send-key {domain_id} --codeset {codeset}"
    if holdtime is not None:
        virsh_cmd += f" --holdtime {holdtime}"
    virsh_cmd += f" --keycode {keycode}"
    cmd_ret = get_output(virsh_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while send_key {keycode} to domain {domain_id}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return f"Successfully send_key {keycode} to domain {domain_id}"


@states_response
def get_domain_hostname(domain_id, flags=0):
    return virt_get_hostname(domain_id, flags=flags)


@states_response
def get_domain_interfaces(domain_id):
    return virt_get_interfaces(domain_id)


@states_response
def get_spice_port(domain_id):
    res = get_domain_dumpxml(domain_id)
    if res['result']:
        xml_dump = res['comment']
        root = ET.fromstring(xml_dump)
        for tag in root.iterfind('devices/graphics'):
            if tag.attrib.get('type', '') == 'spice' and 'port' in tag.attrib:
                return tag.attrib['port']
    else:
        raise CommandExecutionError(res['comment'])
    raise CommandExecutionError(f"Not found spice port for domain {domain_id}")


@states_response
def get_vnc_port(domain_id):
    res = get_domain_dumpxml(domain_id)
    if res['result']:
        xml_dump = res['comment']
        root = ET.fromstring(xml_dump)
        for tag in root.iterfind('devices/graphics'):
            if tag.attrib.get('type', '') == 'vnc' and 'port' in tag.attrib:
                return tag.attrib['port']
    else:
        raise CommandExecutionError(res['comment'])
    raise CommandExecutionError(f"Not found vnc port for domain {domain_id}. "
                                f"Maybe need force restart or shutdown and start domain.")


@states_response
def update_resources(**kwargs):
    domain_id = kwargs['id']
    comment = 'success.'
    if 'cpu_count' in kwargs.keys():
        cmd_ret = update_vcpu(domain_id, kwargs['cpu_count'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    if 'memory_xml' in kwargs.keys():
        cmd_ret = attach_device(domain_id, kwargs['memory_xml'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    if 'cpu_map' in kwargs.keys():
        cmd_ret = update_vcpu_pinning(domain_id, kwargs['cpu_map'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    if 'cpu_priority' in kwargs.keys():
        cmd_ret = update_priority(domain_id, kwargs['cpu_priority'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    if 'cpu_shares' in kwargs.keys():
        cmd_ret = set_domain_sched_params(domain_id,
                                          params={'cpu_shares': kwargs['cpu_shares']})
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    if 'cpu_period' in kwargs.keys():
        cmd_ret = set_domain_sched_params(domain_id,
                                          params={'vcpu_quota': kwargs['cpu_period']})
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    if 'memory_limit' in kwargs.keys():
        cmd_ret = set_domain_mem_params(domain_id,
                                        params={'soft_limit': kwargs['memory_limit']})
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    if 'metadata' in kwargs.keys():
        cmd_ret = update_metadata(domain_id, kwargs['metadata'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    return comment


@states_response
def full_shutdown(**kwargs):
    domain_id = kwargs['domain_id']
    data_operator = CommonDataOperator(domain_id, libvirt_operator=False)
    if not _is_domain_running(domain_id) and not _is_domain_paused(domain_id):
        data_operator.update_power_state_expected('OFF')
        return f"Domain {domain_id} was already turned off."
    if kwargs['safety']:
        cmd_ret = ptrace(domain_id, kwargs['clean_type'], kwargs['clean_count'], kwargs['task'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        data_operator.update_power_state_expected('OFF')
        return cmd_ret['comment']
    else:
        cmd_ret = shutdown(domain_id, kwargs['force'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        data_operator.update_power_state_expected('OFF')
        return cmd_ret['comment']


@states_response
def full_reboot(**kwargs):
    domain_id = kwargs['id']
    comment = ''
    if kwargs['safety']:
        cmd_ret = ptrace(domain_id, kwargs['clean_type'], kwargs['clean_count'], kwargs['task'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
        cmd_ret = start(domain_id, kwargs['xml'], kwargs['cpu_priority'],
                        kwargs['domain_infs_map'], kwargs['first_boot'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    else:
        if kwargs['force']:
            shutdown_ret = shutdown(name=domain_id, force=True)
            if not shutdown_ret['result']:
                raise CommandExecutionError(shutdown_ret['comment'])
            comment += shutdown_ret['comment']
            cmd_ret = start(domain_id, kwargs['xml'], kwargs['cpu_priority'],
                            kwargs['domain_infs_map'], kwargs['first_boot'])
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            comment += cmd_ret['comment']
        else:
            cmd_ret = reboot(domain_id)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            comment += cmd_ret['comment']
    return comment


@states_response
def full_delete(**kwargs):
    domain_id = kwargs['id']
    uefi_file = kwargs['uefi_boot_vars']
    clean_list = []
    comment = ''
    if kwargs['safety']:
        cmd_ret = ptrace(domain_id, kwargs['clean_type'], kwargs['clean_count'], kwargs['task'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

        if kwargs['snapshots']:
            for source in kwargs['remove_sources']:
                clean_list.append(source)
        if kwargs['cloud_iso']:
            for cloud_iso in kwargs['cloud_iso']:
                clean_list.append(cloud_iso)
        clean_list.append(uefi_file)

        cmd_ret = clean_guaranteed(sources=clean_list, clean_type=kwargs['clean_type'],
                                   clean_count=kwargs['clean_count'], task=kwargs['task'],
                                   redis=kwargs.get('redis', None))
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    else:
        cmd_ret = shutdown(domain_id, force=True)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

        if kwargs['snapshots']:
            from .snapshot import remove_snapshots
            cmd_ret = remove_snapshots(kwargs['snapshots'], kwargs['user'])
            comment += cmd_ret['comment']
        if kwargs['cloud_iso']:
            # pylint: disable=redefined-outer-name
            from .iso import delete
            for cloud_iso in kwargs['cloud_iso']:
                cmd_ret = delete(iso_path=cloud_iso['source'], owner=kwargs['user'])
                comment += cmd_ret['comment']
        delete_file(uefi_file)
    delete_xml(domain_id)
    return comment


@states_response
def gen_cloud_init_iso(iso_name, iso_path, user_data, meta_data):
    """Создает iso образ cloud_init."""
    from .directory_datapool import get_file_size
    tmp_file_path = f'/tmp/{uuid.uuid4()}'
    os.mkdir(tmp_file_path)
    tmp_user_data_path = f'{tmp_file_path}/user-data'
    with open(tmp_user_data_path, 'w') as f:
        f.write(user_data)
    tmp_meta_data_path = f'{tmp_file_path}/meta-data'
    with open(tmp_meta_data_path, 'w') as f:
        f.write(meta_data)
    iso_path = f'{iso_path}/{iso_name}'
    cmd = f"genisoimage -output {iso_path} -volid cidata -joliet -rock {tmp_user_data_path} {tmp_meta_data_path}"
    cmd_ret = get_output(cmd)
    shutil.rmtree(tmp_file_path, ignore_errors=False)
    if not cmd_ret['result']:
        comment = f"Can't gen_cloud_init_image. Error: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    info_ret = get_file_size(iso_path)
    if not info_ret['result']:
        return 0
    return info_ret['comment']


@states_response
def xml_to_dict(xml_str):
    """
    XML (str) to dict magic
    Просто раскладывает по ключам:
        аттрибуты в attrib,
        текстовое значение в text_value,
        список детишек в children (даже если ребенок один)
    """

    def rec(node):
        node_dict = dict()
        node_dict[node.tag] = dict(attrib=node.attrib,
                                   text_value=node.text.strip() if isinstance(node.text, str) else '',
                                   children=list())
        for child in node:
            node_dict[node.tag]['children'].append(rec(child))
        return node_dict

    try:
        root = ET.fromstring(xml_str)
    except Exception:
        return {}

    return rec(root)


@states_response
def xml_to_dict2(xml_str):
    """
    XML (str) to dict magic
    Упрощенная структура, без суррогатных ключей типа attrib и children.
    Аттрибуты и вложенные теги (детишки) находятся на одном уровне для каждого тега.
    Только повторяющиеся теги (детишки) собираются в список.
    По идее такой структурой будет удобней пользоваться - меньше итераций по спискам одинаковых тег.
    Один нюанс: если название аттрибута и вложенного тега совпадут, то название тега
    будет изменено на <имя_тега>_tag_<порядковый_номер> что бы не превратить значение аттрибута в
    список значений.
    """

    def rec(node):
        tag_dict = dict()
        node_text = node.text.strip() if isinstance(node.text, str) else ''
        if node_text:
            try:
                tag_dict['_text_'] = int(node_text)
            except ValueError:
                if node_text == 'True':
                    tag_dict['_text_'] = True
                elif node_text == 'False':
                    tag_dict['_text_'] = False
                else:
                    tag_dict['_text_'] = node_text
        if node.attrib:
            for attr, value in node.attrib.items():
                try:
                    tag_dict[attr] = int(value)
                except ValueError:
                    if value == 'True':
                        tag_dict[attr] = True
                    elif value == 'False':
                        tag_dict[attr] = False
                    else:
                        tag_dict[attr] = value
        num = 1
        for child in node:
            try:
                tag_dict[child.tag].append(rec(child))
            except KeyError:
                tag_dict[child.tag] = rec(child)
            except AttributeError:
                if isinstance(tag_dict[child.tag], dict):
                    tag_dict[child.tag] = [tag_dict[child.tag], rec(child)]
                else:
                    tag_dict[child.tag + '_tag_{}'.format(num)] = rec(child)
                    num += 1
        return tag_dict

    try:
        one_line_xml_str = ' '.join(xml_str.split())
        root = ET.fromstring(one_line_xml_str)
    except Exception:
        return {}

    return {root.tag: rec(root)}


@states_response
def backup(domain_id, filename, xml, datapool, vdisks, isos, snapshots, memory, compress_type, user, task, **kwargs):
    """
    Функция сохраняет в tar архив XML файл подготовленный для libvirt, имя файла uuid ВМ + текущее время,
    образы дисков взятые из оригинальных мест и их снимки, эти файлы должны лежать в корне архива.
    Возвращается словарь, для завершения создания объекта "файл".
    """
    # pylint: disable = reimported, redefined-outer-name
    import shutil
    from veilrpc.utils.file import chown_directory
    from .copy import popen_with_progress

    redis = kwargs.get('redis', None)
    zfs = kwargs.get('zfs')
    nvram = kwargs.get('nvram', [])
    powered = kwargs.get('powered')
    progress_data = dict(type='progress', jid=task, id='', name='backup', result=True)
    finish_backup_path = f'{datapool}/{PATH_FILE_STORAGE}/{filename}'
    backup_path = f'{datapool}/{PATH_TMP}/{filename}'
    # tmp_dir = f'{datapool}/{PATH_FILE_STORAGE}/{domain_id}'
    tmp_dir = f'{datapool}/{PATH_TMP}/{domain_id}'
    timestamp = int(dt.timestamp(dt.strptime(re.search('.*-(\\d{4}\\.\\d{2}\\.\\d{2}-\\d{2}\\.\\d{2}\\.\\d{2}).*',
                                                       filename).group(1), '%Y.%m.%d-%H.%M.%S')))
    tmp_file_path = f'{tmp_dir}/{timestamp}.{domain_id}.xml'
    os.makedirs(f'{tmp_dir}/xml', exist_ok=True)
    with open(tmp_file_path, 'w') as f:
        f.write(xml)
    if powered and zfs and vdisks:
        for i, vdisk in enumerate(vdisks):
            if f'{ZPOOL_STORAGES_PATH}/' in vdisk:
                pool, dataset = tuple(vdisk.split(f'{ZPOOL_STORAGES_PATH}/')[-1].split('/vdisk.qcow2')[0].split('/'))
                get_snapshots_disk = get_output(f'sudo zfs list -Ht snapshot -o name -s creation -r {pool}/{dataset}')
                if not get_snapshots_disk['result'] or not get_snapshots_disk['comment'].strip().split():
                    raise CommandExecutionError(
                        f'Could not get list snapshots from dataset {pool}/{dataset}!. '
                        f'Error: {get_snapshots_disk["comment"]}')
                else:
                    fist_snapshot = get_snapshots_disk['comment'].strip().split()[0]
                # делаем временный клон от первого снапшота
                cmd_ret = get_output(f'sudo zfs clone {fist_snapshot} {pool}/{dataset}_tmp_clone')
                if not cmd_ret['result']:
                    raise CommandExecutionError(
                        f'An error occurred while creating clone {pool}/{dataset}_tmp_clone: {cmd_ret["comment"]}')
                vdisks[i] = f'{ZPOOL_STORAGES_PATH}/{pool}/{dataset}_tmp_clone/vdisk.qcow2'

    files = ' '.join(vdisks + [f'\"{i}\"' for i in isos] + snapshots + memory + nvram) + f' {tmp_file_path}'
    total_size_cmd = f'set -o pipefail; du -cb {files} | tail -n1 |cut -f1'
    total_size_ret = get_output(total_size_cmd)
    if not total_size_ret['result'] or not total_size_ret['comment'].strip().isdigit():
        shutil.rmtree(tmp_dir, ignore_errors=True)
        raise CommandExecutionError(f'Failed to get total size files for backup: {files}. '
                                    f'Error: {total_size_ret["comment"]}')
    compress_cmd = ''
    if compress_type is not None and compress_type == LIBRARY_ZIP_TYPES.zstd.split()[0]:
        compress_cmd = '2>&3 | /usr/local/sbin/veil-zstd '
    elif compress_type is not None:
        raise CommandExecutionError(f'Unsupported format {compress_type}.')
    cmd = f'set -o pipefail; exec 3>&1; '\
          f'tar cf - --transform \"s:.*/zfs/\\(.*\\)_tmp_clone/vdisk.qcow2:\\1.qcow2:;' \
          f's:.*/zfs/\\(.*\\)/vdisk.qcow2:\\1.qcow2:;s:.*/::\" -P {files} | '\
          f'pv -ns {total_size_ret["comment"].strip()} {compress_cmd}> {backup_path} {"" if compress_cmd else "2>&3"}'
    p = popen_with_progress(cmd, "pv", progress_data, redis)
    if p.returncode:
        shutil.rmtree(tmp_dir, ignore_errors=True)
        raise CommandExecutionError(f'Failed to create backup {filename}. Error: {p.stderr.read()}')
    try:
        shutil.move(backup_path, finish_backup_path)
    except Exception as e:
        raise CommandExecutionError(f'Failed to move backup to {finish_backup_path}. Error: {e}')
    shutil.rmtree(tmp_dir, ignore_errors=True)
    if powered and zfs and vdisks:
        for i, vdisk in enumerate(vdisks):
            if f'{ZPOOL_STORAGES_PATH}/' in vdisk:
                pool, dataset = tuple(vdisk.split(f'{ZPOOL_STORAGES_PATH}/')[-1].split('/vdisk.qcow2')[0].split('/'))
                # удаляем временный клон
                cmd_ret = get_output(f'sudo zfs destroy -rR {pool}/{dataset}')
                if not cmd_ret['result']:
                    raise CommandExecutionError(
                        f'An error occurred while deleting dataset {pool}/{dataset}: {cmd_ret["comment"]}')
    try:
        chown_directory(finish_backup_path, user)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None

    return {'path': finish_backup_path,
            'size': os.path.getsize(finish_backup_path),
            'additional_flags': {'can_be_unpacked': True,
                                 'can_be_domained': True}}


@states_response
def increment_backup(domain_id, backup, datapool, xml, vdisks, isos, snapshots, memory, compress_type, user, task,
                     **kwargs):  # pylint: disable=redefined-outer-name
    # pylint: disable = redefined-outer-name, reimported
    import shutil
    from .copy import popen_with_progress

    files = vdisks + isos + snapshots + memory
    redis = kwargs.get('redis', None)
    progress_data = dict(type='progress', jid=task, id='', name='automated_restore', result=True, steps={})
    progress_data['steps']['total_steps'] = len(files) + 1 if compress_type else len(files)
    progress_data['steps']['step'] = 0

    if not os.path.isfile(backup):
        raise CommandExecutionError(f'File {backup} not found.')
    # если резервная копия была сжата, то распаковываем(разжимаем?)
    decompress_cmd = f'set -o pipefail; exec 3>&1; pv -n {backup} 2>&3 | '
    ret_type = get_output(f'file {backup}')
    if not ret_type['result']:
        raise CommandExecutionError(f'Failed to get type from backup "{backup}". Error: {ret_type["comment"]}')
    new_decompress_backup = tmp_new_decompress_backup = None
    if LIBRARY_ZIP_TYPES.zstd in ret_type['comment']:
        new_decompress_backup = '.'.join(backup.split('.')[:-1])
        tmp_new_decompress_backup = new_decompress_backup.replace(PATH_FILE_STORAGE, PATH_TMP)
        decompress_cmd += f'zstd --rm -d -c > "{tmp_new_decompress_backup}"'
    # если передан тип сжатия, то сжимаем резервную копию.
    # если резервная копия была сжата но тип сжатия не передан, то оставляем в несжатом состоянии
    compress_cmd = None
    new_compress_backup = tmp_new_compress_backup = None
    if compress_type is not None and compress_type == LIBRARY_ZIP_TYPES.zstd.split()[0]:
        new_compress_backup = f'{new_decompress_backup if new_decompress_backup else backup}{EXTENSION_ZIP_TYPES.zst}'
        tmp_new_compress_backup = new_compress_backup.replace(PATH_FILE_STORAGE, PATH_TMP)
        compress_cmd = f'set -o pipefail; exec 3>&1; ' \
                       f'pv -n {new_decompress_backup if new_decompress_backup else backup} 2>&3 ' \
                       f'| /usr/local/sbin/veil-zstd > "{tmp_new_compress_backup}"'
    elif compress_type is not None:
        raise CommandExecutionError(f'Unsupported format {compress_type}.')
    # запуск распаковки(разжатия?)
    if new_decompress_backup:
        progress_data['steps']['total_steps'] += 1
        progress_data['steps']['step'] += 1
        p = popen_with_progress(decompress_cmd, "pv", progress_data, redis)
        if p.returncode:
            if os.path.exists(new_decompress_backup):
                try:
                    os.remove(new_decompress_backup)
                except OSError:
                    pass
            raise CommandExecutionError(
                f'Failed to decompress backup from {backup}. Error: {p.stderr.read()}')
        if os.path.exists(backup):
            try:
                os.remove(backup)
            except OSError:
                pass
        try:
            chown_directory(new_decompress_backup, user)
        except Exception as e:
            raise CommandExecutionError(str(e)) from None
        try:
            shutil.move(tmp_new_decompress_backup, new_decompress_backup)
        except Exception as e:
            raise CommandExecutionError(f'Failed to move backup to {new_decompress_backup}. Error: {e}')
        backup = new_decompress_backup
    # добавление файлов в резервную копию
    for f in files:
        cmd = f'set -o pipefail; exec 3>&1; pv -n {f} 2>&3 | veil-tarappend -t {backup} -f {f.split("/")[-1]}'
        progress_data['steps']['step'] += 1
        p = popen_with_progress(cmd, "pv", progress_data, redis)
        if p.returncode:
            raise CommandExecutionError(
                f'Failed to append file {f} to backup "{backup}". Error: {p.stderr.read()}')
    # обновление конфигурации ВМ
    tmp_dir = f'{datapool}/{PATH_TMP}/{domain_id}'
    timestamp = int(dt.timestamp(dt.strptime(re.search('.*-(\\d{4}\\.\\d{2}\\.\\d{2}-\\d{2}\\.\\d{2}\\.\\d{2}).*',
                                                       kwargs['filename']).group(1), '%Y.%m.%d-%H.%M.%S')))
    tmp_file_path = f'{tmp_dir}/{timestamp}.{domain_id}.xml'
    os.makedirs(tmp_dir, exist_ok=True)
    with open(tmp_file_path, 'w') as f:
        f.write(xml)
    ret_append_xml = get_output(f'set -o pipefail; exec 3>&1; pv -n {tmp_file_path} 2>&3 | '
                                f'veil-tarappend -t {backup} -f {timestamp}.{domain_id}.xml')
    if not ret_append_xml['result']:
        shutil.rmtree(tmp_dir, ignore_errors=True)
        raise CommandExecutionError(f'Failed to append libvirtXML {domain_id}.xml to backup {backup}. '
                                    f'Error: {ret_append_xml["comment"]}')
    shutil.rmtree(tmp_dir, ignore_errors=True)
    # запуск процесса сжатия
    if compress_cmd:
        progress_data['steps']['step'] += 1
        p = popen_with_progress(compress_cmd, "pv", progress_data, redis)
        if p.returncode:
            if os.path.exists(new_compress_backup):
                try:
                    os.remove(new_compress_backup)
                except OSError:
                    pass
            raise CommandExecutionError(f'Failed to compress backup to {new_compress_backup}. Error: {p.stderr.read()}')
        if os.path.exists(new_decompress_backup if new_decompress_backup else backup):
            try:
                os.remove(new_decompress_backup if new_decompress_backup else backup)
            except OSError:
                pass
        try:
            chown_directory(new_compress_backup, user)
        except Exception as e:
            raise CommandExecutionError(str(e)) from None
        try:
            shutil.move(tmp_new_compress_backup, new_compress_backup)
        except Exception as e:
            raise CommandExecutionError(f'Failed to move backup to {new_compress_backup}. Error: {e}')
        backup = new_compress_backup
    # переименовываем файл бакапа с актуальной датой
    new_backup = '/'.join(backup.split('/')[:-1] + [kwargs['filename']])
    try:
        os.rename(backup, new_backup)
    except Exception as e:
        raise CommandExecutionError(f'Faild rename backup file {backup} to {new_backup}. Error: {e}') from None

    return {'path': new_backup,
            'size': os.path.getsize(new_backup)}


@contextmanager
def ratamount(m_src, task=None, **kwargs):
    from .copy import popen_with_progress

    redis = kwargs.get('redis', None)
    progress_data = kwargs.get('progress_data', dict(type='progress', jid=task, id='', name='ratamount', result=True))
    directory = False
    if os.path.isfile(m_src):
        temp_dir = tempfile.mkdtemp()
        mount_cmd = f'/usr/local/sbin/veil-ratarmount "{m_src}" {temp_dir} -o allow_other,umask=022'
        if 'steps' in progress_data and 'step' in progress_data['steps']:
            progress_data['steps']['step'] += 1
        p = popen_with_progress(mount_cmd, "pv", progress_data, redis)
        if p.returncode:
            raise CommandExecutionError(f'Failed to mount backup "{m_src}". Error: {p.stderr.read()}')
    elif os.path.isdir(m_src):
        directory = True
        temp_dir = m_src
    else:
        raise CommandExecutionError(f'Failed path "{m_src}".')
    yield temp_dir
    if not directory:
        umount_cmd = f'fusermount -u {temp_dir}'
        umount_cmd_ret = get_output(umount_cmd)
        if not umount_cmd_ret['result']:
            raise CommandExecutionError(f'Failed to umount backup file "{m_src}". '
                                        f'Error: {umount_cmd_ret["comment"]}')
        shutil.rmtree(temp_dir, ignore_errors=True)
        try:
            os.remove(f'{m_src}.index.sqlite')
        except OSError:
            pass
        try:
            os.remove(f'/root/.ratarmount/{m_src.replace("/", "_")}.index.sqlite')
        except OSError:
            pass


def get_info_from_backup(path, **kwargs):
    """
    Функция возвращает информацию о файлах внутри архива.
    """
    # pylint: disable = reimported, redefined-outer-name
    import json

    redis = kwargs.get('redis', None)
    task = kwargs.get('task', None)
    progress_data = kwargs.get('progress_data',
                               dict(type='progress', jid=task, id='', name='get_info_from_backup', result=True))

    directory = kwargs.get('directory', False)
    ovf_files = kwargs.get('ovf_files')
    msg = ''
    issnapshots = False
    list_qemu_info = []
    memories = {}
    root = None
    with ratamount(os.path.dirname(path) if directory else path,
                   task, redis=redis, progress_data=progress_data) as temp_dir:
        for f in sorted(os.listdir(temp_dir)):
            if directory and ovf_files is not None and f not in ovf_files:
                continue
            if '.memory' in f:
                ret_xml = get_output(f"dd if={temp_dir}/{f} bs=32k count=1 status=none", universal_newlines=False)
                if not ret_xml['result']:
                    raise CommandExecutionError(f'Failed to get memory xml  from backup "{path}". '
                                                f"Error: {ret_xml['comment']}")
                xml = f"<domain " \
                      f"{ret_xml['comment'].decode(errors='ignore').split('</domain>')[0].split('<domain')[-1]}" \
                      f" </domain>\n"
                memories[f.split('.')[0]] = xml
            if '.snapshot' in f:
                issnapshots = True
            if '.xml' in f:
                with open(f'{temp_dir}/{f}') as xml_f:
                    xml = xml_f.read()
                try:
                    root = ET.fromstring(xml)
                except Exception as e:
                    raise CommandExecutionError(f'Failed to parse xml from backup "{path}". Error: {e}') from None
                if not root.find('metadata/veil:instance', {'veil': "http://controller/"}):
                    return {}
            if os.path.isdir(f'{temp_dir}/{f}'):
                continue
            ret = get_output(f'qemu-img info --output=json "{temp_dir}/{f}"')
            if not ret['result']:
                msg += f'Failed to get info about "{temp_dir}/{f}". Error: {ret["comment"]}'
            ret_dict = json.loads(ret["comment"])
            list_qemu_info.append(ret_dict)
    files_map = {}
    vdisks_size = {}
    isos_size = {}
    map_snapshots = {}
    nvram = ''
    for f in list_qemu_info:
        if 'format' in f and f['format'] in ['vmdk', 'qcow2']:
            if '.snapshot' not in f['filename']:
                vdisks_size.update({'.'.join(f['filename'].split('/')[-1].split('.')[:-1]): f['virtual-size']})
                if ovf_files and f['filename'].split('/')[-1] in ovf_files:
                    files_map.update(
                        {f['filename'].split('/')[-1]: '.'.join(f['filename'].split('/')[-1].split('.')[:-1])})
            if issnapshots and f['format'] == 'qcow2':
                map_snapshots.update(
                    {f['filename'].split('/')[-1]: f['backing-filename'].split(
                        '/')[-1] if 'backing-filename' in f else None})
        elif 'format' in f and f['format'] == 'raw' and '_VARS.fd' in f['filename']:
            nvram = f['filename'].split('/')[-1]
        elif 'format' in f and f['format'] == 'raw' and f['filename'].split('.')[-1].lower() in ['raw', 'img', 'bin']:
            vdisks_size.update({'.'.join(f['filename'].split('/')[-1].split('.')[:-1]): f['virtual-size']})
            if ovf_files and f['filename'].split('/')[-1] in ovf_files:
                files_map.update({f['filename'].split('/')[-1]: '.'.join(f['filename'].split('/')[-1].split('.')[:-1])})
        elif 'format' in f and f['format'] == 'raw' and f['filename'].split('.')[-1].lower() == 'iso':
            isos_size.update({f['filename'].split('/')[-1]: f['virtual-size']})
            if ovf_files and f['filename'].split('/')[-1] in ovf_files:
                files_map.update({f['filename'].split('/')[-1]: f['filename'].split('/')[-1]})

    result = {'vdisks_size': vdisks_size,
              'isos_size': isos_size,
              'memories': memories,
              'map_snapshots': map_snapshots,
              'files_map': files_map,
              'nvram': nvram}
    if root:
        result.update({'xml': ET.tostring(root, encoding='UTF-8').decode()})
    return result


def check_backup(path, ext):
    check_cmd = f'tar -tf "{path}" 2>/dev/null | grep ".{ext}$" || echo "is not backup"'
    check_ret = get_output(check_cmd)
    if not check_ret['result']:
        raise CommandExecutionError(f'Failed to check backup file {path}. '
                                    f'Error: {check_ret["comment"]}')
    check = check_ret['comment'].strip()
    if len(check.split('\n')) == 1 and check != 'is not backup':
        return check
    elif len(check.split('\n')) > 1 and ext == 'xml':
        return sorted(check.split('\n'))[-1]
    else:
        return None


@states_response
def get_domain_from_backup(path, **kwargs):
    """
    Функция возвращает XML с конфикурацией ВМ и виртуальные размеры дисков и другие параметры о бэкапе.
    """
    # pylint: disable=reimported, redefined-outer-name
    from .library import EXTENSION_ZIP_TYPES

    try:
        os.path.exists(path=path)
    except Exception as e:
        raise CommandExecutionError(f'Path "{path}" don\'t exists. Error: {e}') from None
    redis = kwargs.get('redis', None)
    task = kwargs.get('task', None)
    progress_data = kwargs.get('progress_data',
                               dict(type='progress', jid=task, id='', name='get_domain_from_backup', result=True))
    ext_file = path.split('.')[-1]
    if ext_file.lower() == 'tar' or f'.{ext_file}' in EXTENSION_ZIP_TYPES.list():
        if f'.{ext_file}' in EXTENSION_ZIP_TYPES.list():
            ext = path.split('.')[-2]
            if ext != 'tar':
                return {}
        else:
            xml_file = check_backup(path, 'xml')
            if not xml_file:
                return {}
            xml_cmd = f'tar -xOf "{path}" --wildcards "{xml_file}"'
            xml_cmd_ret = get_output(xml_cmd)
            if not xml_cmd_ret['result']:
                raise CommandExecutionError(f'Failed to get xml from backup "{path}". '
                                            f'Error: {xml_cmd_ret["comment"]}')
            try:
                root = ET.fromstring(xml_cmd_ret['comment'])
            except Exception as e:
                raise CommandExecutionError(f'Failed to parse xml from backup "{path}". Error: {e}') from None
            if not root.find('metadata/veil:instance', {'veil': "http://controller/"}):
                return {}
        backup_info = get_info_from_backup(path, task=task, redis=redis, progress_data=progress_data)
        backup_info.update({'is_xml': False, 'is_ova': False, 'is_ovf': False, 'ovf_version': None})
    elif ext_file.lower() == 'xml':
        with open(path, 'r') as f:
            xml = f.read()
        try:
            root = ET.fromstring(xml)
        except Exception as e:
            raise CommandExecutionError(f'Failed to parse xml from "{path}". Error: {e}') from None
        if not root.find('metadata/veil:instance', {'veil': "http://controller/"}):
            return {}
        backup_info = {'xml': ET.tostring(root, encoding='UTF-8').decode(), 'is_xml': True, 'is_ova': False,
                       'is_ovf': False, 'vdisks_size': None, 'isos_size': None, 'memories': None,
                       'map_snapshots': None, 'nvram': '', 'ovf_version': None}
    elif ext_file.lower() == 'ova':
        ovf_file = check_backup(path, 'ovf')
        if not ovf_file:
            return {}
        # на всякий случай имя ВМ берем из имени файла OVF
        domain_name = ovf_file.split('.ovf')[0].replace(' ', '_')
        domain_id = str(uuid.uuid4())
        tmp_path = tempfile.mkdtemp()
        extract_ovf = f'tar -xOf "{path}" \"{ovf_file}\" > \"{tmp_path}/{ovf_file}\"'
        ret_extract_ovf = get_output(extract_ovf)
        if not ret_extract_ovf['result']:
            shutil.rmtree(tmp_path, ignore_errors=True)
            raise CommandExecutionError(f'Failed to extract ovf from ova backup "{path}". '
                                        f'Error: {ret_extract_ovf["comment"]}')
        try:
            namespaces = dict([n for _, n in ET.iterparse(f'{tmp_path}/{ovf_file}', events=['start-ns'])])
        except Exception as e:
            raise CommandExecutionError(f'Failed to parse xml from ova backup "{path}". Error: {e}') from None
        tree = ET.parse(f'{tmp_path}/{ovf_file}')
        root = tree.getroot()
        ovf_version = root.attrib.get(f"{{{namespaces['ovf']}}}version", '1.0')
        tags = None
        for k in root:
            key_regex = re.compile(f"{{{namespaces['ovf']}}}tags$", re.IGNORECASE)
            if key_regex.match(k.tag):
                tags = k.text
                break
        os_version = None
        if root.find(f'{{{namespaces["ovf"]}}}VirtualSystem') is not None \
                and root.find(f'{{{namespaces["ovf"]}}}VirtualSystem/'
                              f'{{{namespaces["ovf"]}}}OperatingSystemSection') is not None \
                and root.find(f'{{{namespaces["ovf"]}}}VirtualSystem/{{{namespaces["ovf"]}}}OperatingSystemSection/'
                              f'{{{namespaces["ovf"]}}}Description') is not None:
            os_version = root.find(f'{{{namespaces["ovf"]}}}VirtualSystem/{{{namespaces["ovf"]}}}OperatingSystemSection'
                                   f'/{{{namespaces["ovf"]}}}Description').text
        if ovf_version == '1.0':
            ovf_files = [e.attrib[f'{{{namespaces["ovf"]}}}href'] for e in root.findall('ovf:References/ovf:File',
                                                                                        namespaces)]
            virt_convert = f'veil-convert-from-ovf \"{tmp_path}/{ovf_file}\"'
            ret_virt_convert = get_output(virt_convert)
            if not ret_virt_convert['result']:
                shutil.rmtree(tmp_path, ignore_errors=True)
                raise CommandExecutionError(f'Failed to convert ovf to libvirtxml from ova backup "{path}". '
                                            f'Error: {ret_virt_convert["comment"]}')
            shutil.rmtree(tmp_path, ignore_errors=True)
            # правим полученный XML в соответствии с SpaceVM форматом
            xml = ret_virt_convert["comment"]
            try:
                root = ET.fromstring(xml)
            except Exception as e:
                raise CommandExecutionError(
                    f'Failed to parse xml after convert from backup ova "{path}". Error: {e}') from None
            # имя ВМ берем из OVF файла, если не получается, то берем из имени файла
            tmp_name = root.find('name').text
            domain_name = tmp_name if tmp_name else domain_name
            root.find('name').text = domain_id
            root.find('uuid').text = domain_id
            index_tag_uuid = list(root).index(root.findall('uuid')[0])
            metadata_str = '<metadata><veil:instance xmlns:veil="http://controller/">' \
                           f'<veil:verbose_name>{domain_name}</veil:verbose_name>'
            if tags:
                metadata_str += f'<veil:tags>{tags}</veil:tags>'
            if os_version:
                # pylint: disable=reimported, redefined-outer-name
                from veilrpc.utils.domain.os_templates import OS_VERSIONS, OS_TYPES
                if os_version in OS_VERSIONS[OS_TYPES.windows].values():
                    metadata_str += f'<veil:os_type>{OS_TYPES.windows}</veil:os_type>' \
                                    f'<veil:os_version>{os_version}</veil:os_version>'
                elif os_version in OS_VERSIONS[OS_TYPES.linux].values():
                    metadata_str += f'<veil:os_type>{OS_TYPES.linux}</veil:os_type>' \
                                    f'<veil:os_version>{os_version}</veil:os_version>'
                elif os_version in OS_VERSIONS[OS_TYPES.other].values():
                    metadata_str += f'<veil:os_type>{OS_TYPES.other}</veil:os_type>' \
                                    f'<veil:os_version>{os_version}</veil:os_version>'
            metadata_str += '</veil:instance></metadata>'
            ET.register_namespace('veil', "http://controller/")
            metadata_xml = ET.fromstring(metadata_str)
            root.insert(index_tag_uuid, metadata_xml)

            backup_info = get_info_from_backup(path, ovf_files=ovf_files,
                                               task=task, redis=redis, progress_data=progress_data)
            is_zfs_pool = True if path.split('/')[-4] == 'zfs' else False
            local_datapool = path.split(f"/{PATH_FILE_STORAGE}")[0]
            backup_info_vdisk_keys = list(backup_info.get('vdisks_size', {}).keys())
            for disk in root.findall('devices/disk'):
                if disk.get('type') == 'file' and disk.get('device') == 'disk':
                    if disk.find('source') is None:
                        root.find('devices').remove(disk)
                        continue
                    vdisk_name = os.path.basename(disk.find('source').get('file')).split('.qcow2')[0]
                    if vdisk_name in backup_info_vdisk_keys:  # part_name:
                        vdisk_id = str(uuid.uuid4())
                        # Все пути меняются на текущий датапул, где расположен файл бакапа
                        source = f'{local_datapool}/{vdisk_id}/vdisk.qcow2' if is_zfs_pool \
                            else f'{local_datapool}/{vdisk_id}.qcow2'
                        disk.find('source').set('file', source)
                        backup_info['vdisks_size'][vdisk_id] = backup_info['vdisks_size'].pop(vdisk_name)
                        vdisk_in_files_map = [k for k, v in backup_info['files_map'].items() if v == vdisk_name]
                        if vdisk_in_files_map:
                            backup_info['files_map'][vdisk_in_files_map[0]] = vdisk_id
                elif disk.findall('source') and disk.get('type') == 'file' and disk.get('device') == 'cdrom':
                    iso_name = os.path.basename(disk.find('source').get('file'))
                    if os.path.basename(disk.find('source').get('file')) in backup_info.get('isos_size', {}).keys():
                        # Все пути меняются на текущий датапул, где расположен файл бакапа
                        source = f'{local_datapool}/_ISO/{iso_name}'
                        disk.find('source').set('file', source)
                        if iso_name in backup_info['files_map'].values():
                            backup_info['files_map'][iso_name] = source
        else:
            shutil.rmtree(tmp_path, ignore_errors=True)
            backup_info = {'vdisks_size': None, 'isos_size': None, 'memories': None, 'map_snapshots': None,
                           'nvram': ''}
        backup_info.update({'xml': ET.tostring(root, encoding='UTF-8').decode(), 'is_xml': False, 'is_ova': True,
                            'is_ovf': False, 'ovf_version': ovf_version})
    elif ext_file.lower() == 'ovf':
        # на всякий случай имя ВМ берем из имени файла OVF
        domain_name = path.split('/')[-1].split('.ovf')[0].replace(' ', '_')
        domain_id = str(uuid.uuid4())
        namespaces = dict([n for _, n in ET.iterparse(path, events=['start-ns'])])
        try:
            tree = ET.parse(path)
        except Exception as e:
            raise CommandExecutionError(f'Failed to parse ovf from "{path}". Error: {e}') from None
        root = tree.getroot()
        ovf_version = root.attrib.get(f"{{{namespaces['ovf']}}}version", '1.0')
        tags = None
        for k in root:
            key_regex = re.compile(f"{{{namespaces['ovf']}}}tags$", re.IGNORECASE)
            if key_regex.match(k.tag):
                tags = k.text
                break
        os_version = None
        if root.find(f'{{{namespaces["ovf"]}}}VirtualSystem') is not None \
                and root.find(f'{{{namespaces["ovf"]}}}VirtualSystem/'
                              f'{{{namespaces["ovf"]}}}OperatingSystemSection') is not None \
                and root.find(f'{{{namespaces["ovf"]}}}VirtualSystem/{{{namespaces["ovf"]}}}OperatingSystemSection/'
                              f'{{{namespaces["ovf"]}}}Description') is not None:
            os_version = root.find(f'{{{namespaces["ovf"]}}}VirtualSystem/{{{namespaces["ovf"]}}}OperatingSystemSection'
                                   f'/{{{namespaces["ovf"]}}}Description').text
        backup_info = {}
        if ovf_version == '1.0':
            ovf_files = [e.attrib[f'{{{namespaces["ovf"]}}}href'] for e in root.findall('ovf:References/ovf:File',
                                                                                        namespaces)]
            virt_convert = f'veil-convert-from-ovf \"{path}\"'
            ret_virt_convert = get_output(virt_convert)
            if not ret_virt_convert['result']:
                raise CommandExecutionError(f'Failed to convert ovf to libvirtxml from ova backup "{path}". '
                                            f'Error: {ret_virt_convert["comment"]}')
            # правим полученный XML в соответствии с SpaceVM форматом
            xml = ret_virt_convert["comment"]
            try:
                root = ET.fromstring(xml)
            except Exception as e:
                raise CommandExecutionError(
                    f'Failed to parse xml after convert ovf from "{path}". Error: {e}') from None
            # имя ВМ берем из OVF файла, если не получается, то берем из имени файла
            tmp_name = root.find('name').text
            domain_name = tmp_name if tmp_name else domain_name
            root.find('name').text = domain_id
            root.find('uuid').text = domain_id
            index_tag_uuid = list(root).index(root.findall('uuid')[0])
            metadata_str = '<metadata><veil:instance xmlns:veil="http://controller/">' \
                           f'<veil:verbose_name>{domain_name}</veil:verbose_name>'
            if tags:
                metadata_str += f'<veil:tags>{tags}</veil:tags>'
            if os_version:
                # pylint: disable=reimported
                from veilrpc.utils.domain.os_templates import OS_VERSIONS, OS_TYPES
                if os_version in OS_VERSIONS[OS_TYPES.windows].values():
                    metadata_str += f'<veil:os_type>{OS_TYPES.windows}</veil:os_type>' \
                                    f'<veil:os_version>{os_version}</veil:os_version>'
                elif os_version in OS_VERSIONS[OS_TYPES.linux].values():
                    metadata_str += f'<veil:os_type>{OS_TYPES.linux}</veil:os_type>' \
                                    f'<veil:os_version>{os_version}</veil:os_version>'
                elif os_version in OS_VERSIONS[OS_TYPES.other].values():
                    metadata_str += f'<veil:os_type>{OS_TYPES.other}</veil:os_type>' \
                                    f'<veil:os_version>{os_version}</veil:os_version>'
            metadata_str += '</veil:instance></metadata>'
            ET.register_namespace('veil', "http://controller/")
            metadata_xml = ET.fromstring(metadata_str)
            root.insert(index_tag_uuid, metadata_xml)
            backup_info = get_info_from_backup(path, ovf_files=ovf_files, directory=True,
                                               task=task, redis=redis, progress_data=progress_data)
            is_zfs_pool = True if path.split('/')[-4] == 'zfs' else False
            local_datapool = path.split(f"/{PATH_FILE_STORAGE}")[0] if f"/{PATH_FILE_STORAGE}" in path else os.path.basename(path)
            backup_info_vdisk_keys = [k.replace(' ', '_') for k in backup_info.get('vdisks_size', {}).keys()]
            # в отличии от OVA могут быть пробелы в именах файлов
            backup_info_vdisk_keys_space = list(backup_info.get('vdisks_size', {}).keys())
            for disk in root.findall('devices/disk'):
                if disk.get('type') == 'file' and disk.get('device') == 'disk':
                    if disk.find('source') is None:
                        root.find('devices').remove(disk)
                        continue
                    vdisk_name = os.path.basename(disk.find('source').get('file')).split('.qcow2')[0]
                    if vdisk_name in backup_info_vdisk_keys:  # part_name:
                        index = backup_info_vdisk_keys.index(vdisk_name)
                        vdisk_id = str(uuid.uuid4())
                        # Все пути меняются на текущий датапул, где расположен файл бакапа
                        source = f'{local_datapool}/{vdisk_id}/vdisk.qcow2' if is_zfs_pool \
                            else f'{local_datapool}/{vdisk_id}.qcow2'
                        disk.find('source').set('file', source)
                        backup_info['vdisks_size'][vdisk_id] = backup_info['vdisks_size'].pop(
                            backup_info_vdisk_keys_space[index])
                        vdisk_in_files_map = [k for k, v in backup_info['files_map'].items()
                                              if v == backup_info_vdisk_keys_space[index]]
                        if vdisk_in_files_map:
                            backup_info['files_map'][vdisk_in_files_map[0]] = vdisk_id
                    else:
                        root.find('devices').remove(disk)
                elif disk.findall('source') and disk.get('type') == 'file' and disk.get('device') == 'cdrom':
                    iso_name = os.path.basename(disk.find('source').get('file'))
                    if os.path.basename(disk.find('source').get('file')) in backup_info.get('isos_size', {}).keys():
                        # Все пути меняются на текущий датапул, где расположен файл бакапа
                        source = f'{local_datapool}/_ISO/{iso_name}'
                        disk.find('source').set('file', source)
                        if iso_name in backup_info['files_map'].values():
                            backup_info['files_map'][iso_name] = source
                    else:
                        root.find('devices').remove(disk)
        backup_info.update({'xml': ET.tostring(root, encoding='UTF-8').decode(), 'is_xml': False, 'is_ova': False,
                            'is_ovf': True, 'ovf_version': ovf_version})
    else:
        return {}

    if 'files_map' not in backup_info:
        backup_info['files_map'] = None
    return backup_info


@states_response
def automated_restore(backup_path, map_vdisks, assignment_type, user, task, **kwargs):
    """
    Функция извлекает диски, снепщоты и iso образы из tar архива или директории с ovf файлом
    в соответствии с переданными путями
    """
    # pylint: disable=reimported, redefined-outer-name
    from .copy import create_local_zfs_dataset, rsync, convert
    from .snapshot import rebase_unsafe
    from veilrpc.utils.file import chown_directory

    redis = kwargs.get('redis', None)
    isos = kwargs.get('isos', [])
    snapshots_matching = kwargs.get('snapshots_matching', [])
    memories_matching = kwargs.get('memories_matching', {})
    nvram = kwargs.get('nvram')
    progress_data = dict(type='progress', jid=task, id='', name='automated_restore', result=True, steps={})
    progress_data['steps']['total_steps'] = \
        len(map_vdisks) + len(isos) + len(snapshots_matching) + len(memories_matching) + 1
    if nvram:
        progress_data['steps']['total_steps'] += 1
    progress_data['steps']['step'] = 0
    with ratamount(os.path.dirname(backup_path) if assignment_type == 'domain_ovf' else backup_path,
                   task=task, redis=redis, progress_data=progress_data) as temp_dir:
        for vdisk_id_old, vdisk in map_vdisks.items():
            if vdisk['zpool_name']:
                create_local_zfs_dataset(vdisk['zpool_name'], vdisk['id'], vdisk['source'].split('vdisk.qcow2')[0], user)
            if vdisk.get("convert"):
                ret = convert(original_source=f'{temp_dir}/{vdisk_id_old}',
                              destination_source=vdisk['source'],
                              destination_format='qcow2',
                              owner=user,
                              preallocate=False,
                              compress=True,
                              progress_data=progress_data,
                              redis=redis)
                if ret['retcode']:
                    raise CommandExecutionError(f'An error occurred while convert file {vdisk_id_old}. '
                                                f'Error: {ret["error"]}')
            else:
                rsync_ret = rsync(True, f'{temp_dir}/{vdisk_id_old}.qcow2', vdisk['source'], {}, user,
                                  progress_data, redis)
                if rsync_ret['retcode']:
                    raise CommandExecutionError(f'An error occurred while rsync: {rsync_ret["error"]}.')
            try:
                chown_directory(vdisk['source'], user)
            except Exception as e:
                raise CommandExecutionError(str(e)) from None
        for iso in isos:
            namefile = os.path.basename(iso)
            rsync_ret = rsync(True, f'{temp_dir}/{namefile}', iso, {}, user, progress_data, redis)
            if rsync_ret['retcode']:
                raise CommandExecutionError(f'An error occurred while rsync: {rsync_ret["error"]}.')
            try:
                chown_directory(iso, user)
            except Exception as e:
                raise CommandExecutionError(str(e)) from None
        for s in snapshots_matching:
            rsync_ret = rsync(True, f'{temp_dir}/{s["old_name"]}', s["snapshot"], {}, user, progress_data, redis)
            if rsync_ret['retcode']:
                raise CommandExecutionError(f'An error occurred while rsync: {rsync_ret["error"]}.')
            try:
                chown_directory(s['snapshot'], user)
            except Exception as e:
                raise CommandExecutionError(str(e)) from None
        if snapshots_matching:
            ret = rebase_unsafe(snapshots_matching, True, {}, user)
            if not ret['result']:
                raise CommandExecutionError(ret['comment'])
        for m in memories_matching.keys():
            rsync_ret = rsync(True, f'{temp_dir}/{m}', memories_matching[m], {}, user, progress_data, redis)
            if rsync_ret['retcode']:
                raise CommandExecutionError(f'An error occurred while rsync: {rsync_ret["error"]}.')
            try:
                chown_directory(memories_matching[m], user)
            except Exception as e:
                raise CommandExecutionError(str(e)) from None
        if nvram:
            from veilrpc.utils.bazalt import NODE_UEFI_NVRAM_DIR
            new_nvram = os.path.join(NODE_UEFI_NVRAM_DIR, kwargs.get('domain') + '_VARS.fd')
            rsync_ret = rsync(True, f'{temp_dir}/{nvram}', new_nvram, {}, user, progress_data, redis)
            if rsync_ret['retcode']:
                raise CommandExecutionError(f'An error occurred while rsync: {rsync_ret["error"]}.')
            try:
                chown_directory(new_nvram, user)
                set_permission([new_nvram], int('600', 8))
            except Exception as e:
                raise CommandExecutionError(str(e)) from None
    if memories_matching:
        from veilrpc.states.snapshot import replace_domain_id_into_memory
        cmd_ret = replace_domain_id_into_memory(kwargs.get('domain'), kwargs.get('parent'),
                                                kwargs.get('list_memory_snapshots', {}), True, {}, user)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])

    return "Successfully extracted disks."


@states_response
def convert_backup_from_ova(domain_id, backup_path, current_date, compress_type, user, task, **kwargs):
    """
    Функция конвертирует OVA бакап в SpaceVM формат (libvirt XML + qcow2 образы).
    Возвращается словарь, для завершения создания объекта "файл".
    """
    # pylint: disable=reimported, redefined-outer-name
    import shutil
    from .copy import popen_with_progress

    redis = kwargs.get('redis', None)
    datapool = kwargs.get('datapool')
    outside = kwargs.get('outside')
    if datapool:
        is_zfs_pool = True if datapool.split('/')[-2] == 'zfs' else False
        local_datapool = datapool
    elif outside:
        raise CommandExecutionError('If backup located in the outside datapool when datapool should be passed.')
    else:
        is_zfs_pool = True if backup_path.split('/')[-4] == 'zfs' else False
        local_datapool = f'{backup_path.split(f"/{PATH_FILE_STORAGE}")[0]}'
    compress_cmd = ''
    compress_ext = ''
    if compress_type is not None and compress_type == LIBRARY_ZIP_TYPES.zstd.split()[0]:
        compress_cmd = '2>&3 | /usr/local/sbin/veil-zstd '
        compress_ext = EXTENSION_ZIP_TYPES.zst
    elif compress_type is not None:
        raise CommandExecutionError(f'Unsupport format {compress_type}.')
    progress_data = dict(type='progress', jid=task, id='', name='convert_backup_from_ova', result=True, steps={})

    get_ovf = f'tar -tf "{backup_path}" --wildcards *.ovf'
    ret_get_ovf = get_output(get_ovf)
    if not ret_get_ovf['result'] or 'ovf' not in ret_get_ovf["comment"].strip():
        raise CommandExecutionError(f'Failed to get ovf file from ova backup "{backup_path}". '
                                    f'Error: {ret_get_ovf["comment"]}')
    # такого не должно быть и ovf файл должен быть один, но на всякий случай берем первый
    ovf_file = ret_get_ovf["comment"].strip().split('\n')[0]
    # на всякий случай имя ВМ берем из имени файла OVF
    domain_name = ovf_file.split('.ovf')[0].replace(' ', '_')
    if outside:
        tmp_path = f'{local_datapool}/{PATH_TMP}/{domain_id}'
    else:
        tmp_path = f'{os.path.dirname(backup_path).replace(PATH_FILE_STORAGE, PATH_TMP)}/{domain_id}'
    os.makedirs(tmp_path, exist_ok=True)
    try:
        chown_directory(tmp_path, user)
    except Exception as e:
        shutil.rmtree(tmp_path, ignore_errors=True)
        raise CommandExecutionError(str(e)) from None
    extract_ovf = f'tar -xOf "{backup_path}" \"{ovf_file}\" > \"{tmp_path}/{ovf_file}\"'
    ret_extract_ovf = get_output(extract_ovf)
    if not ret_extract_ovf['result']:
        shutil.rmtree(tmp_path, ignore_errors=True)
        raise CommandExecutionError(f'Failed to extract ovf from ova backup "{backup_path}". '
                                    f'Error: {ret_extract_ovf["comment"]}')
    virt_convert = f'veil-convert-from-ovf \"{tmp_path}/{ovf_file}\"'
    ret_virt_convert = get_output(virt_convert)
    if not ret_virt_convert['result']:
        shutil.rmtree(tmp_path, ignore_errors=True)
        raise CommandExecutionError(f'Failed to convert ovf to libvirtxml from ova backup "{backup_path}". '
                                    f'Error: {ret_virt_convert["comment"]}')
    # veil-convert-from-ovf не берет из ovf cdrom и приатаченные к нему iso, поэтому делаем это сами
    my_namespaces = dict([node for _, node in ET.iterparse(f'{tmp_path}/{ovf_file}', events=['start-ns'])])
    tree = ET.parse(f'{tmp_path}/{ovf_file}')
    root = tree.getroot()
    iso_ids = [e.find('rasd:HostResource', my_namespaces).text
               for e in root.findall('ovf:VirtualSystem/ovf:VirtualHardwareSection/ovf:Item', my_namespaces)
               if e.find('rasd:HostResource', my_namespaces) is not None
               and 'CD/DVD Drive' in e.find('rasd:ElementName', my_namespaces).text]
    isos_ovf = []
    for e in root.findall('ovf:References/ovf:File', my_namespaces):
        if e.attrib[f'{{{my_namespaces["ovf"]}}}id'] in [i.split('/')[-1] for i in iso_ids]:
            iso = e.attrib[f'{{{my_namespaces["ovf"]}}}href']
            if iso and iso.split('.')[-1] in ['iso', 'ISO', 'iSo', 'isO', 'Iso', 'ISo']:
                isos_ovf.append(e.attrib[f'{{{my_namespaces["ovf"]}}}href'])
    os.remove(f'{tmp_path}/{ovf_file}')
    # правим полученный XML в соответствии с SpaceVM форматом, а также извлекаем диски во временный каталог
    # если требуется конвертируем в qcow2
    xml = ret_virt_convert["comment"]
    root = ET.fromstring(xml)
    # имя ВМ берем из OVF файла, если не получается, то берем из имени файла
    tmp_name = root.find('name').text
    domain_name = tmp_name if tmp_name else domain_name
    root.find('name').text = domain_id
    root.find('uuid').text = domain_id
    index_tag_uuid = list(root).index(root.find('uuid'))
    metadata_str = '<metadata><veil:instance xmlns:veil="http://controller/">'\
                   f'<veil:verbose_name>{domain_name}</veil:verbose_name></veil:instance></metadata>'
    ET.register_namespace('veil', "http://controller/")
    metadata_xml = ET.fromstring(metadata_str)
    root.insert(index_tag_uuid, metadata_xml)
    files_in_ova = f'tar -tf "{backup_path}"'
    ret_files_in_ova = get_output(files_in_ova)
    if not ret_files_in_ova['result']:
        shutil.rmtree(tmp_path, ignore_errors=True)
        raise CommandExecutionError(f'Failed to list files from ova backup {backup_path}. '
                                    f'Error: {ret_files_in_ova["comment"]}')
    vdisks = [v for v in ret_files_in_ova['comment'].strip().split('\n') if v.split('.')[-1] in
              ['qcow2', 'raw', 'img', 'vmdk']]
    isos = [i for i in ret_files_in_ova['comment'].strip().split('\n') if i in isos_ovf]
    progress_data['steps']['total_steps'] = \
        len(vdisks) + len([True for _ in vdisks if _.split('.')[-1] != 'qcow2']) + len(isos) + 1
    progress_data['steps']['step'] = 0
    for disk in list(root.findall('devices/disk')):
        if disk.get('type') == 'file' and disk.get('device') == 'disk':
            if disk.find('source') is None:
                root.find('devices').remove(disk)
                continue
            part_name = os.path.basename(disk.find('source').get('file')).split('.qcow2')[0]
            v = [v for v in vdisks if '.'.join(v.split('.')[:-1]).replace(' ', '_') == part_name]
            if v:
                v = v[0]
                vdisk_id = str(uuid.uuid4())
                # Все пути меняются на текущий датапул
                source = f'{local_datapool}/{vdisk_id}{"/vdisk.qcow2" if is_zfs_pool else ".qcow2"}'
                disk.find('source').set('file', source)
                if v.split('.')[-1] != 'qcow2':
                    is_qcow2 = False
                    path_to_disk = f'{tmp_path}/{v}'
                else:
                    is_qcow2 = True
                    path_to_disk = f'{tmp_path}/{vdisk_id}.qcow2'
                size_extract_vdisk_cmd = f'tar -tvf "{backup_path}" | grep "{v}" | awk \'{{print $3}}\''
                size_extract_vdisk_ret = get_output(size_extract_vdisk_cmd)
                if not size_extract_vdisk_ret['result'] or not size_extract_vdisk_ret['comment'].strip().isdigit():
                    raise CommandExecutionError(f'Failed to get size vdisk {v} from ova backup "{backup_path}". '
                                                f'Error: {size_extract_vdisk_ret["comment"]}')
                extract_vdisk = f'set -o pipefail; exec 3>&1; tar -xOf "{backup_path}" "{v}"  | pv -ns '\
                                f'{size_extract_vdisk_ret["comment"].strip()} > "{path_to_disk}" 2>&3'
                progress_data['steps']['step'] += 1
                p = popen_with_progress(extract_vdisk, "pv", progress_data, redis)
                if p.returncode:
                    shutil.rmtree(tmp_path, ignore_errors=True)
                    raise CommandExecutionError(
                        f'Failed to extract disk {v} from backup "{backup_path}". CMD: {extract_vdisk}. '
                        f'Error: {p.stderr.read()}')
                if not is_qcow2:
                    from .copy import convert
                    ret = convert(original_source=path_to_disk,
                                  destination_source=f'{tmp_path}/{vdisk_id}.qcow2',
                                  destination_format='qcow2',
                                  owner=user,
                                  preallocate=False,
                                  compress=True,
                                  progress_data=progress_data,
                                  redis=redis)
                    if ret['retcode']:
                        shutil.rmtree(tmp_path, ignore_errors=True)
                        raise CommandExecutionError(f'An error occurred while convert file {v}. '
                                                    f'Error: {ret["error"]}')
                    if os.path.exists(f'{tmp_path}/{v}'):
                        os.remove(f'{tmp_path}/{v}')
                else:
                    root.find('devices').remove(disk)
        elif disk.findall('source') and disk.get('type') == 'file' and disk.get('device') == 'cdrom':
            iso = os.path.basename(disk.find('source').get('file'))
            if iso in isos:
                source = f'{local_datapool}/_ISO/{iso}'
                disk.find('source').set('file', source)
                size_extract_iso_cmd = f'set -o pipefail; tar -tvf "{backup_path}" | ' \
                                       f'grep "{iso}" | awk \'{{print $3}}\''
                size_extract_iso_ret = get_output(size_extract_iso_cmd)
                if not size_extract_iso_ret['result'] or not size_extract_iso_ret['comment'].strip().isdigit():
                    raise CommandExecutionError(f'Failed to get size iso {iso} from ova backup "{backup_path}". '
                                                f'Error: {size_extract_iso_ret["comment"]}')
                extract_iso = f'set -o pipefail; exec 3>&1; tar -xOf "{backup_path}" "{iso}"  | pv -ns ' \
                              f'{size_extract_iso_ret["comment"].strip()} ' \
                              f'> {tmp_path}/{iso.lower().replace(" ", "_")} 2>&3'
                progress_data['steps']['step'] += 1
                p = popen_with_progress(extract_iso, "pv", progress_data, redis)
                if p.returncode:
                    shutil.rmtree(tmp_path, ignore_errors=True)
                    raise CommandExecutionError(
                        f'Failed to extract iso "{iso}" from backup "{backup_path}". Error: {p.stderr.read()}')
    with open(f'{tmp_path}/{domain_name}.xml', 'bw') as f:
        f.write(ET.tostring(root))
    get_size = f'set -o pipefail; du -csb {tmp_path}/*| tail -1 |cut -f1'
    ret_size = get_output(get_size)
    if not ret_size['result'] or not ret_size['comment'].strip().isdigit():
        shutil.rmtree(tmp_path, ignore_errors=True)
        raise CommandExecutionError(f'Failed to size files for converted backup "{domain_name}-{current_date}.tar". '
                                    f'Error: {ret_size["comment"]}')
    new_backup_path = f'{local_datapool}/{PATH_FILE_STORAGE}/{domain_name}-{current_date}.tar{compress_ext}'
    tmp_backup_path = f'{local_datapool}/{PATH_TMP}/{domain_name}-{current_date}.tar{compress_ext}'
    cmd = f'set -o pipefail; exec 3>&1; tar cf - --transform \"s:.*/::\" -P {tmp_path}/* | '\
          f'pv -ns {ret_size["comment"].strip()} {compress_cmd}> "{tmp_backup_path}" {"" if compress_cmd else "2>&3"}'
    progress_data['steps']['step'] += 1
    p = popen_with_progress(cmd, "pv", progress_data, redis)
    if p.returncode:
        shutil.rmtree(tmp_path, ignore_errors=True)
        raise CommandExecutionError(f'Failed to create backup "{tmp_backup_path}". Error: {p.stderr.read()}')
    try:
        shutil.move(tmp_backup_path, new_backup_path)
    except Exception as e:
        raise CommandExecutionError(f'Failed to move to {new_backup_path} after convert backup from ova. Error: {e}')
    shutil.rmtree(tmp_path, ignore_errors=True)
    return {'filename': f'{domain_name}-{current_date}.tar{compress_ext}',
            'path': new_backup_path,
            'size': os.path.getsize(new_backup_path),
            'additional_flags': {'can_be_unpacked': True,
                                 'can_be_domained': True}}


@states_response
def convert_backup_from_ovf(domain_id, backup_path, current_date, compress_type, user, task, **kwargs):
    """
    Функция конвертирует OVF и дисков в бакап SpaceVM формата (libvirt XML + qcow2 образы).
    Возвращается словарь, для завершения создания объекта "файл".
    """
    # pylint: disable=reimported, redefined-outer-name, redefined-outer-name
    import shutil
    from .copy import popen_with_progress
    from veilrpc.utils.file import chown_directory

    redis = kwargs.get('redis', None)
    datapool = kwargs.get('datapool')
    outside = kwargs.get('outside')
    if datapool:
        is_zfs_pool = True if datapool.split('/')[-2] == 'zfs' else False
        local_datapool = datapool
    elif outside:
        raise CommandExecutionError('If backup located in the outside datapool when datapool should be passed.')
    else:
        is_zfs_pool = True if backup_path.split('/')[-4] == 'zfs' else False
        local_datapool = f'{backup_path.split(f"/{PATH_FILE_STORAGE}")[0]}'
    progress_data = dict(type='progress', jid=task, id='', name='convert_backup_from_ovf', result=True, steps={})
    try:
        os.path.exists(backup_path)
    except Exception as e:
        raise CommandExecutionError(f'File "{backup_path}" do not exists. Error: {e}') from None
    ovf_file = os.path.basename(backup_path)
    # на всякий случай имя ВМ берем из имени файла OVF
    domain_name = ovf_file.split('.ovf')[0].replace(' ', '_')
    tmp_path = os.path.dirname(backup_path)
    virt_convert = f'veil-convert-from-ovf \"{backup_path}\"'
    ret_virt_convert = get_output(virt_convert)
    if not ret_virt_convert['result']:
        raise CommandExecutionError(f'Failed to convert ovf to libvirtxml from ovf "{backup_path}". '
                                    f'Error: {ret_virt_convert["comment"]}')
    my_namespaces = dict([node for _, node in ET.iterparse(backup_path, events=['start-ns'])])
    tree = ET.parse(backup_path)
    root = tree.getroot()
    iso_ids = [e.find('rasd:HostResource', my_namespaces).text
               for e in root.findall('ovf:VirtualSystem/ovf:VirtualHardwareSection/ovf:Item', my_namespaces)
               if e.find('rasd:HostResource', my_namespaces) is not None
               and 'CD/DVD Drive' in e.find('rasd:ElementName', my_namespaces).text]
    isos_ovf = []
    files_from_ovf = []
    for e in root.findall('ovf:References/ovf:File', my_namespaces):
        file = e.attrib[f'{{{my_namespaces["ovf"]}}}href']
        if e.attrib[f'{{{my_namespaces["ovf"]}}}id'] in [i.split('/')[-1] for i in iso_ids]:
            if file and file.split('.')[-1] in ['iso', 'ISO', 'iSo', 'isO', 'Iso', 'ISo']:
                isos_ovf.append(file)
        else:
            files_from_ovf.append(file)
    # правим полученный XML в соответствии с SpaceVM форматом, а также извлекаем диски во временный каталог
    # если требуется конвертируем в qcow2
    xml = ret_virt_convert["comment"]
    root = ET.fromstring(xml)
    # имя ВМ берем из OVF файла, если не получается, то берем из имени файла
    tmp_name = root.find('name').text
    domain_name = tmp_name if tmp_name else domain_name
    root.find('name').text = domain_id
    root.find('uuid').text = domain_id
    index_tag_uuid = list(root).index(root.find('uuid'))
    metadata_str = '<metadata><veil:instance xmlns:veil="http://controller/">'\
                   f'<veil:verbose_name>{domain_name}</veil:verbose_name></veil:instance></metadata>'
    ET.register_namespace('veil', "http://controller/")
    metadata_xml = ET.fromstring(metadata_str)
    root.insert(index_tag_uuid, metadata_xml)
    files_into_dir = os.listdir(tmp_path)
    vdisks = [v for v in files_into_dir if v.split('.')[-1] in ['qcow2', 'raw', 'img', 'vmdk'] and v in files_from_ovf]
    isos = [i for i in files_into_dir if i in isos_ovf]
    progress_data['steps']['total_steps'] = \
        len(vdisks) + len([True for _ in vdisks if _.split('.')[-1] != 'qcow2']) + len(isos) + 1
    compress_cmd = ''
    compress_ext = ''
    if compress_type is not None and compress_type == LIBRARY_ZIP_TYPES.zstd.split()[0]:
        compress_cmd = '2>&3 | /usr/local/sbin/veil-zstd '
        compress_ext = EXTENSION_ZIP_TYPES.zst
        progress_data['steps']['total_steps'] += 1
    elif compress_type is not None:
        raise CommandExecutionError(f'Unsupport format {compress_type}.')
    progress_data['steps']['step'] = 0
    tmp_path_local = f'{local_datapool}/{PATH_TMP}/{domain_id}'
    os.makedirs(tmp_path_local, exist_ok=True)
    try:
        chown_directory(tmp_path_local, user)
    except Exception as e:
        shutil.rmtree(tmp_path_local, ignore_errors=True)
        raise CommandExecutionError(str(e)) from None
    append_to_backup = ''
    for disk in list(root.findall('devices/disk')):
        if disk.get('type') == 'file' and disk.get('device') == 'disk':
            if disk.find('source') is None:
                root.find('devices').remove(disk)
                continue
            part_name = os.path.basename(disk.find('source').get('file')).split('.qcow2')[0]
            v = [v for v in vdisks if '.'.join(v.split('.')[:-1]).replace(' ', '_') == part_name]
            if v:
                v = v[0]
                vdisk_id = str(uuid.uuid4())
                # Все пути меняются на текущий датапул
                source = f'{local_datapool}/{vdisk_id}{"/vdisk.qcow2" if is_zfs_pool else ".qcow2"}'
                disk.find('source').set('file', source)
                if v.split('.')[-1] != 'qcow2':
                    path_to_disk = f'{tmp_path}/{v}'
                    from .copy import convert
                    ret = convert(original_source=path_to_disk,
                                  destination_source=f'{tmp_path_local}/{vdisk_id}.qcow2',
                                  destination_format='qcow2',
                                  owner=user,
                                  preallocate=False,
                                  compress=True,
                                  progress_data=progress_data,
                                  redis=redis)
                    if ret['retcode']:
                        shutil.rmtree(tmp_path_local, ignore_errors=True)
                        raise CommandExecutionError(f'An error occurred while convert file {v}. '
                                                    f'Error: {ret["error"]}')
                    append_to_backup = f'set -o pipefail; exec 3>&1; ' \
                                       f'pv -n {tmp_path_local}/{vdisk_id}.qcow2 2>&3 | ' \
                                       f'/usr/local/sbin/veil-tarappend -t ' \
                                       f'{local_datapool}/{PATH_TMP}/{domain_name}-{current_date}.tar ' \
                                       f'-f {vdisk_id}.qcow2'
                else:
                    append_to_backup = f'set -o pipefail; exec 3>&1; ' \
                                       f'pv -n "{tmp_path}/{v}" 2>&3 | ' \
                                       f'/usr/local/sbin/veil-tarappend -t ' \
                                       f'{local_datapool}/{PATH_TMP}/{domain_name}-{current_date}.tar ' \
                                       f'-f {vdisk_id}.qcow2'
            else:
                root.find('devices').remove(disk)
        elif disk.findall('source') and disk.get('type') == 'file' and disk.get('device') == 'cdrom':
            iso = os.path.basename(disk.find('source').get('file'))
            if iso in isos:
                iso_replace = iso.replace(' ', '_')
                source = f'{local_datapool}/_ISO/{iso_replace}'
                disk.find('source').set('file', source)
                append_to_backup = f'set -o pipefail; exec 3>&1; pv -n "{tmp_path}/{iso}" 2>&3 | ' \
                                   f'/usr/local/sbin/veil-tarappend -t ' \
                                   f'{local_datapool}/{PATH_TMP}/{domain_name}-{current_date}.tar -f "{iso_replace}"'
        else:
            continue
        if append_to_backup:
            progress_data['steps']['step'] += 1
            p = popen_with_progress(append_to_backup, "pv", progress_data, redis)
            if p.returncode:
                shutil.rmtree(tmp_path_local, ignore_errors=True)
                raise CommandExecutionError(
                    f'Failed to append file to backup "{domain_name}-{current_date}.tar". '
                    f'CMD: {append_to_backup}. Error: {p.stderr.read()}')

    with open(f'{tmp_path_local}/{domain_name}.xml', 'bw') as f:
        f.write(ET.tostring(root))
    tmp_backup_path = f'{local_datapool}/{PATH_TMP}/{domain_name}-{current_date}.tar'
    new_backup_path = f'{local_datapool}/{PATH_FILE_STORAGE}/{domain_name}-{current_date}.tar'
    cmd = f'cat {tmp_path_local}/{domain_name}.xml | /usr/local/sbin/veil-tarappend -t ' \
          f'{local_datapool}/{PATH_TMP}/{domain_name}-{current_date}.tar -f {domain_name}.xml'
    ret = get_output(cmd)
    if not ret['result']:
        shutil.rmtree(tmp_path_local, ignore_errors=True)
        raise CommandExecutionError(f'Failed to append file to backup "{domain_name}-{current_date}.tar". '
                                    f'CMD: {cmd}. Error: {ret["comment"]}')
    shutil.rmtree(tmp_path_local, ignore_errors=True)
    if compress_cmd:
        old_tmp_backup_path = tmp_backup_path
        tmp_backup_path = f'{tmp_backup_path}{compress_ext}'
        new_backup_path = f'{new_backup_path}{compress_ext}'
        progress_data['steps']['step'] += 1
        conmpress_cmd = f'set -o pipefail; exec 3>&1; ' \
                        f'pv -n {old_tmp_backup_path} {compress_cmd} > {tmp_backup_path}' \
                        f' {"" if compress_cmd else "2>&3"}'
        p = popen_with_progress(conmpress_cmd, "pv", progress_data, redis)
        if p.returncode:
            raise CommandExecutionError(
                f'Failed to compress the backup file "{domain_name}-{current_date}.tar". '
                f'CMD: {conmpress_cmd}. Error: {p.stderr.read()}')
        try:
            os.remove(f'{local_datapool}/{PATH_TMP}/{domain_name}-{current_date}.tar')
        except OSError:
            pass
    try:
        shutil.move(tmp_backup_path, new_backup_path)
    except Exception as e:
        raise CommandExecutionError(f'Failed to move to {new_backup_path} after convert backup from ovf. Error: {e}')

    return {'filename': f'{domain_name}-{current_date}.tar{compress_ext}',
            'path': new_backup_path,
            'size': os.path.getsize(new_backup_path),
            'additional_flags': {'can_be_unpacked': True,
                                 'can_be_domained': True}}


@states_response
def list_domains(**kwargs) -> list:
    domains = get_domains()
    result = list()
    alive_domain_ids = list()
    for _, dom in enumerate(domains, start=1):
        state, _ = dom.state()
        dom_xml = dom.XMLDesc()
        dom_id, meta_values = get_meta_values(dom_xml, {'verbose_name': '',
                                                        'template': False,
                                                        'parent': False,
                                                        'thin': False})
        if not dom_id:
            continue
        alive_domain_ids.append(dom_id)
        result.append(dict(dom_id=dom_id,
                           state=STATES[state],
                           dom_name=str(meta_values['verbose_name']),
                           dom_template=meta_values.get('template'),
                           dom_parent=meta_values.get('parent'),
                           dom_thin=meta_values.get('thin'),
                           dom_xml=dom_xml))
    result.extend(get_shutoff_domains(alive_domain_ids))
    return result


@states_response
def list_powered_domains_and_remove_doubles(other_nodes_known_domains: list) -> list:
    cmd_ret = list_domains()
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    all_node_domain_ids = set([dom['dom_id'] for dom in cmd_ret['comment']])
    doubles = set(all_node_domain_ids.intersection(set(other_nodes_known_domains)))
    if doubles:
        delete(list(doubles))
    all_powered_domains = set([dom['dom_id'] for dom in cmd_ret['comment'] if dom['state'] == STATES[1]])
    return list(all_powered_domains - doubles)


@states_response
def list_powered_domains() -> list:
    domains = get_domains()
    result = list()
    for dom in domains:
        state, _ = dom.state()
        if state == 1:
            result.append(dom.name())
    return result


def get_shutoff_domains_configurations(alive_domain_ids: list) -> dict:
    result = dict()
    if os.path.exists(DOMAINS_FOLDER):
        for file_name in os.listdir(DOMAINS_FOLDER):
            dom_id, _ = os.path.splitext(file_name)
            if dom_id not in alive_domain_ids:
                try:
                    file_path = f'{DOMAINS_FOLDER}/{file_name}'
                    with open(file_path, 'r') as f:
                        dom_xml = f.read()
                    if not dom_xml:
                        os.remove(file_path)
                        continue
                    dom_conf_ret = get_draft_domain_from_xml(dom_xml, show=True, domain_id=dom_id)
                    if dom_conf_ret['result']:
                        configuration = dom_conf_ret['comment']
                        configuration['state'] = STATES[5]
                        result[dom_id] = configuration
                except (OSError, UnicodeDecodeError):
                    continue
    return result


def get_shutoff_domains(
        alive_domain_ids: list,
        id_list: bool = False,
        node_domains: bool = False,
        with_expected_state: bool = False,
        get_xml: bool = False) -> Union[list, tuple]:
    result = list()
    tmp_buffer = dict()
    if os.path.exists(DOMAINS_FOLDER):
        extensions = [f'.{DOM_XML_EXT}']
        if not node_domains:
            extensions.insert(0, '.xml')
        for file_name in os.listdir(DOMAINS_FOLDER):
            dom_id, extension = os.path.splitext(file_name)
            if extension in extensions:
                if dom_id not in alive_domain_ids:
                    if id_list:
                        result.append(dom_id)
                    else:
                        try:
                            with open(f'{DOMAINS_FOLDER}/{file_name}', 'r') as f:
                                dom_xml = f.read()
                            _id, meta_values = get_meta_values(
                                dom_xml, {'verbose_name': dom_id, 'power_state_expected': ''})
                            power_state_expected = meta_values.pop('power_state_expected', None)
                            if with_expected_state:
                                tmp_buffer[_id] = power_state_expected
                            if get_xml:
                                dom_conf_ret = get_draft_domain_from_xml(dom_xml, show=True, domain_id=dom_id)
                                if dom_conf_ret['result']:
                                    result.append(dom_conf_ret['comment'])
                            else:
                                _id, meta_values = get_meta_values(dom_xml,
                                                                   {'verbose_name': dom_id,
                                                                    'template': False,
                                                                    'parent': False,
                                                                    'thin': False})
                                result.append(dict(dom_id=dom_id,
                                                   state=STATES[5],
                                                   dom_name=str(meta_values['verbose_name']),
                                                   dom_template=meta_values.get('template'),
                                                   dom_parent=meta_values.get('parent'),
                                                   dom_thin=meta_values.get('thin'),
                                                   dom_xml=dom_xml))
                        except OSError:
                            continue
                if with_expected_state:
                    if dom_id not in tmp_buffer:
                        cmd_ret = get_domain_dumpxml(dom_id)
                        if cmd_ret['result']:
                            dom_xml = cmd_ret['comment']
                        _id, meta_values = get_meta_values(dom_xml, {'power_state_expected': ''})
                        tmp_buffer[_id] = meta_values.pop('power_state_expected', None)
    if with_expected_state:
        return result, [i for i in tmp_buffer if tmp_buffer[i] == 'OFF']
    return result


def get_meta_values(domain_xml: str, meta_values: dict = None) -> tuple:
    if meta_values is None:
        meta_values = dict()
    cmd_ret = xml_to_dict2(domain_xml)
    if not cmd_ret['result']:
        return None, meta_values
    domain_dict = cmd_ret['comment'].get('domain')
    if not domain_dict:
        return None, meta_values
    domain_id = domain_dict['name']['_text_']
    metadata = domain_dict['metadata']
    meta = next(iter(metadata.values()))
    for meta_value in meta_values.keys():
        for k, v in meta.items():
            if meta_value in k:
                meta_values[meta_value] = v.get('_text_')
                break
    return domain_id, meta_values


def get_domain_interfaces_from_xml(domain_xml: str) -> list:
    cmd_ret = xml_to_dict2(domain_xml)
    if not cmd_ret['result']:
        return []
    interface_configs = cmd_ret['comment'].get('domain', {}).get('devices', {}).get('interface')
    if isinstance(interface_configs, list):  # type checking так как xml_to_dict2 может выдать список
        return interface_configs
    return [interface_configs] if interface_configs else []


@states_response
def start_node(domain_id: str, **kwargs) -> str:
    cmd_ret = get_domain_dumpxml(domain_id)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    domain_xml = cmd_ret['comment']
    _id, meta_values = get_meta_values(domain_xml, {'cpu_priority': 10, 'template': False})
    if meta_values['template']:
        raise CommandExecutionError('Domain template can not be started.')

    # получаем информацию о интерфейсах вм из xml
    domain_interfaces = get_domain_interfaces_from_xml(domain_xml)
    domain_infs_map = list()

    # перебираем полученные из xml интерфейсы вм, заполняя список интерфейс vlan конфигов
    for domain_inf in domain_interfaces:
        vlan_info = domain_inf.get('vlan')
        vlan_mode = None
        vlan_tag = None
        vlan_trunks = None
        # если у интерфейс содержит vlan информацию -> заполняем trunks, vlan_mode и tag id
        if vlan_info:
            if vlan_info.get("trunk"):
                # если vlan mode не access -> ищем тип nativeMode и заполняем trunks тагами
                if isinstance(vlan_info['tag'], list):
                    # если несколько тагов -> значит есть транки, перебираем их
                    for tag in vlan_info['tag']:
                        if tag.get('nativeMode'):
                            # если указан nativeMode -> значит это vlan id сети
                            vlan_tag = tag['id']
                            vlan_mode = 'native-tagged' if tag['nativeMode'] == 'tagged' else 'native-untagged'
                        else:
                            # если не указан nativeMode -> значит это транк, добавляем его в список
                            if vlan_trunks is None:
                                vlan_trunks = []
                            vlan_trunks.append(tag['id'])
                else:
                    # если один таг, значит транков нету, находим только tag id и nativeMode
                    tag = vlan_info['tag']
                    vlan_tag = tag['id']
                    vlan_mode = 'native-tagged' if tag['nativeMode'] == 'tagged' else 'native-untagged'
            else:
                # если vlan mode access -> берём id тага
                vlan_mode = 'access'
                vlan_tag = vlan_info['tag']['id']
                vlan_trunks = None
        # добавляем заполненный конфиг интерфейса в список
        domain_infs_map.append(dict(interface=domain_inf['target']['dev'],
                                    config=dict(vlan_mode=vlan_mode,
                                                vlan_tag=vlan_tag,
                                                vlan_trunks=vlan_trunks,
                                                mtu=domain_inf['mtu']['size'])))
    cmd_ret = start(**{'domain_id': domain_id, 'xml': domain_xml,
                       'cpu_priority': meta_values['cpu_priority'],
                       'domain_infs_map': domain_infs_map, 'first_boot': False, 'paused': False})
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f"Domain {domain_id} successfully started."


@states_response
def remove_node(domain_id: str, **kwargs) -> str:
    fargs = {'id': domain_id, 'safety': False, 'snapshots': False,
             'cloud_iso': False, 'uefi_boot_vars': ''}
    ret = full_delete(**fargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return f"Domain {domain_id} successfully removed."


def delete_xml(domain_id: str) -> str:
    try:
        file_path = xml_path_by_id(domain_id)
        os.remove(file_path)
        return file_path
    except (CommandExecutionError, OSError):
        pass


@states_response
def delete_xml_after_recovery(domain_id: str, **kwargs) -> str:
    try:
        deleted_file = delete_xml(domain_id)
        if deleted_file and not os.path.exists(deleted_file):
            return f"The file {deleted_file} has been deleted."
        else:
            raise Exception(f"The removal data for domain_id {domain_id} was not successful.")
    except Exception as err:
        raise CommandExecutionError(err) from None


@states_response
def delete_all_xml(**kwargs) -> None:
    try:
        shutil.rmtree(DOMAINS_FOLDER)
        os.mkdir(DOMAINS_FOLDER)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None


@states_response
def save_xml(domains_data: dict, extension: str = 'xml', **kwargs) -> None:
    try:
        for domain_id, domain_xml in domains_data.items():
            xml_path = f'{DOMAINS_FOLDER}/{domain_id}.{extension}'
            node_xml_path = f'{DOMAINS_FOLDER}/{domain_id}.{DOM_XML_EXT}'
            if extension == 'xml' and os.path.isfile(node_xml_path):
                os.remove(node_xml_path)
            if os.path.isfile(xml_path):
                with open(xml_path, 'r+') as f:
                    file_xml = f.read()
                    if domain_xml != file_xml:
                        f.seek(0)
                        f.write(domain_xml)
                        f.truncate()
            else:
                with open(xml_path, 'w') as f:
                    f.write(domain_xml)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None


@states_response
def node_create(domain_id, verbose_name, xml):
    cmd_ret = list_domains()
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    for domain in cmd_ret['comment']:
        if domain_id == domain['dom_id']:
            raise CommandExecutionError(f'Domain with id {domain_id} already exists.')
        elif verbose_name == domain['dom_name']:
            raise CommandExecutionError(f'Domain with name {verbose_name} already exists.')
    save_xml(domains_data={domain_id: xml}, extension=DOM_XML_EXT, send_event=True)
    return f'VM {verbose_name} ({domain_id}) created.'


def domain_xml_root(domain_id: str):
    xml_path = xml_path_by_id(domain_id)
    if os.path.isfile(xml_path):
        ET.register_namespace('veil', "http://controller/")
        ET.register_namespace('qemu', "http://libvirt.org/schemas/domain/qemu/1.0")
        tree = ET.parse(xml_path)
        root = tree.getroot()
        return root
    else:
        raise CommandExecutionError(f"Can`t read domains {domain_id} XML file.")


def node_cdrom_target_dev_by_iso_path(domain_id, iso_path):
    root = domain_xml_root(domain_id)
    for tag in root.iterfind('devices/disk'):
        if tag.attrib['device'] == 'cdrom':
            source = tag.find('source')
            if ET.iselement(source) and source.attrib['file'] == iso_path:
                target = tag.find('target')
                return target.attrib['dev']
    raise CommandExecutionError(f"Iso {iso_path} is not found.")


def node_free_cdrom_target_dev(domain_id):
    root = domain_xml_root(domain_id)
    for tag in root.iterfind('devices/disk'):
        if tag.attrib['device'] == 'cdrom':
            source = tag.find('source')
            if ET.iselement(source):
                continue
            target = tag.find('target')
            return target.attrib['dev']
    raise CommandExecutionError("No free cdrom.")


@states_response
def node_attach_iso(domain_id: str, iso_path: str):
    root = domain_xml_root(domain_id)
    for tag in root.iterfind('devices/disk'):
        if tag.attrib['device'] == 'cdrom':
            source = tag.find('source')
            if ET.iselement(source):
                continue
            source = ET.SubElement(tag, 'source')
            source.set('file', iso_path)
            ET.indent(root)
            with open(xml_path_by_id(domain_id), 'bw') as f:
                f.write(ET.tostring(root))
            return f"Iso {iso_path} successfully attached to domains {domain_id} cdrom."
    raise CommandExecutionError(f"No free cdrom to attach iso {iso_path}.")


@states_response
def node_detach_iso(domain_id: str, iso_path: str):
    root = domain_xml_root(domain_id)
    for tag in root.iterfind('devices/disk'):
        if tag.attrib['device'] == 'cdrom':
            source = tag.find('source')
            if ET.iselement(source) and source.attrib['file'] == iso_path:
                tag.remove(source)
                ET.indent(root)
                with open(xml_path_by_id(domain_id), 'bw') as f:
                    f.write(ET.tostring(root))
                return f"Iso {iso_path} successfully detached from domains {domain_id} cdrom."
    raise CommandExecutionError(f"Iso {iso_path} is not attached to domains {domain_id} cdrom.")


@states_response
def node_attach_vmachine_interface(domain_id: str, xml: str):
    if is_domain_xml_exists(domain_id):
        root = domain_xml_root(domain_id)
        vinterface = ET.fromstring(xml)
        vinterface_name = vinterface.find('target').attrib['dev']
        devices = root.find('devices')
        for device in devices.iterfind('interface'):
            if device.attrib['type'] == 'bridge':
                target = device.find('target')
                if ET.iselement(target) and target.attrib['dev'] == vinterface_name:
                    raise CommandExecutionError(f"Vinterface {vinterface_name} is "
                                                f"already attached to domain {domain_id}.")
        devices.append(vinterface)
        ET.indent(root)
        with open(xml_path_by_id(domain_id), 'bw') as f:
            f.write(ET.tostring(root))
        return "Virtual interface created successfully."


@states_response
def node_detach_vmachine_interface(domain_id: str, xml: str):
    if is_domain_xml_exists(domain_id):
        root = domain_xml_root(domain_id)
        vinterface = ET.fromstring(xml)
        vinterface_name = vinterface.find('target').attrib['dev']
        devices = root.find('devices')
        for device in devices.iterfind('interface'):
            if device.attrib['type'] == 'bridge':
                target = device.find('target')
                if ET.iselement(target) and target.attrib['dev'] == vinterface_name:
                    devices.remove(device)
                    ET.indent(root)
                    with open(xml_path_by_id(domain_id), 'bw') as f:
                        f.write(ET.tostring(root))
                    return "Virtual interface detached successfully."
        raise CommandExecutionError(f"Vinterface {vinterface_name} is not attached to domain {domain_id}.")


@states_response
def node_attach_vdisk(domain_id: str, vdisk_id: str, vdisk_xml: str):
    root = domain_xml_root(domain_id)
    vdisk = ET.fromstring(vdisk_xml)
    devices = root.find('devices')
    for device in devices.iterfind('disk'):
        if device.attrib['device'] == 'disk':
            source = device.find('source')
            if ET.iselement(source) and vdisk_id in source.attrib.get('file', source.attrib.get('name', '')):
                raise CommandExecutionError(f"Vdisk {vdisk_id} is already attached to domain {domain_id}.")
    devices.append(vdisk)
    ET.indent(root)
    with open(xml_path_by_id(domain_id), 'bw') as f:
        f.write(ET.tostring(root))
    return f"Vdisk successfully attached to domain {domain_id}."


@states_response
def node_detach_vdisk(domain_id: str, vdisk_id: str):
    root = domain_xml_root(domain_id)
    devices = root.find('devices')
    for device in devices.iterfind('disk'):
        if device.attrib['device'] == 'disk':
            source = device.find('source')
            if ET.iselement(source) and vdisk_id in source.attrib.get('file', source.attrib.get('name', '')):
                devices.remove(device)
                ET.indent(root)
                with open(xml_path_by_id(domain_id), 'bw') as f:
                    f.write(ET.tostring(root))
                return f"Vdisk {vdisk_id} successfully detached from domain {domain_id}."
    raise CommandExecutionError(f"Vdisk {vdisk_id} is not attached to domain {domain_id}.")


@states_response
def node_remove_cdrom(domain_id: str, target_dev: str):
    root = domain_xml_root(domain_id)
    devices = root.find('devices')
    for device in devices.iterfind('disk'):
        if device.attrib['device'] == 'cdrom':
            target = device.find('target')
            if ET.iselement(target) and target.attrib['dev'] == target_dev:
                devices.remove(device)
                ET.indent(root)
                with open(xml_path_by_id(domain_id), 'bw') as f:
                    f.write(ET.tostring(root))
                return f"Cd-rom {target_dev} successfully removed from domain {domain_id}."
    raise CommandExecutionError(f"Cd-rom {target_dev} is not attached to domain {domain_id}.")


@states_response
def domain_xml_extension(domain_id: str) -> str:
    xml_path = xml_path_by_id(domain_id)
    _, ext = os.path.splitext(xml_path)
    return ext[1:]


@states_response
def node_only(domain_id: str) -> bool:
    cmd_ret = domain_xml_extension(domain_id)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment'] == DOM_XML_EXT


@states_response
def node_set_boot_order(domain_id: str, boot_order: list = None):
    root = domain_xml_root(domain_id)
    devices = root.find('devices')
    if boot_order is not None:
        for device in devices.iterfind('disk'):
            target = device.find('target')
            device_boot_order = device.find('boot')
            if target.attrib['dev'] in boot_order:
                device_boot_order.attrib['order'] = str(boot_order.index(target.attrib['dev']) + 1)
            else:
                device.remove(device_boot_order)
    else:
        for device in devices.iterfind('disk'):
            device_boot_order = device.find('boot')
            if device.attrib['device'] == 'cdrom':
                device_boot_order.attrib['order'] = '1'
            else:
                device_boot_order.attrib['order'] = '0'
    ET.indent(root)
    with open(xml_path_by_id(domain_id), 'bw') as f:
        f.write(ET.tostring(root))
    return f"Domain's {domain_id} boot order successfully updated."


@states_response
def get_draft_domain_from_xml(xml_dump, node=None, restore_from_backup=False, restore_from_xml=False, show=False,
                              domain_id=None):
    """
    Из xml формируется структура, поля которой соответствуют актуальным сериализаторам

    :param xml_dump:
    :param node:
    :param restore_from_backup: bool
    :param restore_from_xml: bool
    :param show: bool - used for showing all possible devices from xml
    :param domain_id: str - for logging only
    """

    domain_configuration = {}  # готовый конфиг для мультифункции создания VM
    additional_info = {}  # дополнительная информация для диалога с пользователем

    def exist_tag(parent, tag):
        """
        Used for bool expressions
        """
        return parent.find(tag) is not None

    def dict_tag(parent, tag):
        """
        Used for tag.get(attr)
        """
        return parent.find(tag) if exist_tag(parent, tag) else {}

    def get_common_disk_info(disk_tag):
        driver_tag = dict_tag(disk_tag, 'driver')
        _target = dict_tag(disk_tag, 'target')
        return dict(driver_type=driver_tag.get('type', 'qcow2'),
                    driver_cache=driver_tag.get('cache', 'none'),
                    target_bus=_target.get('bus'),
                    io=driver_tag.get('io') if driver_tag.get('io') in VDISK_IO_TYPES else VDISK_IO_TYPES[0],
                    target_dev=_target.get('dev'))

    def get_boot(disk_tag):
        boot = dict_tag(disk_tag, 'boot')
        return int(boot.get('order')) if exist_tag(disk_tag, 'boot') else None

    try:
        root = ET.fromstring(xml_dump)
    except Exception as e:
        raise CommandExecutionError(f"error in ET.fromstring for domain {str(domain_id)}: {str(e)}") from None
    controllers = defaultdict(list)
    cpu_topology = {}
    vdisks, luns, isos, cdroms = [], [], [], []
    vmachine_infs, vfunction_infs = [], 0

    usb_devices = []
    pci_devices = []

    bootmenu = False
    bootmenu_timeout = 3
    boot_type = "LegacyMBR"
    ordered_devices = {}
    ordered_devs = {}
    boot_free_devs = []

    try:
        domain_configuration['domain_id'] = root.find('uuid').text
    except Exception as err:
        raise CommandExecutionError('UUID into xml not found. Impossible to create VM.') from err

    domain_configuration.update(verbose_name=f'domain_{str(uuid.uuid4())[:8]}')

    def to_int_bool_if_possible(value: str):
        try:
            return int(value)
        except (TypeError, ValueError):
            if value == 'True':
                return True
            elif value == 'False':
                return False
            elif value == 'None':
                return None
        return value

    goa_options = ['_ha_detect_fail_haltpoll', '_ha_reboot_after_fail', '_ha_detect_fail_ping', '_ha_delay_after_fail',
                   '_ha_domain_boot_delay', '_ha_detect_fail_ga', '_ha_reboot_attempts', '_ha_ping_address']
    ha_options = ['_ha_enabled', '_ha_autoselect', '_ha_nodepolicy', '_ha_boot_order',
                  '_ha_retrycount', '_ha_timeout', '_ha_boot_delay', '_ha_boot_agent'] + goa_options
    for key in ['template', 'start_on_boot', 'os_type', 'os_version', 'service_module', 'start_on_boot', 'parent',
                'disastery_enabled', 'safety', 'verbose_name', 'memory_shares', 'memory_limit', 'compatibility',
                'description', 'created', 'memory_min_guarantee', 'cpu_min_guarantee', 'owners',
                'graphics_password', 'tags', 'power_state_expected', 'first_boot'] + ha_options:
        name = CONTROLLER_PATTERN.format(k=key)
        if root.findall(name, CONTROLLER_TAG):
            domain_configuration[key] = to_int_bool_if_possible(root.find(name, CONTROLLER_TAG).text)

    domain_configuration['_ha_options_cluster_sync'] = all(option is None for option in
                                                           [domain_configuration.get(name) for name in ha_options])
    name = CONTROLLER_PATTERN.format(k='cpu_priority')
    if root.findall(name, CONTROLLER_TAG):
        cpu_topology['cpu_priority'] = to_int_bool_if_possible(root.find(name, CONTROLLER_TAG).text)

    if domain_configuration.get('owners') is not None:
        owners = domain_configuration['owners'].split(',')
        if owners:
            domain_configuration['owners'] = owners
    if not isinstance(domain_configuration.get('owners'), list):
        domain_configuration['owners'] = []

    if 'os_type' in domain_configuration.keys():
        if domain_configuration['os_type'] not in list(OS_TYPES):
            domain_configuration['os_type'] = OS_TYPES.other
        if 'os_version' in domain_configuration.keys():
            if domain_configuration['os_version'] not in list(OS_VERSIONS[domain_configuration['os_type']].values()):
                domain_configuration['os_version'] = next(iter(OS_VERSIONS[domain_configuration['os_type']].values()))
        else:
            domain_configuration['os_version'] = next(iter(OS_VERSIONS[domain_configuration['os_type']].values()))
    else:
        domain_configuration['os_type'] = OS_TYPES.other
        if 'os_version' in domain_configuration.keys():
            if domain_configuration['os_version'] not in list(OS_VERSIONS[OS_TYPES.other].values()):
                domain_configuration['os_version'] = OS_VERSIONS[OS_TYPES.other]['unknown']
        else:
            domain_configuration['os_version'] = OS_VERSIONS[OS_TYPES.other]['unknown']

    if not domain_configuration.get('description'):
        domain_configuration['description'] = ''

    for ctrl_tag in root.iterfind('devices/controller'):
        type_ = ctrl_tag.get('type')
        model_ = ctrl_tag.get('model')
        order_ = ctrl_tag.get('index')

        if type_ != 'usb' or (type_ == 'usb' and model_ in USB_CONTROLLER_TYPES):
            controllers[type_].append(dict(model=model_ if model_ else {}, order=order_))

    if controllers:
        domain_configuration.update(controllers=dict(controllers))

    cpu_topology['cpu_map'] = {}
    cpu_topology['cpu_count_max'] = 0

    if exist_tag(root, 'cputune/shares'):
        cpu_shares_tag = root.find('cputune/shares')
        cpu_topology['cpu_shares'] = int(cpu_shares_tag.text) if cpu_shares_tag.text else 10000

    if exist_tag(root, 'cputune/quota'):
        cpu_quota_tag = root.find('cputune/quota')
        cpu_topology['cpu_period'] = int(cpu_quota_tag.text) if cpu_quota_tag.text else CPU_PERIOD

    if exist_tag(root, 'vcpu'):
        vcpu_tag = root.find('vcpu')
        cpu_topology['cpu_count_max'] = int(vcpu_tag.text) if vcpu_tag.text else 0
        cpu_topology['cpu_count'] = int(vcpu_tag.get('current')) \
            if vcpu_tag.get('current') else cpu_topology['cpu_count_max']
        cpu_topology['cpu_sockets'] = cpu_topology['cpu_count']
        cpu_topology['cpu_cores'] = 1
        cpu_topology['cpu_threads'] = 1

    # пример "cpu_map": {"0": ["0", "1"],}, в xml <vcpupin vcpu='0' cpuset='0-1'/>
    for pin_tag in root.iterfind('cputune/vcpupin'):
        vcpu = pin_tag.get('vcpu')
        if vcpu is not None and int(vcpu) in list(range(cpu_topology['cpu_count_max'])):
            cpu_topology['cpu_map'][vcpu] = pin_tag.get('cpuset').split("-")

    cpu_tag = dict_tag(root, 'cpu')
    cpu_topology['cpu_mode'] = cpu_tag.get('mode') if cpu_tag.get('mode') else 'default'
    cpu_topology['cpu_model'] = root.find('cpu/model').text if \
        exist_tag(root, 'cpu/model') and cpu_topology['cpu_mode'] == 'custom' else 'default'

    if exist_tag(root, 'cpu/topology'):
        topology_tag = root.find('cpu/topology')
        cpu_topology['cpu_sockets'] = int(topology_tag.get('sockets'))
        cpu_topology['cpu_cores'] = int(topology_tag.get('cores'))
        cpu_topology['cpu_threads'] = int(topology_tag.get('threads'))

    domain_configuration.update(machine=DEFAULT_MACHINE)
    domain_configuration.update(arch=DEFAULT_ARCH)
    if exist_tag(root, 'os/type'):
        os_tag = root.find('os/type')
        machine = os_tag.get('machine')
        if machine and 'q35' in machine:
            domain_configuration['machine'] = 'q35'
        arch = os_tag.get('arch')
        if arch and 'e2k' in arch:
            domain_configuration['arch'] = arch

    if exist_tag(root, 'memory'):
        unit = root.find('memory').get('unit', 'KiB')
        memory_count = memory_units_convertor(int(root.find('memory').text), unit, DEFAULT_MEMORY_UNIT)
        domain_configuration.update(memory_count=memory_count)
    elif exist_tag(root, 'cpu/numa/cell'):
        unit = root.find('cpu/numa/cell').get('unit', 'KiB')
        memory_count = memory_units_convertor(int(root.find('cpu/numa/cell').get('memory')), unit, DEFAULT_MEMORY_UNIT)
        domain_configuration.update(memory_count=memory_count)

    if exist_tag(root, 'devices/tpm'):
        domain_configuration['tpm_device'] = True

    if exist_tag(root, 'devices/video/model'):
        video = {}
        video_model_tag = root.find('devices/video/model')
        video['type'] = video_model_tag.get('type')
        video['vram'] = int(video_model_tag.get('vram') if exist_tag(video_model_tag, 'vram') else 16 * 1024)
        video['heads'] = int(video_model_tag.get('heads') if exist_tag(video_model_tag, 'heads') else 1)

        domain_configuration.update(video=video)

    if exist_tag(root, 'devices/sound'):
        sound = {'model': root.find('devices/sound').get('model'), 'enabled': True}
        # для ich6 обязателен codec
        if sound['model'] == 'ich6':
            if root.find('devices/sound/codec'):
                sound['codec'] = root.find('devices/sound/codec').get('type')
            else:
                sound['codec'] = 'micro'
        domain_configuration.update(sound=sound)
    else:
        domain_configuration.update(sound={'enabled': False})

    if exist_tag(root, 'devices/graphics'):
        domain_configuration.update(remote_access=True)
        remote_access_port = root.find('devices/graphics').get('port', 'unknown')
        domain_configuration.update(remote_access_port=remote_access_port)

    if exist_tag(root, 'os/bootmenu'):
        bootmenu_tag = root.find('os/bootmenu')
        if bootmenu_tag and bootmenu_tag.get('enable') == 'yes':
            bootmenu = True
            bootmenu_timeout = int(bootmenu_tag.get('timeout')) / 1000 \
                if bootmenu_tag.get('timeout') else bootmenu_timeout

    if exist_tag(root, 'os/loader'):
        loader_tag = root.find('os/loader')
        if loader_tag.get('type') == 'pflash' and exist_tag(root, 'os/nvram'):
            boot_type = "UEFI"
            domain_configuration.update(boot_type=boot_type)

    for disk_tag in root.iterfind('devices/disk'):
        if disk_tag.get('type') in ['file', 'network']:
            # виртуальные диски
            if disk_tag.get('device') == 'disk':
                common_info = get_common_disk_info(disk_tag)
                source_tag = disk_tag.find('source')
                if source_tag.get('file'):
                    source = source_tag.get('file')
                elif source_tag.get('dev'):
                    source = source_tag.get('dev')
                elif disk_tag.get('type') == 'network' and source_tag.get('name'):
                    source = f"/storages/gluster/{source_tag.get('name')}"
                else:
                    continue
                disk_info = dict(source=source)
                disk_info.update(common_info)

                # ищем диск среди существующих
                disk_id = ''
                if restore_from_backup or show:
                    if '/vdisk.qcow2' in disk_info['source']:
                        disk_id = disk_info['source'].split('/')[-2]
                    elif show and '.snapshot' in disk_info['source']:
                        continue
                    elif '.snapshot' in disk_info['source']:
                        disk_id = disk_info['source'].split('/')[-1].split('__')[0]
                    else:
                        disk_id = disk_info['source'].split('/')[-1].split('.qcow2')[0]
                if disk_id:
                    boot_order = get_boot(disk_tag)
                    vdisks.append({'vdisk': disk_id,
                                   'source': source,
                                   'driver_cache': common_info.get('driver_cache'),
                                   'io': common_info.get('io'),
                                   'target_bus': common_info.get('target_bus'),
                                   'target_dev': common_info.get('target_dev'),
                                   'boot_order': boot_order
                                   })
                    if boot_order:
                        ordered_devices.update({disk_id: int(boot_order)})
                        ordered_devs.update({common_info['target_dev']: int(boot_order)})
                    else:
                        boot_free_devs.append(common_info['target_dev'])
            # cdrom-ы и примонтированные iso
            if disk_tag.get('device') == 'cdrom':
                cdrom_info = {}
                iso_info = {}

                # чтобы запомнить позицию устройства в порядке загрузки или обработать подключение iso
                # генерим временный id, при создании cdrom временный id в структуре boot должен быть
                # заменен на реальный id
                cdrom_id = str(uuid.uuid4())
                target_dev = disk_tag.find('target').get('dev')
                cdrom_info.update(cdrom=cdrom_id,
                                  target_dev=target_dev)

                # iso
                source = disk_tag.find('source').get('file') if exist_tag(disk_tag, 'source') else None
                if source and (restore_from_backup or show):
                    iso_id = str(uuid.uuid4())
                    iso_info.update(iso=iso_id, cdrom=cdrom_id, source=source)
                    cdrom_info.update(iso=iso_id, source=source)

                boot_order = get_boot(disk_tag)
                if boot_order:
                    ordered_devices.update({cdrom_id: int(boot_order)})
                    ordered_devs.update({target_dev: int(boot_order)})
                else:
                    boot_free_devs.append(target_dev)

                if iso_info:
                    isos.append(iso_info)
                cdroms.append(cdrom_info)

        # luns
        if disk_tag.get('type') == 'block' and disk_tag.get('device') in ['disk', 'lun']:
            common_info = get_common_disk_info(disk_tag)
            device = disk_tag.find('source').get('dev') if exist_tag(disk_tag, 'source') else None
            disk_info = dict(device=device)
            disk_info.update(common_info)
            boot_order = get_boot(disk_tag)

            # ищем LUN среди существующих
            if show:
                lun_id = str(uuid.uuid4())
                luns.append({'lun': lun_id,
                             'source': device,
                             'driver_cache': common_info.get('driver_cache'),
                             'io': common_info.get('io'),
                             'target_bus': common_info.get('target_bus'),
                             'target_dev': common_info.get('target_dev'),
                             })
                if boot_order:
                    ordered_devices.update({lun_id: int(boot_order)})
                    ordered_devs.update({common_info['target_dev']: int(boot_order)})
                else:
                    boot_free_devs.append(common_info['target_dev'])

    for hostdev_tag in root.iterfind('devices/hostdev'):
        # примонтированные usb устройства
        # сохраняем эту информацию для дальнейшей обработки на этапе диалога с пользователем
        # возможно понадобится дополнительный endpoint
        if hostdev_tag.get('mode') == 'subsystem':
            if hostdev_tag.get('type') == 'usb':
                if exist_tag(hostdev_tag, 'source/vendor') and exist_tag(hostdev_tag, 'source/product'):
                    usb_devices.append(dict(vendor=hostdev_tag.find('source/vendor').get('id').split('x')[1],
                                            product=hostdev_tag.find('source/product').get('id').split('x')[1]))
                elif exist_tag(hostdev_tag, 'source/address'):
                    address_tag = hostdev_tag.find('source/address')
                    bus, device = address_tag.get('bus'), address_tag.get('device')
                    usb_devices.append(dict(bus=bus, device=device))
            elif hostdev_tag.get('type') == 'pci':
                if exist_tag(hostdev_tag, 'source/address'):
                    bus = hostdev_tag.find('source/address').get('bus').split('x')[1]
                    slot = hostdev_tag.find('source/address').get('slot').split('x')[1]
                    function = hostdev_tag.find('source/address').get('function').split('x')[1]
                    pci_devices.append(dict(bus=bus,
                                            slot=slot,
                                            function=function))

    for interface_tag in root.iterfind('devices/interface'):
        # для подключения сетевых интерфейсов необходима информация о интерфейсных группах и, соответственно,
        # информация о ноде - требуется диалог с пользователем, на этапе конвертации ничего не делаем
        if interface_tag.get('type') == 'bridge':
            vmachine_inf_id = interface_tag.find('virtualport/parameters').get('interfaceid')
            boot_order = interface_tag.find('boot').get('order') if exist_tag(interface_tag, 'boot') else None
            nic_driver = interface_tag.find('model').get('type') if exist_tag(interface_tag, 'model') else None
            mac_addr = interface_tag.find('mac').get('address')
            switch = interface_tag.find('source').get('bridge')
            mtu = interface_tag.find('mtu').get('size')
            link_state = interface_tag.find('link').get('state')
            target_dev = interface_tag.find('target').get('dev')
            if exist_tag(interface_tag, 'address'):
                pci = dict(
                    domain=interface_tag.find('address').get('domain'),
                    bus=interface_tag.find('address').get('bus'),
                    slot=interface_tag.find('address').get('slot'),
                    function=interface_tag.find('address').get('function'))
                if not all(pci.values()):
                    pci = {}
            else:
                pci = {}

            # информация для определения группы интерфейсов
            # bridge_name = interface_tag.find('source').get('bridge') if exist_tag(interface_tag, 'source') else None
            if boot_order:
                ordered_devices.update({vmachine_inf_id: int(boot_order)})
                ordered_devs.update({target_dev: int(boot_order)})
            else:
                boot_free_devs.append(target_dev)
            vmachine_infs.append(dict(vmachine_inf=vmachine_inf_id,
                                      nic_driver=nic_driver,
                                      mac_address=mac_addr,
                                      pci=pci,
                                      switch=switch,
                                      mtu=mtu,
                                      link_state=link_state,
                                      target_dev=target_dev))

        # количество sriov
        if interface_tag.get('type') == 'hostdev':
            vfunction_infs += 1

    qemu_state = False
    for channel in root.iterfind('devices/channel'):
        if channel.get('type') == 'unix':
            target = channel.find('target')
            if target.get('name') == BASE_QEMU_CHANNEL and target.get('state') == 'connected':
                qemu_state = True
                break

    boot = dict(bootmenu=bootmenu,
                bootmenu_timeout=bootmenu_timeout,
                boot_type=boot_type)

    domain_configuration.update(vdisks=vdisks,
                                luns=luns,
                                cdroms=cdroms,
                                isos=isos,
                                vmachine_infs=vmachine_infs)

    boot['devices'] = [x[0] for x in sorted(ordered_devices.items(), key=itemgetter(1))] if ordered_devices else []
    boot['devs'] = [x[0] for x in sorted(ordered_devs.items(), key=itemgetter(1))] if ordered_devices else []
    boot['free_devs'] = boot_free_devs
    domain_configuration.update(boot=boot)
    domain_configuration.update(boot_devices=ordered_devices)

    additional_info.update(vfunction_infs_info=vfunction_infs,
                           pci_devices_info=pci_devices,
                           usb_devices_info=usb_devices,
                           cpu_topology_info=cpu_topology,
                           qemu_state=qemu_state)
    return dict(domain_configuration=domain_configuration, additional_info=additional_info)


def get_screenshot(domain_id):
    tmp_path = f'/tmp/{domain_id}'
    tmp_screenshot_path = f"{tmp_path}.screenshot"
    screenshot(domain_id, tmp_screenshot_path)
    tmp_png_path = f"{tmp_path}.png"
    convert_cmd = f"convert -resize 342x256 {tmp_screenshot_path} {tmp_png_path}"
    cmd_ret = get_output(convert_cmd)
    try:
        os.remove(tmp_screenshot_path)
    except OSError:
        pass
    if not cmd_ret['result']:
        comment = f"An error occurred while convert domain screenshot: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    with open(tmp_png_path, 'rb') as f:
        byte = f.read(1024000)
        yield byte
        while byte != b"":
            # Do stuff with byte.
            yield f.read(1024000)


@states_response
def wait_for_ga(domain_id: str, timeout: int = 300, raise_on_timeout: bool = True):
    step = 5
    spended_time = 0
    while True:
        if spended_time >= timeout:
            if raise_on_timeout:
                raise CommandExecutionError(
                    f'Domain {domain_id} qemu agent is not active for {str(timeout)} seconds.')
            else:
                break
        # Начало цикла со слипа неслучайно
        sleep(step)
        spended_time += step
        res = qemu_guest_command(domain_id, command='guest-info')
        if res['result']:
            break
    return spended_time


@states_response
def wait_for_delay(delay: int):
    sleep(delay)


@states_response
def get_domain_configuration(domain_id: str, show: bool = True):
    cmd_ret = get_domain_dumpxml(domain_id)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    domain_xml = cmd_ret['comment']
    dom_conf_ret = get_draft_domain_from_xml(domain_xml, show=show, domain_id=domain_id)
    if not dom_conf_ret['result']:
        raise CommandExecutionError(dom_conf_ret['comment'])
    return dom_conf_ret['comment']


def get_domains_haltpoll_time():
    """
    При kernel panic / BSOD фризится параметр cpu.haltpoll.fail.time
    мы это используем для детекта повисшей ВМ
    Нас интересуют строки:
        Domain: 'b17c8efe-af58-4800-b464-faeed6fd011f'
        cpu.haltpoll.fail.time=490062565
    """
    cmd = 'virsh domstats --cpu-total | grep "Domain\|fail.time"'
    raw_lines = get_output(cmd)['comment'].splitlines()
    result = list()
    for index, line in enumerate(raw_lines):
        if line.startswith('Domain'):
            domain = line.split()[-1].strip("'")
            fail_time = raw_lines[index + 1].split('=')[-1]
            result.append(dict(domain=domain, fail_time=fail_time))
    return result


def ping(address: str, interval: float = 0.002, size: int = 1, count: int = 5, wait_answer_sec: float = 0.5):
    """
    Функция пингует адрес
    Возравщает True если нет 100% потери пакетов
    """
    cmd = f'ping {address} -i {interval} -s {size} -c {count} -W {wait_answer_sec}'
    raw_lines = get_output(cmd, suppress_err=False)['comment'].splitlines()
    try:
        raw_packets, raw_received, raw_loss, raw_time = raw_lines[-2].split(',')
        return '100%' not in raw_loss
    except Exception:
        return False


def check_domain_availability(domain_conf, domain_statuses, redis_client):
    """
    Функция проверяет доступность VM по параметрам:
        - доступность Guest Agent
        - Пинг хоста, если задано
        - Параметр cpu.haltpoll.fail.time
    Если задано, то VM согласно настройкам будет перезагружаться. По достижению порога попыток - прекратит
    В  callback передаем метод отправки эвента
    """
    from veilrpc.states.lvm import _convert_size_mb
    current_time = int(time())
    domain_id = domain_conf.get('domain_id')
    uptime = 0
    if domain_id in domain_statuses.get('domain_uptime', dict()):
        uptime = int(domain_statuses['domain_uptime'][domain_id].get('uptime', 0))
    reboot_after_fail = domain_conf.get('_ha_reboot_after_fail')
    reboot_attempts = domain_conf.get('_ha_reboot_attempts')
    delay_after_fail = domain_conf.get('_ha_delay_after_fail')
    domain_boot_delay = domain_conf.get('_ha_domain_boot_delay')
    ping_address = domain_conf.get('_ha_ping_address')
    detect_fail_ga = domain_conf.get('_ha_detect_fail_ga')
    detect_fail_ping = domain_conf.get('_ha_detect_fail_ping')
    detect_fail_haltpoll = domain_conf.get('_ha_detect_fail_haltpoll')
    ga_status = domain_statuses['domain_ga_status'].get(domain_id, {})
    event = None, None

    if not any((detect_fail_haltpoll, detect_fail_ga, detect_fail_ping)):
        return None, event

    def is_os_present(config: dict) -> bool:
        """
        Проверка, есть ли ОС
        - Проверяется память на изменение после запуска вм
        - СDrom проверяется на наличие смонтированного iso
        - Проверка наличия загруочного диска, где может быть ОС
          (Условно ОС должна весить не менее 10мб, поэтому загрузочный vdisk проверяется на размер)
        :param config:
        :return:
        """
        if boot_devices := config.get('boot_devices'):
            mem = virt_memory_stats(config['domain_id'])
            if mem.get('last_update', 0) > 0:
                return True
            if isos := config.get('isos'):
                for iso in isos:
                    if iso.get('cdrom') in boot_devices:
                        return True
            if vdisks := config.get('vdisks'):
                for vdisk in vdisks:
                    if vdisk['vdisk'] in boot_devices:
                        cmd = f"qemu-img info \'{vdisk.get('source')}\' -U --output json"
                        try:
                            response = get_output(cmd)
                            if not response.get('retcode'):
                                size = _convert_size_mb(int(json.loads(response['comment']).get('actual-size')))
                                if size > 10:
                                    return True
                        except json.JSONDecodeError as e:
                            pass
                        except Exception as e:
                            pass
        return False

    def get_fail_duration(success_state: bool, redis_key: str) -> int:
        if not success_state and not redis_client.exists(redis_key):
            expire_time = 60 * 60
            redis_client.setex(redis_key, expire_time, current_time)
        elif success_state and redis_client.exists(redis_key):
            redis_client.delete(redis_key)
        return current_time - int(redis_client.get(redis_key) or current_time)

    def reboot_or_shutdown(domain_name: str, attempts: int):
        redis_key = REBOOT_ATTEMPTS_PREFIX + domain_name
        reboot_count = int(redis_client.get(redis_key) or 0)
        _event = None
        if reboot_count < attempts:
            _event = (domain_name, 'It seems like guest OS not responding. Restarting VM')
            reset(domain_name)
            redis_client.incr(redis_key)
            expire_time = 60 * 60
            redis_client.expire(redis_key, expire_time)
        else:
            _event = (domain_name, 'It seems like guest OS not responding. Reboot limit exceeded. Shutting down VM')
            shutdown(domain_name, force=True)
            redis_client.delete(redis_key)
        return _event

    def get_domains_haltpoll_changed() -> dict:
        result = dict()
        source = get_domains_haltpoll_time()
        for item in source:
            _domain_id = item['domain']
            key = HALTPOLL_FAIL_VALUE_PREFIX + _domain_id
            prev_value = int(redis_client.get(key) or 0)
            current_value = int(item['fail_time'])
            result[item['domain']] = current_value != prev_value
            redis_client.setex(key, 60 * 10, current_value)
        return result

    haltpoll_time_changes = get_domains_haltpoll_changed()
    fails = dict(GA=[bool(value) for value in ga_status.values()].count(True) > 1,
                 HALTPOLL=haltpoll_time_changes.get(domain_id))
    if ping_address:
        fails['PING'] = ping(ping_address)
        duration_of_no_ping = get_fail_duration(success_state=fails.get("PING"), redis_key=PING_FAIL_PREFIX + domain_id)
    failed_haltpoll_duration = get_fail_duration(success_state=fails["HALTPOLL"], redis_key=HALTPOLL_FAIL_PREFIX + domain_id)
    ga_duration = get_fail_duration(success_state=fails["GA"], redis_key=GA_FAIL_PREFIX + domain_id)

    if uptime > domain_boot_delay:
        reasons_to_reboot = []
        if detect_fail_ga:
            reasons_to_reboot.append(ga_duration >= delay_after_fail)
        if detect_fail_ping and ping_address:
            reasons_to_reboot.append(duration_of_no_ping >= delay_after_fail)
        if detect_fail_haltpoll:
            reasons_to_reboot.append(failed_haltpoll_duration >= delay_after_fail)
        if all(reasons_to_reboot):
            if reboot_after_fail:
                event = reboot_or_shutdown(domain_id, reboot_attempts)
            else:
                if is_os_present(config=domain_conf):
                    event = (domain_id, 'It seems like guest OS not responding')
    statuses = {'ga_fail_time': int(redis_client.get(GA_FAIL_PREFIX + domain_id) or 0),
                'ping_fail_time': int(redis_client.get(PING_FAIL_PREFIX + domain_id) or 0),
                'haltpoll_fail_time': int(redis_client.get(HALTPOLL_FAIL_PREFIX + domain_id) or 0),
                'reboots_count': int(redis_client.get(REBOOT_ATTEMPTS_PREFIX + domain_id) or 0)}
    return statuses, event


class CommonDataOperator:
    """
    Класс для реализации единой точки доступа к metadata по ВМ.
    """
    def __init__(self, domain_id: str, xml_operator: bool = True, libvirt_operator: bool = True):
        self.domain_id = domain_id
        self.libvirt_operator = LibvirtDataOperator(self.domain_id) if libvirt_operator else None
        self.xml_operator = XMLDataOperator(self.domain_id) if xml_operator else None
        self.operators = [operator for operator in (self.libvirt_operator, self.xml_operator) if operator]
        self.frequently_used_fields = {
            "power_state_expected": [
                DataOperator.StepTup('metadata', False, 0),
                DataOperator.StepTup('veil:instance', True, 0),
                DataOperator.StepTup('veil:power_state_expected', True, 0)],
            "metadata": [
                DataOperator.StepTup('metadata', False, 0),
                DataOperator.StepTup('veil:instance', True, 0)],
        }
        self.log = logging.getLogger('bazalt')

    def create_domain_info(self, xml_data: str) -> None:
        """
        Создание/перезапись xml файла.
        :param xml_data: xml строка.
        """
        for operator in self.operators:
            try:
                operator.create_domain_info(xml_data)
            except NotImplementedError:
                pass
            except Exception as err:
                self.log.error(err)
                return None

    def update_power_state_expected(self, value: str) -> None:
        """
        Изменение состояния поля power_state_expected
        :param value: новое значение поля.
        """
        for operator in self.operators:
            try:
                operator.update_power_state_expected(value)
            except NotImplementedError:
                pass
            except Exception as err:
                self.log.error(err)
                return None

    def update_metadata(self, xml: str) -> None:
        """
        Обновление поля metadata как одиночного поля (one_field)
        :param xml: новое значение для существующей ранее записи.
        """
        for operator in self.operators:
            try:
                operator.update_one_field('metadata', xml)
            except NotImplementedError:
                pass
            except Exception as err:
                self.log.error(err)
                return None

    def get_field_info(self, path: list[DataOperator.StepTup], update_raw_data: bool = False) -> Any:
        """
        Получение информации по отдельному полю. Из всех источников информации наиболее доверенным считается XML файл.
        :param path: путь к полю в xml структуре.
        :param update_raw_data: обновить данные, полученные при инициализации объекта.
        :return: значение поля.
        """
        try:
            return self.xml_operator.get_field_info(path, update_raw_data)
        except NotImplementedError:
            pass
        except Exception as err:
            self.log.error(err)
            return None

    def update_ha_enabled(self, var: bool):
        """
        Изменение состояния поля ha_enabled
        :param var: новое значение поля.
        """
        for operator in self.operators:
            try:
                operator.update_ha_enabled(var)
            except NotImplementedError:
                pass
            except Exception as err:
                self.log.error(err)
                return None
