"""
(c) 2023 LLC «DACOM M»
"""

import os
import re
import textwrap

from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output, states_result

FC_PATH = '/sys/class/fc_transport/'
FC_HOST_PATH = '/sys/class/fc_host/'
FC_WWN_REGEX = re.compile(r'(0x)?([0-9a-fA-F]{16})')


@states_response
def connected(sources, vendor=None, vendor_options=None,
              multipathing=False, multipathing_mode='failover'):
    # TODO: реализовать vendor specific options
    wwns = get_connected_wwns()
    if multipathing and multipathing_mode == 'failover':
        if all([source not in wwns for source in sources]):
            comment = f"The node is not connected to storage with sources: {', '.join(sources)}."
            raise CommandExecutionError(comment)
    else:
        for source in sources:
            if source not in wwns:
                comment = f'The node is not connected to storage with source: {source}.'
                raise CommandExecutionError(comment)
    return f"The node is connected to storage with sources: {', '.join(sources)}."


def get_connected_wwns(**kwargs):
    result = []
    if os.path.exists(FC_PATH) and os.listdir(FC_PATH):
        cmd = f'cat {FC_PATH}*/port_name'
        cmd_ret = get_output(cmd, **kwargs)
        if cmd_ret['result']:
            for wwn in cmd_ret['comment'].splitlines():
                # wwn - строка в 16ричной системе счисления, выполняем проверку ее корректности
                if re.match(FC_WWN_REGEX, wwn):
                    result.append(wwn.lower())
    return result


@states_response
def connected_wwns(**kwargs):
    return get_connected_wwns(**kwargs)


@states_response
def discovery():
    cmd = "for i in $(find /sys/class/scsi_host/* -maxdepth 1); do echo '- - -' > $i/scan; done"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return 'Successfully scan scsi hosts.'


def get_local_wwns(**kwargs):
    result = []
    if os.path.exists(FC_HOST_PATH) and os.listdir(FC_HOST_PATH):
        cmd = f'cat {FC_HOST_PATH}*/port_name'
        cmd_ret = get_output(cmd, **kwargs)
        if cmd_ret['result']:
            for wwn in cmd_ret['comment'].splitlines():
                res = re.match(FC_WWN_REGEX, wwn).groups()
                result.append('-'.join(textwrap.wrap(res[-1], 2)).lower())
    return result


@states_response
def local_wwns(**kwargs):
    return get_local_wwns(**kwargs)


@states_response
def scan_scsi_bus():
    return states_result(get_output, cmd="rescan-scsi-bus.sh")


@states_response
def scan_vhba():
    to_rescan = list()
    out = ''
    lsscsi = get_output('lsscsi -g -D')
    if lsscsi['result']:
        report_to_scan = "\nVHBAs detected on ports:"
        for line in lsscsi['comment'].splitlines():
            words = line.split()    # scsi_addr ([h:c:t:l]), type (hex), ..., sg_name (/dev/sgXX)
            if words[1] == '0xd':   # if device type is enclosure
                inquiry = get_output(f'sg_inq -u {words[-1]}')  # scsi_generic device inquiry in KEY=val format
                if inquiry['result'] and 'SCSI_TYPE=disk' in inquiry['comment'].splitlines():
                    to_rescan.append(words[0].strip('[]'))
                    report_to_scan += f'\n{words[0]}'
        out += report_to_scan
    else:
        return "Unable to list scsi devices (lsscsi error)\n"

    if to_rescan:
        get_output(f'rescan-scsi-bus.sh -r --hosts={",".join({x.split(":")[0] for x in to_rescan})}')
        report_new_devices = "\nNew disks found:\n"
        for device in to_rescan:
            rep = get_output(f'lsscsi -s {device}')
            if rep['result']:
                report_new_devices += f'{rep["comment"]}'
        out += report_new_devices
    else:
        return "No suitable devices found\n"
    return out
