"""
(c) 2023 LLC «DACOM M»
"""
import subprocess

import os
import re
import time
import logging
from subprocess import run, CalledProcessError, PIPE, DEVNULL

from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output, call, \
    TransportProtocols, get_result, action_service, states_result, RUN_PATH, LUN_PREFIX
from veilrpc.utils.size import get_megabytes

ISCSIADM = 'sudo /sbin/iscsiadm -m'
TIMEOUT = 'timeout -s 9 10'
REMOVE_BLOCK_DEVICE_CMD = "echo 1 > /sys/block/{}/device/delete 2>&-"
MULTIPATHD = "multipathd show paths format"
LUNS_CMD = f"{MULTIPATHD} \'%P %s %m %w %z %S %t %n %r %N %R %d %a %i %o %p %C %D %T\'"
CLI_ISCSI_LUNS_CMD = f"{MULTIPATHD} \'%P %s %m %w %z %S %t %n %d %a %i %o %p %C %D %T\' | grep " \
                     f"-e iscsi -e protocol || true"
CLI_FC_LUNS_CMD = f"{LUNS_CMD} | grep -e fcp -e protocol || true"
ISCSI_LUNS_CMD = f"{MULTIPATHD} '%w %S %t %n %z %d %s'"
FC_LUNS_CMD = f"{MULTIPATHD} '%w %S %t %r %z %d %s'"
FC_WWN_REGEX = re.compile(r'(0x)?[0-9a-fA-F]{16}')
SCSI_WWN_REGEX = re.compile(r'\[undef\]')
ISCSI_TARGET_REGEX = re.compile(r'iqn\.[\d]{4}-[\d]{2}\.[A-Za-z0-9-_.]+(:[A-Za-z0-9-_.]+){,2}|'
                                r'eui\.[0-9a-fA-F]{16}|'
                                r'naa\.[0-9a-fA-F]{32}|naa\.[0-9a-fA-F]{16}')
MULTIPATHD_MAPS_CMD = "multipathd show maps format '%w'"
LOG = logging.getLogger('bazalt')


def _connect(target, source, username, password):
    error = _create_iscsi_target_record(target, source, username, password)
    if not error:
        cmd = f'{ISCSIADM} node --targetname {target} --portal {source} -l'
        cmd_ret = get_output(cmd)
        if not cmd_ret['result'] and 'already present' not in cmd_ret['comment']:
            cmd = f'{ISCSIADM} node --targetname {target} --portal {source} -o delete'
            call(cmd)
            comment = f"Can not connect to the target {target} with source {source}: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
        cmd = f'{ISCSIADM} node --targetname {target} --portal {source} -o update -n node.startup -v automatic'
        call(cmd)
    else:
        raise CommandExecutionError(error)
    return f"Successfully connected to the target {target} with source {source}"


@states_response
def connected(target, sources, username=None, password=None, vendor=None, vendor_options=None,
              multipathing=False, multipathing_mode='failover'):
    # TODO: реализовать vendor specific options
    res = ''
    if vendor in ['Yadro', 'Other_Multipath']:
        if not vendor_options:
            raise CommandExecutionError("Please write vendor_options in database for Yadro vendor")
        if isinstance(vendor_options, dict):
            for real_target, real_sources in vendor_options.items():
                for real_source in real_sources:
                    res += _connect(real_target, real_source, username, password)
        elif isinstance(vendor_options, list):
            for option in vendor_options:
                real_target = option['target']
                for real_source in option['address']:
                    res += _connect(real_target, real_source, username, password)
    else:
        for source in sources:
            res += _connect(target, source, username, password)
    return res


def _disconnect(target, source):
    cmd = f'{ISCSIADM} node --targetname {target} --portal {source} -u'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result'] and cmd_ret.get('retcode', 0) != 21:
        comment = f"No connections to the target {target} with source {source} found: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    cmd = f'{ISCSIADM} node --targetname {target} --portal {source} -o delete'
    call(cmd)
    return f"Successfully disconnected from the target {target} with source {source}"


@states_response
def disconnected(target, sources, vendor=None, vendor_options=None):
    # TODO: реализовать vendor specific options
    # TODO: реализовать multipathing
    res = ''
    if vendor in ['Yadro', 'Other_Multipath']:
        if not vendor_options:
            raise CommandExecutionError("Please write vendor_options in database for Yadro vendor")
        if isinstance(vendor_options, dict):
            for real_target, real_sources in vendor_options.items():
                for real_source in real_sources:
                    res += _disconnect(real_target, real_source)
        elif isinstance(vendor_options, list):
            for option in vendor_options:
                real_target = option['target']
                for real_source in option['address']:
                    res += _disconnect(real_target, real_source)
    else:
        for source in sources:
            res += _disconnect(target, source)

    return res


def _create_iscsi_target_record(target, source, username, password):
    # очистка базы на контроллере (нужна была при каких-то катаклизмах на таргете) приводит к удалению из базы текущих
    # записей, т.к. на контроллере после discover восстанавливается автостарт. Изменённые записи имеют node.tpgt = 1,
    # в отличие от свежесозданных с node.tpgt = -1, которые не трогаются удалением. на нодах всё норм. Заменяем
    # очистку всей базы на удаление соответствующей записи
    cleandb = f'{ISCSIADM} node --targetname {target} --portal {source} -o delete'
    call(cleandb)
    cmd = f'{ISCSIADM} node --targetname {target} --portal {source} -o new'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        return cmd_ret['comment']
    if username and password:
        cmd = f"{ISCSIADM} node --targetname {target} --portal {source} " \
              f"-o update -n node.session.auth.authmethod -v CHAP " \
              f"-o update -n node.session.auth.username -v {username} " \
              f"-o update -n node.session.auth.password -v {password}"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            return cmd_ret['comment']


@states_response
def iscsi_session_rescan(as_root=True, **kwargs):
    if as_root:
        cmd = f"{ISCSIADM} session --rescan"
    else:
        cmd = "/sbin/iscsiadm -m session --rescan"
    call(cmd, **kwargs)
    time.sleep(5)


@states_response
def multipath_topology(**kwargs):
    cmd_ret = get_output('multipath -ll', **kwargs)
    if not cmd_ret['result']:
        comment = f"An error occurred while getting multipath_topology: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def get_luns(**kwargs):
    cmd_ret = get_output(LUNS_CMD, **kwargs)
    if not cmd_ret['result']:
        comment = f"An error occurred while scanning luns: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def get_iscsi_luns(**kwargs):
    cmd_ret = get_output(CLI_ISCSI_LUNS_CMD, **kwargs)
    if not cmd_ret['result']:
        comment = f"An error occurred while scanning iscsi luns: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def get_fc_luns(**kwargs):
    cmd_ret = get_output(CLI_FC_LUNS_CMD, **kwargs)
    if not cmd_ret['result']:
        comment = f"An error occurred while scanning fc luns: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def device_fs_type(device):
    cmd_ret = run(['blkid', '-p', '-o', 'value', '-s', 'TYPE', device],
                  shell=False, text=True, check=False, timeout=50, stderr=DEVNULL, stdout=PIPE)
    if cmd_ret.returncode != 0:
        if cmd_ret.returncode == 8:  # ambivalent result -> more than one FS traces found
            try:
                wipefs_ret = run(['wipefs', '-p', device], shell=False, text=True, check=True, stdout=PIPE)
            except CalledProcessError as e:  # this should not occur when blkid returned 8 but ...
                raise CommandExecutionError(e.stderr)
            fs_list = [line.split(',') for line in wipefs_ret.stdout.splitlines(keepends=False)]
            if [line for line in fs_list if line[-1] == 'gfs2']:
                return 'gfs2'  # here we give priority to gfs2 since other choices seem useless:
                # ocfs2 is deprecated,
                # ext4 etc. is not used in on luns in Space,
                # zfs if found along with other FS must be already overwritten & broken. zfs has many superblocks
                # and if some are alive and there's gfs2 superblock probably gfs2 was last created
        else:
            return ''
    else:
        if cmd_ret.stdout is None:
            return ''
        else:
            return 'zfs' if ((fs_type := cmd_ret.stdout.strip()) == 'zfs_member') else fs_type


@states_response
def multipath_maps(**kwargs):
    maps = states_result(get_output, cmd=MULTIPATHD_MAPS_CMD, **kwargs).splitlines()
    return maps[1:] if len(maps) > 1 else []


@states_response
def scan_luns(db_target, protocol, rescan=True, vendor=None, get_fs_type=False, **kwargs):
    sys_paths = os.listdir('/etc/sys_mpath') if os.path.isdir('/etc/sys_mpath') else []
    luns_found = {}
    if protocol == TransportProtocols.iscsi:
        target_regex = ISCSI_TARGET_REGEX
        if rescan:
            iscsi_session_rescan(**kwargs)
        find_luns_cmd = ISCSI_LUNS_CMD
    elif protocol == TransportProtocols.fc:
        target_regex = FC_WWN_REGEX
        find_luns_cmd = FC_LUNS_CMD
    else:
        return luns_found
    cmd_ret = get_output(find_luns_cmd, **kwargs)
    if not cmd_ret['result']:
        comment = f"An error occurred while scanning luns: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)

    luns_info = cmd_ret['comment'].splitlines()[1:]
    for lun_info in luns_info:
        try:
            if lun_info:
                try:
                    (device_id,
                     size,
                     state,
                     target,
                     serial_number,
                     device,
                     vend_prod_rev) = re.split(r'\s+', lun_info, maxsplit=6)
                except ValueError:
                    continue
                if device_id in sys_paths:
                    continue
                if not re.match(target_regex, target):
                    continue
                if (isinstance(db_target, list) and target in db_target) or target == db_target:
                    device = f"{LUN_PREFIX}/{device_id}"
                    mount_cmd = "mount -l | awk '{{print $1}}' | grep -e '^{}$'".format(device)
                    mounted = get_result(mount_cmd, **kwargs)
                    fs_type = ''
                    if get_fs_type:
                        fs_type_res = device_fs_type(device)
                        if fs_type_res['result']:
                            fs_type = fs_type_res['comment']
                    max_size = size
                    max_size_convert = get_megabytes(size)
                    if max_size_convert <= luns_found.get(device_id, {}).get('size_convert', 0):
                        max_size = luns_found.get(device_id, {}).get('size', '0K')
                        max_size_convert = luns_found.get(device_id, {}).get('size_convert', 0)
                    verbose_name = device
                    try:
                        # VPD82 is an optional page reporting some text info about SAN & LUN
                        # (say, mgmt IP, LUN name in SAN etc) so we place it into verbose_name if available
                        p82 = subprocess.check_output(['sg_vpd', '-p', '0x82', '-r', device], stderr=subprocess.DEVNULL,
                                                      timeout=5)
                        if (p82_text := p82[5:ord(p82[4:5]) + 5].decode(errors='replace')).strip():
                            verbose_name = p82_text
                    except (subprocess.TimeoutExpired, subprocess.CalledProcessError):
                        pass
                    luns_found[device_id] = dict(size=max_size,
                                                 size_convert=max_size_convert,
                                                 state=state,
                                                 serial_number=serial_number,
                                                 mounted=mounted,
                                                 fs_type=fs_type,
                                                 vend_prod_rev=vend_prod_rev,
                                                 verbose_name=verbose_name)
        except Exception as error:
            comment = f"An error occurred while scanning luns. Data: {lun_info}. Error: {error}"
            raise CommandExecutionError(comment) from None
    return luns_found


@states_response
def transport_luns(protocol, rescan=True, **kwargs):
    luns_found = list()
    if protocol == TransportProtocols.iscsi:
        target_regex = ISCSI_TARGET_REGEX
        if rescan:
            iscsi_session_rescan(**kwargs)
        find_luns_cmd = ISCSI_LUNS_CMD
    elif protocol == TransportProtocols.fc:
        target_regex = FC_WWN_REGEX
        find_luns_cmd = FC_LUNS_CMD
    elif protocol == TransportProtocols.scsi:
        target_regex = SCSI_WWN_REGEX
        find_luns_cmd = FC_LUNS_CMD
    else:
        return luns_found
    cmd_ret = get_output(find_luns_cmd, **kwargs)
    if not cmd_ret['result']:
        comment = f"An error occurred while scanning luns: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)

    luns_info = cmd_ret['comment'].splitlines()[1:]
    for lun_info in luns_info:
        try:
            if lun_info:
                try:
                    (device_id,
                     size,
                     state,
                     target,
                     serial_number,
                     device,
                     vend_prod_rev) = re.split(r'\s+', lun_info, maxsplit=6)
                except ValueError:
                    continue
                if not re.match(target_regex, target):
                    continue
                cmd = "mount -l | awk '{{print $1}}' | grep -e '^{}/{}$'".format(LUN_PREFIX, device_id)
                mounted = get_result(cmd, **kwargs)
                luns_found.append(dict(size=size,
                                       size_convert=get_megabytes(size),
                                       state=state,
                                       target=target,
                                       serial_number=serial_number,
                                       mounted=mounted,
                                       device_id=device_id,
                                       vend_prod_rev=vend_prod_rev,
                                       serial_and_id=f"{serial_number}___{device_id}"))
        except Exception as error:
            comment = f"An error occurred while scanning luns. Data: {lun_info}. Error: {error}"
            raise CommandExecutionError(comment) from None
    return luns_found


@states_response
def discovery(hostname, port='3260'):
    ret = set()
    cmd_ret = call(f'netcat -w 3 -z {hostname} {port}')
    if not cmd_ret['result']:
        raise CommandExecutionError(f'Hostname {hostname} is not available')

    # сохраняем список таргетов на ноде и их состояний автостарта (node.startup)
    # WARNING! если нет записей, команда возвращает 21, это нормально
    get_nodes_cmd = f'{TIMEOUT} {ISCSIADM} node 2>&-'
    tgt_status_dict = {}
    try:
        cmd_ret = get_output(get_nodes_cmd)
    except CommandExecutionError as e:
        if cmd_ret['retcode'] != 21:
            raise CommandExecutionError(str(e)) from None
    if (cmd_ret.get('retcode') in [None, 0]) and cmd_ret['comment']:
        for tgt in [x.split()[1] for x in cmd_ret['comment'].splitlines()]:
            get_tgt_state_cmd = f'{ISCSIADM} node -T {tgt} | grep node.startup'
            tgt_status_dict[tgt] = get_output(get_tgt_state_cmd)['comment'].split()[-1]

    cmd = f'{TIMEOUT} {ISCSIADM} discovery -t st -p {hostname}:{port}'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(f'Can not scan {hostname}:{port} for targets. error: {cmd_ret["comment"]}')
    out = cmd_ret['comment'].splitlines()

    # удаляем ненужные таргеты из базы iscsiadm
    for line in out:
        _, target = line.split()
        ret.add(target)
        # не удаляем из базы, если таргеты были в базе до дискавера
        if not tgt_status_dict.get(target):
            cmd = f'{ISCSIADM} node --targetname {target} -o delete'
            call(cmd)

    # восстанавливаем состояние автостарта
    for tgt in tgt_status_dict:
        restore_cmd = f'{ISCSIADM} node -T {tgt} -o update -n node.startup -v {tgt_status_dict[tgt]}'
        call(restore_cmd)

    return sorted(list(ret))


def parse_host_port(s, default_port=None):
    if s[-1] == ']':
        # ipv6 literal (with no port)
        return s, default_port

    out = s.rsplit(":", 1)
    if len(out) == 1:
        # No port
        port = default_port
    else:
        try:
            port = int(out[1])
        except ValueError:
            raise ValueError(f"Invalid host:port {s}") from None

    return out[0], port


@states_response
def discovery_many(hostnames: list):
    ret = set()
    for hostname in hostnames:
        try:
            host, port = parse_host_port(hostname, 3260)
        except ValueError as e:
            raise CommandExecutionError(str(e)) from None
        cmd_ret = call(f'netcat -w 3 -z {host} {port}')
        if not cmd_ret['result']:
            raise CommandExecutionError(f'Hostname {hostname} is not available')

    # сохраняем список таргетов на ноде и их состояний автостарта (node.startup)
    # WARNING! если нет записей, команда возвращает 21, это нормально
    get_nodes_cmd = f'{TIMEOUT} {ISCSIADM} node 2>&-'
    tgt_status_dict = {}
    try:
        cmd_ret = get_output(get_nodes_cmd)
    except CommandExecutionError as e:
        if cmd_ret['retcode'] != 21:
            raise CommandExecutionError(str(e)) from None
    if (cmd_ret.get('retcode') in [None, 0]) and cmd_ret['comment']:
        for tgt in [x.split()[1] for x in cmd_ret['comment'].splitlines()]:
            get_tgt_state_cmd = f'{ISCSIADM} node -T {tgt} | grep node.startup'
            tgt_status_dict[tgt] = get_output(get_tgt_state_cmd)['comment'].split()[-1]

    for hostname in hostnames:
        cmd = f'{TIMEOUT} {ISCSIADM} discovery -t st -p {hostname}'
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(f'Can not scan {hostname} for targets. error: {cmd_ret["comment"]}')
        out = cmd_ret['comment'].splitlines()

        # удаляем ненужные таргеты из базы iscsiadm
        for line in out:
            _, target = line.split()
            ret.add(target)
            # не удаляем из базы, если таргеты были в базе до дискавера
            if not tgt_status_dict.get(target):
                cmd = f'{ISCSIADM} node --targetname {target} -o delete'
                call(cmd)

    # восстанавливаем состояние автостарта
    for tgt in tgt_status_dict:
        restore_cmd = f'{ISCSIADM} node -T {tgt} -o update -n node.startup -v {tgt_status_dict[tgt]}'
        call(restore_cmd)

    return sorted(list(ret))


@states_response
def initiator_name(**kwargs):
    if os.path.exists('/etc/iscsi/initiatorname.iscsi'):
        cmd_ret = get_output("cat /etc/iscsi/initiatorname.iscsi | grep InitiatorName=", **kwargs)
        if cmd_ret['result']:
            return cmd_ret['comment'].split('=')[1].strip()
        else:
            raise CommandExecutionError(cmd_ret['comment'])
    else:
        raise CommandExecutionError('initiatorname does not exist')


@states_response
def modify_initiator_name(name, **kwargs):
    with open('/etc/iscsi/initiatorname.iscsi', 'w') as f:
        f.write(f'InitiatorName={name}')
    cmd_ret = action_service('iscsid', 'restart', **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f'Successfully update InitiatorName to {name}'


@states_response
def queue_depth(**kwargs):
    if os.path.exists('/etc/iscsi/iscsid.conf'):
        cmd_ret = get_output("cat /etc/iscsi/iscsid.conf | grep node.session.queue_depth", **kwargs)
        if cmd_ret['result']:
            return cmd_ret['comment'].split('=')[1].strip()
        else:
            raise CommandExecutionError(cmd_ret['comment'])
    else:
        raise CommandExecutionError('iscsid.conf does not exist')


@states_response
def clear_iscsiadm_base(**kwargs):
    ret_1 = states_result(get_output, cmd=f"{ISCSIADM} node --logout")
    ret_2 = states_result(get_output, cmd=f"{ISCSIADM} node -o delete")
    return f"{ret_1}; {ret_2}"


@states_response
def get_iscsi_configs(**kwargs):
    return states_result(get_output, cmd=r"grep -r node\.startup /etc/iscsi/nodes/")


@states_response
def copy_data(source_device, target_device, task, **kwargs):
    from .copy import popen_with_progress
    from tempfile import mkstemp
    fh_stderr, f_stderr = mkstemp(dir=RUN_PATH, text=True)
    cmd = f'(pv -n {source_device} | dd iflag=fullblock conv=sync,noerror of={target_device} bs=8M 2>{f_stderr}) 2>&1'
    progress_data = dict(type='progress', jid=task, id='', name='copy_data', result=True)
    p = popen_with_progress(cmd, "pv", progress_data, kwargs.get('redis', None))
    with os.fdopen(fh_stderr, 'r') as f:
        f.seek(0)
        raw_lines = f.read()
    os.remove(f_stderr)
    lines = raw_lines.splitlines()
    if not lines or (p.returncode and lines[0] != f"dd: error writing '{target_device}': No space left on device"):
        raise CommandExecutionError(f'Failed to copy_data to device {target_device} '
                                    f'from {source_device}: {raw_lines or p.comment}')
    return lines[-1]  # там статистика dd, сколько скопировано и скорость
