"""
(c) 2023 LLC «DACOM M»
"""

import os

from uuid import uuid4
from typing import Optional, List, Union
from rtslib import (RTSRoot, BlockStorageObject, FileIOStorageObject, PSCSIStorageObject, FabricModule, Target, TPG,
                    NetworkPortal, NodeACL, LUN, MappedLUN)
from veilrpc.utils.bazalt import states_response, CommandExecutionError

target_save_directory = '/etc/target'


def check_save_directory():
    try:
        if not os.path.exists(target_save_directory):
            os.makedirs(target_save_directory, exist_ok=True)
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while check or create directory for save a configuration file '
                                    f'Linux-IO Target. Error: {e}') from None


def get_root() -> RTSRoot:
    check_save_directory()
    try:
        root = RTSRoot()
        root.save_to_file()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while get root Linux-IO Target. Error: {e}') from None
    return root


def save_root() -> RTSRoot:
    check_save_directory()
    try:
        root = RTSRoot()
        root.save_to_file()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while get root Linux-IO Target. Error: {e}') from None
    return root


@states_response
def lio_conf() -> dict:
    return get_root().dump()


def get_list_storages() -> List[Union[BlockStorageObject, FileIOStorageObject, PSCSIStorageObject]]:
    return [_ for _ in get_root().storage_objects]


@states_response
def list_storages_conf() -> List[dict]:
    return [_.dump() for _ in get_list_storages()]


def get_list_block_storages() -> List[BlockStorageObject]:
    return [s_obj for s_obj in get_root().storage_objects if isinstance(s_obj, BlockStorageObject)]


@states_response
def list_block_storages_conf() -> List[dict]:
    return [_.dump() for _ in get_list_block_storages()]


def get_storage(name: Optional[str] = None, dev: Optional[str] = None, wwn: Optional[str] = None
                ) -> Union[BlockStorageObject, FileIOStorageObject, PSCSIStorageObject, None]:
    if any((name, dev, wwn)):
        for storage in get_root().storage_objects:
            if any((storage.name == name, storage.udev_path == dev, storage.wwn == wwn)):
                return storage


def delete_storage(name: Optional[str] = None, dev: Optional[str] = None, wwn: Optional[str] = None) -> None:
    if any((name, dev, wwn)):
        for storage in get_root().storage_objects:
            if any((storage.name == name, storage.udev_path == dev, storage.wwn == wwn)):
                storage.delete()


@states_response
def storage_conf(name: Optional[str] = None, dev: Optional[str] = None, wwn: Optional[str] = None) -> dict:
    storage = get_storage(name, dev, wwn)
    if storage:
        return storage.dump()
    else:
        return {}


@states_response
def block_storage_conf(name: Optional[str] = None, dev: Optional[str] = None, wwn: Optional[str] = None) -> dict:
    storage = get_storage(name, dev, wwn)
    if storage and isinstance(storage, BlockStorageObject):
        return storage.dump()
    else:
        return {}


@states_response
def pscsi_storage_conf(name: Optional[str] = None, dev: Optional[str] = None, wwn: Optional[str] = None) -> dict:
    storage = get_storage(name, dev, wwn)
    if storage and isinstance(storage, PSCSIStorageObject):
        return storage.dump()
    else:
        return {}


@states_response
def fileio_storage_conf(name: Optional[str] = None, dev: Optional[str] = None, wwn: Optional[str] = None) -> dict:
    storage = get_storage(name, dev, wwn)
    if storage and isinstance(storage, FileIOStorageObject):
        return storage.dump()
    else:
        return {}


def get_list_targets() -> List[Target]:
    return [_ for _ in get_root().targets]


@states_response
def list_targets_conf() -> List[dict]:
    return [_.dump() for _ in get_list_targets()]


@states_response
def list_targets_wwn() -> List[str]:
    return [t.wwn for t in get_root().targets]


def get_target(wwn: str) -> Optional[Target]:
    for target in get_root().targets:
        if wwn == target.wwn:
            return target


def delete_target(wwn: str) -> None:
    for target in get_root().targets:
        if wwn == target.wwn:
            target.delete()


@states_response
def list_target_port_groups_conf(wwn: str) -> List[str]:
    for target in get_root().targets:
        if target.wwn == wwn:
            return [tpg.dump() for tpg in target.tpgs]
    return []


def get_tpg(wwn: str, tag: int) -> Optional[TPG]:
    for target in get_root().targets:
        if wwn == target.wwn:
            for tpg in target.tpgs:
                if tpg.tag == tag:
                    return tpg


def delete_tpg(wwn: str, tag: int) -> None:
    for target in get_root().targets:
        if wwn == target.wwn:
            for tpg in target.tpgs:
                if tpg.tag == tag:
                    tpg.delete()


def get_list_portals(wwn: str, tag: int) -> List[NetworkPortal]:
    tpg = get_tpg(wwn, tag)
    if tpg:
        return [_ for _ in tpg.network_portals]


@states_response
def list_portals_conf(wwn: str, tag: int) -> List[dict]:
    return [_.dump() for _ in get_list_portals(wwn, tag)]


def get_list_luns(wwn: str, tag: int) -> List[LUN]:
    tpg = get_tpg(wwn, tag)
    if tpg:
        return [_ for _ in tpg.luns]
    return []


def get_lun(tpg: Optional[TPG] = None, wwn: Optional[str] = None, tag: Optional[int] = None,
            lun_index: Optional[int] = None, lun_alias: Optional[str] = None) -> Optional[LUN]:
    if tpg is None:
        tpg = get_tpg(wwn, tag)
    for lun in tpg.luns:
        if (lun_alias and lun.alias == lun_alias) or (lun_index and int(lun.path.split('_')[-1])):
            return lun
        elif not lun_alias and not lun_index:
            return lun


def get_nodeacl(tpg: Optional[TPG] = None, wwn: Optional[str] = None, tag: Optional[int] = None,
                node_wwn: Optional[str] = None) -> Optional[NodeACL]:
    if tpg is None:
        tpg = get_tpg(wwn, tag)
    for node_acl in tpg.node_acls:
        if node_wwn and node_acl.node_wwn == node_wwn:
            return node_acl
        elif not node_wwn:
            return node_acl


def get_mapped_lun(tpg: Optional[TPG] = None, wwn: Optional[str] = None, tag: Optional[int] = None,
                   node_wwn: Optional[str] = None, alias: Optional[str] = None, mapped_lun: Optional[int] = None
                   ) -> Optional[LUN]:
    nodeacl = get_nodeacl(tpg=tpg, wwn=wwn, tag=tag, node_wwn=node_wwn)
    if nodeacl:
        for mlun in nodeacl.mapped_luns:
            if mlun.alias == alias or mlun.mapped_lun == mapped_lun:
                return mlun
            elif not alias and not mapped_lun:
                return mlun


def get_portal(tpg: Optional[TPG] = None, wwn: Optional[str] = None, tag: Optional[int] = None,
               ip_address: Optional[str] = None, port: Optional[int] = None) -> Optional[LUN]:
    if tpg is None:
        tpg = get_tpg(wwn, tag)
    for portal in tpg.network_portals:
        if ip_address and portal.ip_address == ip_address and port and portal.port == port:
            return portal
        elif not ip_address and not port:
            return portal


@states_response
def list_luns_conf(wwn: str, tag: int) -> List[dict]:
    return [_.dump() for _ in get_list_luns(wwn, tag)]


def create_block_storage(name: str, dev: Optional[str] = None, wwn: Optional[str] = None,
                         readonly: Optional[bool] = False, write_back: Optional[bool] = False
                         ) -> BlockStorageObject:
    if readonly is None:
        readonly = False
    if write_back is None:
        write_back = False
    try:
        root = get_root()
        storage = BlockStorageObject(name=name, dev=dev, wwn=wwn, readonly=readonly, write_back=write_back, index=None)
        root.save_to_file()
    except Exception as e:
        raise e
    else:
        return storage


@states_response
def create_block_storage_conf(*args, **kwargs) -> dict:
    try:
        storage = create_block_storage(*args, **kwargs)
    except Exception as e:
        raise e
    else:
        return storage.dump()


@states_response
def create_block_storages(storages: list) -> list:
    try:
        for storage in storages:
            create_block_storage(name=storage.get('name'), dev=storage.get('dev'), wwn=storage.get('wwn'),
                                 readonly=storage.get('readonly'), write_back=storage.get('write_back'))
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while create block storage for iscsi target. '
                                    f'Error: {e}') from None
    return get_root().dump()['storage_objects']


def create_iscsi_target(wwn: Optional[str] = None) -> Target:
    try:
        target = Target(FabricModule("iscsi"), wwn=wwn)
    except Exception as e:
        raise e
    else:
        return target


def create_iscsi_tpg(target: Optional[Target] = None, wwn: Optional[str] = None, tpg_tag: Optional[int] = None,
                     chap_userid: Optional[str] = None, chap_password: Optional[str] = None,
                     chap_mutual_userid: Optional[str] = None, chap_mutual_password: Optional[str] = None) -> TPG:
    if not target:
        target = get_target(wwn=wwn)
    try:
        tpg = TPG(target, tag=tpg_tag)
        if chap_userid:
            tpg.chap_userid = chap_userid
        if chap_password:
            tpg.chap_password = chap_password
        if chap_mutual_userid:
            tpg.chap_mutual_userid = chap_mutual_userid
        if chap_mutual_password:
            tpg.chap_mutual_password = chap_mutual_password
    except Exception as e:
        raise e
    else:
        return tpg


def create_iscsi_portal(tpg: Optional[TPG] = None, wwn: Optional[str] = None, tpg_tag: int = 1,
                        ip_address: str = "0.0.0.0", port: int = 3260) -> NetworkPortal:
    if not tpg:
        tpg = get_tpg(wwn=wwn, tag=tpg_tag)
    try:
        portal = NetworkPortal(tpg, ip_address=ip_address, port=port)
    except Exception as e:
        raise e
    else:
        return portal


def create_iscsi__tpg_lun(tpg: Optional[TPG] = None, wwn: Optional[str] = None, tpg_tag: int = 1,
                          storage: Optional[BlockStorageObject] = None, storage_name: Optional[str] = None,
                          storage_dev: Optional[str] = None, storage_wwn: Optional[str] = None,
                          lun_index: Optional[int] = None, alias: Optional[str] = None) -> Optional[LUN]:
    if not tpg:
        tpg = get_tpg(wwn=wwn, tag=tpg_tag)
    if not storage:
        storage = get_storage(name=storage_name, dev=storage_dev, wwn=storage_wwn)
    if tpg and storage:  # and storage.status == 'deactivated':
        try:
            lun = LUN(parent_tpg=tpg, storage_object=storage, lun=lun_index, alias=alias)
        except Exception as e:
            raise e
        else:
            return lun
    elif not tpg or not storage:
        msg = ''
        if not tpg:
            msg += f'not find tpg: wwn={wwn}, tag={tpg_tag};'
        if not storage:
            msg += f'not find storage: name={storage_name}, dev={storage_dev}, wwn={storage_wwn};'
        raise Exception(msg)


def create_iscsi_nodeacl(tpg: Optional[TPG] = None, wwn: Optional[str] = None, tpg_tag: int = 1,
                         node_wwn: Optional[str] = None, chap_userid: Optional[str] = None,
                         chap_password: Optional[str] = None, chap_mutual_userid: Optional[str] = None,
                         chap_mutual_password: Optional[str] = None) -> Optional[NodeACL]:
    if not tpg:
        tpg = get_tpg(wwn=wwn, tag=tpg_tag)
    if tpg and node_wwn:
        try:
            nodeacl = NodeACL(tpg, node_wwn)
            if chap_userid:
                nodeacl.chap_userid = chap_userid
            if chap_password:
                nodeacl.chap_password = chap_password
            if chap_mutual_userid:
                nodeacl.chap_mutual_userid = chap_mutual_userid
            if chap_mutual_password:
                nodeacl.chap_mutual_password = chap_mutual_password
        except Exception as e:
            raise e
        else:
            return nodeacl
    elif not tpg or not node_wwn:
        msg = ''
        if not tpg:
            msg += f'not find tpg: wwn={wwn}, tag={tpg_tag};'
        if not node_wwn:
            msg += 'not passed down node_wwn;'
        raise Exception(msg)


def create_iscsi_mapped_lun(lun: Optional[LUN] = None, tpg: Optional[TPG] = None, wwn: Optional[str] = None,
                            tag: Optional[int] = None, lun_index: Optional[int] = None, lun_alias: Optional[str] = None,
                            nodeacl: Optional[NodeACL] = None, node_wwn: Optional[str] = None, mapped_lun: int = 0,
                            write_protect: bool = False, alias: Optional[str] = None) -> Optional[MappedLUN]:
    if not lun:
        lun = get_lun(tpg=tpg, wwn=wwn, tag=tag, lun_index=lun_index, lun_alias=lun_alias)
    if not nodeacl:
        nodeacl = get_nodeacl(wwn=wwn, tag=tag, node_wwn=node_wwn)
    if lun and nodeacl:
        try:
            mlun = MappedLUN(nodeacl, mapped_lun, lun, write_protect=write_protect, alias=alias)
        except Exception as e:
            raise e
        else:
            return mlun
    elif not lun or not nodeacl:
        msg = ''
        if not lun:
            msg += f'not find lun: lun_index={lun_index}, lun_alias={lun_alias};'
        if not nodeacl:
            msg += f'not find nodeacl: node_wwn={node_wwn};'
        raise Exception(msg)


@states_response
def create(**kwargs):
    wwn = kwargs.get('wwn')
    storages = kwargs.get('storages', [])
    tpgs = kwargs.get('tpgs')
    try:
        obj_storages = []
        if storages:
            for s in storages:
                storage_name = s.get('name')
                storage_wwn = s.get('wwn', uuid4())
                storage_dev = s.get('dev')
                storage_readonly = s.get('readonly')
                storage_write_back = s.get('write_back')
                storage = get_storage(name=storage_name, dev=storage_dev, wwn=storage_wwn)
                if not storage:
                    storage = create_block_storage(name=storage_name, dev=storage_dev, wwn=storage_wwn,
                                                   readonly=storage_readonly, write_back=storage_write_back)
                obj_storages.append(storage)
        if wwn:
            target = get_target(wwn)
            if not target:
                target = create_iscsi_target(wwn)
        else:
            target = create_iscsi_target()
        for tpg in tpgs:
            tpg_tag = kwargs.get('tpg_tag')
            authentication = tpg.get('authentication', False)
            chap_userid = tpg.get('chap_userid')
            chap_password = tpg.get('chap_password')
            chap_mutual_userid = tpg.get('chap_mutual_userid')
            chap_mutual_password = tpg.get('chap_mutual_password')
            generate_node_acls = tpg.get('generate_node_acls')
            enable = tpg.get('enable')
            obj_tpg = create_iscsi_tpg(target=target, tpg_tag=tpg_tag, chap_userid=chap_userid,
                                       chap_password=chap_password, chap_mutual_userid=chap_mutual_userid,
                                       chap_mutual_password=chap_mutual_password)
            obj_tpg.set_attribute('authentication', '1' if authentication else '0')
            obj_tpg.set_attribute('generate_node_acls', '1' if generate_node_acls else '0')
            if isinstance(enable, bool):
                obj_tpg.enable = enable
            if tpg.get('TargetAlias'):
                obj_tpg.set_parameter('TargetAlias', tpg['TargetAlias'])
            for lun in tpg.get('luns', []):
                obj_storage = get_storage(dev=lun['dev'])
                create_iscsi__tpg_lun(tpg=tpg, storage=obj_storage, lun_index=lun['index'], alias=lun['alias'])
            for portal in tpg.get('portals', []):
                create_iscsi_portal(tpg=obj_tpg, ip_address=portal['ip_address'], port=portal['port'])
            for node_acl in tpg.get('node_acls', []):
                obj_node_acl = create_iscsi_nodeacl(tpg=obj_tpg, node_wwn=node_acl['node_wwn'],
                                                    chap_userid=node_acl['chap_userid'],
                                                    chap_password=node_acl['chap_password'],
                                                    chap_mutual_userid=node_acl['chap_mutual_userid'],
                                                    chap_mutual_password=node_acl['chap_mutual_password'])
                for ml in node_acl['mapped_luns']:
                    create_iscsi_mapped_lun(lun_alias=ml['lun_alias'], nodeacl=obj_node_acl, mapped_lun=ml['index'],
                                            write_protect=ml['write_protect'], alias=ml['alias'])
        root = save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while create iscsi target. Error: {e}') from None
    else:
        return {'wwn': target.wwn, 'dump': root.dump()}


@states_response
def delete(wwn, **kwargs):
    storages = kwargs.get('storages', [])
    try:
        target = get_target(wwn)
        for tpg in target.tpgs:
            for nodeacl in tpg.node_acls:
                for mlun in nodeacl.mapped_luns:
                    mlun.delete()
            for lun in tpg.luns:
                lun.delete()
        for storage in storages:
            if storage.get('wwn') or storage.get('name') or storage.get('dev'):
                storage = get_storage(name=storage.get('name'), dev=storage.get('dev'), wwn=storage.get('wwn'))
                if storage:
                    storage.delete(save=False)
        target.delete()
        root = save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while delete iscsi target. Error: {e}') from None
    else:
        return {'dump': root.dump()}


@states_response
def change_auth(wwn, update_auth):
    try:
        target = get_target(wwn)
        tpg_tags = [int(k) for k in update_auth.keys()]
        for obj_tpg in target.tpgs:
            tpg = None
            if obj_tpg.tag in tpg_tags:
                tpg = update_auth[str(obj_tpg.tag)]
            else:
                continue
            if 'authentication' in tpg and tpg['authentication'] is not None:
                obj_tpg.set_attribute('authentication', '1' if tpg['authentication'] else '0')
                if 'chap_userid' in tpg and tpg['chap_userid'] is not None:
                    obj_tpg.chap_userid = tpg['chap_userid']
                if 'chap_password' in tpg and tpg['chap_password']:
                    obj_tpg.chap_password = tpg['chap_password']
                if 'chap_mutual_userid' in tpg and tpg['chap_mutual_userid'] is not None:
                    obj_tpg.chap_mutual_userid = tpg['chap_mutual_userid']
                if 'chap_mutual_password' in tpg and tpg['chap_mutual_password'] is not None:
                    obj_tpg.chap_mutual_password = tpg['chap_mutual_password']
                if 'generate_node_acls' in tpg and tpg['generate_node_acls'] is not None:
                    obj_tpg.set_attribute('generate_node_acls', '1' if tpg['generate_node_acls'] else '0')
                if 'node_acls' in tpg:
                    node_acl_wwns = [na.node_wwn for na in obj_tpg.node_acls]
                    for node_acl in tpg['node_acls']:
                        node_wwn = tpg['node_wwn']
                        if node_wwn in node_acl_wwns:
                            obj_node_acl = [na for na in obj_tpg.node_acls if na.node_wwn == node_wwn][0]
                            if 'chap_userid' in node_acl and node_acl['chap_userid'] is not None:
                                obj_node_acl.chap_userid = node_acl['chap_userid']
                            if 'chap_password' in node_acl and node_acl['chap_password'] is not None:
                                obj_node_acl.chap_password = node_acl['chap_password']
                            if 'chap_mutual_userid' in node_acl and node_acl['chap_mutual_userid'] is not None:
                                obj_node_acl.chap_mutual_userid = node_acl['chap_mutual_userid']
                            if 'chap_mutual_password' in node_acl and node_acl['chap_mutual_password'] is not None:
                                obj_node_acl.chap_mutual_password = node_acl['chap_mutual_password']
        root = save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while change authentication iscsi target. Error: {e}') from None
    else:
        return {'dump': root.dump()}


@states_response
def tpg_create(wwn, tpg_tag, **kwargs):
    try:
        target = get_target(wwn)
        authentication = kwargs.get('authentication', False)
        chap_userid = kwargs.get('chap_userid')
        chap_password = kwargs.get('chap_password')
        chap_mutual_userid = kwargs.get('chap_mutual_userid')
        chap_mutual_password = kwargs.get('chap_mutual_password')
        generate_node_acls = kwargs.get('generate_node_acls')
        demo_mode_write_protect = kwargs.get('demo_mode_write_protect')
        enable = kwargs.get('enable')
        obj_tpg = create_iscsi_tpg(target=target, tpg_tag=tpg_tag, chap_userid=chap_userid,
                                   chap_password=chap_password, chap_mutual_userid=chap_mutual_userid,
                                   chap_mutual_password=chap_mutual_password)
        obj_tpg.set_attribute('authentication', '1' if authentication else '0')
        obj_tpg.set_attribute('generate_node_acls', '1' if generate_node_acls else '0')
        obj_tpg.set_attribute('demo_mode_write_protect', '1' if demo_mode_write_protect else '0')
        if kwargs.get('TargetAlias'):
            obj_tpg.set_parameter('TargetAlias', kwargs['TargetAlias'])
        if isinstance(enable, bool):
            obj_tpg.enable = enable
        save_root()
    except Exception as e:
        raise CommandExecutionError(
            f'An error occurred while create target port group for iscsi target. Error: {e}') from None
    else:
        return obj_tpg.dump()


@states_response
def storage_delete(name, wwn, dev):
    try:
        delete_storage(name=name, dev=dev, wwn=wwn)
        root = save_root()
    except Exception as e:
        raise CommandExecutionError(
            f'An error occurred while delete storage_object for iscsi target. Error: {e}') from None
    else:
        return root.dump()


@states_response
def tpg_delete(wwn, tpg_tag):
    try:
        delete_tpg(wwn=wwn, tag=tpg_tag)
        root = save_root()
    except Exception as e:
        raise CommandExecutionError(
            f'An error occurred while delete target port group for iscsi target. Error: {e}') from None
    else:
        return root.dump()


@states_response
def lun_create(wwn, tpg_tag, alias, dev, index, storage_wwn):
    try:
        lun = create_iscsi__tpg_lun(wwn=wwn, tpg_tag=tpg_tag, storage_wwn=storage_wwn, storage_dev=dev, lun_index=index,
                                    alias=alias)
        save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while create lun for iscsi target. Error: {e}') from None
    else:
        return lun.dump()


@states_response
def lun_delete(wwn, tpg_tag, alias, index):
    try:
        obj_lun = get_lun(wwn=wwn, tag=tpg_tag, lun_index=index, lun_alias=alias)
        obj_lun.delete()
        root = save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while delete lun for iscsi target. Error: {e}') from None
    else:
        return root.dump()


@states_response
def node_acl_create(wwn, tpg_tag, node_wwn, **kwargs):
    try:
        node_acl = create_iscsi_nodeacl(wwn=wwn, tpg_tag=tpg_tag, node_wwn=node_wwn,
                                        chap_userid=kwargs.get('chap_userid'),
                                        chap_password=kwargs.get('chap_password'),
                                        chap_mutual_userid=kwargs.get('chap_mutual_userid'),
                                        chap_mutual_password=kwargs.get('chap_mutual_password'))
        save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while create node_acl for iscsi target. Error: {e}') from None
    else:
        return node_acl.dump()


@states_response
def node_acl_delete(wwn, tpg_tag, node_wwn):
    try:
        obj_node_acl = get_nodeacl(wwn=wwn, tag=tpg_tag, node_wwn=node_wwn)
        obj_node_acl.delete()
        root = save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while delete node_acl for iscsi target. Error: {e}') from None
    else:
        return root.dump()


@states_response
def mapped_lun_create(wwn, tpg_tag, tpg_lun, lun_alias, write_protect, index, alias):
    try:
        mapped_lun = create_iscsi_mapped_lun(wwn=wwn, tag=tpg_tag, lun_index=tpg_lun, lun_alias=lun_alias,
                                             mapped_lun=index, write_protect=write_protect, alias=alias)
        save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while create mapped_lun for iscsi target. Error: {e}') from None
    else:
        return mapped_lun.dump()


@states_response
def mapped_lun_delete(wwn, tpg_tag, node_wwn, alias, index):
    try:
        obj_mapped_lun = get_mapped_lun(wwn=wwn, tag=tpg_tag, node_wwn=node_wwn, alias=alias, mapped_lun=index)
        obj_mapped_lun.delete()
        root = save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while delete mapped_lun for iscsi target. Error: {e}') from None
    else:
        return {'dump': root.dump()}


@states_response
def portal_create(wwn, tpg_tag, ip_address, port):
    try:
        portal = create_iscsi_portal(wwn=wwn, tpg_tag=tpg_tag, ip_address=ip_address, port=port)
        save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while create portal for iscsi target. Error: {e}') from None
    else:
        return portal.dump()


@states_response
def portal_delete(wwn, tpg_tag, ip_address, port):
    try:
        obj_portal = get_portal(wwn=wwn, tag=tpg_tag, ip_address=ip_address, port=port)
        obj_portal.delete()
        root = save_root()
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while delete portal for iscsi target. Error: {e}') from None
    else:
        return root.dump()
