"""
(c) 2023 LLC «DACOM M»
"""

import os

from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output, \
    delete_file_with_owner, WORKER_USER, states_result
from veilrpc.utils.file import chown_files_in_dir

ISO_TYPES = ['ISO 9660', 'DOS/MBR boot sector', 'data', 'x-iso9660-image']
ISO_CMD = "find /storages -type f -name *.iso -exec du -b {} \\;"


@states_response
def full_iso_discovery(with_filename=False, **kwargs):
    from .directory_datapool import get_file_info
    results = []
    for line in states_result(get_output, cmd=ISO_CMD, **kwargs).splitlines():
        try:
            # pylint: disable=cell-var-from-loop
            size, filename = line.split()
            info_ret = get_file_info(filename, **kwargs)
            if not info_ret['result']:
                continue
            if any(map(lambda x: x in info_ret['comment'], ISO_TYPES)):
                if with_filename:
                    _, tail = os.path.split(filename)
                    results.append({'path': filename,
                                    'filename': tail,
                                    'size': int(size.strip())})
                else:
                    results.append({'path': filename, 'size': int(size.strip())})
        except (KeyError, ValueError):
            continue
    return results


@states_response
def iso_discovery(datapool_path, iso_storage_path, chown_before=False, remove_unknown=False, **kwargs):
    from .directory_datapool import get_file_info, find_with_size_cmd, _create_nested_directory
    path = os.path.join(datapool_path, iso_storage_path)
    try:
        _create_nested_directory(path, WORKER_USER, True)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None
    if chown_before:
        try:
            chown_files_in_dir(path)
        except OSError:
            pass
    results, unknowns = [], []
    ret = find_with_size_cmd(path, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(f"Can not discover ISO in datapool {datapool_path}: {ret['comment']}")
    for line in ret['comment']:
        # pylint: disable=cell-var-from-loop
        try:
            size, filename = line.split()
            info_ret = get_file_info(filename, **kwargs)
            if not info_ret['result']:
                if remove_unknown:
                    unknowns.append(filename)
                continue
            if any(map(lambda x: x in info_ret['comment'], ISO_TYPES)):
                results.append({'path': filename, 'size': int(size.strip())})
            else:
                if remove_unknown:
                    unknowns.append(filename)
        except (KeyError, ValueError):
            continue
    if remove_unknown and unknowns:
        cmd_ret = multiple_delete(unknowns, WORKER_USER)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    return results


@states_response
def attach(domain_id, iso_path, cdrom_target_dev):
    cmd = f"virsh change-media {domain_id} {cdrom_target_dev} '{iso_path}'"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f"Domain {domain_id} iso successfully attached."


@states_response
def detach(domain_id, cdrom_target_dev):
    cmd = f"virsh change-media {domain_id} {cdrom_target_dev} --eject --force"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f"Domain {domain_id} iso successfully detached."


@states_response
def delete(iso_path, owner):
    cmd_ret = delete_file_with_owner(iso_path, owner)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f"Iso with path {iso_path} successfully deleted."


@states_response
def multiple_delete(file_paths, owner):
    for file_path in file_paths:
        cmd_ret = delete_file_with_owner(file_path, owner)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    return f"Iso with paths {', '.join(file_paths)} successfully deleted."
