"""
(c) 2023 LLC «DACOM M»
"""

from veilrpc.utils.bazalt import states_response, is_service_active, remote_command, \
    CommandExecutionError, get_output, action_service
from veilrpc.states.puppet import puppet_modify
from linstor import Linstor, ResourceData
from time import sleep
from typing import Optional, Any, List


def linstor_obj(ctrl_host: str = 'localhost'):
    obj = Linstor(f"linstor://{ctrl_host}")
    obj.connect()
    return obj


def resp_is_success(res, ret, method_name):
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            if method_name in ret['data'] and isinstance(ret['data'][method_name], list):
                ret['data'][method_name].append(resp.data_v1)
            else:
                ret['data'].update({method_name: [resp.data_v1]})
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                if method_name in ret['data'] and isinstance(ret['data'][method_name], list):
                    ret['data'][method_name].append(resp.data_v1)
                else:
                    ret['data'].update({method_name: [resp.data_v1]})
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")


@states_response
def list_node(ctrl_host: str = 'localhost', lin: Linstor = None, human: bool = False, **kwargs):
    if human:
        cmd = "linstor node list --pastable"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return cmd_ret['comment']
    else:
        lin = lin or linstor_obj(ctrl_host)
        res = lin.node_list()
        if res:
            nodes = res[0]
            return nodes.data_v1
        else:
            return []


@states_response
def linstor_controller_action(enable=True, **kwargs):
    ret = ""
    if enable:
        service = 'linstor-controller'
        if not is_service_active(service):
            action_service(service, 'enable', timeout=300)
            cmd_ret = action_service(service, 'start', timeout=300)
            if not cmd_ret['result']:
                raise CommandExecutionError(f"Error in start {service}: {cmd_ret['comment']}")
            else:
                ret += f'start {service}. '
        puppet_modify(module='space_controller', params={'space_controller::linstor_ensure': enable}, apply=False)
        try:
            nodes = list_node()
            sleep(3)
            return nodes
        except Exception:
            sleep(3)
            try:
                return list_node()
            except Exception as e:
                raise CommandExecutionError(f"Error in list_node: {str(e)}") from None
    else:
        puppet_modify(module='space_controller', params={'space_controller::linstor_ensure': enable}, apply=False)
    return ret


@states_response
def create_cluster(nodes, ctrl_host: str = 'localhost', **kwargs):
    ret = ""
    for service in ['linstor-satellite']:
        if not is_service_active(service):
            action_service(service, 'enable', timeout=300)
            cmd_ret = action_service(service, 'start', timeout=300)
            if not cmd_ret['result']:
                raise CommandExecutionError(f"Error in start {service}: {cmd_ret['comment']}")
            else:
                sleep(3)
                if is_service_active(service):
                    ret += f'start {service}. '
                else:
                    cmd_ret = action_service(service, 'start', timeout=300)
                    if is_service_active(service):
                        ret += f'start {service}. '
                    else:
                        raise CommandExecutionError(f"Error in start {service}: {cmd_ret['comment']}")
    puppet_modify(module='space_node', params={'space_node::linstor_ensure': True}, apply=False)

    lin = linstor_obj(ctrl_host)
    for node_dict in nodes:
        node_create_ret = lin.node_create(**node_dict)
        for resp in node_create_ret:
            if resp.is_success():
                ret += resp.message
            else:
                if resp.cause and 'already exists' in resp.cause:
                    ret += resp.cause
                    ret += ". "
                else:
                    sleep(3)
                    node_create_ret = lin.node_create(**node_dict)
                    for resp in node_create_ret:
                        if resp.is_success():
                            ret += resp.message
                        else:
                            if resp.cause and 'already exists' in resp.cause:
                                ret += resp.cause
                                ret += ". "
                            else:
                                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                                            f"correction: {resp.correction}")

    return ret


@states_response
def delete_cluster(ctrl_host: str = 'localhost', **kwargs):
    ret = ""
    lin = linstor_obj(ctrl_host)

    list_node_ret = list_node(ctrl_host, lin=lin)
    if list_node_ret['result']:
        nodes = list_node_ret['comment']
        for node in nodes:
            node_delete_ret = lin.node_delete(node_name=node['name'])
            for resp in node_delete_ret:
                if resp.is_success():
                    ret += resp.message
                    ret += ". "
                else:
                    if resp.cause and 'does not exist' in resp.cause:
                        ret += resp.cause
                        ret += ". "
                    elif resp.cause and 'evicted' in resp.cause:
                        ret += resp.cause
                        ret += ". "
                        node_lost_ret = lin.node_lost(node_name=node['name'])
                        for resp in node_lost_ret:
                            if resp.is_success():
                                ret += resp.message
                                ret += ". "
                            else:
                                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                                            f"correction: {resp.correction}")
                    else:
                        raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                                    f"correction: {resp.correction}")
    puppet_modify(module='space_node', params={'space_node::linstor_ensure': False}, apply=False)
    return ret


@states_response
def delete_nodes(nodes, ctrl_host: str = 'localhost', **kwargs):
    ret = ""
    lin = linstor_obj(ctrl_host)
    for node_name in nodes:
        node_delete_ret = lin.node_delete(node_name=node_name)
        for resp in node_delete_ret:
            if resp.is_success():
                ret += resp.message
                ret += ". "
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def list_storage_pool(ctrl_host: str = 'localhost', lin: Linstor = None, human: bool = False,
                      filter_by_nodes=None, filter_by_stor_pools=None, filter_by_props=None, **kwargs):
    if human:
        cmd = "linstor storage-pool list --pastable"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return cmd_ret['comment']
    else:
        lin = lin or linstor_obj(ctrl_host)
        res = lin.storage_pool_list(filter_by_nodes=filter_by_nodes,
                                    filter_by_stor_pools=filter_by_stor_pools,
                                    filter_by_props=filter_by_props)
        if res:
            return res[0].data_v1
        else:
            return []


@states_response
def storage_pool_create(params: dict, ctrl_host: str = 'localhost', **kwargs):
    lin = linstor_obj(ctrl_host)
    storage_pool_create_ret = lin.storage_pool_create(**params)
    for resp in storage_pool_create_ret:
        raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                    f"correction: {resp.correction}")
    pools_res = list_storage_pool(ctrl_host=ctrl_host, lin=lin)
    if pools_res['result']:
        return pools_res['comment']
    else:
        raise CommandExecutionError(pools_res['comment'])


@states_response
def storage_pools_create(params_list: list, ctrl_host: str = 'localhost', **kwargs):
    ret = ""
    lin = linstor_obj(ctrl_host)
    for params in params_list:
        storage_pool_create_ret = lin.storage_pool_create(**params)
        for resp in storage_pool_create_ret:
            if resp.is_success():
                ret += resp.message
                ret += ". "
            else:
                sleep(3)
                storage_pool_create_ret = lin.storage_pool_create(**params)
                for resp_1 in storage_pool_create_ret:
                    if resp_1.is_success():
                        ret += resp_1.message
                        ret += ". "
                    else:
                        raise CommandExecutionError(f"message: {resp_1.message}, cause: {resp_1.cause}, "
                                                    f"correction: {resp_1.correction}")
    return ret


@states_response
def storage_pool_delete(params: dict, ctrl_host: str = 'localhost', **kwargs):
    ret = ""
    lin = linstor_obj(ctrl_host)
    storage_pool_delete_ret = lin.storage_pool_delete(**params)
    for resp in storage_pool_delete_ret:
        if resp.is_success():
            ret += resp.message
            ret += ". "
        else:
            raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                        f"correction: {resp.correction}")
    return ret


@states_response
def storage_pools_delete(params_list: list, ctrl_host: str = 'localhost', **kwargs):
    ret = ""
    lin = linstor_obj(ctrl_host)
    for params in params_list:
        storage_pool_delete_ret = lin.storage_pool_delete(**params)
        for resp in storage_pool_delete_ret:
            if resp.is_success():
                ret += resp.message
                ret += ". "
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def delete_cluster_volume(storage_pool_name: str, ctrl_host: str = 'localhost', **kwargs):
    ret = ""
    lin = linstor_obj(ctrl_host)
    pools_res = list_storage_pool(ctrl_host=ctrl_host, lin=lin)
    if pools_res['result']:
        pools = pools_res['comment']
        for pool in pools:
            if pool['storage_pool_name'] == storage_pool_name:
                params = {"node_name": pool['node_name'], "storage_pool_name": storage_pool_name}
                storage_pool_delete_ret = lin.storage_pool_delete(**params)
                for resp in storage_pool_delete_ret:
                    if resp.is_success():
                        ret += resp.message
                        ret += ". "
                    else:
                        raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                                    f"correction: {resp.correction}")

    else:
        raise CommandExecutionError(pools_res['comment'])
    return ret


@states_response
def create_resource_group(name: str, storage_pool: str, ctrl_host: str = 'localhost', lin: Linstor = None, **kwargs):
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.resource_group_create(name=name, place_count=len(lin.node_list()[0].nodes),
                                    storage_pool=[storage_pool])
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def delete_resource_group(name: str, ctrl_host: str = 'localhost', lin: Linstor = None, **kwargs):
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.resource_group_delete(name=name)
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def list_resource_dfns(ctrl_host: str = 'localhost', lin: Linstor = None, human: bool = False,
                       query_volume_definitions=True, filter_by_resource_definitions=None, filter_by_props=None,
                       **kwargs):
    if human:
        cmd = "linstor resource-definition list --pastable"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return cmd_ret['comment']
    else:
        lin = lin or linstor_obj(ctrl_host)
        res = lin.resource_dfn_list(query_volume_definitions=query_volume_definitions,
                                    filter_by_resource_definitions=filter_by_resource_definitions,
                                    filter_by_props=filter_by_props)
        if res:
            resource_dfns = res[0]
            return resource_dfns.data_v1
        else:
            return []


@states_response
def list_resource_groups(ctrl_host: str = 'localhost', lin: Linstor = None, grpc: bool = False,
                         human: bool = False,
                         filter_by_resource_groups=None,
                         filter_by_props=None, **kwargs):
    if human:
        cmd = "linstor resource-group list --pastable"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return cmd_ret['comment']
    else:
        lin = lin or linstor_obj(ctrl_host)
        res = lin.resource_group_list_raise(filter_by_resource_groups=filter_by_resource_groups,
                                            filter_by_props=filter_by_props)
        if res:
            resource_groups = res[0].resource_groups
            if grpc:
                return [{"name": rg.name,
                         "description": rg.description,
                         "properties": rg.properties} for rg in resource_groups]
            else:
                return resource_groups
        else:
            return []


@states_response
def list_resources(ctrl_host: str = 'localhost', lin: Linstor = None, human: bool = False,
                   filter_by_nodes=None,
                   filter_by_resources=None,
                   filter_by_props=None, **kwargs):
    if human:
        cmd = "linstor resources list --pastable"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return cmd_ret['comment']
    else:
        lin = lin or linstor_obj(ctrl_host)
        res = lin.resource_list(filter_by_nodes=filter_by_nodes,
                                filter_by_resources=filter_by_resources,
                                filter_by_props=filter_by_props)
        ret = res[0].data_v1
        return ret


@states_response
def list_storage_pool_props(node_name, storage_pool_name, ctrl_host: str = 'localhost', lin: Linstor = None, **kwargs):
    cmd = f"linstor storage-pool list-properties {node_name} {storage_pool_name}"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def create_volume_group(name: str, ctrl_host: str = 'localhost', lin: Linstor = None, volume_nr: Optional[int] = None,
                        propertu_dict: Optional[dict] = None, gross: bool = False, **kwargs):
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.volume_group_create(resource_grp_name=name, volume_nr=volume_nr, property_dict=propertu_dict, gross=gross)
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def delete_volume_group(name: str, ctrl_host: str = 'localhost', lin: Linstor = None, volume_nr: Optional[int] = None,
                        **kwargs):
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.volume_group_delete(resource_grp_name=name, volume_nr=volume_nr)
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def list_volume_groups(resource_grp_name: str, ctrl_host: str = 'localhost', lin: Linstor = None, human: bool = False,
                       **kwargs):
    if human:
        cmd = f"linstor volume-group list {resource_grp_name} --pastable"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return cmd_ret['comment']
    else:
        lin = lin or linstor_obj(ctrl_host)
        res = lin.volume_group_list_raise(resource_grp_name=resource_grp_name)
        if res:
            return res.data_v1
        else:
            return []


@states_response
def resource_group_spawn(rsc_dfn_name: str, rsc_grp_name: str, vlm_sizes: str, ctrl_host: str = 'localhost',
                         lin: Linstor = None, partial: Optional[bool] = False,
                         definitions_only: Optional[bool] = False, external_name: Any = None, **kwargs):
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.resource_group_spawn(rsc_grp_name=rsc_grp_name, rsc_dfn_name=rsc_dfn_name, vlm_sizes=[vlm_sizes],
                                   partial=partial, definitions_only=definitions_only, external_name=external_name)
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def volume_list(ctrl_host: str = 'localhost', lin: Linstor = None, filter_by_nodes=None, filter_by_stor_pools=None,
                filter_by_resources=None, filter_by_props=None, human=False, **kwargs):
    if human:
        cmd = "linstor volume list --pastable"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return cmd_ret['comment']
    else:
        lin = lin or linstor_obj(ctrl_host)
        res = lin.volume_list(filter_by_nodes=filter_by_nodes, filter_by_stor_pools=filter_by_stor_pools,
                              filter_by_resources=filter_by_resources, filter_by_props=filter_by_props)
        if res:
            return res[0].data_v1
        return []


@states_response
def volume_dfn_create(rsc_name: str, size: int, ctrl_host: str = 'localhost', lin: Linstor = None,
                      volume_nr: Optional[int] = None, minor_nr: Optional[int] = None,
                      encrypt: bool = False, storage_pool: Any = None, gross: bool = False, **kwargs):

    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.volume_dfn_create(rsc_name=rsc_name, size=size, volume_nr=volume_nr, minor_nr=minor_nr,
                                encrypt=encrypt, storage_pool=storage_pool, gross=gross)
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def volume_dfn_delete(rsc_name: str, ctrl_host: str = 'localhost', lin: Linstor = None,
                      volume_nr: Optional[int] = None, **kwargs):
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.volume_dfn_delete(rsc_name=rsc_name, volume_nr=volume_nr, async_msg=False)
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def resource_dfn_create(name: str, port: Optional[int] = None, ctrl_host: str = 'localhost',
                        lin: Linstor = None, layer_list: List[str] = None,
                        resource_group: Optional[str] = None, external_name: Any = None, **kwargs):

    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.resource_dfn_create(name=name, port=port, external_name=external_name, layer_list=layer_list,
                                  resource_group=resource_group)
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def resource_dfn_delete(name: str, ctrl_host: str = 'localhost', lin: Linstor = None, **kwargs):
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.resource_dfn_delete(name=name, async_msg=False)
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def resource_create(name: str, nodes: List[str] = None, ctrl_host: str = 'localhost',
                    lin: Linstor = None, layer_list: List[str] = None,
                    storage_pool: Optional[str] = None, place_count: Any = None, **kwargs):

    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)

    if nodes:
        rscs = []
        for node in nodes:
            rsc = ResourceData(node_name=node, rsc_name=name, storage_pool=storage_pool, layer_list=layer_list)
            rscs.append(rsc)
        res = lin.resource_create(rscs=rscs)
    elif place_count:
        res = lin.resource_auto_place(rsc_name=name, place_count=place_count, storage_pool=[storage_pool])
    else:
        raise CommandExecutionError("One of argument 'nodes' or 'place_count' have to by.")
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def resource_delete(rsc_name: str, node: str, ctrl_host: str = 'localhost', lin: Linstor = None, **kwargs):
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    res = lin.resource_delete(node_name=node, rsc_name=rsc_name, async_msg=False)
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret


@states_response
def create_disk(name: str, size: str, ctrl_host: str = 'localhost',
                lin: Linstor = None, layer_list: List[str] = None,
                storage_pool: Optional[str] = None, place_count: Any = None, **kwargs):

    from veilrpc.utils.size import get_kibibytes
    size = get_kibibytes(size)

    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)

    crt_res_resp = resource_dfn_create(name=name, lin=lin, layer_list=layer_list)
    if crt_res_resp['result']:
        ret['comment'] += crt_res_resp['comment']['comment']
        ret['data']['create_resource_definition'] = crt_res_resp['comment']['data']
    else:
        raise CommandExecutionError(crt_res_resp['comment'])
    crt_vol_resp = volume_dfn_create(rsc_name=name, size=size, lin=lin)
    if crt_vol_resp['result']:
        ret['comment'] += '\n' + crt_vol_resp['comment']['comment']
        ret['data']['create_volume_definition'] = crt_vol_resp['comment']['data']
    else:
        raise CommandExecutionError(crt_vol_resp['comment'])
    crt_vol_resp = resource_create(name=name, lin=lin, storage_pool=storage_pool, place_count=place_count)
    if crt_vol_resp['result']:
        ret['comment'] += '\n' + crt_vol_resp['comment']['comment']
        ret['data']['create_volume_definition'] = crt_vol_resp['comment']['data']
    else:
        raise CommandExecutionError(crt_vol_resp['comment'])
    list_res_resp = list_resources(lin=lin, filter_by_resources=[name])
    if list_res_resp['result']:
        ret['data']['list_resource'] = list_res_resp['comment']
    else:
        raise CommandExecutionError(list_res_resp['comment'])
    return ret


@states_response
def delete_disk(resource: str, ctrl_host: str = 'localhost', lin: Linstor = None, volume_nr: Optional[int] = None,
                **kwargs):
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    # del_vol_resp = volume_dfn_delete(rsc_name=name, lin=lin, volume_nr=volume_nr)
    # if del_vol_resp['result']:
    #    ret['comment'] += del_vol_resp['comment']['comment']
    #    ret['data']['delete_volume_definition'] = del_vol_resp['comment']['data']
    # else:
    #    raise CommandExecutionError(del_vol_resp['comment'])
    del_res_resp = resource_dfn_delete(name=resource, lin=lin)
    if del_res_resp['result']:
        # ret['comment'] += '\n' + del_res_resp['comment']['comment']
        ret['comment'] += del_res_resp['comment']['comment']
        ret['data']['delete_resource_definition'] = del_res_resp['comment']['data']
    else:
        raise CommandExecutionError(del_res_resp['comment'])
    return ret


@states_response
def create_datapool(**kwargs):
    return 'The linstor datapool was successfully created.'


@states_response
def create_snapshot(snapshot_uuid: str, resource: str, ctrl_host: str = 'localhost', lin: Linstor = None, **kwargs):

    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)

    res = lin.snapshot_create(rsc_name=resource, snapshot_name=f"_{snapshot_uuid}", node_names=[])
    if res:
        resp = res[0]
        if resp.is_success():
            ret['comment'] += resp.message
            ret['data'] = resp.data_v1
        else:
            if resp.cause and 'already exists' in resp.cause:
                ret['comment'] += resp.cause
                ret['comment'] += ". "
                ret['data'] = resp.data_v1
            else:
                raise CommandExecutionError(f"message: {resp.message}, cause: {resp.cause}, "
                                            f"correction: {resp.correction}")
    return ret['comment']


@states_response
def delete_snapshot(snapshot_uuid: str, resource: str, ctrl_host: str = 'localhost', lin: Linstor = None, **kwargs):

    ret = {'comment': '', 'data': {'map_resrc_snap': []}}
    lin = lin or linstor_obj(ctrl_host)
    if kwargs.get('merge'):
        list_resrcs = [_['name'] for _ in lin.resource_list(filter_by_nodes=[kwargs.get('node_uuid')])[0].data_v1
                       if resource != _['name'] and resource[:-8] == _['name'][:-8]]
        resrc_on_nodes = lin.resource_list(filter_by_resources=[resource])[0].data_v1
        list_snap_of_resrc = [{'snapshot_name': _['name'],
                               'resource_name': _['resource_name']} for _ in lin.snapshot_dfn_list()[0].data_v1
                              if resource[:-8] == _['resource_name'][:-8]]
        for resrc_on_node in resrc_on_nodes:
            zfs_dev = resrc_on_node['layer_object']['children'][0]['storage']['storage_volumes'][0]['device_path'] \
                .replace('/dev/zvol/', '')
            cmd = f'sudo zfs promote {zfs_dev}; ' \
                  f'sudo zfs list -H -t snapshot -o name | grep -E "{resource}*" | xargs -r -n1 sudo zfs destroy -R'
            for ssh_options in kwargs.get('list_ssh_options', []):
                if ssh_options['node_uuid'] == resrc_on_node['node_name']:
                    cmd_kwargs = {
                        'cmd': cmd,
                        'ssh_options': ssh_options
                    }
                    if 'owner' in kwargs and kwargs['owner']:
                        cmd_kwargs.update({'owner': kwargs['owner']})
                    cmd_ret = remote_command(**cmd_kwargs)
                    if not cmd_ret['result']:
                        raise CommandExecutionError(f'{ret["comment"]}\n{cmd_ret["comment"]}')
            if resrc_on_node['node_name'] == kwargs.get('node_uuid'):
                cmd_ret = get_output(cmd)
                if not cmd_ret['result']:
                    raise CommandExecutionError(f'{ret["comment"]}\n'
                                                f'An error occurred while executing command {cmd}:'
                                                f' {cmd_ret["comment"]}')
        for s in list_snap_of_resrc:
            res = lin.snapshot_delete(rsc_name=s['resource_name'], snapshot_name=s['snapshot_name'])
            resp_is_success(res, ret, 'snapshot_delete')
        for resrc in list_resrcs:
            res = lin.resource_dfn_delete(name=resrc)
            resp_is_success(res, ret, 'resource_dfn_delete')
        return ret
    last_of_branch = kwargs.get('last_of_branch', False)
    parent_uuid = kwargs.get('parent_uuid', '')
    if last_of_branch and parent_uuid:
        clone_resrc = [_ for _ in lin.resource_list()[0].data_v1
                       if parent_uuid[:8] == _['name'][-8:] and resource[:-8] == _['name'][:-8]]
    else:
        clone_resrc = [_ for _ in lin.resource_list()[0].data_v1
                       if snapshot_uuid[:8] == _['name'][-8:] and resource[:-8] == _['name'][:-8]]
    clone_resrc_is_uptodate = [_ for _ in clone_resrc if 'in_use' in _['state']]
    if clone_resrc_is_uptodate:
        for cln_resrc in clone_resrc_is_uptodate:
            zfs_dev = cln_resrc['layer_object']['children'][0]['storage']['storage_volumes'][0]['device_path'] \
                .replace('/dev/zvol/', '')
            if last_of_branch and parent_uuid:
                cmd = f'sudo zfs promote {zfs_dev}'
            else:
                cmd = f'sudo zfs promote {zfs_dev}; sudo zfs destroy -R {zfs_dev}@_{snapshot_uuid}'
            for ssh_options in kwargs.get('list_ssh_options', []):
                if ssh_options['node_uuid'] == cln_resrc['node_name']:
                    cmd_kwargs = {
                        'cmd': cmd,
                        'ssh_options': ssh_options
                    }
                    if 'owner' in kwargs and kwargs['owner']:
                        cmd_kwargs.update({'owner': kwargs['owner']})
                    cmd_ret = remote_command(**cmd_kwargs)
                    if not cmd_ret['result']:
                        raise CommandExecutionError(f'{ret["comment"]}\n{cmd_ret["comment"]}')
            if cln_resrc['node_name'] == kwargs.get('node_uuid'):
                cmd_ret = get_output(cmd)
                if not cmd_ret['result']:
                    raise CommandExecutionError(f'{ret["comment"]}\n'
                                                f'An error occurred while executing command {cmd}:'
                                                f' {cmd_ret["comment"]}')
    resrc_snapshots = lin.snapshot_dfn_list(filter_by_resources=[resource])[0].data_v1
    resrc_is_inactive = True if [_ for _ in lin.resource_list(filter_by_resources=[resource])[0].data_v1
                                if 'flags' in _ and _['flags'] == ['INACTIVE']] else False
    res = lin.snapshot_delete(rsc_name=resource, snapshot_name=f"_{snapshot_uuid}")
    resp_is_success(res, ret, 'snapshot_delete')
    cmd = f'sudo zfs list -H -t snapshot -o name | grep -E "{resource[:-8]}\\w{{8}}_00000*@_{snapshot_uuid}" | ' \
          f'xargs -r -n1 sudo zfs destroy -R'
    for ssh_options in kwargs.get('list_ssh_options', []):
        cmd_kwargs = {
            'cmd': cmd,
            'ssh_options': ssh_options
        }
        if 'owner' in kwargs and kwargs['owner']:
            cmd_kwargs.update({'owner': kwargs['owner']})
        cmd_ret = remote_command(**cmd_kwargs)
        if not cmd_ret['result']:
            raise CommandExecutionError(f'{ret["comment"]}\n{cmd_ret["comment"]}')
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(f'{ret["comment"]}\n'
                                    f'An error occurred while executing command {cmd}:'
                                    f' {cmd_ret["comment"]}')
    if len(resrc_snapshots) == 1 and resrc_is_inactive:
        res = lin.resource_dfn_delete(name=resource)
        resp_is_success(res, ret, 'resource_dfn_delete')
    elif len(resrc_snapshots) > 1 and resrc_is_inactive and clone_resrc_is_uptodate:
        new_resrc_of_snap = clone_resrc_is_uptodate[0]['name']
        for s in resrc_snapshots:
            if s['name'] != f'_{snapshot_uuid}':
                res = lin.snapshot_delete(rsc_name=resource, snapshot_name=s['name'])
                resp_is_success(res, ret, 'snapshot_delete')
                res = lin.snapshot_create(rsc_name=new_resrc_of_snap, snapshot_name=s['name'], node_names=[])
                resp_is_success(res, ret, 'snapshot_create')
                ret['data']['map_resrc_snap'].append({'snapshot': s['name'],
                                                      'old_resrc': resource,
                                                      'new_resrc': new_resrc_of_snap})
        res = lin.resource_dfn_delete(name=resource)
        resp_is_success(res, ret, 'resource_dfn_delete')
    return ret


@states_response
def revert_snapshot(snapshot_uuid: str, resource: str, ctrl_host: str = 'localhost',
                    lin: Linstor = None, **kwargs):
    resource_prev = kwargs.get('resource_prev')
    snap_resource = kwargs.get('snap_resource')
    ret = {'comment': '', 'data': {}}
    lin = lin or linstor_obj(ctrl_host)
    nodes = [n.get('name') for n in lin.node_list()[0].data_v1]
    last_snapshot_in_old_branch = None
    try:
        if resource_prev:
            snapshots = lin.snapshot_dfn_list(filter_by_resources=[resource_prev])
            snapshots = snapshots[0].data_v1
            if snapshots:
                last_snapshot_in_old_branch = [_['name'] for _ in sorted(
                    snapshots,
                    key=lambda d: int(d['props']['SequenceNumber']),
                    reverse=True)][0]
    except KeyError as e:
        raise CommandExecutionError(f"The name of last snapshot do not got. Error: {e}")
    ret['data'].update({'old_resource_name': snap_resource,
                        'old_source': f"/dev/drbd/by-res/{snap_resource}/0",
                        'resource_name': resource,
                        'source': f"/dev/drbd/by-res/{resource}/0"})
    if resource != resource_prev and resource != snap_resource:
        res = lin.resource_dfn_create(name=resource)
        resp_is_success(res, ret, 'resource_dfn_create')
        res = lin.snapshot_volume_definition_restore(from_resource=snap_resource,
                                                     from_snapshot=f"_{snapshot_uuid}",
                                                     to_resource=resource)
        resp_is_success(res, ret, 'snapshot_volume_definition_restore')
        res = lin.snapshot_resource_restore(from_resource=snap_resource,
                                            from_snapshot=f"_{snapshot_uuid}",
                                            to_resource=resource,
                                            node_names=nodes)
        resp_is_success(res, ret, 'snapshot_resource_restore')
    if resource != resource_prev:
        for node in nodes:
            res = lin.resource_deactivate(node_name=node, rsc_name=resource_prev)
            resp_is_success(res, ret, 'resource_deactivate')
            if resource == snap_resource:
                res = lin.resource_activate(node_name=node, rsc_name=resource)
                resp_is_success(res, ret, 'resource_activate')
    if last_snapshot_in_old_branch:
        res = lin.snapshot_rollback(rsc_name=resource_prev, snapshot_name=last_snapshot_in_old_branch)
        resp_is_success(res, ret, 'snapshot_rollback')
    else:
        res = lin.resource_dfn_delete(name=resource_prev)
        resp_is_success(res, ret, 'resource_dfn_delete')
    return ret
