"""
(c) 2023 LLC «DACOM M»
"""
import subprocess
import json
import os
import re
import six

from veilrpc.utils.bazalt import states_response, get_output, CommandExecutionError, states_result
from .node import _zerofill_flag
from .iscsi_target import list_block_storages_conf

# Типы устройств, доступных для выбора при добавлении в группу томов
AVAILABLE_BLOCK_DEVICE_TYPES = ['disk', 'part', 'loop']
LVM_DATAPOOL_CMD = "head -n1 /dev/{meta_file}"
# префиксы id блочных локальных устройств, которые отфильтровываем, т.к. на /dev/sdX может указывать несколько id,
# нам нужен ata/scsi
DEV_PREF_FILTER_OUT = ['dm', 'lvm', 'wwn', 'usb']


def _map_blkdev_id():
    # список блочных устройств (разделы не включаем)
    parent_devs = [rec['name'] for rec in _fix_lsblk_json('-JM')['blockdevices']]
    symlinks = dict()
    for dev in parent_devs:
        # изредка lsblk отдаёт то, что в /dev/mapper, как имя из /dev (пропало физ.устройство?)
        if not os.path.exists(f'/dev/{dev}'):
            continue
        # команда отдаёт список параметров udev устройства, в т.ч. симлинки в /dev/disk
        dev_props = get_output(f'udevadm info /dev/{dev}')['comment']
        for line in dev_props.splitlines():
            if line.startswith('S: '):
                symlink = line[3:]
                if symlink.startswith('disk/by-id'):
                    if symlink in symlinks:
                        symlinks[symlink] += 1
                    else:
                        symlinks[symlink] = 1
    # создаём список симлинков, ссылающихся более чем на 1 реальное устройство (напр. неуникальный SCSI S/N, привет HP)
    # в виде шаблонов для сравнения включая разделы
    non_uniq = [re.compile(re.escape(symlink[11:]) + r'(-part\d+)*')  # откусываем 'disk/by-id/'
                for symlink in symlinks if symlinks[symlink] > 1]
    # список сортируем, т.к. одному dev могут соответствовать несколько id, иногда они отдаются не в том порядке
    ls_id = sorted([x for x in os.listdir('/dev/disk/by-id')
                    if x.split('-')[0] not in DEV_PREF_FILTER_OUT
                    # тут удаляем неуникальные dev-id
                    and not [_ for _ in non_uniq if re.fullmatch(_, x)]
                    ])
    # тут мы получаем последнее по алфавиту соответствие
    return {os.path.realpath(x): x for x in map(lambda z: '/dev/disk/by-id/' + z, ls_id)}


def _fix_lsblk_json(params):
    # lsblk -Jpo (util-linux 2.36) внезапно умеет путать строки и отдавать словарь без запятой в конце,
    # где надо с ней, и наоборот
    outstring = ''
    res = get_output(f'lsblk {params}')['comment']
    spl = res.splitlines()
    for ind, line in enumerate(spl):
        ls = line.strip()
        if ls.endswith('}') and ind < len(spl) - 1 and spl[ind + 1].strip().startswith('{'):
            outstring += ls + ',\n'
        elif ls.endswith('},') and ind < len(spl) - 1 and spl[ind + 1].strip().startswith(']'):
            outstring += ls[:-1] + '\n'
        else:
            outstring += ls + '\n'
    return json.loads(outstring)


def _get_blkdev_by_id(restricted_tran=None, allow_children=True):
    dict_id = _map_blkdev_id()
    # ключик M отдаёт диски включая вирт.устройства, т.е. можно использовать softRAID
    devices = [x for x in _fix_lsblk_json('-M -Jpo NAME,SIZE,MOUNTPOINT,PARTTYPE,TYPE,UUID,TRAN,ROTA')['blockdevices']
               if not isinstance(restricted_tran, list) or x['tran'] not in restricted_tran]
    for dev in devices:
        dev['short_name'] = dev['name']
        dev['name'] = dict_id.get(dev['name'], dev['name'])
        if allow_children:
            for child in dev.get('children', []):
                child['short_name'] = child['name']
                child['name'] = dict_id.get(child['name'], child['name'])
        elif dev.get('children'):
            del dev['children']
    return devices


def get_json_lvm_data(vgname=None, **kwargs):
    list_cmd = ['lvm', 'fullreport', '--units', 'B', '--no-suffix', '--reportformat', 'json']
    if vgname:
        list_cmd.append(vgname)
    p = subprocess.Popen(list_cmd,
                         stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=False)
    stdout, stderr = p.communicate(timeout=50)
    if stderr:
        raise CommandExecutionError(stderr.decode() if isinstance(stderr, bytes) else stderr)
    return json.loads(stdout)


def list_vgs_names(**kwargs):
    """
    List all volume groups
    """
    results = []
    out = states_result(get_output, cmd='vgdisplay').splitlines()
    for item in out:
        if item.find("VG Name") != -1:
            results.append(item.split()[2])
    results = list(set(results))
    return results


@states_response
def get_lvm_data(**kwargs):

    lvm_meta, lvm_vdisks = [], []
    vgs_names = list_vgs_names(**kwargs)
    lvm_data = get_json_lvm_data(**kwargs)

    for vgname in vgs_names:
        try:
            for value_vg in lvm_data['report']:
                if "vg" in value_vg and value_vg["vg"][0].get("vg_name") == vgname:

                    # резервируем под мету 256 мегабайт
                    vg_free_size = int(round(_convert_size_mb(value_vg["vg"][0]["vg_free"]) - 256, 0))

                    for number_lv, value_lv in enumerate(value_vg["lv"]):

                        lv_name = value_lv["lv_name"]
                        if re.match("^datapool-[\w]{8}.meta$", lv_name):
                            thin_pool_name = value_lv["pool_lv"] or None
                            meta_data = dict(path=os.path.join(vgname, lv_name), pool_lv=thin_pool_name, vgname=vgname)
                            meta_data["space"] = dict(size=int(round(_convert_size_mb(value_vg["vg"][0]["vg_size"]) - 256, 0)))

                            if not thin_pool_name:
                                meta_data["space"]["free"] = vg_free_size
                            else:
                                # по имени тонкого пула находим к какому пулу он принадлежит
                                for value_lv_thin in value_vg["lv"]:
                                    if value_lv_thin["lv_name"] == thin_pool_name:
                                        lv_size = _convert_size_mb(value_lv_thin["lv_size"])
                                        meta_data_size = 256 - _convert_size_mb(value_lv_thin["lv_metadata_size"])
                                        data_percent = round(float(value_lv_thin["data_percent"]) / 1000 / 1000 / 100, 2)
                                        data_size = data_percent * lv_size
                                        meta_data["space"]["free"] = int(round(lv_size - meta_data_size - data_size, 0))
                            lvm_meta.append(meta_data)

                        elif re.match(r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-4[0-9a-fA-F]{3}"
                                      "-[89abAB][0-9a-fA-F]{3}-[0-9a-fA-F]{12}$", lv_name):
                            vdisk = dict(id=lv_name, vgname=vgname, size=_convert_size_mb(value_vg["lv"][number_lv]["lv_size"]))
                            lvm_vdisks.append(vdisk)
        except Exception as e:
            raise CommandExecutionError(str(e)) from None
    return lvm_meta, lvm_vdisks


@states_response
def get_free_space_thin_pool(vgname, pool_name, **kwargs):
    """
    Return information about free space on the thin pool.
    """
    results = dict()
    lvm_data = get_json_lvm_data(**kwargs)

    for value_vg in lvm_data['report']:
        if "vg" in value_vg and value_vg["vg"][0].get("vg_name") == vgname:

            for value_lv in value_vg["lv"]:
                if value_lv["lv_name"] == pool_name:
                    lv_size = _convert_size_gb(value_lv["lv_size"])
                    meta_data_size = 0.25 - _convert_size_gb(value_lv["lv_metadata_size"])
                    data_percent = round(float(value_lv["data_percent"]) / 1000 / 1000 / 100, 2)
                    data_size = data_percent * lv_size
                    results['free'] = round(lv_size - meta_data_size - data_size, 2)

    return results


@states_response
def vgs(**kwargs):
    """
    Return information about the volume group(s):
    name, free size, size and list of used physical volumes.
    """
    results = []

    # перед сканированием перечитываем разделы на лунах для достижения актуальности данных
    ret = get_output('multipathd show maps json')
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    try:
        mp_maps = json.loads(ret['comment'])
    except (json.JSONDecodeError, TypeError):
        pass
    else:
        for mpath_dev_id in [mp['uuid'] for mp in mp_maps['maps']]:
            subprocess.call(f'partprobe /dev/mapper/{mpath_dev_id}', shell=True)

    vgs_names = list_vgs_names(**kwargs)
    dict_id = _map_blkdev_id()
    lvm_data = get_json_lvm_data(**kwargs)

    for vgname in vgs_names:
        skip_vg = False

        check_mount_cmd = "mount -l | awk '{{print $1 \" \" $3}}' | " \
                          "grep {vgname}".format(vgname=vgname.replace('-', '--'))

        cmd_ret = get_output(check_mount_cmd, **kwargs)

        if cmd_ret['result']:
            for line in cmd_ret['comment'].splitlines():
                if line:
                    _, mount_point = line.split()
                    if mount_point == '/':
                        skip_vg = True
        if not skip_vg:

            for value_vg in lvm_data['report']:
                if "vg" in value_vg and value_vg["vg"][0].get("vg_name") == vgname:
                    res = dict(name=vgname, size=_convert_size_mb(value_vg["vg"][0]["vg_size"]), free=_convert_size_mb(value_vg["vg"][0]["vg_free"]))
                    res['pvs'] = []

                    for value_pv in value_vg["pv"]:
                        pvs = dict(name=dict_id.get(value_pv["pv_name"], value_pv["pv_name"]), size=_convert_size_mb(value_pv["dev_size"]))
                        res['pvs'].append(pvs)
                    results.append(res)
    return results


def _convert_size_gb(size):
    return round(float(size) / 1024 / 1024 / 1024, 2)


def _convert_size_mb(size):
    """
    Конвертирует байты в мб.
    Возвращает 0, если мегабайтов мало
    :param size: байты.
    :return: мегабайты.
    """
    return int(float(size) / 1024 / 1024) if size else 0


@states_response
def get_unused_block_devices():
    """
    Return information about the block device(s) which are not used:
    name (path), size and type ('hdd/ssd').
    """
    # получаем список устройств и разделов, задействованных в активных на узле softRAID, чтоб их не трогать
    md_slaves = set()
    VIRTBLK = '/sys/devices/virtual/block'
    if os.path.isdir(VIRTBLK):
        for virtdev in os.listdir(VIRTBLK):
            slv_dir = os.path.join(VIRTBLK, virtdev, 'slaves')
            if os.path.isdir(slv_dir):
                for slave in os.listdir(slv_dir):
                    uev = os.path.join(slv_dir, slave, 'uevent')
                    if os.path.isfile(uev):
                        with open(uev, 'r') as uevent:
                            line = uevent.readline()
                            while line:
                                spl = line.split('=')
                                if spl[0] == 'DEVNAME':
                                    md_slaves.add(f'/dev/{spl[1].strip()}')
                                    break
                                line = uevent.readline()
    # получаем список устройств, задействованных в ISCSI storage
    ret = list_block_storages_conf()
    iscsi_storages = set()
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    if isinstance(ret['comment'], list):
        for storage in ret['comment']:
            iscsi_storages.add(storage['dev'])

    # получаем список устройств, задействованных в linstor и zvol
    zvol_drbd_cmd = 'readlink -f /dev/zvol/*/* ; readlink -f /dev/drbd/by-res/*/* ; true'  # returned false if not found
    zvol_drbd_ret = get_output(zvol_drbd_cmd)
    if not zvol_drbd_ret['result']:
        raise CommandExecutionError(zvol_drbd_ret['comment'])
    zvol_drbd_stors = zvol_drbd_ret['comment'].split()
    devices = _get_blkdev_by_id()
    umount_disks = [dev for dev in _search_umount_disks(devices, _used_pvs())
                    # отбрасываем разделы
                    if
                    not (re.search(r'-part\d+$', dev.get('name', '') or re.search(r'\d+$', dev.get('short_name', ''))))
                    # и диски из активных virt.blkdev(softRAID)
                    and not (dev.get('name', '') in md_slaves or dev.get('short_name', '') in md_slaves)
                    # и диски используемые в ISCSI storage
                    and not (dev.get('name', '') in iscsi_storages or dev.get('short_name', '') in iscsi_storages)
                    # и диски из zvol и drbd
                    and not (dev.get('name', '') in zvol_drbd_stors or dev.get('short_name', '') in zvol_drbd_stors)]

    unique_umount_disks = list({v['name']: v for v in umount_disks}.values())
    results = []
    for disk in unique_umount_disks:
        if not os.path.exists(_zerofill_flag(disk['name'])):
            results.append({'name': disk['name'],
                            'size': disk['size'],
                            'type': 'ssd' if disk.get('rota') is False else 'hdd'})
    return results


@states_response
def get_block_devices():
    """
    Return information about the block device(s):
    """
    return _get_blkdev_by_id(restricted_tran=['iscsi', 'fc'])


def _used_pvs(**kwargs):
    results = {}
    cmd = 'pvdisplay --readonly'
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    out = cmd_ret['comment'].splitlines()

    pv_name = ""
    pv_uuid = ""
    pv_drop = True

    for item in out:

        if item.find("PV Name") != -1:
            pv_name = item.split()[2]

        if item.find("VG Name") != -1:
            vg_item = item.split()
            if len(vg_item) > 2:
                pv_drop = False

        if item.find("PV UUID") != -1:
            pv_uuid = item.split()[2]

            if not pv_drop:
                results.update({pv_name: pv_uuid})
            pv_drop = True

    return results


def _search_umount_disks(devices, used_pvs):
    """
    The function searches for in the received list of unused devices,
    except extended partitions, devices like the rom and thin pools.
    ПРИ ИЗМЕНЕНИИ ЛОГИКИ РАБОТЫ С УСТРОЙСТВАМИ НУЖНО ПРОВЕРЯТЬ
    РАБОТОСПОСОБНОСТЬ ZFS ФУНКЦИОНАЛА
    :param devices: list

    """
    list_unused_disks = []
    from .zfs import zpool_used_devices
    zp_devices = zpool_used_devices()
    if not zp_devices['result']:
        raise Exception("Get ZFS pool device list error.")

    for dev in devices:
        # Убираем из списка iscsi, fc, usb устройства.
        if dev['tran'] in ['iscsi', 'fc', 'usb']:
            continue
        # Убираем из списка устройства, включенные в zfs пулы
        # Тема с parttype в lsblk НЕ РАБОТАЕТ - после удаления ZFS пула раздел остается с parttype ZFS
        if dev['name'] in zp_devices['comment']:
            continue
        # убираем EFI разделы, т.к. они должны быть на каждом устройстве, входящем в softRAID, для надёжности загрузки,
        # а замонтирован к /boot/efi будет только один из них
        if dev['parttype'] == 'c12a7328-f81f-11d2-ba4b-00a0c93ec93b':
            continue
        if dev['mountpoint'] is None and _is_block_dev(dev) and \
                _parttype_is_not_extended(dev['parttype']):
            if _filtered_children(dev):
                unused_disks = _search_umount_disks(_filtered_children(dev), used_pvs)
                list_unused_disks.extend(unused_disks)
            elif not (dev['name'] in used_pvs or dev.get('short_name') in used_pvs or dev['uuid'] in used_pvs.values()):
                list_unused_disks.append(dev)
    return list_unused_disks


def _parttype_is_not_extended(part_type):
    """Checks that part_type is not Extended.
    :param part_type: hex string
    """
    if part_type:
        try:
            return int(part_type, 16) != 5
        except ValueError:
            return True
    else:
        return True


def _is_block_dev(dev):
    """
    Checks if device is of available type
    :param dev: dict with device attrs
    """

    return dev['type'] in AVAILABLE_BLOCK_DEVICE_TYPES


def _filtered_children(dev):
    """
    Метод удаляет из списка children multipath-устройства.
    """

    only_usable_devices = []
    if dev.get('children'):
        for device in dev['children']:
            if _is_block_dev(device):
                only_usable_devices.append(device)
    return only_usable_devices


@states_response
def lv_absent_all(vgname):
    """
    Remove existing logicals volumes from a named existing volume group
    """

    lv_list = _get_lvs(vgname)
    if lv_list:
        for lv in lv_list:
            cmd_ret = lv_absent(name=lv, vgname=vgname)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
    return f'Logical Volumes for vg {vgname} removed.'


def _get_lvs(vgname):
    lv_list = []
    out = states_result(get_output, cmd='lvs').splitlines()
    for item in out[1:]:
        if item.find(vgname) != -1:
            lv_list.append(item.split()[0])
    return lv_list


@states_response
def create(vdisk_name, vgname, size, owner, thinvolume):
    cmd_ret = lv_present(name=vdisk_name, vgname=vgname, size=size, thinvolume=thinvolume)
    if not cmd_ret['result']:
        comment = f"An error occurred while creating lvm volume {vdisk_name}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        # Изменение owner-а
        source = f'/dev/{vgname}/{vdisk_name}'
        chown_cmd = "chown {user}:{group} $(readlink -f {source}".format(user=owner, group=owner,
                                                                         source=source)
        cmd_ret = get_output(chown_cmd)
        if not cmd_ret['result']:
            comment = f"An error occurred while changing owner of lvm {source}: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
    return 'success'


@states_response
def lvdisplay(lvname=''):
    """
    Return information about the logical volume(s)
    """
    ret = {}
    if lvname:
        cmd = f'lvdisplay -c {lvname}'
    else:
        cmd = 'lvdisplay -c'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        return ret

    out = cmd_ret['comment'].splitlines()
    for line in out:
        if 'open failed' in line:
            continue
        comps = line.strip().split(':')
        if len(comps) < 13:
            raise CommandExecutionError(f'Not all fields in {cmd}: {str(comps)}')
        ret[comps[0]] = {
            'Logical Volume Name': comps[0],
            'Volume Group Name': comps[1],
            'Logical Volume Access': comps[2],
            'Logical Volume Status': comps[3],
            'Internal Logical Volume Number': comps[4],
            'Open Logical Volumes': comps[5],
            'Logical Volume Size': comps[6],
            'Current Logical Extents Associated': comps[7],
            'Allocated Logical Extents': comps[8],
            'Allocation Policy': comps[9],
            'Read Ahead Sectors': comps[10],
            'Major Device Number': comps[11],
            'Minor Device Number': comps[12]}
    return ret


@states_response
def vgdisplay(vgname=''):
    """
    Return information about the volume group(s)
    """
    ret = {}
    cmd = f'vgdisplay -c {vgname}' if vgname else 'vgdisplay -c'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        return ret

    out = cmd_ret['comment'].splitlines()
    for line in out:
        if 'open failed' in line:
            continue
        comps = line.strip().split(':')
        if len(comps) < 17:
            raise CommandExecutionError(f'Not all fields in {cmd}: {str(comps)}')
        ret[comps[0]] = {
            'Volume Group Name': comps[0],
            'Volume Group Access': comps[1],
            'Volume Group Status': comps[2],
            'Internal Volume Group Number': comps[3],
            'Maximum Logical Volumes': comps[4],
            'Current Logical Volumes': comps[5],
            'Open Logical Volumes': comps[6],
            'Maximum Logical Volume Size': comps[7],
            'Maximum Physical Volumes': comps[8],
            'Current Physical Volumes': comps[9],
            'Actual Physical Volumes': comps[10],
            'Volume Group Size (kB)': comps[11],
            'Physical Extent Size (kB)': comps[12],
            'Total Physical Extents': comps[13],
            'Allocated Physical Extents': comps[14],
            'Free Physical Extents': comps[15],
            'UUID': comps[16]}
    return ret


@states_response
def pvdisplay(pvname='', real=False):
    """
    Return information about the physical volume(s)

    pvname
        physical device name
    real
        dereference any symlinks and report the real device
    """
    ret = {}
    cmd = f'pvdisplay -c {pvname}' if pvname else 'pvdisplay -c'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        return ret

    out = cmd_ret['comment'].splitlines()
    for line in out:
        if ('is a new physical volume' in line) or ('open failed' in line) or ('Error reading device' in line):
            continue
        comps = line.strip().split(':')
        if len(comps) < 11:
            raise CommandExecutionError(f'Not all fields in {cmd}: {str(comps)}')
        if real:
            device = os.path.realpath(comps[0])
        else:
            device = comps[0]
        ret[device] = {
            'Physical Volume Device': comps[0],
            'Volume Group Name': comps[1],
            'Physical Volume Size (kB)': comps[2],
            'Internal Physical Volume Number': comps[3],
            'Physical Volume Status': comps[4],
            'Physical Volume (not) Allocatable': comps[5],
            'Current Logical Volumes Here': comps[6],
            'Physical Extent Size (kB)': comps[7],
            'Total Physical Extents': comps[8],
            'Free Physical Extents': comps[9],
            'Allocated Physical Extents': comps[10]}
        if real:
            ret[device]['Real Physical Volume Device'] = device
    return ret


@states_response
def pvcreate(devices, override=True, **kwargs):
    """
    Set a physical device to be used as an LVM physical volume

    override
        Skip devices, if they are already an LVM physical volumes
    """
    if not devices:
        raise CommandExecutionError('Error: at least one device is required')
    if isinstance(devices, str):
        devices = devices.split(',')

    cmd = 'pvcreate -y'
    cmd_devices = ''
    for device in devices:
        if not os.path.exists(device):
            raise CommandExecutionError('{} does not exist'.format(device))
        # Verify pvcreate was successful
        cmd_ret = pvdisplay(device)
        if cmd_ret['result'] and not cmd_ret['comment']:
            cmd_devices += ' {}'.format(device)
        elif not override:
            raise CommandExecutionError('Device "{}" is already an LVM physical volume.'.format(device))
    cmd += cmd_devices
    valid = ('metadatasize', 'dataalignment', 'dataalignmentoffset',
             'pvmetadatacopies', 'metadatacopies', 'metadataignore',
             'restorefile', 'norestorefile', 'labelsector',
             'setphysicalvolumesize')
    for var in kwargs:
        if kwargs[var] and var in valid:
            cmd += ' --{} {}'.format(var, kwargs[var])

    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    sync_ret = data_sync(cmd_devices)
    if not sync_ret['result']:
        raise CommandExecutionError(sync_ret['comment'])

    partprobe_ret = partprobe(cmd_devices)
    if not partprobe_ret['result']:
        raise CommandExecutionError(partprobe_ret['comment'])

    check_ret = check_if_partitions(cmd_devices)
    if not check_ret['result']:
        raise CommandExecutionError(check_ret['comment'])

    # Verify pvcreate was successful
    pvscan()
    for device in devices:
        if not pvdisplay(device)['result']:
            raise CommandExecutionError(f'Device {device} was not affected.')
    return True


@states_response
def pvremove(devices, override=True):
    """
    Remove a physical device being used as an LVM physical volume

    override
        Skip devices, if they are already not used as an LVM physical volumes
    """
    if isinstance(devices, str):
        devices = devices.split(',')
    cmd = 'pvremove -y'
    for device in devices:
        if pvdisplay(device)['result']:
            cmd += ' {}'.format(device)
        elif not override:
            raise CommandExecutionError(f'{device} is not a physical volume')

    if not cmd[2:]:
        # Nothing to do
        return True

    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    # Verify pvcremove was successful
    pvscan()
    for device in devices:
        if pvdisplay(device):
            raise CommandExecutionError(f'Device {device} was not affected.')

    return True


@states_response
def vgcreate(vgname, devices, **kwargs):
    """
    Create an LVM volume group
    """
    if not vgname or not devices:
        return 'Error: vgname and device(s) are both required'
    if isinstance(devices, str):
        devices = devices.split(',')
    cmd = 'vgcreate {}'.format(vgname)
    for device in devices:
        cmd += ' {}'.format(device)
    valid = ('clustered', 'maxlogicalvolumes', 'maxphysicalvolumes',
             'vgmetadatacopies', 'metadatacopies', 'physicalextentsize')
    for var in kwargs:
        if kwargs[var] and var in valid:
            cmd += ' --{} {}'.format(var, kwargs[var])
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    out = cmd_ret['comment'].splitlines()
    vgdata = vgdisplay(vgname)['comment']
    vgdata['Output from vgcreate'] = out[0].strip()
    return vgdata


@states_response
def vgextend(vgname, devices):
    """
    Add physical volumes to an LVM volume group
    """
    if not vgname or not devices:
        raise CommandExecutionError('Error: vgname and device(s) are both required')
    if isinstance(devices, str):
        devices = devices.split(',')
    cmd = 'vgextend {}'.format(vgname)
    for device in devices:
        cmd += ' {}'.format(device)
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    out = cmd_ret['comment'].splitlines()
    vgdata = {'Output from vgextend': out[0].strip()}
    return vgdata


@states_response
def lvcreate(lvname,
             vgname,
             size=None,
             extents=None,
             snapshot=None,
             pv=None,
             thinvolume=False,
             thinpool=False,
             **kwargs):
    """
    Create a new logical volume, with option for which physical volume to be used
    """

    if size and extents:
        raise CommandExecutionError('Error: Please specify only one of size or extents')
    if thinvolume and thinpool:
        raise CommandExecutionError('Error: Please set only one of thinvolume or thinpool to True')

    valid = ('activate', 'chunksize', 'contiguous', 'discards', 'stripes',
             'stripesize', 'minor', 'persistent', 'mirrors', 'noudevsync',
             'monitor', 'ignoremonitoring', 'permission', 'poolmetadatasize',
             'readahead', 'regionsize', 'type',
             'virtualsize', 'zero')
    no_parameter = ('noudevsync', 'ignoremonitoring', 'thin',)

    extra_arguments = []
    if kwargs:
        for k, v in six.iteritems(kwargs):
            if k in no_parameter:
                extra_arguments.append(f'--{k}')
            elif k in valid:
                extra_arguments.extend([f'--{k}', f'{v}'])

    cmd = 'lvcreate --yes'

    if thinvolume:
        cmd += f' --thin -n {lvname}'
    elif thinpool:
        cmd += f' --thinpool {lvname}'
    else:
        cmd += f' -n {lvname}'

    if snapshot:
        cmd += f' -s {vgname}/{snapshot}'
    else:
        cmd += f' {vgname}'

    if size and thinvolume:
        cmd += f' -V {size}'
    elif extents and thinvolume:
        raise CommandExecutionError('Error: Thin volume size cannot be specified as extents')
    elif size:
        cmd += f' -L {size}'
    elif extents:
        cmd += f' -l {extents}'
    else:
        raise CommandExecutionError('Error: Either size or extents must be specified')

    if pv:
        cmd += f' {pv}'
    if extra_arguments:
        cmd += f" {' '.join(extra_arguments)}"

    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    out = cmd_ret['comment'].splitlines()
    if thinvolume:
        lvdev = f"/dev/{vgname.split('/')[0]}/{lvname}"
    else:
        lvdev = f'/dev/{vgname}/{lvname}'
    lv_ret = lvdisplay(lvdev)
    if not lv_ret['result']:
        raise CommandExecutionError(lv_ret['comment'])
    lvdata = lv_ret['comment']
    lvdata['Output from lvcreate'] = out[0].strip()
    return lvdata


@states_response
def vgremove(vgname):
    """
    Remove an LVM volume group
    """
    return states_result(get_output, cmd=f'vgremove -f {vgname}').strip()


@states_response
def lvremove(lvname, vgname):
    """
    Remove a given existing logical volume from a named existing volume group
    """
    cmd = f'lvremove -f {vgname}/{lvname}' if lvname else f'lvremove -f {vgname}'
    return states_result(get_output, cmd=cmd).strip()


@states_response
def lvresize(size, lvpath):
    """
    Return information about the logical volume(s)
    """
    return states_result(get_output, cmd=f'lvresize -L {str(size)} {lvpath}')


@states_response
def pv_present(name, **kwargs):
    """
    Set a physical device to be used as an LVM physical volume

    name
        The device name to initialize.

    kwargs
        Any supported options to pvcreate.
    """
    cmd_ret = pvdisplay(name)
    if cmd_ret['result'] and cmd_ret['comment']:
        return f'Physical Volume {name} already present'
    else:
        cmd_ret = pvcreate(name, **kwargs)
        if not cmd_ret['result']:  # pylint: disable=unsubscriptable-object
            raise CommandExecutionError(cmd_ret['comment'])
        cmd_ret = pvdisplay(name)
        if cmd_ret['result'] and cmd_ret['comment']:
            return f'Created Physical Volume {name}'
        else:
            raise CommandExecutionError(f'Failed to create Physical Volume {name}')


@states_response
def pv_absent(name):
    """
    Ensure that a Physical Device is not being used by lvm

    name
        The device name to initialize.
    """
    # pylint: disable=unsubscriptable-object
    cmd_ret = pvdisplay(name)
    if cmd_ret['result'] and not cmd_ret['comment']:
        return 'Physical Volume {0} does not exist'.format(name)

    cmd_ret = pvremove(name)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    cmd_ret = pvdisplay(name)
    if cmd_ret['result'] and cmd_ret['comment']:
        raise CommandExecutionError('Failed to remove Physical Volume {0}'.format(name))
    return 'Removed Physical Volume {0}'.format(name)


@states_response
def vg_present(name, devices=None, **kwargs):
    """
    Create an LVM volume group

    name
        The volume group name to create

    devices
        A list of devices that will be added to the volume group

    kwargs
        Any supported options to vgcreate.
    """

    cmd_ret = vgdisplay(name)
    if cmd_ret['result'] and cmd_ret['comment']:
        comment = f'Volume Group {name} already present'
        for device in devices:
            realdev = os.path.realpath(device)
            cmd_ret = pvdisplay(realdev, real=True)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            pvs = cmd_ret['comment']
            if pvs and pvs.get(realdev, None):
                if pvs[realdev]['Volume Group Name'] == name:
                    comment += '\n{}'.format(f'{device} is part of Volume Group')
                else:
                    comment += '\n{}'.format(f"{device} is part of {pvs[realdev]['Volume Group Name']}")
                    raise CommandExecutionError(comment)
            else:
                comment += '\n{}'.format('pv {} is not present'.format(device))
                raise CommandExecutionError(comment)
    else:
        cmd_ret = vgcreate(name, devices, **kwargs)
        if not cmd_ret['result']:
            raise CommandExecutionError(f'Failed to create volume group {name}: {cmd_ret["comment"]}')
        cmd_ret = vgdisplay(name)
        if cmd_ret['result'] and cmd_ret['comment']:
            return f'Created Volume Group {name}'
        else:
            raise CommandExecutionError(f'Failed to check volume group {name}: {cmd_ret["comment"]}')


@states_response
def vg_absent(vgname):
    """
    Remove an LVM volume group

    vgname
        The volume group to remove
    """
    cmd_ret = vgdisplay(vgname)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    cmd_ret = vgremove(vgname)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    cmd_ret = vgdisplay(vgname)
    if cmd_ret['result']:
        return f'Removed Volume Group {vgname}'
    else:
        raise CommandExecutionError(cmd_ret['comment'])


@states_response
def lv_present(name, vgname=None, size=None, extents=None,
               snapshot=None, pv='', thinvolume=False, thinpool=False, **kwargs):
    """
    Create a new logical volume

    name
        The name of the logical volume
    vgname
        The volume group name for this logical volume
    size
        The initial size of the logical volume
    extents
        The number of logical extents to allocate
    snapshot
        The name of the snapshot
    pv
        The physical volume to use
    kwargs
        Any supported options to lvcreate. See
    thinvolume
        Logical volume is thinly provisioned
    thinpool
        Logical volume is a thin pool
    """
    _snapshot = None

    if snapshot:
        _snapshot = name
        name = snapshot

    if thinvolume:
        lvpath = f"/dev/{vgname.split('/')[0]}/{name}"
    else:
        lvpath = f'/dev/{vgname}/{name}'

    cmd_ret = lvdisplay(lvpath)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if cmd_ret['comment']:
        return f'Logical Volume {name} already present'
    cmd_ret = lvcreate(name, vgname, size=size, extents=extents,
                       snapshot=_snapshot, pv=pv,
                       thinvolume=thinvolume, thinpool=thinpool,
                       **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    cmd_ret = lvdisplay(lvpath)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if cmd_ret['comment']:
        return f'Created Logical Volume {name}'
    else:
        raise CommandExecutionError(f'Failed to create Logical Volume {name}')


@states_response
def lv_absent(vgname, name=None):
    """
    Remove a given existing logical volume from a named existing volume group

    name
        The logical volume to remove

    vgname
        The volume group name
    """
    if name:
        lvpath = f'/dev/{vgname}/{name}'
    else:
        lvpath = f'/dev/{vgname}'
    cmd_ret = lvdisplay(lvpath)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if not cmd_ret['comment']:
        return f'Logical Volume {name} already absent'

    cmd_ret = lvremove(name, vgname)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    cmd_ret = lvdisplay(lvpath)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if lvpath in cmd_ret['comment']:
        raise CommandExecutionError(f'Couldn\'t remove Logical Volume {lvpath}')
    return f'Removed Logical Volume {name}'


@states_response
def wipefs(device):
    cmd = f'wipefs {device} -a'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def vgs_create(vgname, dev_names):
    """
    Функция создания группы томов на узле.
    """

    for device in dev_names:
        cmd_ret = wipefs(device)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        cmd_ret = pv_present(device)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    cmd_ret = vg_present(vgname, dev_names)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def vgs_extend(vgname, dev_names, thin_pool_name):
    """
    Функции расширения группы томов на ноде.
    """

    for device in dev_names:
        cmd_ret = pv_present(device)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])

    cmd_ret = vgextend(vgname, dev_names)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if thin_pool_name:
        cmd = 'lvextend {} {} -l100%VG'.format(vgname, thin_pool_name)
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def created(block_shared, vgname, meta_volume, metadata, **kwargs):
    if block_shared:
        cmd_ret = vgs_create(vgname, dev_names=kwargs['dev_names'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    cmd_ret = lv_absent_all(vgname)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    cmd_ret = lv_present(meta_volume, vgname, size='4M', thinpool=False)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    metadata = '\n'.join(metadata)
    cmd = "echo '{}' > {}".format(metadata, f'/dev/{vgname}/{meta_volume}')
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return 'Successfully created lvm datapool.'


@states_response
def created_thin(vgname, meta_volume, metadata, thin_pool_name, **kwargs):
    cmd_ret = lv_absent_all(vgname)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    cmd_ret = lv_present(thin_pool_name, vgname, extents='100%FREE', thinpool=True)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    cmd_ret = lv_present(meta_volume, '{}/{}'.format(vgname, thin_pool_name), size='4M', thinvolume=True)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    metadata = '\n'.join(metadata)
    cmd = "echo '{}' > {}".format(metadata, f'/dev/{vgname}/{meta_volume}')
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return 'Successfully created thin lvm datapool.'


@states_response
def deleted(block_shared, vgname, **kwargs):
    cmd_ret = lv_absent(vgname)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if block_shared:
        cmd_ret = vg_absent(vgname)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    return 'Successfully deleted lvm datapool.'


@states_response
def deleted_thin(vgname, **kwargs):
    cmd_ret = lv_absent(vgname)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return 'Successfully deleted thin lvm datapool.'


@states_response
def lvscan():
    return states_result(get_output, cmd='lvscan -q --readonly')


@states_response
def pvscan():
    return states_result(get_output, cmd='pvscan')


@states_response
def vgscan():
    return states_result(get_output, cmd='vgscan -q --nolocking')


@states_response
def data_sync(devices):
    if isinstance(devices, list):
        devices = ' '.join(devices)
    cmd = f'sync {devices}'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(f'Failed into: {cmd}. Error: {cmd_ret["comment"]}')
    return cmd_ret['comment']


@states_response
def check_if_partitions(devices):
    if isinstance(devices, list):
        devices = ' '.join(devices)
    cmd = f'lsblk -J {devices}'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(f'Failed into: {cmd}. Error: {cmd_ret["comment"]}')
    inf_devices = json.loads(cmd_ret['comment'])
    for dev in inf_devices['blockdevices']:
        if 'children' in dev and dev['children']:
            raise CommandExecutionError(f'Device {dev["name"]} has partitions.')
    return inf_devices


@states_response
def partprobe(devices):
    if isinstance(devices, list):
        devices = ' '.join(devices)
    cmd = f'partprobe {devices}'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(f'Failed into: {cmd}. Error: {cmd_ret["comment"]}')
    return cmd_ret['comment']


@states_response
def get_lvm_datapool_id_from_meta(meta_file, **kwargs):
    cmd = LVM_DATAPOOL_CMD.format(meta_file=meta_file)
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret["comment"])
    return json.loads(cmd_ret['comment'])["datapool_id"]
