"""
(c) 2023 LLC «DACOM M»
"""
import os
import tempfile
from defusedxml import ElementTree as ET

from veilrpc.states.domain import get_domain_dumpxml, delete_xml
from veilrpc.utils.bazalt import states_response, CommandExecutionError, call, ZPOOL_STORAGES_PATH, \
    DOMAINS_FOLDER
from veilrpc.utils.file import chown_directory
from veilrpc.states.copy import popen_with_progress


@states_response
def shared(domain_id, target_node_ip, ssh_user, private_ssh_key,
           cdroms_xml_map, vmachine_infs_xml_map, vdisks_xml_map, unsafe=False,
           user_opts=None, **kwargs):
    # Получение и изменение текущей xml
    cmd_ret = get_domain_dumpxml(domain_id, migratable=True)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    tmp_ssh_fh, tmp_ssh_key_path = tempfile.mkstemp()
    tmp_xml_fh, tmp_xml_path = tempfile.mkstemp()
    ssh_cmd = "ssh -o \"StrictHostKeyChecking no\" -i {} {}@{} ".format(tmp_ssh_key_path, ssh_user, target_node_ip)
    try:
        with os.fdopen(tmp_ssh_fh, 'w') as ssh_key_file:
            ssh_key_file.write(private_ssh_key)
        xml = mutate_xml(cmd_ret['comment'], cdroms_xml_map, vmachine_infs_xml_map, vdisks_xml_map)
        with os.fdopen(tmp_xml_fh, 'w') as xml_file:
            xml_file.write(xml)
        opts = "--live"
        if unsafe:
            opts += " --unsafe"
        if user_opts:
            opts += user_opts if user_opts.startswith(' ') else f" {user_opts}"
        # Миграция
        virsh_cmd = "virsh migrate --domain {domain_id} --migrateuri tcp://{target_node_ip} {opts} --xml {xml_file} " \
                    "--verbose qemu+ssh://{ssh_user}@{target_node_ip}/system?keyfile={keyfile}\\&no_verify=1 2>&1" \
            .format(domain_id=domain_id, ssh_user=ssh_user, opts=opts,
                    keyfile=tmp_ssh_key_path, target_node_ip=target_node_ip,
                    xml_file=tmp_xml_path)
        progress_data = dict(type='progress', jid=kwargs.get('task', ''), id='', name='shared', result=True)
        p = popen_with_progress(virsh_cmd, "blockcommit", progress_data, kwargs.get('redis'))
        # так как прогресс отправляется в stderr, то в virsh_cmd мы переводим его в stdout "2>&1",
        # соответственно ошибка выполнения тоже попадет в stdout
        cmd_ret = {'retcode': p.returncode, 'error': str(p.stderr.read()) + str(p.stdout.read()) + str(
            p.comment if hasattr(p, "comment") else "")}
        if cmd_ret['retcode']:
            comment = f"An error with retcode {cmd_ret['retcode']} occurred while migration: {cmd_ret['error']}," \
                      f" mutate_xml: {xml}"
            raise CommandExecutionError(comment)
        delete_xml(domain_id)
        xml_cmd = f'echo \"{xml}\" > {DOMAINS_FOLDER}/{domain_id}.xml'
        cmd = f"{ssh_cmd}'{xml_cmd}'"
        cmd_ret = call(cmd, timeout=60)
    except KeyError as e:
        comment = f"KeyError in mutating xml: {e}, domain_xml: {cmd_ret['comment']}"
        raise CommandExecutionError(comment) from None
    except Exception as e:
        comment = f"Unknown error: {e}"
        raise CommandExecutionError(comment) from None
    finally:
        os.remove(tmp_ssh_key_path)
        os.remove(tmp_xml_path)

    return f"Domain {domain_id} successfully shared migrated."


@states_response
def non_shared(domain_id, target_node_ip, ssh_user, private_ssh_key,
               cdroms_xml_map, vmachine_infs_xml_map, vdisks_xml_map, vdisks, unsafe=False,
               user_opts=None, **kwargs):
    comment = 'Success.'
    # Получение и изменение текущей xml
    cmd_ret = get_domain_dumpxml(domain_id, migratable=True)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    tmp_ssh_fh, tmp_ssh_key_path = tempfile.mkstemp()
    tmp_xml_fh, tmp_xml_path = tempfile.mkstemp()
    ssh_cmd = "ssh -o \"StrictHostKeyChecking no\" -i {} {}@{} ".format(tmp_ssh_key_path, ssh_user, target_node_ip)
    try:
        with os.fdopen(tmp_ssh_fh, 'w') as ssh_key_file:
            ssh_key_file.write(private_ssh_key)
        chown_directory(tmp_ssh_key_path, ssh_user)
        xml = mutate_xml(cmd_ret['comment'], cdroms_xml_map, vmachine_infs_xml_map, vdisks_xml_map)
        with os.fdopen(tmp_xml_fh, 'w') as xml_file:
            xml_file.write(xml)
        vdisk_pattern = 'qemu-img create -f {format} -o preallocation={preallocation} "{path}/{vdisk_name}" ' \
                        '{virtual_size}'
        for vdisk_arg in vdisks:
            if vdisk_arg['virtual_size'] == 0:
                vdisk_arg['virtual_size'] = vdisk_arg['size'] * 1024 ** 3
            if f'{ZPOOL_STORAGES_PATH}/' in vdisk_arg['path']:
                from .copy import create_delete_remote_zfs_dataset
                zfs_ret = create_delete_remote_zfs_dataset(
                    True, vdisk_arg['path'].split(f'{ZPOOL_STORAGES_PATH}/')[-1], vdisk_arg['vdisk_name'].split('.qcow2')[0],
                    {'private_ssh_key': private_ssh_key, 'user': ssh_user, 'destination_ip': target_node_ip}, ssh_user)
                if not zfs_ret['result']:
                    raise CommandExecutionError(zfs_ret['comment'])
                vdisk_arg['path'] = f'{vdisk_arg["path"]}/{vdisk_arg["vdisk_name"].split(".qcow2")[0]}'
                vdisk_arg['vdisk_name'] = 'vdisk.qcow2'
            vdisk_cmd = vdisk_pattern.format(**vdisk_arg)
            cmd = ssh_cmd + vdisk_cmd
            cmd_ret = call(cmd, timeout=60)
            if not cmd_ret['result']:
                comment = f"An error occurred while creating disks. Error: {cmd_ret['comment']}. command: {cmd}"
                raise CommandExecutionError(comment)
        opts = "--live"
        if unsafe:
            opts += " --unsafe"
        if user_opts:
            opts += user_opts if user_opts.startswith(' ') else f" {user_opts}"
        # Миграция
        virsh_cmd = "virsh migrate --domain {domain_id} --migrateuri tcp://{target_node_ip} " \
                    "{opts} --copy-storage-all --xml {xml_file} --verbose " \
                    "qemu+ssh://{ssh_user}@{target_node_ip}/system?keyfile={keyfile}\\&no_verify=1 2>&1" \
            .format(domain_id=domain_id, ssh_user=ssh_user, opts=opts,
                    keyfile=tmp_ssh_key_path, target_node_ip=target_node_ip,
                    xml_file=tmp_xml_path)
        progress_data = dict(type='progress', jid=kwargs.get('task', ''), id='', name='non_shared', result=True)
        p = popen_with_progress(virsh_cmd, "blockcommit", progress_data, kwargs.get('redis'))
        # так как прогресс отправляется в stderr, то в virsh_cmd мы переводим его в stdout "2>&1",
        # соответственно ошибка выполнения тоже попадет в stdout
        cmd_ret = {'retcode': p.returncode, 'error': str(p.stderr.read()) + str(p.stdout.read()) + str(
            p.comment if hasattr(p, "comment") else "")}
        if cmd_ret['retcode']:
            comment = f"An error occurred while migration: {cmd_ret['error']}, mutate_xml: {xml}"
            raise CommandExecutionError(comment)
        delete_xml(domain_id)
        xml_cmd = f'echo \"{xml}\" > {DOMAINS_FOLDER}/{domain_id}.xml'
        cmd = f"{ssh_cmd}'{xml_cmd}'"
        cmd_ret = call(cmd, timeout=60)
    except KeyError as e:
        for vdisk_arg in vdisks:
            cmd = ssh_cmd + f"rm -f {vdisk_arg['path']}/{vdisk_arg['vdisk_name']}"
            cmd_ret = call(cmd, timeout=60)
            if not cmd_ret['result']:
                comment += f"An error occurred while remote deleting vdisk " \
                           f"{vdisk_arg['path']}/{vdisk_arg['vdisk_name']}: {cmd_ret['comment']}. command: {cmd}"
            else:
                comment += f"Successfully deleting remote vdisk {vdisk_arg['path']}/{vdisk_arg['vdisk_name']}."

        raise CommandExecutionError(f"KeyError in mutating xml: {e}, domain_xml: "
                                    f"{cmd_ret['comment']}. Comment: {comment}") from None
    except Exception as e:
        for vdisk_arg in vdisks:
            cmd = ssh_cmd + f"rm -f {vdisk_arg['path']}/{vdisk_arg['vdisk_name']}"
            cmd_ret = call(cmd, timeout=60)
            if not cmd_ret['result']:
                comment += f"An error occurred while deleting remote vdisk " \
                           f"{vdisk_arg['path']}/{vdisk_arg['vdisk_name']}: {cmd_ret['comment']}. command: {cmd}"
            else:
                comment += f"Successfully deleting remote vdisk {vdisk_arg['path']}/{vdisk_arg['vdisk_name']}."

        raise CommandExecutionError(f"Unknown error: {e}. Comment: {comment}") from None
    else:
        from .vdisk import deleted, deleted_meta
        for vdisk_arg in vdisks:
            vdisk_path = os.path.join(vdisk_arg['old_path'], vdisk_arg['vdisk_name'])
            del_ret = deleted(path=vdisk_path, owner=ssh_user)
            if not del_ret['result']:
                comment += f"An error occurred while deleting disks: {cmd_ret['comment']}"
            vdisk_meta_path = os.path.join(vdisk_arg['old_path'], vdisk_arg['vdisk_name'].replace('qcow2', 'meta'))
            del_ret = deleted_meta(path=vdisk_meta_path, owner=ssh_user)
            if not del_ret['result']:
                comment += f"An error occurred while deleting disk meta: {cmd_ret['comment']}"
            if ZPOOL_STORAGES_PATH in vdisk_arg['old_path']:
                from .copy import delete_local_zfs_dataset
                del_zfs_d_ret = delete_local_zfs_dataset(vdisk_arg['old_path'].split(f'{ZPOOL_STORAGES_PATH}/')[-1],
                                                         vdisk_arg['vdisk_name'].split('.qcow2')[0])
                if not del_zfs_d_ret['result']:
                    comment += del_zfs_d_ret['comment']
    finally:
        os.remove(tmp_ssh_key_path)
        os.remove(tmp_xml_path)
    comment += f"Domain {domain_id} successfully non_shared migrated."
    return comment


# Ниже идут функции преобразующие xml полученную через dumpxml в желаемое состояние.
# Код заимствован из модуля xml_mutator подсистемы снепшотов. Там аналогичная задача
# возникала при ревёрте к снепшоту.

def mutate_xml(xml_dump, cdroms_xml_map, vinterfaces_xml_map, vdisks_xml_map):
    root = ET.fromstring(xml_dump)
    cdroms, vinterfaces, vdisks = [], [], []
    for disk_tag in root.iterfind('devices/disk'):
        if disk_tag.attrib['device'] == 'cdrom':
            cdroms.append(disk_tag)
        elif disk_tag.attrib['device'] == 'disk':
            vdisks.append(disk_tag)
    for vinterface in root.iterfind('devices/interface'):
        vinterfaces.append(vinterface)

    usb_mutation(root)
    cdroms_mutation(cdroms, cdroms_xml_map)
    vdisks_mutation(vdisks, vdisks_xml_map)
    vinterfaces_mutation(vinterfaces, vinterfaces_xml_map)
    return ET.tostring(root, encoding='unicode')


def vdisks_mutation(vdisks_xml_objects, vdisks_xml_map):
    for vdisk in vdisks_xml_objects:
        target_dev = vdisk.find('target').attrib['dev']
        address = vdisk.find('address')
        desired_xml = ET.fromstring(vdisks_xml_map[target_dev])
        desired_xml.append(address)
        vdisk.clear()
        vdisk.extend(list(desired_xml))
        vdisk.attrib.update(desired_xml.attrib)


def cdroms_mutation(cdroms_xml_objects, cdroms_xml_map):
    for cdrom in cdroms_xml_objects:
        target_dev = cdrom.find('target').attrib['dev']
        address = cdrom.find('address')
        desired_xml = ET.fromstring(cdroms_xml_map[target_dev])
        desired_xml.append(address)
        cdrom.clear()
        cdrom.extend(list(desired_xml))
        cdrom.attrib.update(desired_xml.attrib)


def vinterfaces_mutation(vinterfaces_xml_objects, vinterfaces_xml_map):
    for vinterface in vinterfaces_xml_objects:
        mac = vinterface.find('mac').attrib['address']
        address = vinterface.find('address')
        desired_xml = ET.fromstring(vinterfaces_xml_map[mac])
        desired_xml.append(address)
        vinterface.clear()
        vinterface.extend(list(desired_xml))
        vinterface.attrib.update(desired_xml.attrib)


def usb_mutation(root):
    # Удаление usb устройств
    for device in root.findall("devices"):
        for usb in device.findall("hostdev"):
            device.remove(usb)
