"""
(c) 2023 LLC «DACOM M»
"""
from typing import Optional, Union

import os
import yaml
import tempfile
import shutil

from veilrpc.utils.bazalt import CommandExecutionError, states_response, get_output, states_result,\
    DATA_PATH, LOCAL_MODULES_PATH, MODULES_PATH, DATA_PUPPET_PATH, BASE_PUPPET_PATH, ROLE_SCRIPT_PATH, \
    CONTROLLER_USER, BASE_NAME
from veilrpc.utils.file import chown_directory, set_mode

EAML_KEYS_STR = "--pkcs7-private-key=/etc/puppetlabs/puppet/eyaml/private_key.pkcs7.pem " \
                "--pkcs7-public-key=/etc/puppetlabs/puppet/eyaml/public_key.pkcs7.pem"
PUPPET_COMMON_DEFAULT_EYAML_PATH = f'{DATA_PUPPET_PATH}/common.eyaml'
PUPPET_COMMON_EYAML_PATH = f'{DATA_PATH}/common.eyaml'


class QuotedYaml(str):
    pass


class QuotedYaml2(str):
    pass


def _quoted_presenter(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='"')


def _quoted_presenter2(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style="'")


yaml.add_representer(QuotedYaml, _quoted_presenter)

yaml.add_representer(QuotedYaml2, _quoted_presenter2)


def _update_dict(d: dict, u: dict):
    """Deep update, merge dict to another dict."""
    for k, v in u.items():
        if isinstance(v, dict):
            d.setdefault(k, {})
            _update_dict(d[k], v)
        else:
            d[k] = v


def _qupdate_yaml(value):
    if isinstance(value, str) and "%{" in value:
        return QuotedYaml(value)
    if isinstance(value, str) and "%quote%" in value:
        return QuotedYaml2(value.replace('%quote%', ''))
    return value


def _clear_dict(d: dict):
    """
    It's return "dict" replace "str" to "QuotedYaml" if "%{".
    AND remove None params.
    Example:
    $ d
    {'a': 1, 'b': {'b1': 22}, 'c': None, 'd': {'d1': None}}
    $ _clear_dict(d)
    {'a': 1, 'b': {'b1': 22}}
    """
    if isinstance(d, dict):
        tmp_d = dict(filter(lambda _: _[1] is not None,
                            map(lambda _: (_qupdate_yaml(_[0]), _clear_dict(_[1])),
                                d.items())))
        return tmp_d if tmp_d else None
    elif isinstance(d, list):
        tmp_d = list(filter(lambda x: x is not None, map(_clear_dict, d)))
        return tmp_d if tmp_d else None
    else:
        return _qupdate_yaml(d)


@states_response
def puppet_apply(modules: Union[str, list], **kwargs) -> str:
    if isinstance(modules, str):
        modules = [modules]
    if 'apt' in modules:
        modules.append('space_repos')
    modules = ','.join(set(modules))
    cmd = f'{ROLE_SCRIPT_PATH} -a {modules} 2>&1'
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        msg = '\n'.join(cmd_ret['comment'].splitlines()[-4:])
        comment = f"Unable to apply {modules} config: {msg}"
        raise CommandExecutionError(comment)
    return f"Modules {modules} successfully applied"


@states_response
def puppet_modify(module: str = 'common', params: Optional[dict] = None, local: bool = True, encrypt: bool = False,
                  apply: bool = True, reset_default: bool = False,
                  apply_modules: list = [], **kwargs) -> Union[dict, str]:
    """
    return dict if params=None else modify params and return string
    """

    if module in ['common', 'veilrole']:
        pref_ext_enc = 'e' if module == 'common' and encrypt else ''
        write_yaml_path = f'{DATA_PATH}/{module}.{pref_ext_enc}yaml'
        if os.path.exists(f'{DATA_PATH}/{module}.{pref_ext_enc}yaml'):
            read_yaml_path = f'{DATA_PATH}/{module}.{pref_ext_enc}yaml'
            if reset_default:
                try:
                    os.remove(read_yaml_path)
                except OSError as e:
                    raise CommandExecutionError(e)
        else:
            read_yaml_path = f'{DATA_PUPPET_PATH}/{module}.{pref_ext_enc}yaml'
    else:
        write_yaml_path = f'{LOCAL_MODULES_PATH}/{module}.yaml' if local else f'{MODULES_PATH}/{module}.yaml'
        if local and os.path.exists(f'{LOCAL_MODULES_PATH}/{module}.yaml'):
            read_yaml_path = f'{LOCAL_MODULES_PATH}/{module}.yaml'
            if reset_default:
                try:
                    os.remove(read_yaml_path)
                except OSError as e:
                    raise CommandExecutionError(e)
        elif not local and os.path.exists(f'{MODULES_PATH}/{module}.yaml'):
            read_yaml_path = f'{MODULES_PATH}/{module}.yaml'
            if reset_default:
                try:
                    os.remove(read_yaml_path)
                except OSError as e:
                    raise CommandExecutionError(e)
        elif os.path.exists(f'{BASE_PUPPET_PATH}/{module}.yaml'):
            read_yaml_path = f'{BASE_PUPPET_PATH}/{module}.yaml'
        else:
            read_yaml_path = None
    if not reset_default:
        conf = {}
        if read_yaml_path:
            try:
                with open(read_yaml_path, 'r') as f:
                    conf = yaml.safe_load(f)
            except (OSError, yaml.YAMLError) as e:
                raise CommandExecutionError(e)
        if params is None:
            return conf
        else:
            _update_dict(conf, params)
            conf = _clear_dict(conf)
        if not conf:
            conf = {}
        try:
            with open(write_yaml_path, 'w') as config:
                yaml.dump(conf, config)
        except Exception as err:
            comment = f'Unable to write {module} configuration: {err}'
            raise CommandExecutionError(comment) from None
        chown_directory(write_yaml_path, kwargs.get('owner', CONTROLLER_USER), kwargs.get('group', BASE_NAME))
        set_mode(write_yaml_path, '0664')
    if apply:
        modules = list(set([module] + apply_modules))
        ret = puppet_apply(modules, **kwargs)
        if not ret['result']:
            raise CommandExecutionError(ret['comment'])
        if reset_default:
            return f'Module {module} reset default state and applied'
        return f'Module {module} successfully modified and applied'
    if reset_default:
        return f'Module {module} reset default state'
    return f'Module {module} successfully modified'


@states_response
def puppet_eyaml_decrypt(value, **kwargs):
    ret = get_output(f"eyaml decrypt {EAML_KEYS_STR} -s '{value}'", **kwargs)
    if ret['result']:
        return ret['comment'].strip()
    return None


@states_response
def puppet_eyaml_encrypt(key, value, **kwargs):
    res = states_result(get_output,
                        cmd=f"eyaml encrypt {EAML_KEYS_STR} -l '{key}' -s '{value}' --quiet --output=string",
                        **kwargs)
    _, hash_value = res.split(":")
    hash_value = hash_value.strip()
    return hash_value


@states_response
def puppet_eyaml_common(key, value, **kwargs) -> bool:
    ret = puppet_eyaml_encrypt(key, value)
    if ret['result']:
        ret_2 = puppet_modify(params={key: ret['comment']}, encrypt=True, apply=kwargs.get('apply', False), **kwargs)
        if ret_2['result']:
            return True
    return False


@states_response
def puppet_modify_from_file(file: str, local: bool = True, apply_modules: list = [], **kwargs) -> Union[dict, str]:
    """
    Modify yaml module from passed file.
    """

    if os.path.exists(file):
        try:
            with open(file, 'r') as tmp:
                tmp_conf = yaml.safe_load(tmp)
        except (OSError, yaml.YAMLError) as e:
            raise CommandExecutionError(e)
    else:
        raise CommandExecutionError(f'File {file} not found!')
    if not tmp_conf:
        raise CommandExecutionError('File is empty, nothing modified.')
    modules_dicts = {}
    for k, v in tmp_conf.items():
        module_name = k.split('::')[0]
        if module_name in modules_dicts:
            modules_dicts[module_name].update({k: v})
        else:
            modules_dicts[module_name] = {k: v}
    if not modules_dicts:
        raise CommandExecutionError('Modules not found, nothing modified.')
    modules = list(set(list(modules_dicts.keys()) + apply_modules))
    list_save_files = []
    for module in modules_dicts.keys():
        save_yaml_path = f'{LOCAL_MODULES_PATH}/{module}.yaml' if local else f'{MODULES_PATH}/{module}.yaml'
        if os.path.exists(save_yaml_path):
            list_save_files.append(save_yaml_path)
    tmp_path = None
    map_save_files = {}
    if list_save_files:
        tmp_path = tempfile.mkdtemp()
        map_save_files = {save_file: f'{tmp_path}/{save_file.split("/")[-1]}' for save_file in list_save_files}
        try:
            for src, dst in map_save_files.items():
                shutil.copyfile(src, dst)
        except Exception as e:
            shutil.rmtree(tmp_path, ignore_errors=True)
            raise CommandExecutionError(e)
    for module, conf in modules_dicts.items():
        ret = puppet_modify(module=module, params=conf, local=local, apply=False, **kwargs)
        if not ret['result']:
            if map_save_files:
                try:
                    for dst, src in map_save_files.items():
                        if os.path.exists(dst):
                            os.remove(dst)
                        shutil.copyfile(src, dst)
                except Exception as e:
                    raise CommandExecutionError(ret['comment'] + '\n' + str(e))
                sub_ret = puppet_apply(modules, **kwargs)
                if not sub_ret['result']:
                    raise CommandExecutionError(ret['comment'] + '\n' + sub_ret['comment'])
                if tmp_path:
                    shutil.rmtree(tmp_path, ignore_errors=True)
            raise CommandExecutionError(f"{ret['comment']}\n\nModules were restored.")
    ret = puppet_apply(modules, **kwargs)
    if not ret['result']:
        if map_save_files:
            try:
                for dst, src in map_save_files.items():
                    if os.path.exists(dst):
                        os.remove(dst)
                    shutil.copyfile(src, dst)
            except Exception as e:
                raise CommandExecutionError(ret['comment'] + '\n' + str(e))
            sub_ret = puppet_apply(modules, **kwargs)
            if not sub_ret['result']:
                raise CommandExecutionError(ret['comment'] + '\n' + sub_ret['comment'])
            if tmp_path:
                shutil.rmtree(tmp_path, ignore_errors=True)
        raise CommandExecutionError(f"{ret['comment']}\n\nModules were restored.")
    if tmp_path:
        shutil.rmtree(tmp_path, ignore_errors=True)
    return f'Module {module} successfully modified and applied'


@states_response
def puppet_modify_modules_list() -> list:
    try:
        modules = os.listdir(MODULES_PATH)
        local_modules = os.listdir(LOCAL_MODULES_PATH)
    except Exception as e:
        raise CommandExecutionError(e)
    for module in local_modules:
        if index := next((i for i, m in enumerate(modules) if module == m), None):
            modules.pop(index)
    return [f'{LOCAL_MODULES_PATH}/{m}' for m in local_modules] + [f'{MODULES_PATH}/{m}' for m in modules]


@states_response
def puppet_modify_modules_read(modules: Optional[list] = None, **kwargs) -> str:
    if isinstance(modules, str):
        modules = [modules]
    module_paths = []
    if modules:
        for module in modules:
            if os.path.exists(f"{LOCAL_MODULES_PATH}/{module}.yaml"):
                module_paths.append(f"{LOCAL_MODULES_PATH}/{module}.yaml")
            elif os.path.exists(f"{MODULES_PATH}/{module}.yaml"):
                module_paths.append(f"{MODULES_PATH}/{module}.yaml")
            elif os.path.exists(f"{BASE_PUPPET_PATH}/{module}.yaml"):
                module_paths.append(f"{BASE_PUPPET_PATH}/{module}.yaml")
            else:
                raise CommandExecutionError(f'Module {module} is not exists.')
    else:
        ret = puppet_modify_modules_list()
        if not ret['result']:
            raise CommandExecutionError(ret['comment'])
        module_paths = ret['comment']
    text = ''
    for module in module_paths:
        text += f'# {module}\n'
        with open(module, 'r') as f:
            text += f.read()
        text += '\n'
    return text
