"""
(c) 2023 LLC «DACOM M»
"""

import json
import os
from pwd import getpwnam
from re import sub
import tempfile
import uuid
from xml.etree import ElementTree as ET  # nosec

from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output, usr_cmd, call, clean_guaranteed
from veilrpc.utils.size import convert


@states_response
def check_snapshots_consistency(domain_id, tree, current_state, domain_is_powered):
    if not domain_is_powered:
        return check_tree(tree)
    if check_tree(tree) and check_current_state(domain_id, current_state):
        return True
    else:
        return False


def check_current_state(domain_id, current_state):
    return current_state == get_domblklist(domain_id)


def check_tree(tree):
    for vdisk_snapshot in tree:
        source, backing_file = vdisk_snapshot['source'], vdisk_snapshot['backing_file']
        actual_backing_file = get_actual_backing_file(source)
        if not backing_file == actual_backing_file:
            return False
    return True


def get_actual_backing_file(source):
    if not os.path.exists(source):
        raise IOError
    backing_file = qemu_img_info(source).get('full-backing-filename', '')
    return backing_file


def get_domblklist(domain_id):
    virsh_cmd = f"virsh domblklist {domain_id}"
    cmd_ret = get_output(virsh_cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(f"Can't read domblklist of {domain_id}")
    return parse_domblklist(cmd_ret['comment'])


def parse_domblklist(domblklist):
    """
    vda        /storages/local/localstorage1/244fd112__15289fc3.snapshot
    hda        -
    """
    result = {}
    for line in domblklist.splitlines():
        try:
            # Changed on new qemu on buster: target_dev, source = line.split("        ")
            target_dev, source = line.strip().split("      ")
            if '/' in source:
                result[target_dev] = source
        except ValueError:
            continue
    return result


def qemu_img_info(path):
    if os.path.exists(path):
        cmd = f"qemu-img info -U --output=json {path}"
        cmd_ret = get_output(cmd)
        return json.loads(cmd_ret['comment'])
    else:
        return {}


@states_response
def get_backing_chain(domain, base):
    """
    Возвращает всю цепочку состояний от текущего снимка до образа диска.
    Только для диска в котором есть снимок 'base'.
    """
    get_current_cmd = ("virsh domblklist {domain_id} --details "
                       "| grep disk | awk '{{print $NF}}'").format(domain_id=domain)
    cmd_ret = get_output(get_current_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while get current snapshots: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    list_current_disks = cmd_ret['comment'].split('\n')
    chain_files_current_disk = []
    for current_disk in list_current_disks:
        qemu_cmd = f"qemu-img info -U --backing-chain --output=json {current_disk.strip()}"
        cmd_ret = get_output(qemu_cmd)
        if not cmd_ret['result']:
            comment = f"An error occurred while get chain of current snapshots: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
        tmp_chain = [str(snap.get('filename')) for snap in json.loads(cmd_ret['comment']) if snap.get('filename')]
        if base in tmp_chain:
            chain_files_current_disk = tmp_chain
            break
    return chain_files_current_disk


@states_response
def create_directory_snapshot(source, backing_file_source, backing_file_format, size, owner, ssh_options=None):
    # if '/storages/gluster/' in backing_file_source:
    #    backing_file_source = backing_file_source.replace('/storages/gluster/', 'gluster+unix:///'
    #                                                      ) + "?socket=/var/run/glusterd.socket"
    # if '/storages/gluster/' in source:
    #    source = source.replace('/storages/gluster/', 'gluster+unix:///') + "?socket=/var/run/glusterd.socket"
    qemu_cmd = "qemu-img create -f qcow2 -b {backing_file} {source}" \
               " -o backing_fmt={backing_fmt},size={size}G"
    qemu_cmd = qemu_cmd.format(backing_file=backing_file_source, source=source,
                               backing_fmt=backing_file_format, size=size)
    if ssh_options:
        tmp_ssh_key_path = None
        try:
            tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
            with os.fdopen(tmp_fh, 'w') as tmp_file:
                tmp_file.write(str(ssh_options['private_ssh_key'].encode('utf-8').decode('unicode_escape')))
            os.chown(tmp_ssh_key_path, getpwnam(owner).pw_uid, getpwnam(owner).pw_gid)
            ssh_cmd = "ssh -o \"StrictHostKeyChecking no\" -i {} {}@{} ".format(
                tmp_ssh_key_path, ssh_options['user'], ssh_options['destination_ip'])
            # alignment the file for gluster, if it need
            # alignment = f"; dd if=/dev/zero bs=1 count=$((512 - ($(du -b {source} " \
            #           f"| sed \"s/\\(.*\\)\\s.*/\\1/\") % 512))) >> {source}"
            # qemu_cmd = f"{ssh_cmd} \'{qemu_cmd}{alignment}\'"
            qemu_cmd = f"{ssh_cmd} \'{qemu_cmd}\'"

            cmd_ret = get_output(qemu_cmd, timeout=300)
        except Exception as e:
            comment = f"Unknown error: {e}"
            raise CommandExecutionError(comment) from None
        finally:
            if tmp_ssh_key_path:
                os.remove(tmp_ssh_key_path)
    else:
        cmd_ret = get_output(usr_cmd(qemu_cmd, owner), timeout=300)
    if not cmd_ret['result']:
        comment = f"An error occurred while creating snapshot {source}: {cmd_ret['comment']}"
        if os.path.exists(source):
            os.remove(source)
        raise CommandExecutionError(comment)
    else:
        # alignment the file for gluster, if it need
        # from .vdisk import align_size
        # align_size(source)
        return f"The snapshot {source} with backing file {backing_file_source} was successfully created."


@states_response
def zfs_create_snapshot(snapshot_uuid, vdisk_uuid, datapool_name, owner):
    zfs_cmd = "sudo zfs snapshot {pool}/{vdisk}@{vdisk}__{snapshot}".format(pool=datapool_name,
                                                                            vdisk=vdisk_uuid,
                                                                            snapshot=snapshot_uuid)
    cmd_ret = get_output(usr_cmd(zfs_cmd, owner), timeout=300)
    if not cmd_ret['result']:
        comment = "An error occurred while creating snapshot {pool}/{vdisk}@{vdisk}__{snapshot}. " \
                  "Error: {ret}".format(pool=datapool_name,
                                        vdisk=vdisk_uuid,
                                        snapshot=snapshot_uuid,
                                        ret=cmd_ret['comment'])
        raise CommandExecutionError(comment)
    else:
        return "The snapshot {pool}/{vdisk}@{vdisk}__{snapshot} was successfully created.".format(
            pool=datapool_name, vdisk=vdisk_uuid, snapshot=snapshot_uuid)


@states_response
def zfs_check_snapshot(snapshot_uuid, vdisk_uuid, datapool_name, owner):
    zfs_cmd = "sudo zfs list -t snapshot -H -o name  {pool}/{vdisk}@{vdisk}__{snapshot}".format(
        pool=datapool_name, vdisk=vdisk_uuid, snapshot=snapshot_uuid)
    cmd_ret = get_output(usr_cmd(zfs_cmd, owner), timeout=300)
    if not cmd_ret['result']:
        comment = "An error occurred while creating snapshot {pool}/{vdisk}@{vdisk}__{snapshot}. " \
                  "Error: {ret}".format(pool=datapool_name,
                                        vdisk=vdisk_uuid,
                                        snapshot=snapshot_uuid,
                                        ret=cmd_ret['comment'])
        raise CommandExecutionError(comment)
    else:
        return "The snapshot {pool}/{vdisk}@{vdisk}__{snapshot} is exists.".format(
            pool=datapool_name, vdisk=vdisk_uuid, snapshot=snapshot_uuid)


@states_response
def zfs_create_clone(clone_vdisk_uuid, orig_vdisk_uuid, datapool_name, owner, snapshot=None):

    comment = ''

    if snapshot:
        ret = zfs_check_snapshot(snapshot, orig_vdisk_uuid, datapool_name, owner)
        if ret['result']:
            comment += ret['comment']
    else:
        ret = zfs_create_snapshot(clone_vdisk_uuid, orig_vdisk_uuid, datapool_name, owner)
        if ret['result']:
            comment += ret['comment']

    zfs_cmd = "sudo zfs clone {pool}/{vdisk}@{vdisk}__{snapshot} {pool}/{clone}".format(
        pool=datapool_name, vdisk=orig_vdisk_uuid, clone=clone_vdisk_uuid,
        snapshot=snapshot if snapshot else clone_vdisk_uuid)
    cmd_ret = get_output(usr_cmd(zfs_cmd, owner))
    if not cmd_ret['result']:
        comment += "\nAn error occurred while creating clone {pool}/{clone}. Error: {ret}".format(
            pool=datapool_name, clone=clone_vdisk_uuid, ret=cmd_ret['comment'])
        raise CommandExecutionError(comment)
    else:
        comment += "\nThe clone {pool}/{clone} was successfully created.".format(
            pool=datapool_name, clone=clone_vdisk_uuid)

    return comment


@states_response
def create_thinlvm_snapshot(name, vgname, thinpool, backing_file_source, backing_file_format,
                            size, owner):

    source = f"/dev/{vgname}/{name}"
    coef = 1.01  # В том записывается qcow2 мета

    # Создание volume
    lv_cmd = "lvcreate --name {name} --thin --thinpool {vgname}/{thinpool} --virtualsize {size}G"
    lv_cmd = lv_cmd.format(name=name, vgname=vgname, thinpool=thinpool, size=size * coef)
    cmd_ret = get_output(lv_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while creating ThinLVM {source}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)

    # Преобразование в qcow2 формат
    qemu_cmd = "qemu-img create -f qcow2 {source} -o backing_fmt={backing_fmt},size={size}G" \
               " -b {backing_file}"
    qemu_cmd = qemu_cmd.format(source=source, size=size, backing_file=backing_file_source,
                               backing_fmt=backing_file_format)
    cmd_ret = get_output(qemu_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while formatting {source} to qcow: {cmd_ret['comment']}"
        # Удаление volume в случае неудачи
        get_output(f"lvremove -f {source}")
        raise CommandExecutionError(comment)

    # Изменение owner-а
    chown_cmd = "chown {user}:{group} {source}".format(user=owner, group=owner, source=source)
    cmd_ret = get_output(chown_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while changing owner of thinlvm {source}: {cmd_ret['comment']}"
        # Удаление volume в случае неудачи
        get_output(f"lvremove -f {source}")
        raise CommandExecutionError(comment)

    return f"ThinLVM {source} successfully created"


@states_response
def create_snapshot_of_powered_domain(domain_id, domain_snapshot_id,
                                      vdisks_snapshots, no_snapshot_vdisks_sources, memory, quiesce=False, debug=False):

    if memory == "None":
        memory = None
    # snapshoting_diskspecs = []
    if memory:
        xml_memspec = f"<memory snapshot='external' file='{memory['source']}'/>"
    else:
        xml_memspec = "<memory snapshot='no'/>"
    begin_xml = f'<domainsnapshot><name>{domain_snapshot_id}</name>' + xml_memspec + '<disks>'
    end_xml = '</disks></domainsnapshot>'
    xml_snapshoting_diskspecs = []
    for snapshot in vdisks_snapshots:
        if debug and snapshot['type_datapool'] == 'gluster':
            xml_diskspec = f"<disk name='{snapshot['dev']}' snapshot='external' type='network'><source " \
                           f"protocol='gluster' name='{snapshot['source'].replace('/storages/gluster/', '')}'>" \
                           f"<host transport='unix' socket='/var/run/glusterd.socket'/></source></disk>"
        elif snapshot['type_datapool'] != 'zfs' and snapshot['type_datapool'] != 'linstor':
            xml_diskspec = f"<disk name='{snapshot['dev']}' snapshot='external'>" \
                           f"<source file='{snapshot['source']}'></source></disk>"
            # diskspec = f"--diskspec {snapshot['backing_file_source']},snapshot=external,file={snapshot['source']}"
            # snapshoting_diskspecs.append(diskspec)
        else:
            xml_diskspec = f"<disk name='{snapshot['dev']}' snapshot='no'></disk>"
            # diskspec = f"--diskspec {snapshot['source']},snapshot=no"
            # snapshoting_diskspecs.append(diskspec)
        xml_snapshoting_diskspecs.append(xml_diskspec)

    xml_not_snapshoting_diskspecs = []
    # not_snapshoting_diskspecs = []
    for no_snapshot_vdisk in no_snapshot_vdisks_sources:
        xml_diskspec = f"<disk name='{no_snapshot_vdisk['dev']}' snapshot='no'></disk>"
        xml_not_snapshoting_diskspecs.append(xml_diskspec)
        # diskspec = f"--diskspec {no_snapshot_vdisk['source']},snapshot=no"
        # not_snapshoting_diskspecs.append(diskspec)

    # diskspecs = " ".join(snapshoting_diskspecs + not_snapshoting_diskspecs)
    xml_diskspecs = begin_xml + " ".join(xml_snapshoting_diskspecs + xml_not_snapshoting_diskspecs) + end_xml
    tmp_xml_path = '/tmp/{}.xml'.format(domain_snapshot_id)
    with open(tmp_xml_path, 'w') as f:
        f.write(xml_diskspecs)
    snapshot_cmd = f"virsh snapshot-create {domain_id} {tmp_xml_path} --reuse-external --no-metadata "
    # snapshot_cmd = f"virsh snapshot-create-as {domain_id} {diskspecs} --reuse-external " \
    #               f"--name {domain_snapshot_id} --no-metadata "
    if memory:
        snapshot_cmd += "--live"  # --memspec {memory['source']},snapshot=external"
    else:
        snapshot_cmd += "--disk-only"
    # if quiesce and not memory:  # quiesce work only with --disk-only
    #    snapshot_cmd += " --quiesce"
    # Для того, чтобы позднее восстановить xml из памяти необходимо добиться
    # совпадения xml записанной в дампе памяти и передаваемой в restore вплоть
    # до слотов, в которые подключены устройства на шинах (см. man virsh -> resore).
    # Для этого мы сохраним в БД текущую xml домена, в которой позднее подменим
    # некоторые поля.
    # dumpxml_cmd = f"virsh dumpxml --inactive --migratable {domain_id}"  # --inactive убирает backingStore
    dumpxml_cmd = f"virsh dumpxml --migratable {domain_id}"  # --inactive убирает backingStore

    call(f'virsh domfsthaw {domain_id}')
    if quiesce and not memory:
        call(f'virsh domfsfreeze {domain_id}', timeout=300)
    cmd_ret = get_output(snapshot_cmd, timeout=300)
    os.remove(tmp_xml_path)
    call(f'virsh domfsthaw {domain_id}')

    if not cmd_ret['result']:
        blk_domain = get_output(f"virsh domblklist {domain_id} | awk 'NR>2&&NF>0{{print $2}}'")
        if blk_domain['result']:
            list_blk = blk_domain['comment'].split()
            for snapshot in vdisks_snapshots:
                if snapshot['type_datapool'] != 'zfs' and snapshot['source'] not in list_blk \
                        and os.path.exists(snapshot['source']):
                    os.remove(snapshot['source'])
        if memory and os.path.exists(memory['source']):
            os.remove(memory['source'])
        comment = f"An error occurred while creating domain snapshot {domain_snapshot_id}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        result = dict()
        result['comment'] = f"The snapshot {domain_snapshot_id} successfully created."
        # comment первой команды был переправлен в null, а comment второй
        # команды соотвествует xml, который будет передан комманде restore
        if memory:
            cmd_ret = get_output(dumpxml_cmd)
            if not cmd_ret['result']:
                result['comment'] += cmd_ret['comment']
            else:
                result['xml_dump'] = cmd_ret['comment']
        return result


@states_response
def restore(domain_id, memory, xml, paused, maxmemory):
    # При восстановлении снимок созданный на другом узле, нужно приводить в соответствие maxMemory
    if maxmemory:
        unit = maxmemory['unit']
        slots = maxmemory['slots']
        memory_count = maxmemory['memory_count']
        size_buffer = '32k'
        get_str_maxmemory = f'dd if={memory} bs={size_buffer} count=1 status=none | ' \
                            f'stdbuf -o{size_buffer} sed -n "s/\(.*maxMemory.*\)/\\1/p"'
        cmd_ret = get_output(get_str_maxmemory)
        if not cmd_ret['result']:
            comment = f"An error get string maxMemory from {memory}. " \
                      f"Error: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
        str_maxmemory = cmd_ret['comment'].rstrip()
        new_str_maxmemory = f"<maxMemory slots='{slots}' unit='{unit}'>{memory_count}</maxMemory>"
        if len(new_str_maxmemory) <= len(str_maxmemory):
            new_str_maxmemory = ' ' * (len(str_maxmemory) - len(new_str_maxmemory)) + \
                                new_str_maxmemory.replace('/', '\\/')
        else:
            raise CommandExecutionError(f'An error occurred while alignment maxMemory for restore from {memory}.')
        set_maxmemory = f'dd if={memory} bs={size_buffer} count=1 status=none | stdbuf -o{size_buffer} ' \
                        f'sed "s/.*maxMemory.*/{new_str_maxmemory}/" | ' \
                        f'dd conv=notrunc bs={size_buffer} count=1 of={memory} status=none'
        cmd_ret = get_output(set_maxmemory)
        if not cmd_ret['result']:
            comment = f"An error set maxMemory into {memory}: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
    tmp_file_path = '/tmp/{}.xml'.format(uuid.uuid4())
    with open(tmp_file_path, 'w') as f:
        f.write(xml)

    state = "--paused" if paused else "--running"
    virsh_cmd = f"virsh restore {memory} --xml {tmp_file_path} {state}"
    cmd_ret = get_output(virsh_cmd, timeout=300)
    os.remove(tmp_file_path)
    if not cmd_ret['result']:
        comment = f"An error occurred while restoring domain: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        return f"Domain {domain_id} successfully restored."


def is_already_commited(base, overlay):
    base_backing = qemu_img_info(base).get('full-backing-filename', None)
    overlay_backing = qemu_img_info(overlay).get('full-backing-filename', None)
    return base_backing == overlay_backing


@states_response
def commit(domain, dev, base, base_file_format, overlay, overlay_children, is_current, snapshot_is_busy, owner,
           task=None, **kwargs):
    if not is_already_commited(base, overlay):
        from .copy import popen_with_progress
        debug = kwargs.get('debug', False)
        if kwargs.get('progress_data'):
            progress_data = kwargs.get('progress_data')
        else:
            progress_data = dict(type='progress', jid=task, id='', name='commit', result=True, steps={})
            progress_data['steps']['total_steps'] = 0
            if snapshot_is_busy:
                files_current_chain = get_backing_chain(domain, base)
                progress_data['steps']['total_steps'] = \
                    len([valid for valid in overlay_children if valid not in files_current_chain['comment']])
            else:
                progress_data['steps']['total_steps'] = len(overlay_children)
            progress_data['steps']['step'] = 0
        progress_data['steps']['step'] += 1
        if snapshot_is_busy:
            base_datapool_type = kwargs.get('base_datapool_type')
            overlay_datapool_type = kwargs.get('overlay_datapool_type')
            # gluster://
            if debug and (base_datapool_type == 'gluster' or overlay_datapool_type == 'gluster'):

                def find_backingstore(bs, path):
                    attrib = None
                    if bs.attrib['type'] == 'network':
                        attrib = 'name'
                    elif bs.attrib['type'] == 'file':
                        attrib = 'file'
                    elif bs.attrib['type'] == 'block':
                        attrib = 'dev'
                    if attrib and bs.find('source') and bs.find('source').attrib[attrib] == path:
                        return bs
                    elif bs.find('backingStore'):
                        return find_backingstore(bs.find('backingStore'), path)
                    else:
                        return None

                dumpxml_cmd = f"virsh dumpxml {domain}"
                ret_dumpxml = get_output(dumpxml_cmd)
                if not ret_dumpxml['result']:
                    raise CommandExecutionError(ret_dumpxml['comment'])
                else:
                    root = ET.fromstring(ret_dumpxml['comment'])
                overlay_index = None
                for disk in root.findall('devices/disk'):
                    if overlay_datapool_type == 'gluster':
                        if is_current and disk.attrib['device'] == 'disk' and \
                                disk.find('target').attrib['dev'] == dev and \
                                disk.find('source').attrib['name'] == overlay.replace('/storages/gluster/', ''):
                            overlay_index = disk.find('source').attrib['index']
                        elif not is_current and disk.attrib['device'] == 'disk' and \
                                disk.find('target').attrib['dev'] == dev and disk.find('backingStore'):
                            bs = find_backingstore(disk.find('backingStore'), overlay.replace('/storages/gluster/', ''))
                            if bs and 'index' in bs.attrib:
                                overlay_index = bs.attrib['index']
                    else:
                        if is_current and disk.attrib['device'] == 'disk' and \
                                disk.find('target').attrib['dev'] == dev and \
                                disk.find('source').attrib['file'] == overlay:
                            overlay_index = disk.find('source').attrib['index']
                        elif not is_current and disk.attrib['device'] == 'disk' and \
                                disk.find('target').attrib['dev'] == dev and disk.find('backingStore'):
                            bs = find_backingstore(disk.find('backingStore'), overlay)
                            if bs and 'index' in bs.attrib:
                                overlay_index = bs.attrib['index']
                base_index = None
                for disk in root.findall('devices/disk'):
                    if base_datapool_type == 'gluster':
                        if disk.attrib['device'] == 'disk' and disk.find('target').attrib['dev'] == dev and \
                                disk.find('backingStore'):
                            bs = find_backingstore(disk.find('backingStore'), base.replace('/storages/gluster/', ''))
                            if bs and 'index' in bs.attrib:
                                base_index = bs.attrib['index']
                    else:
                        if disk.attrib['device'] == 'disk' and disk.find('target').attrib['dev'] == dev \
                                and disk.find('backingStore'):
                            bs = find_backingstore(disk.find('backingStore'), base)
                            if bs and 'index' in bs.attrib:
                                base_index = bs.attrib['index']
                top_overlay = f"--top {dev}[{overlay_index}]" if overlay_index else ''
                virsh_cmd = f'virsh blockcommit --domain {domain} {dev}' \
                            f' --base {dev}[{base_index}] {top_overlay} --wait --verbose'
            else:
                virsh_cmd = f'virsh blockcommit --domain {domain} {dev} --base {base} --top {overlay} --wait --verbose'
            if is_current:
                bandwidth = kwargs.get('bandwidth', 0)
                virsh_cmd += f" --bandwidth {bandwidth} --active  --pivot"
                if kwargs.get('blkdeviotune'):
                    call(f"virsh blkdeviotune {domain} {dev} "
                         f"--total-iops-sec {kwargs['blkdeviotune']['total_iops_sec']} "
                         f"--total-bytes-sec {kwargs['blkdeviotune']['total_bytes_sec']} --live; sleep 3")
            virsh_cmd = virsh_cmd + ' 2>&1'
            p = popen_with_progress(virsh_cmd, "blockcommit", progress_data, kwargs.get('redis', None))
            if is_current and kwargs.get('blkdeviotune'):
                call(f"virsh blkdeviotune {domain} {dev} --total-iops-sec 0 --total-bytes-sec 0 --live")

        else:
            base_datapool_type = kwargs.get('base_datapool_type')
            overlay_datapool_type = kwargs.get('overlay_datapool_type')
            if debug and base_datapool_type == 'gluster':
                base = base.replace('/storages/gluster/', 'gluster+unix:///') + '?socket=/var/run/glusterd.socket'
            if debug and overlay_datapool_type == 'gluster':
                overlay = overlay.replace('/storages/gluster/', 'gluster+unix:///') + '?socket=/var/run/glusterd.socket'
            qemu_cmd = f'qemu-img commit -p -b {base} {overlay}'
            p = popen_with_progress(qemu_cmd, "qemu_commit", progress_data, kwargs.get('redis', None))
        if p.returncode:
            raise CommandExecutionError(
                f'An error occurred while COMMIT snapshot (stage 1): '
                f'{str(p.stderr.read()) + str(p.stdout.read()) + str(p.comment if hasattr(p, "comment") else "")}')
        if snapshot_is_busy:
            # Если ВМ включена, то в текущей цепочке состояний диски блокированы, поэтому диск входящий
            # в неё нужно исключить. К тому же после "virsh blockcommit", qemu-img rebase не требуется
            # (для текущей цепочки).
            files_current_chain = get_backing_chain(domain, base)
            overlay_children = [valid for valid in overlay_children if valid not in files_current_chain['comment']]
        for child in overlay_children:
            progress_data['steps']['step'] += 1
            qemu_rebase_cmd = f"qemu-img rebase -p -F {base_file_format} -b {base} {child}"
            p = popen_with_progress(qemu_rebase_cmd, "qemu_commit", progress_data, kwargs.get('redis', None))
            if p.returncode:
                raise CommandExecutionError(
                    f"An error occurred while REBASE child ({child}) (stage 2): {p.stderr.read()}")
    return "Snapshot committed."


@states_response
def zfs_remove_snapshots(snapshots, owner, merge=False):
    """Eсли удаляемый снимок находится не в текущей ветке, то проверяем есть ли еще снимки, если есть, то
    делаем rollback на последний, чтобы датасет соответствовал последнему сохраненному состоянию, если в той
    ветке больше нет снимков, то удаляем датасет"""

    comment = ''

    full_name = 'sudo zfs list -Ht snapshot -o name | grep @{{}} {}'.format('||true' if merge else '')
    list_chain_snapshots = 'sudo zfs list -Ht snapshot -o name -s creation -r {dataset_name}'
    cmd_destroy = 'sudo zfs destroy {merge} {}'
    cmd_rollback = 'sudo zfs rollback {}'

    for snapshot in snapshots:
        snapshot_name = '{}__{}'.format(snapshot['vdisk_uuid'], snapshot['snapshot_uuid'])
        get_full_name = get_output(
            usr_cmd(full_name.format(snapshot_name), owner))
        if not get_full_name['result']:
            raise CommandExecutionError(f"Snapshot {snapshot_name} not found!. Error: {get_full_name['comment']}")
        else:
            if merge and get_full_name['comment'].strip() == '':
                continue
            get_list_chain_snapshots = get_output(usr_cmd(list_chain_snapshots.format(
                dataset_name=get_full_name['comment'].strip().split('@')[0]), owner))
            ret_destroy = get_output(usr_cmd(cmd_destroy.format(get_full_name['comment'].strip(),
                                                                merge='-R' if merge else ''), owner))
            if not ret_destroy['result']:
                raise CommandExecutionError(
                    "Snapshot {} not destroy! Error: {ret}".format(snapshot_name, ret=ret_destroy['comment']))
            elif get_full_name['comment'].split('@')[0] != snapshot['zfs_pool_name'] + '/' + snapshot['vdisk_uuid'] \
                    and get_list_chain_snapshots['comment'].strip().split()[-1] == get_full_name['comment'].strip() \
                    and len(get_list_chain_snapshots['comment'].strip().split()) > 1:
                ret_rollback = get_output(usr_cmd(cmd_rollback.format(
                    get_list_chain_snapshots['comment'].split()[-2]), owner))
                if not ret_rollback['result']:
                    raise CommandExecutionError("Failed to rollback to snapshot {}! Error: {ret}".format(
                        get_list_chain_snapshots['comment'].strip().split()[-2], ret=ret_rollback['comment']))

            elif get_full_name['comment'].split('@')[0] != snapshot['zfs_pool_name'] + '/' + snapshot['vdisk_uuid'] \
                    and len(get_list_chain_snapshots['comment'].strip().split()) == 1:
                ret_destroy_dataset = get_output(usr_cmd(cmd_destroy.format(get_full_name['comment'].split('@')[0],
                                                                            merge=''), owner))
                if not ret_destroy_dataset['result']:
                    raise CommandExecutionError(
                        "Failed to destroy dataset {dataset}: {ret}".format(
                            dataset=get_full_name['comment'].split('@')[0],
                            ret=ret_destroy_dataset['comment']))
            else:
                comment += "The snapshot {} from {pool} was successfully deleted.".format(
                    snapshot_name, pool=snapshot['zfs_pool_name'])
    return comment


@states_response
def remove_snapshots(snapshots, owner, ssh_options=None, **kwargs):
    """Стейт для удаления снепшотов. result всегда True. Необходимо
    для того, чтобы стейты управления снепшотами не становились failed
    только из-за того, что на последнем шаге не удалось удалить какие-то
    файлы снепшотов."""
    if ssh_options:
        tmp_ssh_key_path = None
        try:
            tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
            with os.fdopen(tmp_fh, 'w') as tmp_file:
                tmp_file.write(str(ssh_options['private_ssh_key'].encode('utf-8').decode('unicode_escape')))
            os.chown(tmp_ssh_key_path, getpwnam(owner).pw_uid, getpwnam(owner).pw_gid)
            ssh_cmd = "ssh -o \"StrictHostKeyChecking no\" -i {} {}@{} ".format(
                tmp_ssh_key_path, ssh_options['user'], ssh_options['destination_ip'])
            for snapshot in snapshots:
                if snapshot['type'] == 'directory':
                    cmd = "rm -f {}".format(snapshot['source'])
                else:
                    cmd = "lvremove -f {}/{}".format(snapshot['vgname'], snapshot['name'])
                cmd = f"{ssh_cmd} \'{cmd}\'"
                get_output(cmd, **kwargs)
        except Exception as e:
            comment = f"Unknown error: {e}"
            raise CommandExecutionError(comment) from None
        finally:
            if tmp_ssh_key_path:
                os.remove(tmp_ssh_key_path)
    else:
        for snapshot in snapshots:
            if os.path.exists(snapshot['source']):
                if snapshot['type'] == 'directory':
                    cmd = "rm -f {}".format(snapshot['source'])
                    cmd = usr_cmd(cmd, owner)
                    get_output(cmd, **kwargs)
                elif snapshot['type'] == 'linstor':
                    from veilrpc.states.linstor import delete_disk
                    ret = delete_disk(snapshot['resource'], volume_nr=kwargs.get('volume_nr', None),
                                      ctrl_host=kwargs['ctrl_host'])
                    if not ret['result']:
                        raise CommandExecutionError(ret['comment'])
                else:
                    cmd = "lvremove -f {}/{}".format(snapshot['vgname'], snapshot['name'])
                    get_output(cmd, **kwargs)
    return "Snapshots deleted."


@states_response
def rebase_unsafe(matched_rebase, is_local, ssh_options, owner):
    assert bool(ssh_options) if not is_local else True
    tmp_ssh_key_path = None
    qemu_rebase_cmd = "qemu-img rebase -u -f qcow2 -b {backing_file} -F qcow2 {snapshot}"
    if not is_local:
        try:
            tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
            with os.fdopen(tmp_fh, 'w') as tmp_file:
                tmp_file.write(str(ssh_options['private_ssh_key'].encode('utf-8').decode('unicode_escape')))
            os.chown(tmp_ssh_key_path, getpwnam(owner).pw_uid, getpwnam(owner).pw_gid)
            ssh_cmd = "ssh -o \"StrictHostKeyChecking no\" -i {} {}@{} ".format(
                tmp_ssh_key_path, ssh_options['user'], ssh_options['destination_ip'])
            rebase_command = ssh_cmd + "'" + ";".join(qemu_rebase_cmd.format(
                backing_file=vs['backing_file'], snapshot=vs['snapshot']) for vs in matched_rebase) + "'"
            cmd_ret = get_output(rebase_command, timeout=300)
        except Exception as e:
            comment = f"Unknown error. Error: {str(e)}"
            raise CommandExecutionError(comment) from None
        finally:
            if tmp_ssh_key_path:
                os.remove(tmp_ssh_key_path)
    else:
        rebase_command = ";".join(qemu_rebase_cmd.format(
            backing_file=vs['backing_file'], snapshot=vs['snapshot']) for vs in matched_rebase)
        cmd_ret = get_output(usr_cmd(rebase_command, owner), timeout=300)
    if not cmd_ret['result']:
        comment = f"An error occurred while REBASE snapshots. Error: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        comment = 'Rebased'
        return comment


@states_response
def replace_domain_id_into_memory(domain, parent, list_memory_snapshots, is_local, ssh_options, owner):
    assert bool(ssh_options) if not is_local else True
    tmp_ssh_key_path = None
    size_buffer = '32k'
    cmd_template = f'dd if={{snapshot}} bs={size_buffer} count=1 status=none | stdbuf -o{size_buffer} ' \
                   f'sed "s/{parent}/{domain}/g;{{replace_mac}}" | ' \
                   f'dd conv=notrunc bs={size_buffer} count=1 of={{snapshot}} status=none'
    cmd = ''
    for snapshot, map_mac in list_memory_snapshots.items():
        if map_mac and any(k != v for k, v in map_mac.items()):
            replace_mac = ';'.join('s/{old_mac}/{new_mac}/g'.format(old_mac=k, new_mac=v) for k, v in map_mac.items())
        else:
            replace_mac = ''
        cmd += cmd_template.format(snapshot=snapshot, replace_mac=replace_mac) + ';'
    if not is_local:
        try:
            tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
            with os.fdopen(tmp_fh, 'w') as tmp_file:
                tmp_file.write(str(ssh_options['private_ssh_key'].encode('utf-8').decode('unicode_escape')))
            os.chown(tmp_ssh_key_path, getpwnam(owner).pw_uid, getpwnam(owner).pw_gid)
            ssh_cmd = "ssh -o \"StrictHostKeyChecking no\" -i {} {}@{} ".format(
                tmp_ssh_key_path, ssh_options['user'], ssh_options['destination_ip'])
            cmd_ret = get_output(ssh_cmd + "'" + cmd + "'")
        except Exception as e:
            comment = "Unknown error: {}".format(e)
            raise CommandExecutionError(comment) from None
        finally:
            if tmp_ssh_key_path:
                os.remove(tmp_ssh_key_path)
    else:
        cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while REPLACE ID snapshots: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        comment = 'Replaced'
        return comment


@states_response
def zfs_revert(snapshot_uuid, vdisk_uuid, zfs_pool_name, owner):
    """Если переход на последнее сохраненное состояние текущей ветки, то делается просто rollback.
    Иначе делаем переключение на клон от снимка, если его нет, то создаем. Если от еще текущего датасета
    больше нет снимков, то удаляем его, иначе rollback на его последнее сохраненное состояние,
    и переименовываем в соответствии с этим состоянием. А датасет на который переключились переименовываем
    в текущий диск"""

    full_name = 'sudo zfs list -Ht snapshot -o name | grep @{}__{}'.format(vdisk_uuid, snapshot_uuid)
    cmd_chain_snapshots = 'sudo zfs list -Ht snapshot -o name -s creation -r '
    current_dataset = '{}/{}'.format(zfs_pool_name, vdisk_uuid)
    clone_dataset = '{}/{}__{}'.format(zfs_pool_name, vdisk_uuid, snapshot_uuid)
    last_snap_curr_chain = cmd_chain_snapshots + '{} | tail -1'.format(current_dataset)
    cmd_umount = 'sudo zfs umount {}'.format(current_dataset)
    cmd_mount = 'sudo zfs mount {}'.format(current_dataset)
    cmd_set_mount_on = 'sudo zfs set canmount=on {}'.format(current_dataset)
    cmd_rollback = 'sudo zfs rollback {}'
    is_exist_clone = 'sudo zfs list -Ht filesystem -o name |grep "{}" ||true'.format(clone_dataset)
    cmd_clone = 'sudo zfs clone -o canmount=noauto {{full_name}} {}'.format(clone_dataset)
    cmd_promote = 'sudo zfs promote {}'.format(clone_dataset)
    is_exists_snapshots = cmd_chain_snapshots + current_dataset
    cmd_destroy = 'sudo zfs destroy {}'
    cmd_rename_from_current = 'sudo zfs rename {} {{}}'.format(current_dataset)
    cmd_rename_to_current = 'sudo zfs rename {} {}'.format(clone_dataset, current_dataset)

    get_full_name = get_output(usr_cmd(full_name, owner))
    if not get_full_name['result']:
        raise CommandExecutionError("Snapshot {vdisk}__{snapshot} not found!. Error: {ret}".format(
            vdisk=vdisk_uuid, snapshot=snapshot_uuid, ret=get_full_name['comment']))
    get_last_snap_curr_chain = get_output(usr_cmd(last_snap_curr_chain, owner))
    if not get_last_snap_curr_chain['result']:
        raise CommandExecutionError(f"Dataset {current_dataset} empty!. "
                                    f"Error: {get_last_snap_curr_chain['comment']}")
    ret_umount = get_output(usr_cmd(cmd_umount, owner))
    if not ret_umount['result']:
        raise CommandExecutionError("Could not umount dataset {}!. Error: {ret}".format(
            current_dataset, ret=ret_umount['comment']))
    if get_last_snap_curr_chain['comment'].strip() == get_full_name['comment'].strip():
        ret_rollback = get_output(usr_cmd(cmd_rollback.format(get_last_snap_curr_chain['comment']), owner))
        if not ret_rollback['result']:
            raise CommandExecutionError("Could not rollback to snapshot {}!. Error: {ret}".format(
                get_last_snap_curr_chain['comment'], ret=ret_rollback['comment']))
    else:
        ret_is_exists_clone = get_output(usr_cmd(is_exist_clone, owner))
        if ret_is_exists_clone['comment'].strip() == '':
            ret_clone = get_output(usr_cmd(cmd_clone.format(full_name=get_full_name['comment'].strip()), owner))
            if not ret_clone['result']:
                raise CommandExecutionError("Could not create clone from snapshot {}!. Error: {ret}".format(
                    get_full_name['comment'], ret=ret_clone['comment']))
        ret_promote = get_output(usr_cmd(cmd_promote, owner))
        if not ret_promote['result']:
            raise CommandExecutionError("Could not promote to dataset {}!. Error: {ret}".format(
                clone_dataset, ret=ret_promote['comment']))
        ret_is_exists_snapshots = get_output(usr_cmd(is_exists_snapshots, owner))
        if not ret_is_exists_snapshots['comment'].strip():
            ret_destroy = get_output(usr_cmd(cmd_destroy.format(current_dataset), owner))
            if not ret_destroy['result']:
                raise CommandExecutionError("Could not destroy dataset {}!. Error: {ret}".format(
                    current_dataset, ret=ret_destroy['comment']))
        else:
            ret_rollback = get_output(usr_cmd(cmd_rollback.format(get_last_snap_curr_chain['comment']), owner))
            if not ret_rollback['result']:
                raise CommandExecutionError("Could not rollback to snapshot {}!. Error: {ret}".format(
                    get_last_snap_curr_chain['comment'], ret=ret_rollback['comment']))
            ret_rename_from_current = get_output(usr_cmd(cmd_rename_from_current.format(
                sub('/.*@', '/', get_last_snap_curr_chain['comment'].strip())), owner))
            if not ret_rename_from_current['result']:
                raise CommandExecutionError("Could not rename dataset from {} to {}!. Error: {ret}".format(
                    current_dataset, sub('/.*@', '/', get_last_snap_curr_chain['comment'].strip()),
                    ret=ret_rename_from_current['comment']))
        ret_rename_to_current = get_output(usr_cmd(cmd_rename_to_current, owner))
        if not ret_rename_to_current['result']:
            raise CommandExecutionError("Could not rename from {} to {}!. Error: {ret}".format(
                clone_dataset, current_dataset, ret=ret_rename_to_current['comment']))
    ret_mount = get_output(usr_cmd(cmd_mount, owner))
    if not ret_mount['result']:
        raise CommandExecutionError("Could not mount dataset {}!. Error: {ret}".format(
            current_dataset, ret=ret_mount['comment']))
    ret_set_mount_on = get_output(usr_cmd(cmd_set_mount_on, owner))
    if not ret_set_mount_on['result']:
        raise CommandExecutionError("Could not set canmount=on to dataset {}!. Error: {ret}".format(
            current_dataset, ret=ret_set_mount_on['comment']))

    return f"Revert to snapshot {clone_dataset} was successful."


@states_response
def zfs_to_zfs_copy(is_local, src_vdisk_uuid, src_pool_name, dst_vdisk_uuid, dst_pool_name, map_snapshots,
                    ssh_options, owner, with_snapshots=False, task=None, redis=None, progress_data=None):
    """
    Функция копирования диска из zfs пула в zfs пул со снимками.

    """

    assert bool(ssh_options) if not is_local else True

    cmd_snapshots_disk = 'sudo zfs list -Ht snapshot -o name -s creation -r {}/{}'.format(src_pool_name, src_vdisk_uuid)
    get_snapshots_disk = get_output(usr_cmd(cmd_snapshots_disk, owner))
    if not get_snapshots_disk['result']:
        raise CommandExecutionError("Could not get list snapshots from dataset {}/{}!. Error: {ret}".format(
            src_pool_name, src_vdisk_uuid, ret=get_snapshots_disk['comment']))
    else:
        snapshots_disk = get_snapshots_disk['comment'].strip().split()

    tmp_ssh_key_path = None
    if is_local:
        cmd_ssh = ''
    else:
        tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
        with os.fdopen(tmp_fh, 'w') as tmp_file:
            tmp_file.write(str(ssh_options['private_ssh_key'].encode('utf-8').decode('unicode_escape')))
        os.chown(tmp_ssh_key_path, getpwnam(owner).pw_uid, getpwnam(owner).pw_gid)
        cmd_ssh = 'ssh -o \"StrictHostKeyChecking no\" -i {} {}@{}'.format(
            tmp_ssh_key_path, ssh_options['user'], ssh_options['destination_ip'])

    cmd_list_all_from_remote_pool = '{ssh} sudo zfs list -Ht all -o name -r {}'.format(dst_pool_name, ssh=cmd_ssh)
    get_list_all_from_remote_pool = get_output(usr_cmd(cmd_list_all_from_remote_pool, owner))
    if not get_list_all_from_remote_pool['result']:
        raise CommandExecutionError("Could not get list all from zfs pool {}!. Error: {ret}".format(
            dst_pool_name, ret=get_list_all_from_remote_pool['comment']))
    else:
        list_all_from_remote_pool = get_list_all_from_remote_pool['comment'].strip().split()

    if snapshots_disk and not map_snapshots:
        map_snapshots[snapshots_disk[0].split('__')[-1]] = 'first_snapshot_for_rename'

    size_cmd = 'sudo zfs get -Ho value used {}'
    from .copy import popen_with_progress
    total_steps = len([s for s in snapshots_disk if s.split('__')[-1] in map_snapshots.keys()])
    if not progress_data:
        progress_data = dict(type='progress', jid=task, id='', name='zfs_to_zfs_copy', result=True, steps={})
        progress_data['steps']['total_steps'] = total_steps
        if '{0}/{1}@{1}__first_snapshot_for_rename'.format(dst_pool_name, dst_vdisk_uuid) in list_all_from_remote_pool:
            progress_data['steps']['total_steps'] -= 1
    if 'step' not in progress_data['steps']:
        progress_data['steps']['step'] = 0

    def recursive_copy(vdisk, snapshots, previous=None):
        for s in snapshots:
            if s == snapshots_disk[-1] and len(snapshots_disk) > 1:
                last_snapshot_delete = True
            else:
                last_snapshot_delete = False
            if s.split('__')[-1] not in map_snapshots.keys():
                continue
            cmd_umount_rollback = ''
            if not previous:
                cmd_copy = ''
                if '{0}/{1}@{1}__first_snapshot_for_rename'.format(dst_pool_name, dst_vdisk_uuid)\
                        in list_all_from_remote_pool:
                    # нужно сделать отмонтирование и откат на это состояние, так как датасет смонтирован на запись и
                    # в нем могут появится изменения, что помешает накатыванию снимков
                    cmd_umount_rollback = '{ssh} sudo zfs umount {pool}/{vdisk}{slash};'\
                                          'sudo zfs rollback {pool}/{vdisk}@{vdisk}__first_snapshot_for_rename'.format(
                        ssh=cmd_ssh, pool=dst_pool_name, vdisk=dst_vdisk_uuid, slash='\\' if cmd_ssh else '')
                    cmd_rename_snapshot = '{ssh} sudo zfs rename {pool}/{vdisk}@{vdisk}__first_snapshot_for_rename '\
                                          '{pool}/{vdisk}@{vdisk}__{new_snapshot_uuid}'.format(
                        ssh=cmd_ssh, pool=dst_pool_name, vdisk=dst_vdisk_uuid,
                        new_snapshot_uuid=map_snapshots[s.strip().split('__')[-1]])
                else:
                    # если сопируется первый раз, то вначале нужно скопировать датасет от первого снимка и сам снимок
                    size_format_cmd = size_cmd.format('{}/{}'.format(src_pool_name, src_vdisk_uuid))
                    size_ret = get_output(size_format_cmd)
                    if not size_ret['result'] or not size_ret['comment'].strip():
                        raise CommandExecutionError(
                            f'Failed to get size vdisk {src_pool_name}/{src_vdisk_uuid}: {size_ret["comment"]}')
                    else:
                        size = convert(size_ret['comment'].strip(), 'b')
                        if not size['result']:
                            raise CommandExecutionError(f'Filed convert size. Error: {size["comment"]}')
                        else:
                            size = size['comment']
                            cmd_copy = 'set -o pipefail; exec 3>&1; sudo zfs send {snapshot} | '\
                                       'pv -ns {size} 2>&3 |{ssh} sudo zfs recv -u {pool}/{new_vdisk}'.format(
                                snapshot=s, pool=dst_pool_name, new_vdisk=dst_vdisk_uuid, ssh=cmd_ssh, size=size)
                            cmd_rename_snapshot = '{ssh} sudo zfs rename {pool}/{new_vdisk}@{old_vdisk}' \
                                                  '__{old_snapshot_uuid} {pool}/{new_vdisk}@{new_vdisk}' \
                                                  '__{new_snapshot_uuid}'.format(
                                ssh=cmd_ssh, old_snapshot_uuid=s.split('__')[-1], pool=dst_pool_name,
                                new_vdisk=dst_vdisk_uuid, old_vdisk=src_vdisk_uuid,
                                new_snapshot_uuid=map_snapshots[s.strip().split('__')[-1]])
            else:
                # последующее копирование осуществляется инкрементально, поэтому нужно указывать предыдущий снимок
                if '__' in vdisk.split('/')[-1]:
                    size_format_cmd = size_cmd.format(s.split('@')[0])
                    size_ret = get_output(size_format_cmd)
                    if not size_ret['result'] or not size_ret['comment'].strip():
                        raise CommandExecutionError(
                            f'Failed to get size snapshot {s.split("@")[0]}. Error: {size_ret["comment"]}')
                    else:
                        size = convert(size_ret['comment'].strip(), 'b')
                        if not size['result']:
                            raise CommandExecutionError(f'Filed convert size. Error: {size["comment"]}')
                        else:
                            size = size['comment']
                            cmd_copy = 'set -o pipefail; exec 3>&1; sudo zfs send -i {previous} {snapshot} | '\
                                       'pv -ns {size}  2>&3 | {ssh} sudo zfs recv -u -o canmount=noauto ' \
                                       '{pool}/{vdisk}'.format(
                                previous=previous, snapshot=s, pool=dst_pool_name, ssh=cmd_ssh, size=size,
                                vdisk=dst_vdisk_uuid + '__' + map_snapshots[vdisk.split('/')[-1].split('__')[-1]])
                            cmd_rename_snapshot = '{ssh} sudo zfs rename {pool}/{vdisk}@{old_snapshot_name} ' \
                                                  '{pool}/{vdisk}@{new_snapshot_name}'.format(
                                ssh=cmd_ssh, old_snapshot_name=s.split('@')[-1], pool=dst_pool_name,
                                vdisk=dst_vdisk_uuid + '__' + map_snapshots[vdisk.split('/')[-1].split('__')[-1]],
                                new_snapshot_name=dst_vdisk_uuid + '__' + map_snapshots[s.strip().split('__')[-1]])
                else:
                    size_format_cmd = size_cmd.format(s.split('@')[0])
                    size_ret = get_output(size_format_cmd)
                    if not size_ret['result'] or not size_ret['comment'].strip():
                        raise CommandExecutionError(
                            f'Failed to get size snapshot {s.split("@")[0]}. Error: {size_ret["comment"]}')
                    else:
                        size = convert(size_ret['comment'].strip(), 'b')
                        if not size['result']:
                            raise CommandExecutionError(f'Filed convert size. Error: {size["comment"]}')
                        else:
                            size = size['comment']
                            cmd_copy = 'set -o pipefail; exec 3>&1; sudo zfs send -i {previous} {snapshot} | '\
                                       'pv -ns {size} 2>&3 | {ssh} sudo zfs recv -u {pool}/{vdisk}'.format(
                                previous=previous, snapshot=s, pool=dst_pool_name,
                                vdisk=dst_vdisk_uuid, ssh=cmd_ssh, size=size)
                            cmd_rename_snapshot = '{ssh} sudo zfs rename {pool}/{vdisk}@{old_snapshot_name} ' \
                                                  '{pool}/{vdisk}@{new_snapshot_name}'.format(
                                ssh=cmd_ssh, old_snapshot_name=s.split('@')[-1], pool=dst_pool_name,
                                vdisk=dst_vdisk_uuid,
                                new_snapshot_name=dst_vdisk_uuid + '__' + map_snapshots[s.strip().split('__')[-1]])
            if cmd_umount_rollback:
                ret_umount_rollback = get_output(usr_cmd(cmd_umount_rollback, owner))
                if not ret_umount_rollback['result']:
                    raise CommandExecutionError(
                        "Could not umount or rollback dataset {}/{}!. Error: {ret}".format(
                            dst_pool_name, dst_vdisk_uuid, ret=ret_umount_rollback['comment']))
            if cmd_copy:
                progress_data['steps']['step'] += 1
                p = popen_with_progress(cmd_copy, "pv", progress_data, redis)
                if p.returncode:
                    raise CommandExecutionError(
                        "Could not copy snapshot {}!. CMD: {cmd}. Error: {ret}".format(
                            s, cmd=cmd_copy, ret=p.stderr.read()))
            ret_rename_snapshot = get_output(usr_cmd(cmd_rename_snapshot, owner))
            if not ret_rename_snapshot['result']:
                raise CommandExecutionError(
                    "Could not rename snapshot {}!. Error: {ret}".format(s, ret=ret_rename_snapshot['comment']))
            if last_snapshot_delete or not with_snapshots:
                # последний снимок создавался чтобы сопировать текущее состояние, больше он не нужен,
                # так как у zfs текущем состоянием является сам диск, или удаляем если не собираемся в дальнейшем
                # копировать снимки
                cmd_delete = '{ssh} sudo zfs destroy {pool}/{vdisk}@{vdisk}__{snapshot}'.format(
                    ssh=cmd_ssh, pool=dst_pool_name, vdisk=dst_vdisk_uuid,
                    snapshot=map_snapshots[s.split('__')[-1]] if with_snapshots else 'first_snapshot_for_rename')
                ret_delete = get_output(usr_cmd(cmd_delete, owner))
                if not ret_delete['result']:
                    raise CommandExecutionError(
                        "Could not delete last snapshot {}!. Error: {ret}".format(
                            s if with_snapshots else '{0}/{1}@{1}__first_snapshot_for_rename'.format(
                                dst_pool_name, dst_vdisk_uuid),
                            ret=ret_delete['comment']))
            cmd_clones_of_snapshot = 'sudo zfs list -Ho clones {}'.format(s)
            get_clones_of_snapshot = get_output(usr_cmd(cmd_clones_of_snapshot, owner))
            if not get_clones_of_snapshot['result']:
                raise CommandExecutionError("Could not get list clones from snapshot {}!. Error: {ret}".format(
                    s, ret=get_clones_of_snapshot['comment']))
            else:
                clones_of_snapshot = get_clones_of_snapshot['comment'].strip().split(',')
            for c in clones_of_snapshot if clones_of_snapshot[0] not in ['', '-'] else []:
                cmd_snapshots_of_clone = 'sudo zfs list -Ht snapshot -o name -s creation -r {}'.format(c)
                get_snapshots_of_clone = get_output(usr_cmd(cmd_snapshots_of_clone, owner))
                if not get_snapshots_of_clone['result']:
                    raise CommandExecutionError("Could not get list snapshots from clone {}!. Error: {ret}".format(
                        c, ret=get_snapshots_of_clone['comment']))
                recursive_copy(c, get_snapshots_of_clone['comment'].strip().split(), s)
            previous = s
    try:
        recursive_copy(src_vdisk_uuid, snapshots_disk)

        cmd_mount = '{ssh} sudo zfs mount {pool}/{vdisk}'.format(ssh=cmd_ssh, pool=dst_pool_name, vdisk=dst_vdisk_uuid)
        ret_mount = get_output(usr_cmd(cmd_mount, owner))
        if not ret_mount['result']:
            raise CommandExecutionError(
                "Could not mount dataset {}/{}!. Error: {ret}".format(dst_pool_name, dst_vdisk_uuid,
                                                                      ret=ret_mount['comment']))
    finally:
        if not is_local and tmp_ssh_key_path:
            os.remove(tmp_ssh_key_path)

    return f"Vdisk successfully copied from {src_pool_name}/{src_vdisk_uuid} {dst_pool_name}/{dst_vdisk_uuid}."


@states_response
def create(**kwargs):
    from .vdisk import created, DEFAULT_UNIT
    res = {'comment': ''}

    domain_id = kwargs['domain_id']
    creating_snapshots = kwargs['creating_snapshots']
    no_snapshots = kwargs['no_snapshots']
    domain_snapshot_id = kwargs['domain_snapshot_id']
    memory_snapshot = kwargs['memory_snapshot']
    domain_is_powered = kwargs['domain_is_powered']
    user = kwargs['user']
    debug = kwargs.get('debug', False)
    quiesce = kwargs.get('quiesce', False)
    ctrl_host = kwargs.get('ctrl_host')
    unit = kwargs.get('unit', DEFAULT_UNIT)
    for snapshot in creating_snapshots:
        if snapshot['type_datapool'] == 'zfs':
            cmd_ret = zfs_create_snapshot(snapshot['snapshot_uuid'], snapshot['vdisk_uuid'],
                                          snapshot['zfs_pool_name'], user)
        elif snapshot['type_datapool'] == 'linstor':
            from veilrpc.states.linstor import create_snapshot as lin_snap_create
            cmd_ret = lin_snap_create(snapshot['snapshot_uuid'], snapshot['resource'], ctrl_host)
        elif snapshot['type'] == 'directory':
            cmd_ret = create_directory_snapshot(snapshot['source'], snapshot['backing_file_source'],
                                                snapshot['backing_file_format'], snapshot['size'], user)
        elif snapshot['type'] == 'thinlvm':
            cmd_ret = create_thinlvm_snapshot(snapshot['name'], snapshot['vgname'], snapshot['thin_pool_name'],
                                              snapshot['backing_file_source'], snapshot['backing_file_format'],
                                              snapshot['size'], user)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        res['comment'] += cmd_ret['comment']

    if memory_snapshot:
        from .lvm import lv_present
        if memory_snapshot['type'] == 'directory':
            cmd_ret = created(memory_snapshot['path'], memory_snapshot['name'], memory_snapshot['id'],
                              'qcow2', memory_snapshot['size'] / 1024, 'off', user, None, unit)
        elif memory_snapshot['type'] == 'thinlvm':
            cmd_ret = lv_present(memory_snapshot['name'],
                                 vgname=memory_snapshot['vgname'] + '/' + memory_snapshot['thin_pool_name'],
                                 size=str(memory_snapshot['size']) + 'M',
                                 thinvolume=True)
            cmd = 'chown {0}:{0} $(readlink -f /dev/{1}/{2})'.format(user, memory_snapshot['vgname'],
                                                                     memory_snapshot['name'])
            call(cmd)
        elif memory_snapshot['type'] == 'lvm':
            cmd_ret = lv_present(memory_snapshot['name'],
                                 vgname=memory_snapshot['vgname'],
                                 size=str(memory_snapshot['size']) + 'M')
            cmd = 'chown {0}:{0} $(readlink -f /dev/{1}/{2})'.format(user, memory_snapshot['vgname'],
                                                                     memory_snapshot['name'])
            call(cmd)
        elif memory_snapshot['type'] == 'linstor':
            from veilrpc.states.linstor import create_disk
            cmd_ret = create_disk(name=memory_snapshot['resource'],
                                  size=f"{str(memory_snapshot['size'])}M",
                                  ctrl_host=ctrl_host,
                                  layer_list=kwargs.get('layer_list', ['drbd', 'storage']),
                                  storage_pool=memory_snapshot['path'],
                                  place_count=memory_snapshot['place_count'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        res['comment'] += str(cmd_ret['comment']) if cmd_ret['comment'] else ''

    if domain_is_powered:
        cmd_ret = create_snapshot_of_powered_domain(domain_id, domain_snapshot_id, creating_snapshots,
                                                    no_snapshots, memory_snapshot, quiesce, debug)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        res['comment'] += cmd_ret['comment']['comment']
        if 'xml_dump' in cmd_ret['comment']:
            res['xml_dump'] = cmd_ret['comment']['xml_dump'].strip()
        # get size the memory snapshot
        if memory_snapshot:
            if memory_snapshot['type'] == 'directory':
                res_memory_size = get_output(f'du -b {memory_snapshot["path"]}/{memory_snapshot["name"]}')
                if res_memory_size['result']:
                    memory_size = res_memory_size['comment'].strip().split()[0]
                    res.update({'memory_size': int(memory_size)} if memory_size.isdigit() else {})
            elif memory_snapshot['type'] == 'thinlvm':
                cmd_memory = 'LVM_SUPPRESS_FD_WARNINGS=1 lvs --noheadings --units b -o lv_size,data_percent'
                res_memory_size = get_output(f'{cmd_memory} {memory_snapshot["vgname"]}/{memory_snapshot["name"]}')
                if res_memory_size['result']:
                    # s_p: [<lv_size>, <data_percent>]
                    s_p = res_memory_size['comment'].strip().split()
                    try:
                        memory_size = int(float(s_p[0][:-1]) * float(s_p[1]) / 100)
                    except ValueError:
                        pass
                    else:
                        res.update({'memory_size': memory_size})
    return res


@states_response
def revert(**kwargs):
    from .domain import shutdown, start
    comment = ''

    creating_snapshots = kwargs['creating_snapshots']
    domain_should_be_powered = kwargs['domain_should_be_powered']
    domain_id = kwargs['domain_id']
    domain_xml = kwargs['domain_xml']
    first_boot = kwargs['first_boot']
    domain_cpu_priority = kwargs['domain_cpu_priority']
    memory = kwargs['memory']
    maxmemory = kwargs.get('maxmemory')
    domain_should_be_suspended = kwargs['domain_should_be_suspended']
    domain_infs_map = kwargs['domain_infs_map']
    user = kwargs['user']
    remove = kwargs['remove']
    zfs_snapshots = kwargs.get('zfs_snapshots', [])
    linstor_snapshots = kwargs.get('linstor_snapshots', [])
    ctrl_host = kwargs.get('ctrl_host')

    cmd_ret = shutdown(domain_id, True)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    comment += cmd_ret['comment']

    for snapshot in creating_snapshots:
        if snapshot['type'] == 'directory':
            cmd_ret = create_directory_snapshot(snapshot['source'], snapshot['backing_file_source'],
                                                snapshot['backing_file_format'], snapshot['size'], user)
        elif snapshot['type'] == 'thinlvm':
            cmd_ret = create_thinlvm_snapshot(snapshot['name'], snapshot['vgname'], snapshot['thin_pool_name'],
                                              snapshot['backing_file_source'], snapshot['backing_file_format'],
                                              snapshot['size'], user)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

    for snapshot in zfs_snapshots:
        cmd_ret = zfs_revert(snapshot['snapshot_uuid'], snapshot['vdisk_uuid'], snapshot['zfs_pool_name'], user)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']
    if linstor_snapshots:
        from veilrpc.states.linstor import revert_snapshot
        for snapshot in linstor_snapshots:
            cmd_ret = revert_snapshot(snapshot['snapshot_uuid'], snapshot['resource'],
                                      ctrl_host=ctrl_host, resource_prev=snapshot['resource_prev'],
                                      snap_resource=snapshot['snap_resource'])
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            comment += cmd_ret['comment']['comment']

    if domain_should_be_powered:
        if memory:
            cmd_ret = restore(domain_id, memory, domain_xml, domain_should_be_suspended, maxmemory)
        else:
            cmd_ret = start(domain_id, domain_xml, domain_cpu_priority, domain_infs_map,
                            first_boot, domain_should_be_suspended)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

    if remove:
        cmd_ret = remove_snapshots(remove, user)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

    return comment


@states_response
def delete(**kwargs):

    res = {'comment': ''}

    if kwargs['commit']:
        progress_data = dict(type='progress', jid=kwargs['task'], id='', name='delete', result=True, steps={})
        progress_data['steps']['total_steps'] = len(kwargs['commit'])
        for vs in kwargs['commit']:
            if kwargs['snapshot_is_busy']:
                files_current_chain = get_backing_chain(kwargs['domain_id'], vs['base'])
                progress_data['steps']['total_steps'] += \
                    len([valid for valid in vs['overlay_children'] if valid not in files_current_chain['comment']])
            else:
                progress_data['steps']['total_steps'] += len(vs['overlay_children'])
        progress_data['steps']['step'] = 0

    for vdisk_snapshot in kwargs['commit']:
        cmd_ret = commit(kwargs['domain_id'], vdisk_snapshot['dev'], vdisk_snapshot['base'],
                         vdisk_snapshot['base_file_format'], vdisk_snapshot['overlay'],
                         vdisk_snapshot['overlay_children'], vdisk_snapshot['is_current'], kwargs['snapshot_is_busy'],
                         kwargs['user'], kwargs['task'], redis=kwargs.get('redis', None), progress_data=progress_data,
                         bandwidth=kwargs.get('bandwidth', 0),
                         base_datapool_type=vdisk_snapshot.get('base_datapool_type'),
                         overlay_datapool_type=vdisk_snapshot.get('overlay_datapool_type'),
                         debug=kwargs.get('debug', False))
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        res['comment'] += cmd_ret['comment']

    if kwargs['remove']:
        if kwargs['guaranteed']:
            cmd_ret = clean_guaranteed(kwargs['remove_sources'], kwargs['clean_type'], kwargs['clean_count'],
                                       kwargs['task'], redis=kwargs.get('redis', None))
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            res['comment'] += cmd_ret['comment']

        for snapshot in kwargs.get('grpc_rollback', []):
            cmd_ret = zfs_revert(snapshot['snapshot_uuid'], snapshot['vdisk_uuid'], snapshot['zfs_pool_name'],
                                 kwargs['user'])
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            res['comment'] += cmd_ret['comment']

        zfs_snapshots = [s for s in kwargs['remove'] if s.get('zfs_vdisk_snapshot')]
        if zfs_snapshots:
            cmd_ret = zfs_remove_snapshots(zfs_snapshots, kwargs['user'], kwargs.get('merge', False))
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
            res['comment'] += cmd_ret['comment']
        linstor_snapshots = [s for s in kwargs['remove'] if s.get('linstor_vdisk_snapshot')]
        if linstor_snapshots:
            from veilrpc.states.linstor import delete_snapshot as lin_snap_delete
            if kwargs.get('merge') or kwargs.get('consolidate'):
                s_current = [_ for _ in linstor_snapshots if _['is_current']]
                if not s_current:
                    raise CommandExecutionError('Not current snapshot.')
                for s in s_current:
                    cmd_ret = lin_snap_delete(None, s['resource'], kwargs.get('ctrl_host'),
                                              list_ssh_options=kwargs.get('list_ssh_options', []),
                                              owner=kwargs.get('user'),
                                              node_uuid=kwargs.get('node_uuid'),
                                              merge=True)
                    if not cmd_ret['result']:
                        raise CommandExecutionError(cmd_ret['comment']['comment'])
                    res['comment'] += cmd_ret['comment']['comment']
                    if cmd_ret['comment']['data']['map_resrc_snap']:
                        res['map_resrc_snap'] = cmd_ret['comment']['data']['map_resrc_snap']
            else:
                for s in linstor_snapshots:
                    if s['is_current']:
                        continue
                    cmd_ret = lin_snap_delete(s['snapshot_uuid'], s['resource'], kwargs.get('ctrl_host'),
                                              list_ssh_options=kwargs.get('list_ssh_options', []),
                                              owner=kwargs.get('user'),
                                              node_uuid=kwargs.get('node_uuid'),
                                              merge=False,
                                              last_of_branch=s.get('last_of_branch', False),
                                              parent_uuid=s.get('parent_uuid', ''))
                    if not cmd_ret['result']:
                        raise CommandExecutionError(cmd_ret['comment']['comment'])
                    res['comment'] += cmd_ret['comment']['comment']
                    if cmd_ret['comment']['data']['map_resrc_snap']:
                        res['map_resrc_snap'] = cmd_ret['comment']['data']['map_resrc_snap']
        cmd_ret = remove_snapshots([s for s in kwargs['remove'] if not s.get('zfs_vdisk_snapshot')
                                    and not s.get('linstor_vdisk_snapshot')], kwargs['user'],
                                   ctrl_host=kwargs.get('ctrl_host'))
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        res['comment'] += cmd_ret['comment']

    return res


@states_response
def copy(**kwargs):
    from .domain import shutdown
    comment = ''

    matched_vdisks_snapshots = kwargs.get('matched_vdisks_snapshots')
    is_local = kwargs['is_local']
    ssh_options = kwargs['ssh_options']
    user = kwargs['user']
    domain = kwargs['domain']
    parent = kwargs['parent']
    map_snapshots = kwargs['map_snapshots']
    map_zfs_vdisks = kwargs.get('map_zfs_vdisks', [])
    remove = kwargs.get('remove')
    redis = kwargs.get('redis', None)
    task = kwargs.get('task', None)

    cmd_ret = shutdown(domain, True)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    comment += cmd_ret['comment']

    progress_data = dict(type='progress', jid=task, id='', name='copy', result=True, steps={})
    if map_zfs_vdisks:
        progress_data['steps']['total_steps'] = len(map_snapshots) * len(map_zfs_vdisks) - len(map_zfs_vdisks)
    else:
        progress_data['steps']['total_steps'] = len(matched_vdisks_snapshots)
    progress_data['steps']['step'] = 0
    for item in map_zfs_vdisks:
        cmd_ret = zfs_to_zfs_copy(item['is_local'], item['src_vdisk_uuid'], item['src_pool_name'],
                                  item['dst_vdisk_uuid'], item['dst_pool_name'], map_snapshots, ssh_options, user, True,
                                  task=task, redis=redis, progress_data=progress_data)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

    from .copy import dir_to_dir_copy
    for vs in matched_vdisks_snapshots:
        cmd_ret = dir_to_dir_copy(vs['is_local'], vs['original_source'], vs['destination_source'],
                                  task, ssh_options, user, redis=redis, progress_data=progress_data)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

    if kwargs['matched_rebase']:
        cmd_ret = rebase_unsafe(kwargs['matched_rebase'], is_local, ssh_options, user)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

    if kwargs.get('list_memory_snapshots'):
        cmd_ret = replace_domain_id_into_memory(domain, parent, kwargs['list_memory_snapshots'],
                                                is_local, ssh_options, user)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

    if 'new_current_state' in kwargs:
        for vdisk_snapshot in kwargs['new_current_state']:
            cmd_ret = create_directory_snapshot(vdisk_snapshot['source'], vdisk_snapshot['backing_file_source'],
                                                vdisk_snapshot['backing_file_format'], vdisk_snapshot['size'],
                                                user, ssh_options=ssh_options)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

    if remove:
        cmd_ret = remove_snapshots(remove, user, ssh_options=ssh_options)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']

    return comment


@states_response
def status(**kwargs):
    if kwargs['snapshots']:
        full_name = 'sudo zfs list -Ht snapshot -o name | grep @{} ||true'
        for snapshot in [s for s in kwargs['snapshots'] if s.get('zfs_vdisk_snapshot')]:
            snapshot_name = '{}__{}'.format(snapshot['vdisk_uuid'], snapshot['snapshot_uuid'])
            get_full_name = get_output(usr_cmd(full_name.format(snapshot_name), kwargs['user']))
            if not get_full_name['result']:
                continue
            else:
                if get_full_name['comment'].strip() != '':
                    snapshot['status'] = 'ACTIVE'
                else:
                    snapshot['status'] = 'FAILED'
        for snapshot in [s for s in kwargs['snapshots'] if not s.get('zfs_vdisk_snapshot')]:
            if os.path.exists(snapshot['source']):
                snapshot['status'] = 'ACTIVE'
            else:
                snapshot['status'] = 'FAILED'
    return kwargs
