"""
(c) 2023 LLC «DACOM M»
"""

import os
import time
import shutil
import uuid
import crypt

from veilrpc.utils.file_based_lock import lock
from veilrpc.utils.bazalt import states_response, CommandExecutionError, SSH_GROUP, \
    get_output
from veilrpc.states.puppet import puppet_modify, puppet_eyaml_common, puppet_apply

COMMENT_TAG = "VEIL_KEY"


def _tmp_file_path(file_path):
    path = os.path.dirname(file_path)
    file_name = os.path.basename(file_path)
    tmp_file_name = f"_{str(uuid.uuid4())}_{file_name}"
    return os.path.join(path, tmp_file_name)


def _add_authorized_key(file_path, key, key_id):
    file_path = str(file_path)
    key = str(key)
    key_id = str(key_id)
    tmp_file_path = _tmp_file_path(file_path)
    shutil.copy2(file_path, tmp_file_path)
    src_file_stat = os.stat(file_path)
    os.chown(tmp_file_path, src_file_stat.st_uid, src_file_stat.st_gid)
    with open(tmp_file_path, 'a') as tmp_file:
        tmp_file.write(f"{key} {COMMENT_TAG}:{key_id}:{time.time()}\n")
    shutil.move(tmp_file_path, file_path)


def _delete_authorized_key(file_path, key_id):
    file_path = str(file_path)
    src_file_stat = os.stat(file_path)
    key_id = str(key_id)
    tmp_file_path = _tmp_file_path(file_path)
    with open(file_path, 'r') as src_file:
        with open(tmp_file_path, 'w') as dst_file:
            for line in src_file:
                if COMMENT_TAG in line:
                    try:
                        veil_comment_position = line.find(COMMENT_TAG)
                        veil_comment_line = line[veil_comment_position:]
                        line_key_id = veil_comment_line.split(":")[1]
                        if line_key_id == key_id:
                            # didn't write to dst tmp file
                            continue
                    except IndexError:
                        # bad formated veil line? copy as-is to dst
                        pass
                dst_file.write(line)
    shutil.copystat(file_path, tmp_file_path)
    os.chown(tmp_file_path, src_file_stat.st_uid, src_file_stat.st_gid)
    shutil.move(tmp_file_path, file_path)


@states_response
def add_authorized_key(file_path, key, key_id):
    # параметр file_path используется как для конфигурации декоратора,
    # так и для передачи в функцию
    return lock(file_path)(_add_authorized_key)(
        file_path, key, key_id)


def sshd_config() -> str:
    if os.path.exists("/etc/ssh/sshd_config"):
        return "/etc/ssh/sshd_config"
    elif os.path.exists("/etc/openssh/sshd_config"):
        return "/etc/openssh/sshd_config"
    else:
        raise CommandExecutionError("no sshd config")


@states_response
def ssh_get_param(param, **kwargs):
    cmd = f"grep -r '^{param}' {sshd_config()} | awk '{{print $2}}'"
    resp = get_output(cmd, **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    value = resp['comment'].strip()
    try:
        return int(value)
    except ValueError:
        return value


@states_response
def security_limits_get_param(param, **kwargs):
    cmd = f"grep -r '^{param}' /etc/security/sshd_config | awk '{{print $2}}'"
    resp = get_output(cmd, **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    value = resp['comment'].strip()
    try:
        return int(value)
    except ValueError:
        return value


@states_response
def delete_authorized_key(file_path, key_id):
    return lock(file_path)(_delete_authorized_key)(
        file_path, key_id)


@states_response
def ssh_status(**kwargs):
    resp = get_output(f"cat {sshd_config()}", **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    return True if f'AllowGroups {SSH_GROUP}' in resp['comment'] else False


@states_response
def ssh_root_to_group(value=True, **kwargs):
    cmd = f"usermod root -G {SSH_GROUP}" if value else "usermod root -G ''"
    resp = get_output(cmd, **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    return f'The root user now has{"" if value else " no"} SSH access'


@states_response
def ssh_root_access(value=True, **kwargs):
    ret = puppet_modify(module='ssh',
                        params={'ssh::server::permit_root_login': 'yes' if value else 'prohibit-password'},
                        **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return f'The root user now has{"" if value else " no"} SSH access by password'


@states_response
def ssh_get_users(info=False, **kwargs):
    resp = get_output(f"getent group {SSH_GROUP} | awk -F: '{{print $4}}'", **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    users = resp['comment'].strip().split(',') if resp['comment'].strip() else []
    if info:
        users_with_info = []
        for user in users:
            if user == 'root':
                resp = ssh_get_param('PermitRootLogin', **kwargs)
                if not resp['result']:
                    raise CommandExecutionError(resp['comment'])
                user_status = False if resp['comment'].strip() in ['without-password', 'prohibit-password'] else True
            else:
                resp = get_output(f"grep '{user}:' /etc/passwd", **kwargs)
                if not resp['result']:
                    raise CommandExecutionError(resp['comment'])
                user_status = False if resp['comment'].strip().split(':')[-1] == '/usr/sbin/nologin' else True
            users_with_info.append(dict(name=user, status=user_status))
        return users_with_info
    return users


def _sshd_config_settings():
    settings = dict()
    resp = ssh_get_param(param='ClientAliveInterval')
    if resp['result']:
        settings['alive_interval'] = resp['comment']
    resp = ssh_get_param(param='ClientAliveCountMax')
    if resp['result']:
        settings['alive_count_max'] = resp['comment']
    resp = ssh_get_param(param='MaxAuthTries')
    if resp['result']:
        settings['maxauthtries'] = resp['comment']
    resp = ssh_get_param(param='MaxSessions')
    if resp['result']:
        settings['maxsessions'] = resp['comment']
    return settings


def _security_limits_settings():
    settings = dict()
    resp = get_security_limits(param='maxlogins')
    if resp['result']:
        settings['maxlogins'] = resp['comment']
    resp = get_security_limits(param='maxsyslogins')
    if resp['result']:
        settings['maxsyslogins'] = resp['comment']
    return settings


@states_response
def ssh_settings(**kwargs):
    settings = _sshd_config_settings()
    settings.update(_security_limits_settings())
    return settings


@states_response
def ssh_unlock_users(ssh_users=None, **kwargs):
    """
    Разблокирует пользователей SSH. Если список пользователей не передан, разблокирует всех.
    """
    error_list = []
    if not ssh_users:
        cmd_ret = ssh_get_users()
        if not cmd_ret['result']:
            error_list.append(cmd_ret['comment'])
            ssh_users = []
        else:
            ssh_users = cmd_ret['comment']
    modify = False
    update_conf = {'vaccounts::user_list': {}}
    for user in ssh_users:
        user = user if isinstance(user, str) else user['name']
        if user == 'root':
            cmd_ret = ssh_root_access(**kwargs)
            if not cmd_ret['result']:
                error_list.append(cmd_ret['comment'])
        else:
            update_conf['vaccounts::user_list'].update({user: {'locked': False}})
            modify = True
    if modify:
        ret = puppet_modify(module='vaccounts', params=update_conf)
        if not ret['result']:
            error_list.append(ret['comment'])
    if error_list:
        raise CommandExecutionError(
            f'An error occurred while enabling SSH users. Errors: {", ".join(error_list)}')


@states_response
def ssh_lock_users(ssh_users, **kwargs):
    """
    Блокирует пользователей SSH.
    """
    error_list = []
    if ssh_users:
        modify =False
        update_conf = {'vaccounts::user_list': {}}
        for user in ssh_users:
            user = user if isinstance(user, str) else user['name']
            if user == 'root':
                cmd_ret = ssh_root_access(value=False, **kwargs)
                if not cmd_ret['result']:
                    error_list.append(cmd_ret['comment'])
            else:
                update_conf['vaccounts::user_list'].update({user: {'locked': True}})
                modify = True
        if modify:
            ret = puppet_modify(module='vaccounts', params=update_conf)
            if not ret['result']:
                error_list.append(ret['comment'])
        if error_list:
            raise CommandExecutionError(
                f'An error occurred while disabling SSH users. Errors: {", ".join(error_list)}')


@states_response
def ssh_remove_users(ssh_users=None, **kwargs):
    """
    Удаляет пользователей SSH. Если список пользователей не передан, удаляет всех.
    На этапе тестирования ни один root не пострадал.
    """
    error_list = []
    if not ssh_users:
        cmd_ret = ssh_get_users(**kwargs)
        if not cmd_ret['result']:
            error_list.append(cmd_ret['comment'])
            ssh_users = []
        else:
            ssh_users = cmd_ret['comment']
    if ssh_users:
        update_conf = {'vaccounts::user_list': {}}
        update_encrypt_conf = {}
        modify = False
        for user in ssh_users:
            user = user if isinstance(user, str) else user['name']
            if user != 'root':
                modify = True
                cmd_ret = get_output(f"userdel -f -r {user}", **kwargs)
                if not cmd_ret['result']:
                    error_list.append(cmd_ret['comment'])
                update_conf['vaccounts::user_list'].update({user: None})
                update_encrypt_conf.update({f"{user}_account_password": None})
        if modify:
            if error_list:
                raise CommandExecutionError(
                    f'An error occurred while removing SSH users. Errors: {", ".join(error_list)}')
            ret = puppet_modify(module='vaccounts', params=update_conf)
            if not ret['result']:
                raise CommandExecutionError(ret['comment'])
            puppet_modify(params=update_encrypt_conf, encrypt=True, apply=False)
            return ret['comment']


@states_response
def ssh_add_users(ssh_users, set_pass=True, **kwargs):
    if ssh_users:
        update_conf = {'vaccounts::user_list': {}}
        modify = False
        for user in ssh_users:
            if user['name'] == 'root':
                continue
            modify = True
            update_conf['vaccounts::user_list'].update({
                user['name']: {
                    'shell': '/bin/bash',
                    'group': "%{lookup('veil_global_group')}",
                    'groups': ["%{lookup('veil_admins_group')}"],
                    'bashrc_source': '/root/.bashrc',
                    'purge_sshkeys': True}})
            if set_pass:
                update_conf['vaccounts::user_list'][user['name']].update(
                    {'password': f"%{{lookup('{user['name']}_account_password')}}"})
                password = crypt.crypt(user['password'], crypt.METHOD_SHA512)
                puppet_eyaml_common(f"{user['name']}_account_password", password)
        if modify:
            ret = puppet_modify(module='vaccounts', params=update_conf)
            if not ret['result']:
                raise CommandExecutionError(ret['comment'])
            return ret['comment']


@states_response
def ssh_set_password(ssh_users, **kwargs):
    for user in ssh_users:
        password = crypt.crypt(user['password'], crypt.METHOD_SHA512)
        puppet_eyaml_common(f"{user['name']}_account_password", password)
    ret = puppet_apply('vaccounts')
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def ssh_remove_entry_restrictions(**kwargs):
    puppet_modify(module='ssh',
                  params={'ssh::server::allow_groups': None,
                          'ssh::server::deny_groups': None},
                  **kwargs)
    ret = ssh_root_access(**kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return 'SSH entry restrictions have been removed'


@states_response
def ssh_enable(**kwargs):
    ret = puppet_modify(module='ssh',
                        params={'ssh::server::deny_groups': None,
                                'ssh::server::allow_groups': [SSH_GROUP, 'veil']},
                        **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    ret = puppet_modify(module='ttyd', params={'ttyd::service_ensure': 'running',
                                               'ttyd::service_enable': True})
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return 'SSH successfully enabled'


@states_response
def ssh_disable(**kwargs):
    ret = puppet_modify(module='ssh',
                        params={'ssh::server::deny_groups': [SSH_GROUP],
                                'ssh::server::allow_groups': ['veil']},
                        **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    ret = puppet_modify(module='ttyd', params={'ttyd::service_ensure': 'stopped',
                                               'ttyd::service_enable': False})
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return 'SSH successfully disabled'


@states_response
def ssh_alive_interval(value, **kwargs):
    ret = puppet_modify(module='ssh',
                        params={'ssh::server::client_alive_interval': value},
                        **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def ssh_alive_count_max(value, **kwargs):
    ret = puppet_modify(module='ssh',
                        params={'ssh::server::client_alive_count_max': value},
                        **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def ssh_maxauthtries(value, **kwargs):
    ret = puppet_modify(module='ssh',
                        params={'ssh::server::max_auth_tries': value},
                        **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def ssh_maxsessions(value, **kwargs):
    ret = puppet_modify(module='ssh',
                        params={'ssh::server::max_sessions': value},
                        **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def ssh_sessions(**kwargs):
    ret = get_output('last | grep "still logged in"', **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def ssh_maxlogins(value, **kwargs):
    ret = set_security_limits(param='maxlogins',
                              value=value)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def ssh_maxsyslogins(value, **kwargs):
    ret = set_security_limits(param='maxsyslogins',
                              value=value)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def set_security_limits(param, value, **kwargs):
    ret = puppet_modify(module='limits',
                        params={'limits::entries': {f'@veil-admins/{param}': {'hard': value}}})
    if not ret['result']:
        comment = f"Unable to apply securuty limits config: {ret['comment']}"
        raise CommandExecutionError(comment)
    return 'Security limits successfully updated.'


@states_response
def get_security_limits(param, **kwargs):
    limits_conf = puppet_modify(module='limits')['comment']
    if limit := limits_conf.get('limits::entries', {}).get(f'@veil-admins/{param}', {}).get('hard'):
        return limit
    else:
        return None
