"""
(c) 2023 LLC «DACOM M»
"""

from glob import glob
import json
import os
import platform
import re
from defusedxml import ElementTree as ET

from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output, clean_guaranteed, \
    delete_file_with_owner, usr_cmd, ListVars, states_result, call, RUN_PATH, ZPOOL_STORAGES_PATH
from veilrpc.utils.size import convert_size_gb
from veilrpc.states.linstor import create_disk, delete_disk

VDISKS_CMD = "find {paths} -maxdepth {maxdepth} -type f -name *.qcow2 2>&-"
DEFAULT_UNIT = 'G'


# pylint: disable=redefined-builtin)
@states_response
def created(path, vdisk_name, vdisk_uuid, format, size, preallocation, owner, metadata=None, unit=DEFAULT_UNIT,
            **kwargs):
    """
    Create directory vdisk with qemu-img.
    If there is not enough space in the datapool, an error will be returned and created 'thin' disk
    will be deleted.

    :param path: datapool path
    :param vdisk_name: name of the vdisk
    :param vdisk_uuid: uuid of the vdisk
    :param format: format for qemu-img (usually qcow2)
    :param size: size of the vdisk
    :param preallocation: preallocation attribute (falloc or off)
    :param owner: user
    :param metadata: metadata
    :param unit: size unit
    """

    vdisk_path = os.path.join(path, vdisk_name)
    qemu_cmd = f'qemu-img create -f {format} -o preallocation={preallocation} {vdisk_path} {size}{unit}'
    cmd = usr_cmd(qemu_cmd, owner)
    cmd_ret = get_output(cmd, timeout=3600)
    if not cmd_ret['result']:
        comment = f"An error occurred while creating vdisk {vdisk_name}: {cmd_ret['comment']}"
        if os.path.exists(vdisk_path):
            delete_file_with_owner(vdisk_path, owner)
        raise CommandExecutionError(comment)
    if metadata:
        from .directory_datapool import write_meta
        write_meta(os.path.join(path, vdisk_name.replace('.qcow2', '.meta')),
                   metadata, owner)
    return get_vdisk_virtual_size(vdisk_path, owner, convert=False)


@states_response
def deleted(path, owner, **kwargs):
    """
    Удаляет диск.
    """

    if not os.path.exists(path):
        return f'Disk {path} is already deleted.'
    else:
        cmd_ret = delete_file_with_owner(path, owner, **kwargs)
        if not cmd_ret['result']:
            comment = f"An error occurred while deleting vdisk {path}: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
        return f'The vdisk {path} was successfully deleted.'


@states_response
def deleted_meta(path, owner, **kwargs):
    """
    Удаляет meta.
    """

    if not os.path.exists(path):
        return f'Meta {path} is already deleted.'
    else:
        cmd_ret = delete_file_with_owner(path, owner, **kwargs)
        if not cmd_ret['result']:
            comment = f"An error occurred while deleting meta {path}: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
        return f'The meta {path} was successfully deleted.'


@states_response
def extended(path, size, owner, unit='M', **kwargs):
    """
    Расширяет диск.

    :param path: vdisk path
    :param size: size
    :param owner: user
    :param unit: unit
    """
    size = int(round(size * 1024))  # to Mb
    if not os.path.exists(path):
        raise CommandExecutionError(f'Disk {path} does not exist.')
    else:
        qemu_cmd = f'qemu-img resize {path} +{size}{unit}'
        cmd = usr_cmd(qemu_cmd, owner)
        cmd_ret = get_output(cmd, **kwargs)
        if not cmd_ret['result']:
            comment = f"An error occurred while extending vdisk {path}: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
        return get_vdisk_virtual_size(path, owner, convert=False)


@states_response
def converted(path, dst_path, owner, task, metadata=None, meta_file=None, **kwargs):
    if not os.path.exists(path):
        raise CommandExecutionError(f'Disk {path} does not exist.')
    else:
        from .copy import convert
        progress_data = dict(type='progress', jid=task, id='', name='converted', result=True)

        if kwargs.get('destination_type') == 'zfs':
            mkdataset(dst_path, owner)

        convert_ret = convert(
            original_source=path,
            destination_source=dst_path,
            destination_format=kwargs.get('destination_format', 'qcow2'),
            owner=owner,
            preallocate=kwargs.get('preallocate', None),
            progress_data=progress_data,
            redis=kwargs.get('redis', None),
            options=kwargs.get('additional_options', None))
        if convert_ret['retcode']:
            comment = f"An error occurred while converting file {path}: {convert_ret['error']}"
            if os.path.exists(dst_path):
                os.remove(dst_path)
            raise CommandExecutionError(comment)
        else:
            if metadata and meta_file:
                from .directory_datapool import write_meta
                write_meta(meta_file, metadata, owner)
            if kwargs.get('with_deletion', False):
                os.remove(path)
            return get_vdisk_virtual_size(dst_path, owner, convert=False)


@states_response
def imported(path, dst_path, owner, with_deletion, task, metadata=None, meta_file=None, **kwargs):

    if not os.path.exists(path):
        raise CommandExecutionError(f'Disk {path} does not exist.')

    from .copy import rsync, convert
    preallocate = kwargs.get('preallocate', False)
    progress_data = dict(type='progress', jid=task, id='', name='imported', result=True)

    # если на входе не qcow2 (или qcow2, но надо поменять preallocated на sparse или наоборот), то конвертим.
    # Чтоб сохранить поведение анализа по содержимому, проверяем, что в файлах
    cmd = usr_cmd(f'qemu-img info \"{path}\"', owner)
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(f"Failed to determine disk type: {cmd_ret['comment']}")

    out = cmd_ret['comment'].splitlines()

    src_fmt = [x.split() for x in out if x.startswith('file format:')][0][-1]
    if src_fmt == 'qcow2':
        # virt_size > disk_size, если тонкий файл
        virt_size = [x.split() for x in out if x.startswith('virtual size:')][0][2]
        disk_size = [x.split() for x in out if x.startswith('disk size:')][0][2]
        preallocated = (virt_size == disk_size)

    if kwargs.get('destination_type') == 'zfs':
        mkdataset(dst_path, owner)

    if (src_fmt != 'qcow2') or (preallocated != preallocate):
        ret = convert(original_source=path, destination_source=dst_path,
                      destination_format='qcow2', owner=owner, preallocate=preallocate,
                      progress_data=progress_data, redis=kwargs.get('redis', None))
    else:
        ret = rsync(is_local=True, original_source=path, destination_source=dst_path,
                    ssh_options={}, owner=owner,
                    progress_data=progress_data, redis=kwargs.get('redis', None))

    if ret['retcode']:
        comment = f"An error occurred while importing file {path}: {ret['error']}"
        if os.path.exists(dst_path):
            os.remove(dst_path)
        raise CommandExecutionError(comment)
    else:
        if metadata and meta_file:
            from .directory_datapool import write_meta
            write_meta(meta_file, metadata, owner)
        if with_deletion:
            os.remove(path)
        return get_vdisk_virtual_size(dst_path, owner, convert=False)


def get_vdisk_virtual_size(path, owner=None, convert=True, **kwargs):
    cmd = f"qemu-img info \'{path}\' -U --output json"
    if owner is not None:
        cmd = usr_cmd(cmd, owner)
    cmd_ret = get_output(cmd, **kwargs)
    try:
        v_size = json.loads(cmd_ret['comment'])['virtual-size']
    except json.JSONDecodeError as e:
        if platform.machine() == 'e2k':
            cmd_ret2 = get_output(f"qemu-img info --output=json \'{path}\'", **kwargs)
            try:
                v_size = json.loads(cmd_ret2['comment'])['virtual-size']
            except json.JSONDecodeError as err:
                raise CommandExecutionError(f"Error in json loads ({str(cmd_ret2['comment'])}): {str(err)}") from None
        else:
            raise CommandExecutionError(f"Error in get_vdisk_virtual_size: {e}") from None
    return convert_size_gb(v_size) if convert else v_size


def get_vdisk_virtual_size_lvm(path, owner=None, convert=True, **kwargs):
    cmd = f"lvs --units B --reportformat json {path}"
    if owner is not None:
        cmd = usr_cmd(cmd, owner)
    cmd_ret = get_output(cmd, **kwargs)
    v_size = json.loads(cmd_ret['comment'])['report'][0]['lv'][0]['lv_size'][:-1]
    return convert_size_gb(v_size) if convert else v_size


def get_vdisk_meta(path):
    try:
        with open(path.replace('.qcow2', '.meta')) as f:
            return json.load(f)
    except (OSError, json.JSONDecodeError):
        return {'uuid': os.path.splitext(path)[0]}


@states_response
def check(path, repair, owner, **kwargs):
    if repair:
        qemu_cmd = f'qemu-img check -r all {path}'
    else:
        qemu_cmd = f'qemu-img check {path}'
    cmd = usr_cmd(qemu_cmd, owner)
    cmd_ret = get_output(cmd, **kwargs)

    if not cmd_ret['result']:
        comment = f"An error occurred while check file {path}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def sparse(path, **kwargs):
    return states_result(get_output,
                         cmd=f"virt-sparsify --format qcow2 --in-place {path}",
                         **kwargs)


class TYPE(ListVars):
    local = 'local'
    lvm = 'lvm'
    thinlvm = 'thinlvm'
    lvm_shared = 'lvm_shared'
    nfs = 'nfs'
    glusterfs = 'glusterfs'
    ocfs2 = 'ocfs2'
    gfs2 = 'gfs2'
    zfs = 'zfs'
    gluster = 'gluster'
    cifs = 'cifs'
    linstor = 'linstor'


@states_response
def vdisk_discovery(datapool_type, datapool_options, meta=False, **kwargs):
    results = []
    # в датапуле типа zfs диски лежат по пути <datapool_path>/<disk_name>/
    maxdepth = '1' if datapool_type != TYPE.zfs else '2'
    if datapool_type in [TYPE.local, TYPE.nfs, TYPE.glusterfs, TYPE.zfs, TYPE.ocfs2, TYPE.gfs2,
                         TYPE.gluster, TYPE.cifs]:
        cmd = VDISKS_CMD.format(paths=datapool_options['path'], maxdepth=maxdepth)
        cmd_ret = get_output(cmd, **kwargs)
        if not cmd_ret['result']:
            raise CommandExecutionError(
                f"Can not discover virtual disks in datapool {datapool_options['path']}: {cmd_ret['comment']}.")
        out = cmd_ret['comment'].splitlines()
        for vdisk in out:
            result = {'path': vdisk, 'size': get_vdisk_virtual_size(vdisk),
                      'virtual_size': get_vdisk_virtual_size(vdisk, convert=False)}
            if meta:
                result['meta'] = get_vdisk_meta(vdisk)
            results.append(result)
    else:
        from .lvm import get_json_lvm_data
        if datapool_options.get('vgname'):
            lvm_data = get_json_lvm_data(vgname=datapool_options['vgname'], **kwargs)
            for value_lv in lvm_data['report'][0]['lv']:
                lv_name = value_lv["lv_name"]
                if re.match(r'^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-4[0-9a-fA-F]{3}'
                            '-[89abAB][0-9a-fA-F]{3}-[0-9a-fA-F]{12}$', lv_name):
                    vdisk = {'id': lv_name, 'size': int(value_lv["lv_size"]) // 1073741824,
                             'virtual_size': int(value_lv["lv_size"])}
                    if datapool_type == TYPE.thinlvm:
                        if value_lv["pool_lv"] == datapool_options['thin_pool_name']:
                            results.append(vdisk)
                    else:
                        results.append(vdisk)

    return results


@states_response
def vdisk_discovery_directory(datapools, meta=True, size=True, **kwargs):
    """
    Находит в папках пулов данных .qcow2 файлы, а также их размер
    :param datapools: список путей к пулам данных
    :param meta: сбор метаданных диска
    :param size: сбор размера диска
    """

    vdisks = []
    datapools_zfs, datapools_other = [], []

    for datapool in datapools:
        if ZPOOL_STORAGES_PATH in datapool:
            datapools_zfs.append(datapool)
        else:
            datapools_other.append(datapool)

    if datapools_other:
        vdisk_ret = get_output(VDISKS_CMD.format(paths=" ".join(datapools_other), maxdepth=1), **kwargs)
        if vdisk_ret['result']:
            vdisks_found = [vdisk for vdisk in vdisk_ret['comment'].splitlines() if vdisk]
        else:
            vdisks_found = []
    else:
        vdisks_found = []

    if datapools_zfs:
        vdisk_zfs_ret = get_output(VDISKS_CMD.format(paths=" ".join(datapools_zfs), maxdepth=2), **kwargs)
        if vdisk_zfs_ret['result']:
            vdisks_zfs_found = [vdisk for vdisk in vdisk_zfs_ret['comment'].splitlines() if vdisk]
        else:
            vdisks_zfs_found = []
    else:
        vdisks_zfs_found = []

    vdisks_found.extend(vdisks_zfs_found)

    for vdisk in vdisks_found:
        vdisk_dict = dict(path=vdisk)
        # размер диска в гигабайтах
        if size:
            vdisk_dict['size'] = get_vdisk_virtual_size(vdisk)
        if meta:
            vdisk_dict['meta'] = get_vdisk_meta(vdisk)
        vdisks.append(vdisk_dict)
    return vdisks


@states_response
def full_delete(guaranteed, vdisk_path, vdisk_name, clean_type, clean_count, task, owner, **kwargs):
    if 'snapshots' in kwargs and kwargs['snapshots']:
        if guaranteed:
            cmd_ret = clean_guaranteed(sources=kwargs['remove_sources'], clean_type=clean_type,
                                       clean_count=clean_count, task=task,
                                       redis=kwargs.get('redis', None))
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])

        from .snapshot import remove_snapshots
        cmd_ret = remove_snapshots(kwargs['snapshots'], owner)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])

    if guaranteed:
        cmd_ret = clean_guaranteed(sources=vdisk_path, clean_type=clean_type,
                                   clean_count=clean_count, task=task,
                                   redis=kwargs.get('redis', None))
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])

    cmd_ret = deleted(vdisk_path, owner)
    if not cmd_ret['result']:  # NOQA
        raise CommandExecutionError(cmd_ret['comment'])  # NOQA

    cmd_ret = deleted_meta(vdisk_path.replace('.qcow2', '.meta'), owner)
    if not cmd_ret['result']:  # NOQA
        raise CommandExecutionError(cmd_ret['comment'])  # NOQA

    return f'The vdisk {vdisk_path} was successfully deleted.'


@states_response
def node_full_delete(guaranteed, vdisk_path, vdisk_name, clean_type, clean_count, task, owner, **kwargs):
    from .domain import list_domains

    cmd_ret = list_domains()
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    for domain in cmd_ret['comment']:
        root = ET.fromstring(domain['dom_xml'])
        devices = root.find('devices')
        for device in devices.iterfind('disk'):
            if device.attrib['device'] == 'disk':
                source = device.find('source')
                if source and vdisk_path == source.attrib.get('file', source.attrib.get('name')):
                    raise CommandExecutionError(
                        f'Vdisk {vdisk_path} is attached to domain {domain["dom_name"]}.')
    full_delete(guaranteed, vdisk_path, vdisk_name, clean_type, clean_count, task, owner, **kwargs)


@states_response
def full_copy(destination, owner, **kwargs):
    from .copy import general_copy, create_local_zfs_dataset, write_vdisk_meta
    if destination['type'] == 'directory':
        cmd_ret = created(destination['path'], destination['name'], destination['id'], 'qcow2',
                          destination['size'], 'off', owner, destination['metadata'], destination.get('unit',
                                                                                                      DEFAULT_UNIT))
        if not cmd_ret['result']:  # NOQA
            raise CommandExecutionError(cmd_ret['comment'])  # NOQA
    elif destination['type'] == TYPE.thinlvm:
        from .lvm import lv_present
        cmd_ret = lv_present(name=destination['name'],
                             vgname='{}/{}'.format(destination['vgname'], destination['thin_pool_name']),
                             size=f"{destination['size']}G",
                             thinvolume=True, failhard=True)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    elif destination['type'] in [TYPE.lvm, TYPE.lvm_shared]:
        from .lvm import lv_present
        cmd_ret = lv_present(name=destination['name'],
                             vgname=destination['vgname'],
                             size=f"{destination['size']}G",
                             failhard=True)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    elif destination['type'] == TYPE.zfs:
        cmd_ret = create_local_zfs_dataset(destination['zpool_name'], destination['id'],
                                           os.path.dirname(destination['source']), owner)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    elif destination['type'] == TYPE.linstor:
        cmd_ret = create_disk(name=destination['resource'], size=f"{destination['size']}G",
                              storage_pool=destination['storage_pool'], place_count=destination['place_count'],
                              layer_list=destination.get('layer_list', ['drbd', 'storage']),
                              ctrl_host=destination['ctrl_host'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    cmd_ret = general_copy(destination=destination, owner=owner, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if destination['type'] == TYPE.zfs:
        write_vdisk_meta(True, destination['source'], {}, owner, destination['metadata'])
    return f"The {destination['type']} vdisk was successfully copied to {destination.get('source')}."


@states_response
def full_move(destination, owner, **kwargs):
    from .copy import general_move, create_local_zfs_dataset, write_vdisk_meta
    if destination['type'] == 'directory':
        cmd_ret = created(destination['path'], destination['name'], destination['id'], 'qcow2',
                          destination['size'], 'off', owner, destination['metadata'], destination.get('unit',
                                                                                                      DEFAULT_UNIT))
        if not cmd_ret['result']:  # NOQA
            raise CommandExecutionError(cmd_ret['comment'])  # NOQA
    elif destination['type'] == TYPE.thinlvm:
        from .lvm import lv_present
        cmd_ret = lv_present(name=destination['name'],
                             vgname='{}/{}'.format(destination['vgname'], destination['thin_pool_name']),
                             size=f"{destination['size']}G",
                             thinvolume=True, failhard=True)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    elif destination['type'] in [TYPE.lvm, TYPE.lvm_shared]:
        from .lvm import lv_present
        cmd_ret = lv_present(name=destination['name'],
                             vgname=destination['vgname'],
                             size=f"{destination['size']}G",
                             failhard=True)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    elif destination['type'] == TYPE.zfs:
        cmd_ret = create_local_zfs_dataset(destination['zpool_name'], destination['source'].split('/')[-2],
                                           os.path.dirname(destination['source']), owner)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    elif destination['type'] == TYPE.linstor:
        cmd_ret = create_disk(name=destination['resource'], size=f"{destination['size']}G",
                              storage_pool=destination['storage_pool'], place_count=destination['place_count'],
                              layer_list=destination.get('layer_list', ['drbd', 'storage']),
                              ctrl_host=destination['ctrl_host'])
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    cmd_ret = general_move(destination=destination, owner=owner, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if destination['type'] == TYPE.zfs:
        write_vdisk_meta(True, destination['source'], {}, owner, destination['metadata'])
    return f"The {destination['type']} vdisk was successfully moved to {destination.get('source')}."


@states_response
def lvm_create(vdisk_name, vgname, size, owner, unit=DEFAULT_UNIT, **kwargs):
    from .lvm import lv_present, lvdisplay
    cmd_ret = lv_present(name=vdisk_name,
                         vgname=vgname,
                         size=f'{size}{unit}')
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    path = f"/dev/{vgname}/{vdisk_name}"
    cmd = f'chown {owner}:{owner} $(readlink -f {path})'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return lvdisplay(lvname=path)['comment']


@states_response
def lvm_create_thin(vdisk_name, vgname, thin_pool_name, size, owner, unit=DEFAULT_UNIT, **kwargs):
    from .lvm import lv_present, lvdisplay
    cmd_ret = lv_present(name=vdisk_name,
                         vgname=f'{vgname}/{thin_pool_name}',
                         size=f'{size}{unit}',
                         thinvolume=True)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    path = f"/dev/{vgname}/{vdisk_name}"
    cmd = f'chown {owner}:{owner} $(readlink -f {path})'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return lvdisplay(lvname=path)['comment']


@states_response
def lvm_full_delete(guaranteed, clean_type, clean_count, task, owner,
                    vgname, vdisk_name, **kwargs):
    if guaranteed:
        cmd_ret = clean_guaranteed(sources=kwargs['vdisk_path'], clean_type=clean_type,
                                   clean_count=clean_count, task=task,
                                   redis=kwargs.get('redis', None))
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])

    from .lvm import lv_absent
    cmd_ret = lv_absent(vgname, vdisk_name)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    if 'snapshots' in kwargs and kwargs['snapshots']:
        if guaranteed:
            cmd_ret = clean_guaranteed(sources=kwargs['remove_sources'], clean_type=clean_type,
                                       clean_count=clean_count, task=task,
                                       redis=kwargs.get('redis', None))
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])

        from .snapshot import remove_snapshots
        cmd_ret = remove_snapshots(kwargs['snapshots'], owner)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])

    return 'The lvm vdisk was successfully deleted.'


@states_response
def lvm_extend(path, size, **kwargs):
    cmd = f'lvextend {path} -L +{size}G'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return get_vdisk_virtual_size_lvm(path, convert=False)


@states_response
def lvm_reduce(path, size):
    cmd = f'lvreduce {path} -L -{size}G -f'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return 'The lvm vdisk was successfully reduced.'


# pylint: disable=redefined-builtin)
@states_response
def zfs_create(path, vdisk_name, vdisk_uuid, format, size, preallocation, owner,
               zpool_name, metadata, unit=DEFAULT_UNIT, **kwargs):  # NOQA
    """
    Create directory vdisk with qemu-img.
    If there is not enough space in the datapool, an error will be returned and created 'thin' disk
    will be deleted.

    :param path: datapool path
    :param vdisk_name: name of the vdisk
    :param vdisk_uuid: uuid of the vdisk
    :param format: format for qemu-img (usually qcow2)
    :param size: size of the vdisk in Gb
    :param preallocation: preallocation attribute (falloc or off)
    :param owner: user
    :param zpool_name: for dataset name building like <zpool_name>/<vdisk_name>
    """
    from .copy import create_local_zfs_dataset, delete_local_zfs_dataset
    zfs_path = os.path.join(path, vdisk_uuid)
    cmd_ret = create_local_zfs_dataset(zpool_name, vdisk_uuid, zfs_path, owner)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    created_cmd_ret = created(path=zfs_path, vdisk_name=vdisk_name, vdisk_uuid=vdisk_uuid, format=format,
                              size=size, preallocation=preallocation, owner=owner,
                              metadata=metadata, unit=unit)
    if not created_cmd_ret['result']:  # NOQA
        delete_local_zfs_dataset(zpool_name, vdisk_uuid)
        raise CommandExecutionError(created_cmd_ret['comment'])  # NOQA
    return created_cmd_ret['comment']  # NOQA


@states_response
def zfs_full_delete(guaranteed, clean_type, clean_count, task, owner, vdisk_path, vdisk_name, **kwargs):
    from .copy import delete_local_zfs_dataset

    only_clone = kwargs.get('only_clone', False)
    if not only_clone:
        if guaranteed:
            cmd_ret = clean_guaranteed(sources=vdisk_path, clean_type=clean_type,
                                       clean_count=clean_count, task=task,
                                       redis=kwargs.get('redis', None))
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])

        cmd_ret = deleted(vdisk_path, owner)
        if not cmd_ret['result']:  # NOQA
            raise CommandExecutionError(cmd_ret['comment'])  # NOQA

        cmd_ret = deleted_meta(vdisk_path.replace('.qcow2', '.meta'), owner)
        if not cmd_ret['result']:  # NOQA
            raise CommandExecutionError(cmd_ret['comment'])  # NOQA

        if 'snapshots' in kwargs and kwargs['snapshots']:
            if guaranteed:
                cmd_ret = clean_guaranteed(sources=kwargs['remove_sources'], clean_type=clean_type,
                                           clean_count=clean_count, task=task,
                                           redis=kwargs.get('redis', None))
                if not cmd_ret['result']:
                    raise CommandExecutionError(cmd_ret['comment'])

            from .snapshot import remove_snapshots
            cmd_ret = remove_snapshots(kwargs['snapshots'], owner)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
    cmd_ret = delete_local_zfs_dataset(kwargs['zpool_name'], vdisk_name, only_clone)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    return 'The zfs vdisk was successfully deleted.'


def mkdataset(path, owner):
    from .copy import create_local_zfs_dataset
    path_as_list = path.split('/')
    cmd_ret = create_local_zfs_dataset(path_as_list[-3], path_as_list[-2],
                                       '/' + '/'.join(path_as_list[:-1]), owner)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])


@states_response
def apply_io_tune(domain_id, device_source, io_tune_config):
    cmd = "virsh domblklist %s | grep %s | awk '{print $1}'"
    cmd_ret = get_output(cmd % (domain_id, device_source))
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    device = cmd_ret['comment'].strip()
    cmd = f"virsh blkdeviotune {domain_id} {device}"
    for param, value in io_tune_config.items():
        cmd += f" --{param} {value}"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return "I/O settings changed successfully"


@states_response
def blockresize(domain_id, path, size, target_dev=None):
    """
    Расширяет диск.
    :param domain_id: domain_id
    :param path: vdisk path
    :param size: size in Gb
    :param target_dev: target_dev, for example vda
    """

    if not os.path.exists(path):
        raise CommandExecutionError(f'Disk {path} does not exist.')
    else:
        if target_dev:
            target_path = target_dev
        else:
            target_path = path
        cmd_ret = get_output(f'virsh blockresize {domain_id} {target_path} {size}G')
        if not cmd_ret['result']:
            comment = f"An error occurred while blockresized vdisk {target_path}: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
        return get_vdisk_virtual_size(path, convert=False)


@states_response
def qcow2_to_lun(path, device, task, **kwargs):
    return dd_file_device(path=path, device=device, task=task, device_to_file=False, **kwargs)


@states_response
def lun_to_qcow2(path, device, task, **kwargs):
    return dd_file_device(path=path, device=device, task=task, device_to_file=True, **kwargs)


def dd_file_device(path, device, task, device_to_file, **kwargs):
    from .copy import popen_with_progress
    from tempfile import mkstemp
    dev = None
    for nbd_path in glob('/sys/class/block/nbd*/size'):
        with open(nbd_path) as f:
            size = f.read().strip()
            if size == '0':
                dev = nbd_path.split('/')[-2]
                break
    if not dev:
        raise CommandExecutionError('No free nbd device.')
    dev = f"/dev/{dev}"
    call(f'qemu-nbd -d {dev}')
    cmd_ret = get_output(f'qemu-nbd -c {dev} {path}', timeout=300)
    if not cmd_ret['result']:
        call(f'qemu-nbd -d {dev}')
        raise CommandExecutionError(f"Error in qemu-nbd -c: {cmd_ret['comment']}")
    fh_stderr, f_stderr = mkstemp(dir=RUN_PATH, text=True)
    if device_to_file:
        src, dst, j_name = device, dev, 'lun_to_qcow2'
    else:
        src, dst, j_name = dev, device, 'qcow2_to_lun'
    cmd = f'(pv -n {src} | dd iflag=fullblock conv=sync,noerror of={dst} bs=8M 2>{f_stderr}) 2>&1'
    progress_data = dict(type='progress', jid=task, id='', name=j_name, result=True)
    p = popen_with_progress(cmd, "pv", progress_data, kwargs.get('redis', None))
    with os.fdopen(fh_stderr, 'r') as f:
        f.seek(0)
        raw_lines = f.read()
    os.remove(f_stderr)
    lines = raw_lines.splitlines()
    call(f'qemu-nbd -d {dev}')
    if not lines or (p.returncode and lines[0] != f"dd: error writing '{dev}': No space left on device"):
        raise CommandExecutionError(f'Failed to copy_data to device {dev} from {device}: {raw_lines}')
    return lines[-1]  # там статистика dd, сколько скопировано и скорость


def is_need_preallocation(path, owner=None, **kwargs):
    cmd = f"qemu-img info \'{path}\' -U --output json"
    if owner is not None:
        cmd = usr_cmd(cmd, owner)
    cmd_ret = get_output(cmd, **kwargs)
    ret_dict = json.loads(cmd_ret['comment'])
    is_need = ret_dict['virtual-size'] > ret_dict['actual-size']
    return is_need


@states_response
def linstor_create(resource, storage_pool, size, place_count, unit=DEFAULT_UNIT, ctrl_host='localhost',
                   **kwargs):
    ret = create_disk(name=resource, size=f'{size}{unit}', storage_pool=storage_pool, place_count=place_count,
                      layer_list=kwargs.get('layer_list', ['drbd', 'storage']), ctrl_host=ctrl_host)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def linstor_delete(resource, ctrl_host='localhost', **kwargs):
    ret = delete_disk(resource, volume_nr=kwargs.get('volume_nr', None), ctrl_host=ctrl_host)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']
