"""
(c) 2023 LLC «DACOM M»
"""

from copy import deepcopy
import os
import re

from veilrpc.utils import file_based_lock
from veilrpc.utils.bazalt import states_response, CommandExecutionError, \
    get_output, states_result, RUN_PATH, LUN_PREFIX, DEVICE_PREFIX, ZPOOL_STORAGES_PATH, get_result
from veilrpc.utils.mount import active
HLEVEL_DEVICE_LABELS = ["mirror", "raidz"]
DEFAULT_HLEVEL_DEVICE = "stripe"

LOCAL_PREFIX = '/dev/disk/by-id/'

IGNORED_STATUS = 'UNAVAIL'
SEMAPHORE = RUN_PATH + 'zpools_semaphore'


@file_based_lock.lock(SEMAPHORE)
# чтоб операции изменения ZFS пулов не вклинивались посередине zpool_full_info
def get_output_critical(cmd):
    return get_output(cmd)


@states_response
def zpool_list(**kwargs):
    """
    Return information about ZFS pool(s):
    name, size, free, health, devices info (name,size), pool_type (veil specific label for ZFS pool).
    Source:
        [root@controller ~]# zpool list -o name,size,free,health -vH
        zfsmirror       80M     79,9M   ONLINE
                mirror  80M     94,5K   79,9M   -       1%      0%
                vdd       -       -       -       -       -       -
                vde       -       -       -       -       -       -
        zfsraid5        256M    255M    ONLINE
                raidz1  256M    1018K   255M    -       0%      0%
                vdf       -       -       -       -       -       -
                vdg       -       -       -       -       -       -
                vdh       -       -       -       -       -       -
        zfsraid6        256M    255M    ONLINE
                raidz2  256M    819K    255M    -       0%      0%
                vdi       -       -       -       -       -       -
                vdj       -       -       -       -       -       -
                vdk       -       -       -       -       -       -
        zfsstripe       1,97G   1,97G   ONLINE
                vdb       1008M   95K     1008M   -       0%      0%
                vdc       1008M   178K    1008M   -       0%      0%
    Result:
        [   {   'devices': [   {   'name': 'vdb', 'size': '1008M'},
                               {   'name': 'vdc', 'size': '1008M'}],
                'free': '1,97G',
                'health': 'ONLINE',
                'name': 'zfsstripe',
                'pool_type': 'stripe',
                'size': '1,97G'},
            {   'devices': [   {   'name': 'vdd', 'size': '-', 'spare': False},
                               {   'name': 'vde', 'size': '-', 'spare': False}],
                'free': '79,9M',
                'health': 'ONLINE',
                'name': 'zfsmirror',
                'pool_type': 'mirror',
                'size': '80M'},
            {   'devices': [   {   'name': 'vdf', 'size': '-', 'spare': False},
                               {   'name': 'vdg', 'size': '-', 'spare': False},
                               {   'name': 'vdh', 'size': '-', 'spare': False}],
                'free': '255M',
                'health': 'ONLINE',
                'name': 'zfsraid5',
                'pool_type': 'raidz1',
                'size': '256M'},
            {   'devices': [   {   'name': 'vdi', 'size': '-', 'spare': False},
                               {   'name': 'vdj', 'size': '-', 'spare': False},
                               {   'name': 'vdk', 'size': '-', 'spare': False}],
                'free': '255M',
                'health': 'ONLINE',
                'name': 'zfsraid6',
                'size': '256M'}]

    Spare inuse case:
    [root@34781ba1-ff84-4127-8219-27093a02585a ~]# zpool list -o name,size,free,health -vH
    zp_ITd__TEST    1008M   1004M   DEGRADED
            mirror  1008M   3,72M   1004M   -       0%      0%
            vdc     -       -       -       -       -       -
            spare   -       -       -       -       -       -
            vdb     -       -       -       -       -       -
            vdi     -       -       -       -       -       -
            vde     -       -       -       -       -       -
    spare      -      -      -         -      -      -
            vdi     -       -       -       -       -       -
            vdh     -       -       -       -       -       -

    result:
    [   {   'devices': [   {   'name': 'vdc', 'size': '-', 'spare': False},
                       {   'name': 'vdb', 'size': '-', 'spare': False},
                       {   'name': 'vde', 'size': '-', 'spare': False},
                       {   'name': 'vdi', 'size': '-', 'spare': True},
                       {   'name': 'vdh', 'size': '-', 'spare': True}],
        'free': '1004M',
        'health': 'DEGRADED',
        'name': 'zp_ITd__TEST',
        'pool_type': 'mirror',
        'size': '1008M'}]

    Replacing case - sometimes we can catch in progress report:
            mirror	1008M	160K	1008M	-	0%	0%
        replacing	-	-	-	-	-	-
        vde	-	-	-	-	-	-
        vdi	-	-	-	-	-	-
        vdc	-	-	-	-	-	-
        vdb	-	-	-	-	-	-

    Decision: ignoring line with replacing and next line with device (replaced device)
    """
    err_message = "An error occurred while get parse information about ZFS pools. "
    # берем данные о свободном и занятом месте из zfs list, т.к.zpool list отдаёт тупо сумму объёмов vdev
    zfs_list_info_ret = zfs_list()
    if not zfs_list_info_ret.get('result'):
        comment = f"{err_message}Error: {zfs_list_info_ret.get('comment')}"
        raise CommandExecutionError(comment)
    zfs_list_info = zfs_list_info_ret.get('comment')
    # берем только корневые датасеты и из них свободное и общее, пихаем это в словарь пулов данных
    root_sets = {dct['name']: {'free': f'{dct["size"]}b',
                               'size': f'{int(dct["used"]) + int(dct["size"])}b'}
                 for dct in zfs_list_info if '/' not in dct['name']}

    # flag 'p' возвращает размер в байтах
    flags = kwargs.get('flags', '-vHp')

    zpool_cmd = "zpool list -o name,size,free,health {flags}".format(flags=flags)
    cmd_ret = get_output(zpool_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while get information about node's ZFS pools: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    try:
        data = []
        zpool_dict = {}
        pool_type = DEFAULT_HLEVEL_DEVICE
        is_spare = False
        is_spare_set = 2
        is_cache = False
        is_cache_set = 2
        is_log = False
        is_log_set = 2
        replacing = False

        for line in cmd_ret['comment'].splitlines():
            if 'replacing' in line:
                replacing = True
                continue

            is_device = line[0].startswith('\t')
            line = line.split('\t')

            # убрать пробелы и пустые строки
            line = [elem.strip() for elem in line if elem.strip()]

            is_spares_start = line[0].startswith("spare")
            is_cache_start = line[0].startswith("cache")
            is_log_start = line[0].startswith("logs")
            is_hlevel_device = any([line[0].startswith(label) for label in HLEVEL_DEVICE_LABELS])

            if is_spares_start:
                if not is_device:
                    is_spare = True
                    is_cache = False
                    is_log = False
                else:
                    is_spare_set -= 1
                continue

            if is_cache_start:
                if not is_device:
                    is_cache = True
                    is_spare = False
                    is_log = False
                else:
                    is_cache_set -= 1
                continue

            if is_log_start:
                if not is_device:
                    is_log = True
                    is_spare = False
                    is_cache = False
                else:
                    is_log_set -= 1
                continue

            if is_hlevel_device:
                pool_type = line[0]
                continue

            if not is_device:
                if zpool_dict:
                    zpool_dict['pool_type'] = pool_type
                    data.append(zpool_dict)
                    pool_type = DEFAULT_HLEVEL_DEVICE
                    is_spare = False
                    is_cache = False
                    is_log = False
                name, size, free, health = line  # pylint: disable=unused-variable
                # для дальнейшего использования к размерам нужно добавить формат
                free_2 = f'{root_sets.get(name, {}).get("free", 0)}'
                zpool_dict = dict(name=name, health=health, devices=[],
                                  size=f'{root_sets.get(name, {}).get("size", 0)}b',
                                  free=free_2 if free_2.endswith('b') else f'{free_2}b')
            else:
                if replacing:
                    replacing = False
                    continue
                if not is_spare_set:
                    is_spare_set = 2
                    continue
                if not is_cache_set:
                    is_cache_set = 2
                    continue
                if not is_log_set:
                    is_log_set = 2
                    continue
                if is_spare_set == 1:
                    is_spare_set -= 1
                if is_cache_set == 1:
                    is_cache_set -= 1
                if is_log_set == 1:
                    is_log_set -= 1
                name, size = line[:2]
                zpool_dict['devices'].append({'name': name, 'size': size,
                                              'spare': is_spare, 'cache': is_cache, 'log': is_log})

        if zpool_dict:
            zpool_dict["pool_type"] = pool_type
            data.append(zpool_dict)
        return data
    except Exception as e:
        comment = f"An error occurred while parse information about node's ZFS pools: {str(e)}"
        raise CommandExecutionError(comment) from None


@states_response
def zpool_status(name):
    return states_result(get_output, cmd=f"zpool status {name}")


@states_response
def zfs_list(**kwargs):
    """
    [root@controller ~]# zfs list -o name,available,used,type,mounted,mountpoint -H
    zfsmirror       39,9M   78K     filesystem      yes     /zfsmirror
    zfsraid5        85,1M   103K    filesystem      yes     /zfsraid5

    result:
    [   {   'dtype': 'filesystem',
        'free': '82,5K',
        'mounted': True,
        'mountpoint': '/zfsmirror',
        'name': 'zfsmirror',
        'size': '39,9M'},
    {   'dtype': 'filesystem',
        'free': '82,5K',
        'mounted': True,
        'mountpoint': '/zfsraid5',
        'name': 'zfsraid5',
        'size': '85,1M'}]
    """

    properties = kwargs.get('properties', 'name,available,used,type,mounted,mountpoint')

    zfs_cmd = "zfs list -o {properties} -Hp".format(properties=properties)
    cmd_ret = get_output(zfs_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while get information about ZFS dataset list: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        try:
            data = []

            for line in cmd_ret['comment'].splitlines():
                line = line.split('\t')
                # убрать пробелы и пустые строки
                line = [elem.strip() for elem in line if elem.strip()]

                name, size, used, dataset_type, mounted, mountpoint = line
                zfs_dict = dict(name=name,
                                size=size,
                                used=used,
                                dtype=dataset_type,
                                mounted=True if mounted == 'yes' else False,
                                mountpoint=mountpoint)

                data.append(zfs_dict)
            return data
        except Exception as e:
            comment = f"An error occurred while parse information about ZFS datasets: {str(e)}"
            raise CommandExecutionError(comment) from None


@states_response
def zpools_status_info(**kwargs):
    """
    [root@controller ~]# zpool status
          pool: mymirror
         state: ONLINE
          scan: none requested
        config:

                NAME        STATE     READ WRITE CKSUM
                mymirror    ONLINE       0     0     0
                  mirror-0  ONLINE       0     0     0
                    vdh     ONLINE       0     0     0
                    vdj     ONLINE       0     0     0

        errors: No known data errors

          pool: myzpool
         state: ONLINE
          scan: none requested
        config:

                NAME        STATE     READ WRITE CKSUM
                myzpool     ONLINE       0     0     0
                  vdb       ONLINE       0     0     0
                  vdc       ONLINE       0     0     0

        errors: No known data errors

    result:
        [   {   'devices': [   {   'name': 'vdb', 'status': 'ONLINE'},
                               {   'name': 'vdc', 'status': 'ONLINE'}],
                'name': 'myzpool',
                'status': 'ONLINE'},
            {   'devices': [   {   'name': 'vdh', 'status': 'ONLINE'},
                               {   'name': 'vdj', 'status': 'ONLINE'}],
                'name': 'mymirror',
                'status': 'ONLINE'}]

    Spare inuse case:

    [root@34781ba1-ff84-4127-8219-27093a02585a ~]# zpool status
      pool: zp_ITd__TEST
     state: DEGRADED
    status: One or more devices could not be used because the label is missing or
            invalid.  Sufficient replicas exist for the pool to continue
            functioning in a degraded state.
    action: Replace the device using 'zpool replace'.
       see: http://zfsonlinux.org/msg/ZFS-8000-4J
      scan: resilvered 135K in 0h0m with 0 errors on Mon Jul 15 14:56:19 2019
    config:

            NAME          STATE     READ WRITE CKSUM
            zp_ITd__TEST  DEGRADED     0     0     0
              mirror-0    DEGRADED     0     0     0
                vdc       ONLINE       0     0     0
                spare-1   DEGRADED     0     0     0
                  vdb     UNAVAIL      0   151     0  corrupted data
                  vdi     ONLINE       0     0     0
                vde       ONLINE       0     0     0
            spares
              vdi         INUSE     currently in use
              vdh         AVAIL

    errors: No known data errors

    result:
    [   {   'devices': [   {   'name': 'vdc', 'spare': False, 'status': 'ONLINE'},
                       {   'name': 'vdb', 'spare': False, 'status': 'UNAVAIL'},
                       {   'name': 'vde', 'spare': False, 'status': 'ONLINE'},
                       {   'name': 'vdi', 'spare': True, 'status': 'INUSE'},
                       {   'name': 'vdh', 'spare': True, 'status': 'AVAIL'}],
        'name': 'zp_ITd__TEST',
        'status': 'DEGRADED'}]


    """

    cmd = 'zpool status'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        comment = f"An error (1) occurred while get information about ZFS pools statuses: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        try:
            data = []
            zpool_status_dict = {}
            counter = 0
            is_spare = False
            is_spare_set = 2
            is_cache = False
            is_cache_set = 2
            is_log = False
            is_log_set = 2
            header = True

            for line in cmd_ret['comment'].splitlines():
                if not line or any(map(line.startswith, ['errors:', 'pool:', 'state:', 'status:', 'action:', 'see:',
                                                         'scan:', 'config:'])):
                    header = True
                    continue
                if 'spares' in line:
                    is_spare = True
                    is_cache = False
                    is_log = False
                    continue
                if 'spare-' in line:
                    is_spare_set -= 1
                    continue
                if 'cache' in line:
                    is_spare = False
                    is_cache = True
                    is_log = False
                    continue
                if 'logs' in line:
                    is_log = True
                    is_spare = False
                    is_cache = False
                    continue
                line = line.split(' ')
                # убрать пробелы и пустые строки
                line = [elem.strip() for elem in line if elem.strip()]

                if len(line) >= 3 and line[:2] == ['NAME', 'STATE']:
                    header = False
                    if zpool_status_dict:
                        data.append(zpool_status_dict)
                        zpool_status_dict = {}
                        is_spare = False
                        is_cache = False
                        is_log = False
                    counter = 0
                    continue
                elif header:
                    continue
                else:
                    counter += 1
                    if counter == 1:
                        name, status = line[0:2]

                        zpool_status_dict = dict(name=name, status=status, devices=[])
                    if any([line[0].startswith(label) for label in HLEVEL_DEVICE_LABELS]) and counter == 2:
                        continue
                    if counter >= 2:
                        if len(line) == 2:
                            name, status = line
                        else:
                            name, status = line[:2]

                        if not is_spare_set:
                            is_spare_set = 2
                            continue
                        if is_spare_set == 1:
                            is_spare_set -= 1

                        if not is_cache_set:
                            is_cache_set = 2
                            continue
                        if is_cache_set == 1:
                            is_cache_set -= 1

                        if not is_log_set:
                            is_log_set = 2
                            continue
                        if is_log_set == 1:
                            is_log_set -= 1
                        zpool_status_dict['devices'].append({'name': name, 'status': status,
                                                             'spare': is_spare, 'cache': is_cache, 'log': is_log})
            if zpool_status_dict:
                data.append(zpool_status_dict)
            return data
        except Exception as e:
            import traceback
            comment = f"An error (2) occurred while parse information about ZFS pools statuses: " \
                      f"{str(e)} {traceback.format_exc()}"
            raise CommandExecutionError(comment) from None


@states_response
@file_based_lock.lock(SEMAPHORE)
def zpools_full_info(**kwargs):

    is_dataset_required = kwargs.get('is_dataset_required', True)

    err_message = "An error occurred while get parse information about ZFS pools"

    zpool_list_info = states_result(zpool_list, err_message, **kwargs)

    zpools_status = states_result(zpools_status_info, err_message, **kwargs)

    # берем данные о свободном и занятом месте из zfs list, т.к.zpool list отдаёт тупо сумму объёмов vdev
    zfs_list_info = states_result(zfs_list, err_message, **kwargs)
    # берем только корневые датасеты и из них свободное и общее, пихаем это в словарь пулов данных
    root_sets = {dct['name']: {'free': f'{dct["size"]}b', 'size': f'{int(dct["used"]) + int(dct["size"])}b'}
                 for dct in zfs_list_info if '/' not in dct['name']}
    for zp in zpool_list_info:
        zp.update(root_sets.get(zp['name'], {}))

    if not is_dataset_required:
        zfs_list_info = []

    from .lvm import get_block_devices
    blockdevices = states_result(get_block_devices, err_message)

    try:
        status_map = {}
        for zpool_status in zpools_status:  # pylint: disable=redefined-outer-name
            status_map[zpool_status['name']] = {device.get('name'): device.get('status')
                                                for device in zpool_status['devices']}
        device_size_map_id = {}
        device_size_map_dev = {}
        for blockdevice in blockdevices:
            if blockdevice['tran'] == 'iscsi' and blockdevice.get('children'):
                for child in blockdevice['children']:
                    device_size_map_id[child['name']] = child['size']
            else:
                device_size_map_id[blockdevice['name']] = blockdevice['size']
                device_size_map_dev[blockdevice['short_name']] = blockdevice['size']

        for zfs_pool in zpool_list_info:
            zfs_pool['datasets'] = []
            for dataset in zfs_list_info:
                dataset_name_elements = dataset['name'].split('/')
                if zfs_pool['name'] == dataset_name_elements[0]:
                    zfs_pool['datasets'].append(dataset)
                    if len(dataset_name_elements) == 1:  # корневой датасет ZFS пула
                        mountpoint = dataset['mountpoint']
                        if mountpoint.endswith('/'):
                            zfs_pool['mountpoint'] = mountpoint[:-1]
                        else:
                            zfs_pool['mountpoint'] = mountpoint
            if 'mountpoint' not in zfs_pool.keys():
                mounts = active()
                for k in mounts.keys():
                    if k.endswith(zfs_pool['name']):
                        zfs_pool['mountpoint'] = k
                        break
            for device in zfs_pool['devices']:
                device['health'] = status_map[zfs_pool['name']][device.get('name')]
                # dev names by-id contain '-' as subsystem separator (scsi-, wwn-, lun-, ... ) and /dev/sdXXX don't
                if '-' in device.get('name', ''):
                    dev_dict = device_size_map_id
                    name_prefix = LOCAL_PREFIX
                else:
                    dev_dict = device_size_map_dev
                    name_prefix = DEVICE_PREFIX
                if not dev_dict.get(f'{name_prefix}{device.get("name")}', False):
                    dev_dict = device_size_map_id
                    name_prefix = f"{LUN_PREFIX}/"

                device['size'] = dev_dict.get(f'{name_prefix}{device.get("name")}')
                device['name'] = f'{name_prefix}{device.get("name")}'

        return zpool_list_info
    except Exception as e:
        import traceback
        comment = "{} Error: {}\n{}".format(err_message, e, traceback.format_exc())
        raise CommandExecutionError(comment) from None


@states_response
def zpools_device_map(all_devices=True):
    """
    Result ex.:
    [{'myzpool1': ['/dev/mapper/10000003', '/dev/vdb'],
     {'myzpool2': ['/dev/mapper/10000002', '/dev/vdc']}
    :param: all_devices - возвращать информацию о всех устройствах, в т.ч. недоступных
    """

    ret = zpools_full_info(is_dataset_required=False)

    zps = ret['comment'] if ret['result'] else []
    result = {}

    for zp in zps:
        result[zp.get('name')] = [device.get('name')
                                  for device in zp.get('devices', [])
                                  if (device.get('status') != IGNORED_STATUS if not all_devices else True)]
    return result


@states_response
def zpool_used_devices():
    """
    Used for filtering in  _modules.bazalt_lvm._search_umount_disks.
    :return: aggregated list with full devices names
    """
    from .lvm import _map_blkdev_id
    dict_id = _map_blkdev_id()
    ret = zpools_device_map()

    devices_info = ret['comment'] if ret['result'] else dict()
    zp_used_devices = []

    for devices in devices_info.values():
        zp_used_devices.extend(devices)
    return [dict_id.get(x, x) for x in zp_used_devices]


@states_response
def get_spare_device_candidates(**kwargs):
    """
    Result ex.:
        [{'name': '/dev/vdb', 'zpool': 'myzpool1', 'size': '1G'},
         {'name': '/dev/mapper/10000003', 'zpool': 'myzpool2', 'size': '1.5G'},
         {'name': '/dev/vdc', 'zpool': None, 'size': '1G'}]

    :return: list of dicts
    """

    err_message = "An error occurred while get information about potential spare devices. "

    ret = zpools_full_info(is_dataset_required=False)

    zps = ret['comment'] if ret['result'] else []

    spare_devices_map = {}
    spare_devices = []
    from .lvm import get_block_devices
    blockdevices_ret = get_block_devices()
    if not blockdevices_ret['result']:
        comment = "{}Error: {}".format(err_message, blockdevices_ret['comment'])
        raise CommandExecutionError(comment)

    try:
        blockdevices = blockdevices_ret['comment']

        common_device_map = {}

        # get info about all block devices of the current node
        for blockdevice in blockdevices:
            if blockdevice['tran'] == 'iscsi' and blockdevice.get('children'):
                for child in blockdevice['children']:
                    common_device_map[child['name']] = child['size']
            else:
                common_device_map[blockdevice['name']] = blockdevice['size']

        devices_in_use = []

        for zp in zps:
            devices = deepcopy(zp.get('devices'))
            for device in devices:
                # ignore spare devices in use
                if device.get('spare') and device.get('health') == 'AVAIL':
                    device['zpool'] = zp['name']
                    device.remove('health')
                    spare_devices_map[device.get('name')] = device
                else:
                    # device is used in ZFS pool for data
                    devices_in_use.append(device.get('name'))
                    continue

        for dev_name in common_device_map.keys():
            if not spare_devices_map.get(dev_name):
                spare_devices.append({'name': dev_name,
                                      'size': common_device_map.get(dev_name),
                                      'zpool': None})
            elif dev_name not in devices_in_use:
                spare_devices.append(spare_devices_map.get(dev_name))
        return spare_devices
    except Exception as e:
        comment = "{}Error: {}".format(err_message, e)
        raise CommandExecutionError(comment) from None


@states_response
def zpool_clear(zpool_name):
    clear_cmd = f'zpool clear {zpool_name}'
    cmd_ret = get_output(clear_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while clear errors in ZFS pool: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def create_zfs_pool(zpool_name, raid_type, dev_names, options):

    ret = {'comment': '', 'changes': {}}
    autoreplace = 'on' if options.pop('autoreplace', False) else 'off'
    ashift = options.pop('ashift', 13)
    mountpoint = options.pop('mountpoint', '')

    raid_type = '' if raid_type is None else raid_type
    option_cmd = f'-o autoreplace={autoreplace} -o ashift={ashift} -O atime=off'
    for k, v in options.items():
        option_cmd += f' -o {k}={v}'
    # ashift=13 fixed error 'new device has a different optimal sector size' for most cases
    cmd = f'zpool create -f -m {mountpoint} {option_cmd} {zpool_name} {raid_type} {dev_names}'
    cmd_ret = get_output_critical(cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while creating zfs pool {cmd}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)

    zpools_full_info_ret = zpools_full_info()
    if not zpools_full_info_ret['result']:
        comment = f"An error occurred while get info about zfs pool {zpool_name}: {zpools_full_info_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        # acltype is for gluster healing etc, xattr=sa is for gluster speed (storing xattrs right in inode)
        for opt in ("xattr=sa", "acltype=posix"):
            cmd_ret = get_output(f'zfs set {opt} {zpool_name}')
            if not cmd_ret['result']:
                raise CommandExecutionError(f'Error setting {opt} parm at ZFS {zpool_name}, e={cmd_ret["comment"]}')

        for zpool in zpools_full_info_ret['comment']:
            if zpool['name'] == zpool_name:
                ret['changes'] = zpool
                ret['comment'] = f'The ZFS pool {zpool_name} was successfully created.'
    return ret


@states_response
def extend_zfs_pool(zpool_name, dev_names, spare_flag=False, cache_flag=False, log_flag=False):
    ret = {'comment': '', 'changes': {}}
    if spare_flag:
        pattern = 'zpool add {zpool_name} -f spare {dev_names}'
    elif log_flag:
        pattern = 'zpool add {zpool_name} -f log {dev_names}'
    elif cache_flag:
        pattern = 'zpool add {zpool_name} -f cache {dev_names}'
    else:
        pattern = 'zpool add {zpool_name} -f {dev_names}'
    cmd = pattern.format(zpool_name=zpool_name, dev_names=dev_names)
    cmd_ret = get_output_critical(cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while extending zfs pool {zpool_name}: {cmd_ret['comment']}"
        wipefs_ret = wipefs(dev_names)
        if not wipefs_ret['result']:
            raise CommandExecutionError(wipefs_ret['comment'])
        else:
            ret['comment'] += wipefs_ret['comment']

        raise CommandExecutionError(comment)

    zpools_full_info_ret = zpools_full_info()
    if not zpools_full_info_ret['result']:
        comment = f"An error occurred while get info about zfs pool {zpool_name}: {zpools_full_info_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        for zpool in zpools_full_info_ret['comment']:
            if zpool['name'] == zpool_name:
                ret['changes'] = zpool
                ret['comment'] = f'The ZFS pool {zpool_name} was successfully extended.'
    return ret


@states_response
def destroy_zfs_pool(zpool_name, mountpoint):
    """
    1. Define zfs pool available disk devices
    2. Destroy zfs pool
    3. Clear headers for disk devices (using step 1 data)
    4. Remove old directory of zfs pool
    """

    ret = {'comment': '', 'changes': {}}
    dev_names = []

    zpools_device_map_ret = zpools_device_map(all_devices=False)

    if not zpools_device_map_ret['result']:
        comment = f"An error occurred while get zfs pool device info: {zpools_device_map_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        for label in zpools_device_map_ret['comment'].keys():
            if label == zpool_name:
                dev_names = zpools_device_map_ret['comment'][label]

        destroy_cmd = f'zpool destroy -f {zpool_name}'

        destroy_cmd_ret = get_output_critical(destroy_cmd)
        if not destroy_cmd_ret['result']:
            comment = f"An error occurred while destroying zfs pool {zpool_name}: {destroy_cmd_ret['comment']}. "
            raise CommandExecutionError(comment)
        else:
            ret['comment'] = f'The ZFS pool {zpool_name} was successfully deleted.'

        wipefs_ret = wipefs(dev_names)
        if not wipefs_ret['result']:
            raise CommandExecutionError(wipefs_ret['comment'])
        else:
            ret['comment'] += wipefs_ret['comment']

        if os.path.exists(mountpoint):
            rm_cmd_ret = get_output('rm -rf {}'.format(mountpoint))
            if not rm_cmd_ret['result']:
                ret['comment'] += f"An error occurred while removing old zfs pool " \
                                  f"directory {mountpoint}: {rm_cmd_ret['comment']}. "
    return ret['comment']


@states_response
def attach_device(zpool_name, dev_names):
    """
    :param zpool_name:
    :param dev_names: str, "<mirrored_device> <attaching_device>"
    """
    ret = {'comment': '', 'changes': {}}
    cmd = f'zpool attach {zpool_name} -f {dev_names}'

    cmd_ret = get_output_critical(cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while attaching device in zfs pool {zpool_name}: {cmd_ret['comment']}"
        wipefs_ret = wipefs(dev_names)
        if not wipefs_ret['result']:
            raise CommandExecutionError(comment + wipefs_ret['comment'])
        else:
            ret['comment'] += wipefs_ret['comment']
        raise CommandExecutionError(comment)

    zpools_full_info_ret = zpools_full_info()
    if not zpools_full_info_ret['result']:
        comment = f"An error occurred while get info about zfs pool {zpool_name}: {zpools_full_info_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        for zpool in zpools_full_info_ret['comment']:
            if zpool['name'] == zpool_name:
                ret['changes'] = zpool
                ret['comment'] = f'Devices {dev_names} were successfully attached to the ZFS pool {zpool_name}.'
    return ret


@states_response
def detach_device(zpool_name, dev_name):
    ret = {'comment': '', 'changes': {}}
    cmd = f'zpool detach {zpool_name} {dev_name}'

    cmd_ret = get_output_critical(cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while detaching device {dev_name} from zfs pool " \
                  f"{zpool_name}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        ret['comment'] = f'Device {dev_name} was successfully detached from the ZFS pool {zpool_name}. '
        wipefs_ret = wipefs([dev_name])
        if not wipefs_ret['result']:
            raise CommandExecutionError(wipefs_ret['comment'])
        else:
            ret['comment'] += wipefs_ret['comment']
    return ret['comment']


@states_response
def wipefs(device_list):
    from veilrpc.states.lvm import data_sync, check_if_partitions, partprobe
    if isinstance(device_list, list):
        device_names = ' '.join(device_list)
    else:
        device_names = device_list
    cmd_ret = get_output('wipefs {device_names} -a'.format(device_names=device_names))
    if not cmd_ret['result']:
        comment = f"An error occurred while clear headers of devices {device_names}: {cmd_ret['comment']}. "
        raise CommandExecutionError(comment)
    else:
        comment = f'{cmd_ret["comment"]}\n'
    sync_ret = data_sync(device_list)
    comment += f'{sync_ret["comment"]}\n'
    if not sync_ret['result']:
        raise CommandExecutionError(comment)
    partprobe_ret = partprobe(device_list)
    comment += f'{partprobe_ret["comment"]}\n'
    if not partprobe_ret['result']:
        raise CommandExecutionError(comment)
    check_ret = check_if_partitions(device_list)
    if not check_ret['result']:
        raise CommandExecutionError(comment + check_ret['comment'])

    return comment


@states_response
def zpool_replace_device(zpool_name, replaced_device, target_device):
    ret = {'comment': '', 'changes': {}}
    masked_target_device = '' if target_device == replaced_device else target_device

    replace_cmd = 'zpool replace {zpool_name} ' \
                  '{replaced_device} {target_device}'.format(zpool_name=zpool_name,
                                                             replaced_device=replaced_device,
                                                             target_device=masked_target_device)
    replace_cmd_ret = get_output_critical(replace_cmd)
    if not replace_cmd_ret['result']:
        comment = f"An error occurred while replacing device {replaced_device} on {target_device} " \
                  f"in ZFS pool {zpool_name}: {replace_cmd_ret['comment']}. "
        if masked_target_device:
            wipefs_ret = wipefs([masked_target_device])
            if not wipefs_ret['result']:
                raise CommandExecutionError(wipefs_ret['comment'])
            else:
                ret['comment'] += wipefs_ret['comment']
        raise CommandExecutionError(comment)
    else:
        # drop replaced device headers in not self-substitution case
        if masked_target_device:
            wipefs_ret = wipefs([replaced_device])

            if not wipefs_ret.get('result'):
                # error in clearing headers don't stop common execution
                ret['comment'] += wipefs_ret.get('comment')

        full_info_ret = zpools_full_info()
        if not full_info_ret['result']:
            comment = ret['comment']
            comment += f" An error occurred while get info about zfs pool {zpool_name}: {full_info_ret['comment']}"
            raise CommandExecutionError(comment)
        else:
            for zpool in full_info_ret['comment']:
                if zpool['name'] == zpool_name:
                    ret['changes'] = zpool
                    ret['comment'] += 'Device {replaced_device} were successfully replaced ' \
                                      'on {target_device} in the ZFS pool {zpool_name}. '.format(
                        replaced_device=replaced_device,
                        target_device=target_device,
                        zpool_name=zpool_name)
    return ret


@states_response
def zpool_switch_device(zpool_name, device_name, device_state):

    ret = {'comment': '', 'changes': {}}

    command_map = {'ONLINE': 'offline', 'OFFLINE': 'online'}
    command = command_map.get(device_state, 'offline')

    switch_cmd = f'zpool {command} {zpool_name} {device_name}'
    switch_cmd_ret = get_output_critical(switch_cmd)
    if not switch_cmd_ret['result']:
        comment = f"An error occurred while run command {command} for device {device_name} " \
                  f"in ZFS pool {zpool_name}: {switch_cmd_ret['comment']}. "
        raise CommandExecutionError(comment)
    else:
        zpools_full_info_ret = zpools_full_info()
        if not zpools_full_info_ret['result']:
            comment = f"An error occurred while get info about zfs pool {zpool_name}: " \
                      f"{zpools_full_info_ret['comment']}"
            raise CommandExecutionError(comment)
        else:
            for zpool in zpools_full_info_ret['comment']:
                if zpool['name'] == zpool_name:
                    ret['changes'] = zpool
                    ret['comment'] += f'Device {device_name} were successfully {command}ed ' \
                                      f'in the ZFS pool {zpool_name}.'
    return ret


@states_response
def zpool_remove_spare(zpool_name, device_name):

    ret = {'comment': '', 'changes': {}}

    remove_cmd = f'zpool remove {zpool_name} {device_name}'

    remove_cmd_ret = get_output_critical(remove_cmd)
    if not remove_cmd_ret['result']:
        comment = f"An error occurred while remove device {device_name} in " \
                  f"ZFS pool {zpool_name}: {remove_cmd_ret['comment']}. "
        raise CommandExecutionError(comment)
    else:
        wipefs_ret = wipefs([device_name])
        if not wipefs_ret['result']:
            raise CommandExecutionError(wipefs_ret['comment'])
        else:
            ret['comment'] += wipefs_ret['comment']
        ret['comment'] += f'Spare device {device_name} was successfully removed ' \
                          f'from the ZFS pool {zpool_name}.'
    return ret


@states_response
def zpool_scrub(zpool_name):
    cmd = f"zpool scrub {zpool_name}"
    cmd_ret = get_output(cmd, timeout=3600)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f"zpool {zpool_name} successfully scrubbed."


def get_re_zfs_vdevs_from_cache():
    """
    получаем в виде скомпилированных шаблонов regexp для сравнения список vdevs в формате /dev/sdxxx
    Список берём из кэша, т.к. zpool status может повесить процесс, если zfs подвис, а кэш не виснет и актуален
    способ с regexp поможет в случае /dev/sdx9, которые есть, но нигде не указано, что они принадлежат zpool
    """
    zdb_out = get_output('zdb', suppress_errors=True)
    zdb_dump = zdb_out['comment'] if zdb_out['result'] else ''
    vdevs = set()
    path, is_whole_device = '', False
    for line in zdb_dump.splitlines():
        line_s = line.strip()
        if line_s.startswith('children['):  # начало новой секции vdev
            if path:  # если что-то было в предыдущей секции
                # если whole, шаблон должен подходить к род.устройству и всем его разделам, иначе как есть
                vdevs.add(re.compile(re.sub(r'\d*$', r'\\d*$', path, 1) if is_whole_device else path))

            path, is_whole_device = '', False
        elif line_s.startswith('path: '):
            path = os.path.realpath(line_s.split(': ')[1].strip("'"))  # /dev/disk/by-id -> /dev/sdxxx
        elif line_s.startswith('whole_disk: '):
            is_whole_device = line_s.split(': ')[1] == '1'
    if path:
        vdevs.add(re.compile(re.sub(r'\d*$', r'\\d*$', path, 1) if is_whole_device else path))

    return vdevs


@states_response
def arc_summary():
    summary = states_result(get_output, cmd="arc_summary -r -d")
    res = {}
    name, descr = "", ""
    for line in summary.splitlines()[6:]:
        line = line.strip()
        if not line:
            continue
        elif line.startswith("#"):
            # Ex: # Percentage over dbuf_cache_max_bytes when dbufs must be evicted directly.
            descr = line[2:]
        elif line.endswith(":"):
            # Ex: ARCSTATS:
            name = line[:-1]
            if name not in res.keys():
                res[name] = []
        else:
            # Ex: dbuf_cache_hiwater_pct    10
            lines = line.split()
            if len(lines) == 2:
                k, v = lines
                if name:
                    if descr:
                        res[name].append(dict(key=k, value=v, description=descr))
                        descr = ""
                    else:
                        res[name].append(dict(key=k, value=v))
    return res


@states_response
def zfs_gl_dataset(zpools, volume='', rec_size_kb=None):
    for zpool in zpools:
        mountpoint = os.path.join(ZPOOL_STORAGES_PATH, zpool)
        if not get_result(['zfs', 'destroy', '-fR', zpool], shell=False, timeout=50):  # removing all datasets/vols on zpool
            raise CommandExecutionError(f'Error removing all datasets and volumes on zpool {zpool}')
        if rec_size_kb is None or volume == '':
            if not get_result(['zfs', 'set', f'mountpoint={mountpoint}', zpool], shell=False, timeout=50):
                raise CommandExecutionError(f'Error reverting mountpoint to default on zpool {zpool}')
        else:
            if not get_result(['zfs', 'set', 'mountpoint=none', zpool], shell=False, timeout=50):  # zpool is umounted here
                raise CommandExecutionError(f'Error clearing mountpoint and umounting zpool {zpool}')
            # flag informs storage_status not to fix mountpoint of this zpool within 1 minute timeout
            # (giving gluster a chance to complete operations like vol creation, brick add, brick replace)
            with open(flag_path := os.path.join(RUN_PATH, f'FLG_ZP_{zpool}'), 'w') as _:
                pass
            if not get_result(['zfs', 'create', '-o', f'mountpoint={mountpoint}'] +
                              (['-o', f'recordsize={rec_size_kb}K'] if rec_size_kb else []) +
                              [ds_name := f'{zpool}/gluster_{volume}'], shell=False, timeout=50):
                os.remove(flag_path)
                raise CommandExecutionError(f'Error creating dataset {ds_name}')
