"""
(c) 2023 LLC «DACOM M»
"""

import hashlib
import grp
import errno
import os
import pwd
import six
import stat

from veilrpc.utils.bazalt import CommandExecutionError, WORKER_USER, delete_file, get_output


def mkdir(dir_path,
          user=None,
          group=None,
          mode=None):
    """
    Ensure that a directory is available.
    """
    dir_path = os.path.expanduser(dir_path)

    directory = os.path.normpath(dir_path)

    if not os.path.isdir(directory):
        # If a caller such as managed() is invoked  with makedirs=True, make
        # sure that any created dirs are created with the same user and group
        # to follow the principal of least surprise method.
        makedirs_perms(directory, user, group, mode)

    return True


def makedirs_perms(name,
                   user=None,
                   group=None,
                   mode='0755'):
    """
    Taken and modified from os.makedirs to set user, group and mode for each
    directory created.
    """
    name = os.path.expanduser(name)

    path = os.path
    head, tail = path.split(name)
    if not tail:
        head, tail = path.split(head)
    if head and tail and not path.exists(head):
        try:
            makedirs_perms(head, user, group, mode)
        except OSError as exc:
            # be happy if someone already created the path
            if exc.errno != errno.EEXIST:
                raise
        if tail == os.curdir:  # xxx/newdir/. exists if xxx/newdir exists
            return
    os.mkdir(name)
    check_perms(name,
                None,
                user,
                group,
                int('{0}'.format(mode)) if mode else None)


def check_perms(name, ret, user, group, mode, follow_symlinks=False):
    """
    Check the permissions on files and chown if needed
    """

    name = os.path.expanduser(name)

    if not ret:
        ret = {'name': name,
               'changes': {},
               'comment': [],
               'result': True}
        orig_comment = ''
    else:
        orig_comment = ret['comment']
        ret['comment'] = []

    # Check permissions
    perms = {}
    cur = stats(name, follow_symlinks=follow_symlinks)
    if not cur:
        # NOTE: The file.directory state checks the content of the error
        # message in this exception. Any changes made to the message for this
        # exception will reflect the file.directory state as well, and will
        # likely require changes there.
        raise CommandExecutionError('{0} does not exist'.format(name))
    perms['luser'] = cur['user']
    perms['lgroup'] = cur['group']
    perms['lmode'] = normalize_mode(cur['mode'])

    # Mode changes if needed
    if mode is not None:
        # File is a symlink, ignore the mode setting
        # if follow_symlinks is False
        if os.path.islink(name) and not follow_symlinks:
            pass
        else:
            mode = normalize_mode(mode)
            if mode != perms['lmode']:
                set_mode(name, mode)
                if mode != normalize_mode(get_mode(name)):
                    ret['result'] = False
                    ret['comment'].append(
                        'Failed to change mode to {0}'.format(mode)
                    )
                else:
                    ret['changes']['mode'] = mode

    # user/group changes if needed, then check if it worked
    if user and isinstance(user, int):
        user = uid_to_user(user)
    if group and isinstance(group, int):
        group = gid_to_group(group)
    if 'cuser' in perms or 'cgroup' in perms:
        if os.path.islink(name) and not follow_symlinks:
            chown_func = lchown
        else:
            chown_func = chown
        if user is None:
            user = perms['luser']
        if group is None:
            group = perms['lgroup']
        try:
            chown_func(name, user, group)
        except OSError:
            ret['result'] = False

    if user:
        if isinstance(user, int):
            user = uid_to_user(user)
        if 'cuser' in perms and user != '':
            ret['changes']['user'] = user
    if group:
        if isinstance(group, int):
            group = gid_to_group(group)
        if 'cgroup' in perms and user != '':
            ret['changes']['group'] = group

    if isinstance(orig_comment, six.string_types):
        if orig_comment:
            ret['comment'].insert(0, orig_comment)
        ret['comment'] = '; '.join(ret['comment'])
    return ret, perms


def gid_to_group(gid):
    """
    Convert the group id to the group name on this system
    """
    try:
        gid = int(gid)
    except ValueError:
        # This is not an integer, maybe it's already the group name?
        gid = group_to_gid(gid)

    if gid == '':
        # Don't even bother to feed it to grp
        return ''

    try:
        return grp.getgrgid(gid).gr_name
    except (KeyError, NameError):
        # If group is not present, fall back to the gid.
        return gid


def group_to_gid(group):
    """
    Convert the group to the gid on this system

    group
        group to convert to its gid
    """
    if group is None:
        return ''
    try:
        if isinstance(group, int):
            return group
        return grp.getgrnam(group).gr_gid
    except KeyError:
        return ''


def uid_to_user(uid):
    """
    Convert a uid to a user name
    """
    try:
        return pwd.getpwuid(uid).pw_name
    except (KeyError, NameError):
        # If user is not present, fall back to the uid.
        return uid


def user_to_uid(user):
    """
    Convert user name to a uid

    user
        user name to convert to its uid
    """
    if user is None:
        user = _get_user()
    try:
        if isinstance(user, int):
            return user
        return pwd.getpwnam(user).pw_uid
    except KeyError:
        return ''


def get_user(path, follow_symlinks=True):
    """
    Return the user that owns a given file

    path
        file or directory of which to get the user

    follow_symlinks
        indicated if symlinks should be followed
    """
    return stats(os.path.expanduser(path), follow_symlinks=follow_symlinks).get('user', False)


def stats(path, hash_type=None, follow_symlinks=True):
    """
    Return a dict containing the stats for a given file
    """
    path = os.path.expanduser(path)

    ret = {}
    if not os.path.exists(path):
        try:
            # Broken symlinks will return False for os.path.exists(), but still
            # have a uid and gid
            pstat = os.lstat(path)
        except OSError:
            # Not a broken symlink, just a nonexistent path
            return ret
    else:
        if follow_symlinks:
            pstat = os.stat(path)
        else:
            pstat = os.lstat(path)
    ret['inode'] = pstat.st_ino
    ret['uid'] = pstat.st_uid
    ret['gid'] = pstat.st_gid
    ret['group'] = gid_to_group(pstat.st_gid)
    ret['user'] = uid_to_user(pstat.st_uid)
    ret['atime'] = pstat.st_atime
    ret['mtime'] = pstat.st_mtime
    ret['ctime'] = pstat.st_ctime
    ret['size'] = pstat.st_size
    ret['mode'] = str(oct(stat.S_IMODE(pstat.st_mode)))
    if hash_type:
        ret['sum'] = get_hash(path, hash_type)
    ret['type'] = 'file'
    if stat.S_ISDIR(pstat.st_mode):
        ret['type'] = 'dir'
    if stat.S_ISCHR(pstat.st_mode):
        ret['type'] = 'char'
    if stat.S_ISBLK(pstat.st_mode):
        ret['type'] = 'block'
    if stat.S_ISREG(pstat.st_mode):
        ret['type'] = 'file'
    if stat.S_ISLNK(pstat.st_mode):
        ret['type'] = 'link'
    if stat.S_ISFIFO(pstat.st_mode):
        ret['type'] = 'pipe'
    if stat.S_ISSOCK(pstat.st_mode):
        ret['type'] = 'socket'
    ret['target'] = os.path.realpath(path)
    return ret


def get_group(path, follow_symlinks=True):
    """
    Return the group that owns a given file

    path
        file or directory of which to get the group

    follow_symlinks
        indicated if symlinks should be followed

    """
    return stats(os.path.expanduser(path), follow_symlinks=follow_symlinks).get('group', False)


def get_hash(path, form='sha256', chunk_size=65536):
    """
    Get the hash sum of a file

    This is better than ``get_sum`` for the following reasons:
        - It does not read the entire file into memory.
        - It does not return a string on error. The returned value of
            ``get_sum`` cannot really be trusted since it is vulnerable to
            collisions: ``get_sum(..., 'xyz') == 'Hash xyz not supported'``

    path
        path to the file or directory

    form
        desired sum format

    chunk_size
        amount to sum at once
    """
    return _get_hash(os.path.expanduser(path), form, chunk_size)


def _get_hash(path, form='sha256', chunk_size=65536):
    """
    Get the hash sum of a file

    This is better than ``get_sum`` for the following reasons:
        - It does not read the entire file into memory.
        - It does not return a string on error. The returned value of
            ``get_sum`` cannot really be trusted since it is vulnerable to
            collisions: ``get_sum(..., 'xyz') == 'Hash xyz not supported'``
    """
    hash_type = hasattr(hashlib, form) and getattr(hashlib, form) or None
    if hash_type is None:
        raise ValueError('Invalid hash type: {0}'.format(form))

    with open(path, 'rb') as ifile:
        hash_obj = hash_type()
        # read the file in in chunks, not the entire file
        for chunk in iter(lambda: ifile.read(chunk_size), b''):
            hash_obj.update(chunk)
        return hash_obj.hexdigest()


def normalize_mode(mode):
    """
    Return a mode value, normalized to a string and containing a leading zero
    if it does not have one.
    """
    if mode is None:
        return None
    if not isinstance(mode, six.string_types):
        mode = str(mode)
    # Strip any quotes any initial zeroes, then though zero-pad it up to 4.
    # This ensures that somethign like '00644' is normalized to '0644'
    return mode.strip('"').strip('\'').lstrip('0').zfill(4)


def get_mode(path, follow_symlinks=True):
    """
    Return the mode of a file

    path
        file or directory of which to get the mode

    follow_symlinks
        indicated if symlinks should be followed
    """
    return stats(os.path.expanduser(path), follow_symlinks=follow_symlinks).get('mode', '')


def set_mode(path, mode):
    """
    Set the mode of a file

    path
        file or directory of which to set the mode

    mode
        mode to set the path to
    """
    path = os.path.expanduser(path)

    mode = str(mode).lstrip('0Oo')
    if not mode:
        mode = '0'
    if not os.path.exists(path):
        raise CommandExecutionError('{0}: File not found'.format(path))
    try:
        os.chmod(path, int(mode, 8))
    except Exception:
        return 'Invalid Mode ' + mode
    return get_mode(path)


def lchown(path, user, group):
    """
    Chown a file, pass the file the desired user and group without following
    symlinks.

    path
        path to the file or directory

    user
        user owner

    group
        group owner
    """
    path = os.path.expanduser(path)

    uid = user_to_uid(user)
    gid = group_to_gid(group)
    if uid == '' and not user:
        uid = -1
    if gid == '' and not group:
        gid = -1

    return os.lchown(path, uid, gid)


def chown(path, user, group):
    """
    Chown a file, pass the file the desired user and group

    path
        path to the file or directory

    user
        user owner

    group
        group owner
    """
    path = os.path.expanduser(path)

    uid = user_to_uid(user)
    gid = group_to_gid(group)
    err = ''
    if uid == '':
        if user:
            err += 'User does not exist\n'
        else:
            uid = -1
    if gid == '':
        if group:
            err += 'Group does not exist\n'
        else:
            gid = -1
    if not os.path.exists(path):
        try:
            # Broken symlinks will return false, but still need to be chowned
            return os.lchown(path, uid, gid)
        except OSError:
            pass
        err += 'File not found'
    if err:
        return err
    return os.chown(path, uid, gid)


def _get_user():
    """
    Get the current user
    """

    return pwd.getpwuid(os.geteuid()).pw_name


def get_uid_gid(owner, group=None):
    if group is not None:
        return pwd.getpwnam(owner).pw_uid, grp.getgrnam(group).gr_gid
    return pwd.getpwnam(owner).pw_uid, pwd.getpwnam(owner).pw_gid


def chown_directory(path, owner, group=None):
    uid, gid = get_uid_gid(owner, group)
    os.chown(path, uid, gid)


def chown_directory_uid_gid(path, uid, gid):
    os.chown(path, uid, gid)


def chown_files_in_dir(path, owner=WORKER_USER):
    uid, gid = get_uid_gid(owner)
    for root, dirs, files in os.walk(path):
        for momo in dirs:
            chown_directory_uid_gid(os.path.join(root, momo), uid, gid)
        for momo in files:
            chown_directory_uid_gid(os.path.join(root, momo), uid, gid)


def modify_file(path):
    try:
        cmd = f"su {WORKER_USER} -c \'touch {path}\'"
        cmd_ret = get_output(cmd)
        if not cmd_ret['result']:
            raise Exception(cmd_ret['comment'])
        # open(path, 'w').close()
        # chown_directory(path, WORKER_USER)
    except Exception:
        try:
            delete_file(path)
            open(path, 'w').close()
            chown_directory(path, WORKER_USER)
        except Exception:
            return False
    return True
