"""
(c) 2023 LLC «DACOM M»
"""

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding
from datetime import datetime
import json
from glob import glob
import os
from redis import RedisError

from veilrpc.utils.bazalt import CONTROLLER_OTHER_PATH, CONTROLLER_PATH

SERIAL_KEY = f'{CONTROLLER_PATH}/serial.key'  # built-in serial key path
SERIAL_KEY_UUID = 'b3f2eada-1287-4f47-ad42-11ce3430839e'  # permanent in SERIAL_KEY
UNLICENSED_UUID = '00000001-0000-0000-0000-000000000000'
SERIAL_PEMS = [f"{CONTROLLER_OTHER_PATH}/controller.pem", f"{CONTROLLER_OTHER_PATH}/controller_2.pem"]
REDIS_LICENSE_KEY = 'CONTROLLER_LICENSE_KEY'
LICENSE_EXPTIME = 86400
UnlicenseData = {
    "verbose_name": "Invalid license",
    "nodes_limit": 0,
    "expired_time": 0,
    'support_expired_time': 0,
    "uuid": UNLICENSED_UUID,
    "company": None,
    "email": None,
    "built_in": False,
    "filename": 'N/A',
    "expired": True,
    "support_expired": True,
}
DEFAULT_LICENSE = {
    "verbose_name": "Default",
    "nodes_limit": 1,
    "expired_time": 0,
    "support_expired_time": 0,
    "uuid": SERIAL_KEY_UUID,
    "company": None,
    "email": None,
    "built_in": True,
    "filename": 'service.key',
    "expired": False,
    "support_expired": False,
}


def get_license_params(key_path: str = SERIAL_KEY, **kwargs) -> dict:
    """ Проверяется ключ из файла. """
    redis = kwargs.get('redis', None)
    try:
        if redis and redis.exists(f"{REDIS_LICENSE_KEY}_{key_path}"):
            data = json.loads(redis.get(f"{REDIS_LICENSE_KEY}_{key_path}").decode())
            if data:
                return data
    except (RedisError, json.JSONDecodeError, UnicodeError, AttributeError):
        pass
    unlicense_data = dict(UnlicenseData)
    filename = os.path.basename(key_path)
    unlicense_data["filename"] = filename
    try:
        if os.path.isfile(key_path):
            for p_key_path in SERIAL_PEMS:
                if os.path.isfile(p_key_path):
                    with open(p_key_path, "rb") as key_file:
                        private_key = serialization.load_pem_private_key(
                            key_file.read(),
                            password=None,
                            backend=default_backend())
                    if private_key:
                        with open(key_path, "rb") as f:
                            data = f.read()
                        try:
                            decrypted_data_bytes = private_key.decrypt(
                                data, padding.OAEP(
                                    mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                    algorithm=hashes.SHA256(),
                                    label=None))
                            decrypted_data = json.loads(decrypted_data_bytes)
                            decrypted_data["filename"] = filename
                            decrypted_data['built_in'] = is_built_in(decrypted_data)
                            decrypted_data['expired'] = is_expired(decrypted_data)
                            decrypted_data['support_expired'] = is_expired(decrypted_data, support=True)
                            if redis:
                                try:
                                    redis.setex(name=f"{REDIS_LICENSE_KEY}_{key_path}",
                                                value=json.dumps(decrypted_data),
                                                time=LICENSE_EXPTIME)
                                except RedisError:
                                    pass
                            return decrypted_data
                        except ValueError:
                            pass
        return unlicense_data
    except (json.JSONDecodeError, UnicodeError, OSError):
        return unlicense_data


def get_licenses_params(**kwargs) -> list:
    licenses = []
    _keys = glob(f"{CONTROLLER_PATH}/*.key")
    license_added = len(_keys) > 1
    for path in _keys:
        lic_params = get_license_params(path, **kwargs)
        if license_added and lic_params['built_in']:
            continue
        if lic_params.get('non_expandable'):  # essential license
            return [lic_params]
        else:
            licenses.append(lic_params)
    if not licenses:
        licenses.append(DEFAULT_LICENSE)
    return licenses


def get_merged_licenses_params(**kwargs) -> dict:
    redis = kwargs.get('redis', None)
    try:
        if redis and redis.exists(REDIS_LICENSE_KEY):
            data = json.loads(redis.get(REDIS_LICENSE_KEY).decode())
            if data:
                return data
    except (RedisError, json.JSONDecodeError, UnicodeError, AttributeError):
        pass
    merged_license = {
        'expired': True,
        'support_expired': True
    }
    licenses = get_licenses_params(**kwargs)
    built_in_license_only = False
    _expired = []
    _support_expired = []
    for lic in licenses:
        if lic['built_in']:
            built_in_license_only = True
        license_uuid = lic.get('uuid', '')
        merged_license_uuid = merged_license.get('uuid', '')
        if merged_license_uuid and license_uuid and license_uuid in merged_license_uuid:
            continue
        for k, v in lic.items():
            if k not in merged_license.keys():
                if k in ['expired_time', 'support_expired_time'] and v in ['0', 0]:
                    v = None
                if k == 'nodes_limit' and lic['expired']:
                    merged_license[k] = 0
                else:
                    merged_license[k] = v
            else:
                if k == 'nodes_limit':
                    if isinstance(merged_license[k], int) and isinstance(v, int) and not lic['expired']:
                        merged_license[k] += v
                elif k == 'expired':
                    _expired.append(v)
                elif k == 'support_expired':
                    _support_expired.append(v)
                elif k in ['expired_time', 'support_expired_time']:
                    if v in ['0', 0]:
                        v = None
                    if v and remaining_time(v) > remaining_time(merged_license[k]):
                        merged_license[k] = v
                else:
                    if not v or not merged_license[k] or not isinstance(merged_license[k], str):
                        merged_license[k] = v
                    elif v not in merged_license[k]:
                        merged_license[k] += f', {v}'
    if not built_in_license_only:
        merged_license['nodes_limit'] += 1
    merged_license['expired'] = all(_expired)
    merged_license['support_expired'] = all(_support_expired)
    if redis:
        try:
            redis.setex(name=REDIS_LICENSE_KEY,
                        value=json.dumps(merged_license),
                        time=LICENSE_EXPTIME)
        except (RedisError, json.JSONDecodeError):
            pass
    return merged_license


def get_nodes_count(**kwargs) -> int:
    """
    :return: количество узлов, доступных для создания
    """
    try:
        return int(get_merged_licenses_params(**kwargs).get('nodes_limit', 0))
    except ValueError:
        return 0


def get_expired_time(**kwargs):
    """
    :return: время, до которого будет работать функционал veil
    """
    return get_merged_licenses_params(**kwargs).get('expired_time', None)


def remaining_time(expired_time):
    """
    Возвращает кол-во оставшихся дней до expired_time
    """
    if expired_time:
        try:
            exp_time = datetime.strptime(expired_time, "%Y-%m-%d").date()
            now = datetime.strptime(datetime.utcnow().strftime("%Y-%m-%d"), "%Y-%m-%d").date()
            delta = exp_time - now
            days = delta.days
            return days if days >= 0 else 0
        except ValueError:
            return 0
    return 0


def is_expired(params, support=False):
    if params['built_in']:
        return False
    if _exp_date := params.get('expired_time' if not support else 'support_expired_time'):
        if _exp_date == '0':
            return False
        return remaining_time(_exp_date) == 0
    return False


def get_activated(**kwargs) -> bool:
    """
    Проверяет nodes_limit and expired
    """
    params = get_merged_licenses_params(**kwargs)
    return True if params['nodes_limit'] and not is_expired(params) else False


def get_supported(**kwargs) -> bool:
    """
    Проверяет nodes_limit and support_expired
    """
    params = get_merged_licenses_params(**kwargs)
    return True if params['nodes_limit'] and not is_expired(params, support=True) else False


def remove_license(filename, **kwargs) -> str:
    """
    Удаляет файл лицензии, если есть
    """
    key_path = f'{CONTROLLER_PATH}/{filename}'
    if filename == 'all.key':  # For tests only
        for _lic in get_licenses_params(**kwargs):
            if not _lic['built_in']:
                os.remove(f'{CONTROLLER_PATH}/{_lic["filename"]}')
    elif os.path.exists(key_path):
        _lic = get_license_params(key_path=key_path, **kwargs)
        if not _lic['built_in']:
            os.remove(key_path)
    redis = kwargs.get('redis', None)
    if redis:
        redis.delete(f"{REDIS_LICENSE_KEY}_{key_path}")
        redis.delete(REDIS_LICENSE_KEY)
    return f'Successfully remove {filename}'


def is_built_in(key_data) -> bool:
    """ Если лицензия встроенная, вернет True. """
    return key_data.get('uuid', '') == SERIAL_KEY_UUID
