"""
(c) 2023 LLC «DACOM M»
"""

from veilrpc.utils.bazalt import states_response, CommandExecutionError

# поддерживаются libvirt
MEMORY_UNITS = ['b', 'bytes', 'KB', 'MB', 'GB', 'TB', 'k', 'K', 'KiB', 'M', 'MiB', 'G', 'GiB', 'T', 'TiB']


def memory_units_convertor(value: float, source: str, target: str) -> int:
    """
    Переводит из исходного юнита в целевой. Например из KiB в MiB или наоборот.

    Из документации libvirt:
    The units for this value are determined by the optional attribute unit,
    which defaults to "KiB" (kibibytes, 210 or blocks of 1024 bytes).
    Valid units are "b" or "bytes" for bytes,
    "KB" for kilobytes (103 or 1,000 bytes),
    "k" or "KiB" for kibibytes (1024 bytes),
    "MB" for megabytes (106 or 1,000,000 bytes),
    "M" or "MiB" for mebibytes (220 or 1,048,576 bytes),
    "GB" for gigabytes (109 or 1,000,000,000 bytes),
    "G" or "GiB" for gibibytes (230 or 1,073,741,824 bytes),
    "TB" for terabytes (1012 or 1,000,000,000,000 bytes),
    "T" or "TiB" for tebibytes (240 or 1,099,511,627,776 bytes).
    """

    def get_value_by_subkey(label, dict_):
        for labels in dict_.keys():
            if label in labels:
                return dict_[labels]

    for unit in [source, target]:
        if unit not in MEMORY_UNITS:
            raise CommandExecutionError('Unknown memory unit: {}. Expected one of {}'.format(unit, MEMORY_UNITS))

    bases = {
        ('KB', 'MB', 'GB', 'TB'): 10 ** 3,
        ('k', 'K', 'KiB', 'M', 'MiB', 'G', 'GiB', 'T', 'TiB'): 2 ** 10,
        ('b', 'bytes'): 1
    }

    powers = {
        ('b', 'bytes'): 0,
        ('k', 'KiB', 'KB', 'K'): 1,
        ('M', 'MiB', 'MB'): 2,
        ('G', 'GiB', 'GB'): 3,
        ('T', 'TiB', 'TB'): 4
    }

    source_base = get_value_by_subkey(source, bases)
    target_base = get_value_by_subkey(target, bases)
    source_power = get_value_by_subkey(source, powers)
    target_power = get_value_by_subkey(target, powers)

    return int(value * (source_base ** source_power / target_base ** target_power))  # все что меньше 1 есть 0


def get_megabytes(s: str) -> int:
    """
    Принимает строку типа '<число><elem from MEMORY_UNITS>' и возвращает MB
    """
    if not isinstance(s, str):
        s = str(s)
    s = s.replace(',', '.')
    value = source_unit = None
    for i in range(1, len(s)):
        try:
            value = float(s[:-i])
            source_unit = s[-i:]
            break
        except ValueError:
            continue
    return memory_units_convertor(value, source_unit, 'MB')


def get_mebibytes(s: str) -> int:
    """
    Принимает строку типа '<число><elem from MEMORY_UNITS>' и возвращает M
    """
    if not isinstance(s, str):
        s = str(s)
    s = s.replace(',', '.')
    value = source_unit = None
    for i in range(1, len(s)):
        try:
            value = float(s[:-i])
            source_unit = s[-i:]
            break
        except ValueError:
            continue
    return memory_units_convertor(value, source_unit, 'M')


@states_response
def get_mebibytes_2(value: float, source_unit: str) -> int:
    """
    возвращает M
    """
    if not isinstance(value, float):
        value = float(value)
    return memory_units_convertor(value, source_unit, 'M')


def get_kibibytes(s: str) -> int:
    """
    Принимает строку типа '<число><elem from MEMORY_UNITS>' и возвращает k
    """
    if not isinstance(s, str):
        s = str(s)
    s = s.replace(',', '.')
    value = source_unit = None
    for i in range(1, len(s)):
        try:
            value = float(s[:-i])
            source_unit = s[-i:]
            break
        except ValueError:
            continue
    return memory_units_convertor(value, source_unit, 'k')


@states_response
def convert(s: str, target: str) -> int:
    if not isinstance(s, str):
        s = str(s)
    s = s.replace(',', '.')
    value = source_unit = None
    for i in range(1, len(s)):
        try:
            value = float(s[:-i])
            source_unit = s[-i:]
            break
        except ValueError:
            continue
    return memory_units_convertor(value, source_unit, target)


def convert_size_gb(size):
    try:
        return round(float(size) / 1073741824, 3)
    except Exception:
        return 0
