"""
(c) 2023 LLC «DACOM M»
"""

from contextlib import contextmanager
import libvirt
from defusedxml import minidom

BASE_URI = "qemu:///system"

STATES = dict([
    (libvirt.VIR_DOMAIN_NOSTATE, 'no_state'),
    (libvirt.VIR_DOMAIN_RUNNING, 'running'),
    (libvirt.VIR_DOMAIN_BLOCKED, 'blocked'),
    (libvirt.VIR_DOMAIN_PAUSED, 'paused'),
    (libvirt.VIR_DOMAIN_SHUTDOWN, 'shutdown'),
    (libvirt.VIR_DOMAIN_SHUTOFF, 'shutoff'),
    (libvirt.VIR_DOMAIN_CRASHED, 'crashed'),
    (libvirt.VIR_DOMAIN_PMSUSPENDED, 'paused')])
BASE_QEMU_CHANNEL = 'org.qemu.guest_agent.0'
VEIL_QEMU_CHANNEL = 'org.qemu.guest_agent.1'
VIR_DOMAIN_EVENT_MAPPING = {
    0: "defined",
    1: "undefined",
    2: "started",
    3: "suspended",
    4: "resumed",
    5: "stopped",
    6: "shutdown",
    7: "pmsuspended",
}
EVENT_TO_STATE_MAPPING = {
    0: 0,
    1: 0,
    2: 1,
    3: 3,
    4: 1,
    5: 0,
    6: 0,
    7: 3,
}


@contextmanager
def connect_to_uri_read(name=BASE_URI):
    conn = libvirt.openReadOnly(name)
    yield conn
    conn.close()


@contextmanager
def connect_to_uri(name=BASE_URI):
    conn = libvirt.open(name)
    yield conn
    conn.close()


def skip_errors(ctxt, err):
    pass


def register_error_handler():
    libvirt.registerErrorHandler(skip_errors, 'context')


def _get_domain(domain_name, name=BASE_URI):
    with connect_to_uri(name) as conn:
        return conn.lookupByName(domain_name)


def _get_domain_by_id(domain_id, name=BASE_URI):
    with connect_to_uri(name) as conn:
        return conn.lookupByID(domain_id)


def screenshot(domain_id, path, name=BASE_URI):
    """
    Save domain screenshot to path as image_x-portable-pixmap
    example path
    /var/lib/ecp-veil/node/screenshots/9fe5519a-223c-47ae-9d9d-5bbe5f5e5ea8_image_x-portable-pixmap.screenshot"""
    with connect_to_uri(name) as conn:
        domain = conn.lookupByName(domain_id)
        stream = conn.newStream()
        # imageType = domain.screenshot(stream, 0)
        domain.screenshot(stream, 0)
        with open(path, 'wb') as f:
            stream_bytes = stream.recv(262120)
            while stream_bytes != b'':
                f.write(stream_bytes)
                stream_bytes = stream.recv(262120)
        stream.finish()
    return f"Successfully create domain {domain_id} screenshot in path {path}"


def virt_is_active(domain_name, name=BASE_URI):
    try:
        domain = _get_domain(domain_name, name)
        return domain.isActive() == 1
    except libvirt.libvirtError:
        return False


def virt_is_running(domain_name, name=BASE_URI):
    try:
        domain = _get_domain(domain_name, name)
        return domain.state()[0] in [1, 7]
    except libvirt.libvirtError:
        return False


def virt_is_pmsuspended(domain_name, name=BASE_URI):
    try:
        domain = _get_domain(domain_name, name)
        return domain.state()[0] == 7
    except libvirt.libvirtError:
        return False


def virt_pmwakeup(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.pMWakeup() == 0


def virt_is_paused(domain_name, name=BASE_URI):
    try:
        domain = _get_domain(domain_name, name)
        return domain.state()[0] == 3
    except libvirt.libvirtError:
        return False


def virt_state(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.state()


def virt_undefine(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.undefine() == 0


def virt_shutdown(domain_name, flags=libvirt.VIR_DOMAIN_SHUTDOWN_DEFAULT, name=BASE_URI):
    """
    VIR_DOMAIN_SHUTDOWN_DEFAULT = 0
    VIR_DOMAIN_SHUTDOWN_ACPI_POWER_BTN = 1
    VIR_DOMAIN_SHUTDOWN_GUEST_AGENT = 2
    VIR_DOMAIN_SHUTDOWN_INITCTL = 4
    VIR_DOMAIN_SHUTDOWN_SIGNAL = 8
    VIR_DOMAIN_SHUTDOWN_PARAVIRT = 16
    """
    domain = _get_domain(domain_name, name)
    return domain.shutdownFlags(flags) == 0


def virt_shutdown_by_id(domain_id, name=BASE_URI):
    try:
        domain = _get_domain_by_id(domain_id, name)
        return domain.shutdown() == 0
    except libvirt.libvirtError:
        return False


def virt_suspend(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.suspend() == 0


def virt_resume(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.resume() == 0


def virt_destroy(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.destroy() == 0


def virt_destroy_by_id(domain_id, name=BASE_URI):
    try:
        domain = _get_domain_by_id(domain_id, name)
        return domain.destroy() == 0
    except libvirt.libvirtError:
        return False


def virt_reboot(domain_name, flags=libvirt.VIR_DOMAIN_REBOOT_DEFAULT, name=BASE_URI):
    """
    VIR_DOMAIN_REBOOT_DEFAULT = 0
    VIR_DOMAIN_REBOOT_ACPI_POWER_BTN = 1
    VIR_DOMAIN_REBOOT_GUEST_AGENT = 2
    VIR_DOMAIN_REBOOT_INITCTL = 4
    VIR_DOMAIN_REBOOT_SIGNAL = 8
    VIR_DOMAIN_REBOOT_PARAVIRT = 16
    """
    domain = _get_domain(domain_name, name)
    return domain.reboot(flags) == 0


def virt_reset(domain_name, flags=libvirt.VIR_DOMAIN_REBOOT_DEFAULT, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.reset(flags) == 0


def ctrl_alt_del(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.sendKey(0, 0, [29, 56, 111], 3, 0) == 0


def virt_attach_device(domain_name, xml, flags=0, name=BASE_URI):
    """
    VIR_DOMAIN_AFFECT_LIVE = 1
    """
    domain = _get_domain(domain_name, name)
    return domain.attachDeviceFlags(xml, flags)


def virt_detach_device(domain_name, xml, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.detachDevice(xml)


def virt_update_device(domain_name, xml, flags=0, name=BASE_URI):
    """
    VIR_DOMAIN_AFFECT_LIVE = 1
    """
    domain = _get_domain(domain_name, name)
    return domain.updateDeviceFlags(xml, flags)


# pylint: disable=redefined-builtin
def virt_set_metadata(domain_name, metadata, type=libvirt.VIR_DOMAIN_METADATA_ELEMENT,
                      key='veil', uri='http://controller/', flags=libvirt.VIR_MIGRATE_LIVE, name=BASE_URI):
    """
    types:
        VIR_DOMAIN_METADATA_DESCRIPTION = 0
        VIR_DOMAIN_METADATA_TITLE = 1
        VIR_DOMAIN_METADATA_ELEMENT = 2
    """
    domain = _get_domain(domain_name, name)
    return domain.setMetadata(type, metadata, key, uri, flags)


def virt_get_metadata(domain_name, _type=libvirt.VIR_DOMAIN_METADATA_ELEMENT,
                      uri='http://controller/', flags=libvirt.VIR_MIGRATE_LIVE, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.metadata(_type, uri, flags)


def virt_get_xml(domain_name, flags=0, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.XMLDesc(flags)


def virt_get_sched_type(domain_name, name=BASE_URI):
    return _get_domain(domain_name, name).schedulerType()


def virt_get_sched_params(domain_name, flags=None, name=BASE_URI):
    """
    types:
        VIR_DOMAIN_SCHEDULER_CPU_SHARES = "cpu_shares"
        VIR_DOMAIN_SCHEDULER_EMULATOR_PERIOD = "emulator_period"
        VIR_DOMAIN_SCHEDULER_EMULATOR_QUOTA = "emulator_quota"
        VIR_DOMAIN_SCHEDULER_GLOBAL_PERIOD = "global_period"
        VIR_DOMAIN_SCHEDULER_GLOBAL_QUOTA = "global_quota"
        VIR_DOMAIN_SCHEDULER_IOTHREAD_PERIOD = "iothread_period"
        VIR_DOMAIN_SCHEDULER_IOTHREAD_QUOTA = "iothread_quota"
        VIR_DOMAIN_SCHEDULER_LIMIT = "limit"
        VIR_DOMAIN_SCHEDULER_RESERVATION = "reservation"
        VIR_DOMAIN_SCHEDULER_SHARES = "shares"
        VIR_DOMAIN_SCHEDULER_VCPU_PERIOD = "vcpu_period"
        VIR_DOMAIN_SCHEDULER_VCPU_QUOTA = "vcpu_quota"
        VIR_DOMAIN_SCHEDULER_WEIGHT = "weight"
    """
    domain = _get_domain(domain_name, name)
    if flags:
        return domain.schedulerParametersFlags(flags)
    else:
        return domain.schedulerParameters()


def virt_set_sched_params(domain_name, params, flags=None, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    if flags:
        return domain.setSchedulerParametersFlags(params, flags)
    else:
        return domain.setSchedulerParameters(params)


def virt_set_mem_params(domain_name, params, flags=0, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.setMemoryParameters(params, flags)


def virt_get_vcpupin(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.vcpuPinInfo()


def virt_memory_stats(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    mem_stats = domain.memoryStats()
    mem_stats['max'] = domain.maxMemory()
    mem_stats['balloon'] = mem_stats['max'] - mem_stats.get('actual', 0)
    return mem_stats


def virt_memory_params(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.memoryParameters()


def virt_get_hostname(domain_name, flags=0, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.hostname(flags=flags)


def virt_get_interfaces(domain_name, name=BASE_URI):
    domain = _get_domain(domain_name, name)
    return domain.interfaceAddresses(libvirt.VIR_DOMAIN_INTERFACE_ADDRESSES_SRC_AGENT, 0)


def start_from_xml(xml, flags=0, name=BASE_URI):
    with connect_to_uri(name) as conn:
        return conn.createXML(xml, flags) is not None


def get_domains(name=BASE_URI):
    """Только для чтения"""
    with connect_to_uri_read(name) as conn:
        domains = []

        for dom_id in conn.listDomainsID():
            dom = conn.lookupByID(dom_id)
            domains.append(dom)
        return domains


def get_domains_2(name=BASE_URI):
    """Для чтения и действий"""
    with connect_to_uri(name) as conn:
        domains = []

        for dom_id in conn.listDomainsID():
            dom = conn.lookupByID(dom_id)
            domains.append(dom)
        return domains


def get_domains_3(name=BASE_URI):
    """Только для чтения"""
    with connect_to_uri_read(name) as conn:
        return conn.listAllDomains()


def get_domains_4(name=BASE_URI):
    """Для чтения и действий"""
    with connect_to_uri(name) as conn:
        return conn.listAllDomains()


def list_domains(name=BASE_URI):
    with connect_to_uri_read(name) as conn:
        return conn.listDomainsID()


def node_info(name=BASE_URI):
    with connect_to_uri_read(name) as conn:
        raw = conn.getInfo()
        info = {'cpu_cores': raw[6],
                'cpu_mhz': raw[3],
                'cpu_architecture': str(raw[0]),
                'cpu_count': raw[2],
                'cpu_threads': raw[7],
                'cpu_numa_nodes': raw[4],
                'memory_count': raw[1],
                'cpu_sockets': raw[5]}
        return info


def list_all_storage_pools(flags=0, name=BASE_URI):
    """
    VIR_CONNECT_LIST_STORAGE_POOLS_INACTIVE = 1
    VIR_CONNECT_LIST_STORAGE_POOLS_ACTIVE = 2
    VIR_CONNECT_LIST_STORAGE_POOLS_PERSISTENT = 4
    VIR_CONNECT_LIST_STORAGE_POOLS_TRANSIENT = 8
    VIR_CONNECT_LIST_STORAGE_POOLS_AUTOSTART = 16
    VIR_CONNECT_LIST_STORAGE_POOLS_NO_AUTOSTART = 32
    VIR_CONNECT_LIST_STORAGE_POOLS_DIR = 64
    VIR_CONNECT_LIST_STORAGE_POOLS_FS = 128
    VIR_CONNECT_LIST_STORAGE_POOLS_NETFS = 256
    VIR_CONNECT_LIST_STORAGE_POOLS_LOGICAL = 512
    VIR_CONNECT_LIST_STORAGE_POOLS_DISK = 1024
    VIR_CONNECT_LIST_STORAGE_POOLS_ISCSI = 2048
    VIR_CONNECT_LIST_STORAGE_POOLS_SCSI = 4096
    VIR_CONNECT_LIST_STORAGE_POOLS_MPATH = 8192
    VIR_CONNECT_LIST_STORAGE_POOLS_RBD = 16384
    VIR_CONNECT_LIST_STORAGE_POOLS_SHEEPDOG = 32768
    VIR_CONNECT_LIST_STORAGE_POOLS_GLUSTER = 65536
    VIR_CONNECT_LIST_STORAGE_POOLS_ZFS = 131072
    VIR_CONNECT_LIST_STORAGE_POOLS_VSTORAGE = 262144
    VIR_CONNECT_LIST_STORAGE_POOLS_ISCSI_DIRECT = 524288
    """
    with connect_to_uri_read(name) as conn:
        return conn.listAllStoragePools(flags)


def list_all_vols(flags=0, name=BASE_URI):
    """
    """
    vols = []
    with connect_to_uri_read(name) as conn:
        pools = conn.listAllStoragePools(flags)
        for pool in pools:
            for vol in pool.listAllVolumes(0):
                vols.append(vol)
    return vols


def list_all_devices(flags=0, name=BASE_URI):
    """
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_SYSTEM = 1
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_PCI_DEV = 2
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_USB_DEV = 4
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_USB_INTERFACE = 8
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_NET = 16
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_SCSI_HOST = 32
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_SCSI_TARGET = 64
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_SCSI = 128
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_STORAGE = 256
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_FC_HOST = 512
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_VPORTS = 1024
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_SCSI_GENERIC = 2048
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_DRM = 4096
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_MDEV_TYPES = 8192
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_MDEV = 16384
    VIR_CONNECT_LIST_NODE_DEVICES_CAP_CCW_DEV = 32768
    """
    with connect_to_uri_read(name) as conn:
        return conn.listAllDevices(flags)


def create_node_device(xmlDesc, flags, name=BASE_URI):
    with connect_to_uri(name) as conn:
        return conn.nodeDeviceCreateXML(xmlDesc, flags)


def cpu_baseline(full=False, migratable=False, out='libvirt', name=BASE_URI):
    with connect_to_uri_read(name) as conn:
        caps = conn.getCapabilities()
        caps = minidom.parseString(caps)

        cpu = caps.getElementsByTagName('host')[0].getElementsByTagName('cpu')[0]

        flags = 0
        if migratable:
            # This one is only in 1.2.14+
            if getattr(libvirt, 'VIR_CONNECT_BASELINE_CPU_MIGRATABLE', False):
                flags += libvirt.VIR_CONNECT_BASELINE_CPU_MIGRATABLE
            else:
                raise ValueError

        if full and getattr(libvirt, 'VIR_CONNECT_BASELINE_CPU_EXPAND_FEATURES', False):
            # This one is only in 1.1.3+
            flags += libvirt.VIR_CONNECT_BASELINE_CPU_EXPAND_FEATURES

        cpu = conn.baselineCPU([cpu.toxml()], flags)
        cpu = minidom.parseString(cpu).getElementsByTagName('cpu')
        cpu = cpu[0]

        if full and not getattr(libvirt, 'VIR_CONNECT_BASELINE_CPU_EXPAND_FEATURES', False):
            # Try do it by ourselves
            # Find the models in cpu_map.xml and iterate over them for as long as entries have submodels
            with open('/usr/share/libvirt/cpu_map.xml', 'r') as cpu_map:
                cpu_map = minidom.parse(cpu_map)

            cpu_model = cpu.getElementsByTagName('model')[0].childNodes[0].nodeValue
            while cpu_model:
                cpu_specs = [el for el in cpu_map.getElementsByTagName('model') if el.getAttribute('name') == cpu_model and el.hasChildNodes()]

                if not len(cpu_specs):
                    raise ValueError(f'Model {cpu_model} not found in CPU map')
                elif len(cpu_specs) > 1:
                    raise ValueError(f'Multiple models {cpu_model} found in CPU map')

                cpu_specs = cpu_specs[0]

                cpu_model = cpu_specs.getElementsByTagName('model')
                if not len(cpu_model):
                    cpu_model = None
                else:
                    cpu_model = cpu_model[0].getAttribute('name')

                for feature in cpu_specs.getElementsByTagName('feature'):
                    cpu.appendChild(feature)

        if out == 'libvirt':
            return cpu.toxml()
        elif out == 'salt':
            return {
                'model': cpu.getElementsByTagName('model')[0].childNodes[0].nodeValue,
                'vendor': cpu.getElementsByTagName('vendor')[0].childNodes[0].nodeValue,
                'features': [feature.getAttribute('name') for feature in cpu.getElementsByTagName('feature')]
            }
